//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2019 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page : http://www.silogic.com/trains/RTC_Running.html
//---------------------------------------------------------------------------
/*
Encapsulates the packet format and response for the TIU

Only one thread of this routine runs
*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#pragma package(smart_init)

#include "Message_Thread.h"
#include "Train_Control.h"
#include "Tserial_event.h"
#include "Comm_Thread.h"
#include "Setup_Control.h"
#include "Debug_Controls.h"
#include "RTCMegaMaster.h"

#include <algorithm>	// STL - needed for lower_bound()


// Rocksoft Model CRC Algorithm
#include "crcmodel.h"

// AKA TSerialEventThread
extern Tserial_event * Serial_port;
extern Tserial_event * RFID_serial_port;

// Radio Modem WiFi Send and Receive ports
static TWSocket * RadioSend = NULL, * RadioRecv = NULL, * AIUSend;

// global vars that might be used in an ISR should be tagged volatile
volatile int iTIUbufCount;
volatile int iRFIDbufCount;
volatile byte cTIUbuf[CMDMAX];				// data from the TIU
volatile byte cRFIDbuf[CMDMAX];				// data from the RFID Detectors

// global vars that might be used in an ISR should be tagged volatile
extern volatile bool bRadioCOMConnected;	// Radio Serial Port Connection Complete Flag
extern volatile bool bRFIDCOMConnected;	    // RFID Serial Port Connection Complete Flag
extern volatile bool bRadioWiFiConnected;	// Radio WiFi Port Connection Complete Flag
extern volatile bool bRFIDWiFiConnected;	// RFID WiFi Port Connection Complete Flag
volatile bool RadioDataSent;				// Indicates that the Radio packet was sent either via COM or WiFi
// AIU Emulator
extern volatile bool bAIUWiFiTXConnected;	// AIU WiFi TX Port Connection Complete Flag
volatile bool AIUDataSent;					// Indicates that the AIU packet was sent via WiFi
extern struct AIU_EMU AIUEmulator[TIUMAX];				// Flags to indicate an AIU Emulator is attached to an emulated TIU
extern struct MSG_DISPLAY MessageDisplay[100];          // Flags to indicate a message display ( 1 - 100 ) is available

extern bool Ignore_FF_Char;
extern byte Sequence;			// Sequence number for TX packets

void __fastcall byteEncodeAString(Cmd * CSource, Cmd * CDest);
String __fastcall byteHextoString(byte * str);

extern TCommThread * ptrTIUCommThread;				// pointer to TIU communications thread
extern int PC_Tenths;								// RunTimeinTenths counter (system wide)
extern String sEncodeAString(Cmd CS);

extern bool bEmergencyStop;

HANDLE CommReqEvent;		// for event communications with CommThread

//---------------------------------------------------------------------------
//
// check that d is a valid Encoded digit, that is, is 0x5,0x6 0x9, or 0xA.
//
inline unsigned char ValidDigit(unsigned char d)
{
if (d == 0x05 || d == 0x06 || d == 0x09 || d == 0x0A) return d;
return 0;
}
//---------------------------------------------------------------------------
//
// check that c is a valid encoded byte, that is, is "0x4D" or contains only digits 0x5,0x6 0x9, and 0xA.
//
// W8027 Functions containing some if statements are not expanded inline.
// W8070 Function should return a value
inline unsigned char ValidByte(unsigned char c)
{
if (c == 0x4D) return c;
if (ValidDigit(c & 0x0F))
  if (ValidDigit((c >> 4) & 0x0F))
    return c;
return '\0';
}
/* ======================================================== */
/* ===============  OnCharArrival     ===================== */
/* ======================================================== */

void SerialEventManager(uint32 object, uint32 event)
{
char txt[512];
Tserial_event *com = (Tserial_event *) object;
if (com!=0)
    {
    switch(event)
        {
        default			:	// Handles CTS, DSR, and CD along with Error and Break
            SendMessage(MainForm->Handle, CM_RADIOCOMSTATUS, 0, event);
            break;

        case  SERIAL_CONNECTED  :
            bRadioCOMConnected = true;	// this is the only spot where this is set true
            SendMessage(MainForm->Handle, CM_RADIOCOMSTATUS, 0, event);
            break;

        case  SERIAL_DISCONNECTED  :
            bRadioCOMConnected = false;	// this is the only spot where this is set false
            // pass the disconnect reason to the MainForm
            SendMessage(MainForm->Handle, CM_RADIOCOMSTATUS, com->DoneReason, event);
            break;

        case  SERIAL_DATA_SENT  :
	        RadioDataSent = true;
            // Don't have to do anything here
            break;

        case  SERIAL_ERR			:
            // look at ClearCommError() for values
            SendMessage(MainForm->Handle, CM_RADIOCOMSTATUS, com->getCommError(), event);
            break;

        case  SERIAL_DATA_ARRIVAL  :
            int size   = com->getDataInSize();
            const char * buffer = com->getDataInBuffer();
            if (size > 511) size = 511;	// don't allow buffer overflow
            // Get the Data
            memcpy(txt, buffer, size);
            com->dataHasBeenRead();
            txt[size] = NULL;	// Null terminate
            //
            // Don't take strlen(txt) as the string may contain one
            // character of 0x00 and that appears as an empty string
            //
            // Display as Hex
            for(int i = 0 ; i < size; i ++ ) {
                // get the character from our buffer and convert it to an integer
                const unsigned char Char1 = static_cast<unsigned char>(txt[i]) & 0x00FF ;
                // skip certain invalid characters
                if (Ignore_FF_Char && (Char1 == 0xFF)) {
                    continue;
                    }
                if (!ValidByte(Char1)) {
                    continue;
                    }
                if (iTIUbufCount > (CMDMAX - 10)) {
                	iTIUbufCount = 0;	// dont' allow buffer overflow
                    }
                // State machine to detect the 55554D prefix on each packet returned
                // from the TIU. Code added July 2018
                switch(iTIUbufCount) {
                    case 0:
                        // wait until a 0x55 is detected
                        if (Char1 == 0x55) {
                            cTIUbuf[iTIUbufCount++] = Char1;
                            // Send CM_TIUCHAR message to MainForm which will move Char1 into the TIU Buffer
                            if (DEBUGGING) (void) PostMessage(MainForm->Handle, CM_TIUCHAR, Char1, iTIUbufCount);
                            } else iTIUbufCount = 0;
                        break;
                    case 1:
                        // second character must also be a 0x55
                        if (Char1 == 0x55) {
                            cTIUbuf[iTIUbufCount++] = Char1;
                            // Send CM_TIUCHAR message to MainForm which will move Char1 into the TIU Buffer
                            if (DEBUGGING) (void) PostMessage(MainForm->Handle, CM_TIUCHAR, Char1, iTIUbufCount);
                            } else iTIUbufCount = 0;
                        break;
                    case 2:
                        // Third character must be a 0x4D
                        if (Char1 == 0x4D) {
                            // found the prefix
                            cTIUbuf[iTIUbufCount++] = Char1;
                            // Send CM_TIUCHAR message to MainForm which will move Char1 into the TIU Buffer
                            if (DEBUGGING) (void) PostMessage(MainForm->Handle, CM_TIUCHAR, Char1, iTIUbufCount);
                            } else iTIUbufCount = 0;
                        break;
                    default:
                        // once the prefix is found, process all characters
                        // until the RX is turned off
                        cTIUbuf[iTIUbufCount++] = Char1;
                        // Send CM_TIUCHAR message to MainForm which will move Char1 into the TIU Buffer
                        if (DEBUGGING) (void) PostMessage(MainForm->Handle, CM_TIUCHAR, Char1, iTIUbufCount);
                        break;
                    }
                }
            break;
        }
    }
}
//---------------------------------------------------------------------------
/* ======================================================== */
/* ===============  OnCharArrival 2  ====================== */
/* ==== Runs in the context of the TSerial_event Thread === */
/* ========NO VCL calls in this routine =================== */
/* ======================================================== */
static void SerialEventManager2(uint32 object, uint32 event)
{
char txt[512];
//static DWORD CE;
Tserial_event *com = (Tserial_event *) object;
if (com!=0) {
    switch(event) {
        default			:
            SendMessage(MainForm->Handle, CM_RFIDCOMSTATUS, com->DoneReason, event);
            break;

        case  SERIAL_CONNECTED  :
            bRFIDCOMConnected = true;	// this is the only spot where this is set true
            iRFIDbufCount = 0;	// RFID Port Byte Counter
            cRFIDbuf[0] = '\0';
            SendMessage(MainForm->Handle, CM_RFIDCOMSTATUS, 0, event);
            break;

        case  SERIAL_DISCONNECTED  :
            bRFIDCOMConnected = false;	// this is the only spot where this is set false
            iRFIDbufCount = 0;	// RFID Port Byte Counter
            cRFIDbuf[0] = '\0';           
            // pass the disconnect reason to the MainForm
            SendMessage(MainForm->Handle, CM_RFIDCOMSTATUS, com->DoneReason, event);
            break;

        case  SERIAL_ERR		:
            // look at ClearCommError() for values
            SendMessage(MainForm->Handle, CM_RFIDCOMSTATUS, com->getCommError(), event);
            assert(false);
            break;

        case  SERIAL_DATA_ARRIVAL  :
            int size   = com->getDataInSize();
            const char * buffer = com->getDataInBuffer();
            if (size > 511) size = 511; // don't allow buffer overflow
            // Get the Data
            memcpy(txt, buffer, size);
            txt[size] = NULL;     // NULL terminator
            com->dataHasBeenRead();
            //
            for(int i = 0 ; i < size; i ++ ) {
                // get the character from our buffer and convert it to an integer
                const unsigned char Char1 = static_cast<unsigned char>(txt[i]) & 0x00FF ;
                if (Char1 == '\r') continue;	// ignore Carriage Return 0x0D CR character
                if (iRFIDbufCount > (CMDMAX - 10)) iRFIDbufCount = 0;	// dont' allow buffer overflow
                if (Char1 != '\n') cRFIDbuf[iRFIDbufCount++] = Char1;	// buffer all non 0x0A LF characters
                assert(iRFIDbufCount < 70);
                // Send CM_RFIDASCIICHAR message to MainForm which will move Char1 into the RFID Buffer
                // Not necessary as CM_PROCESSRFIDDATA message will display the complete incoming packet
                //if (DEBUGGING) (void) PostMessage(MainForm->Handle, CM_RFIDASCIICHAR, Char1, iRFIDbufCount);
                if (Char1 == '\n') {  // on Line Feed 0x0A LF character
                    cRFIDbuf[iRFIDbufCount] = '\0';	// null terminated
					//
                    // send tag to MainForm which relays it to all PC threads.
                    char * pBuffer = new char[(CMDMAX - 10)];	// NOTE: buffer is delete[]d in the CM_PROCESSRFIDDATA message handler CMProcessRFIDData()
                    assert(pBuffer != NULL);
                    strncpy(pBuffer, (char *) cRFIDbuf, (CMDMAX - 10));
                    // Send CM_PROCESSRFIDDATA message to MainForm which will handle the incoming data in pBuffer
                    PostMessage(MainForm->Handle, CM_PROCESSRFIDDATA, Char1, (long)pBuffer);    // Serial RFID Full Packet
                    // Message handler sends RFIDReqEvents to all PC threads.
                    // pBuffer is deleted in the MainForm Message Handler
                    // clear the buffer
                    cRFIDbuf[0] = '\0';
                    iRFIDbufCount = 0;
                    }
                }
            break;
        }
    }
}
//---------------------------------------------------------------------------
//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall TCommThread::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------
__fastcall TCommThread::TCommThread(bool CreateSuspended)
	: TThread(CreateSuspended)
{
assert(false);	// never called since I used the overloaded constructor below
}
//---------------------------------------------------------------------------
__fastcall TCommThread::TCommThread(bool CreateSuspended, TForm * Caller, HWND Parent, String PortName, int BaudRate, int ParitySel, char BitsSel, int StopSel,
			bool Radio_COM_Mode, bool Radio_WiFi_Mode, bool RFID_COM_Mode, bool RFID_WiFi_Mode, bool AIU_WiFi_Mode, int * ResultValue)
	: TThread(CreateSuspended)
{
// Calling Form and Handle
MyCaller = Caller;
ParentHandle = Parent;

MyResultValue = ResultValue;

Finished = false;

// free memory on terminate so that parent does not have to worry about it
FreeOnTerminate = true;

// termination
OnTerminate = &RunOnTerminate;

// Priority
Priority = tpHighest;
//Priority = tpTimeCritical;

// Operation Modes
My_Radio_COM_Mode = Radio_COM_Mode;
My_RFID_COM_Mode = RFID_COM_Mode;
My_Radio_WiFi_Mode = Radio_WiFi_Mode;
My_RFID_WiFi_Mode = RFID_WiFi_Mode;
assert(My_Radio_COM_Mode || My_Radio_WiFi_Mode);	// one or the other mode must be selected
My_AIU_WiFi_Mode = AIU_WiFi_Mode;

// Radio Serial port if needed
if (My_Radio_COM_Mode) {
    if (!Serial_port) Serial_port = new Tserial_event;
    assert(Serial_port);
    // Set Event Manager
    Serial_port->setManager(SerialEventManager);
    } else {
    Serial_port = NULL;
    }
// RFID Serial port if needed
if (My_RFID_COM_Mode) {
    if (!RFID_serial_port) RFID_serial_port = new Tserial_event;
    assert(RFID_serial_port);
    // Set Event Manager
    RFID_serial_port->setManager(SerialEventManager2);
    } else {
    RFID_serial_port = NULL;
    }

// RFID Port
iRFIDbufCount = 0;	// RFID Port Byte Counter

RadioWiFiRXConnected = false;
RadioWiFiTXConnected = false;

Initialize();

// indicate data buffer is empty
RadioDataSent = true;
AIUDataSent = true;
*MyResultValue = 0;
}
//---------------------------------------------------------------------------

__fastcall TCommThread::~TCommThread() {
// Destructor
if (DEBUGGINGX(1)) {
	PrintItem("TCommThread::~TCommThread ThreadID " + IntToHex((int)ThreadID, 8), (TColor) clGreen);
    if (My_Radio_COM_Mode)
    	PrintItem("TSerialEventThread:~TSerialEventThread Serial Port ThreadID " + IntToHex((int)Serial_port->getThreadId(), 8), (TColor) clGreen);
    if (My_RFID_COM_Mode)
    	PrintItem("TSerialEventThread:~TSerialEventThread RFID Port ThreadID " + IntToHex((int)RFID_serial_port->getThreadId(), 8), (TColor) clGreen);
	}
CloseHandle(CommReqEvent);		// not really needed - the system closes the handle automatically when the process terminates
// close RFID serial port
if (RFID_serial_port) {
    delete RFID_serial_port;
    RFID_serial_port = NULL;
    };
// close Radio serial port
if (Serial_port) {
    delete Serial_port;
    Serial_port = NULL;
    }
Finished = true;
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::Initialize(void)     // called using Synchronize()
{
/*
I've learned not to intermix calls to MainForm->Delay() and Sleep(). Sleep() is ok
in the Execute() function but this function runs as the main thread which
makes a lot of calls to send commands to the TIU which uses MainForm->Delay().
*/
NumRetries = Setup->NumRetries->Value;			// 0-9 with 10 meaning to disable sending packets
Speed_Mode = Setup->Speed_Mode->Checked;
TXOnlyComm = Setup->TXOnlyComm->Checked;
Echo = Setup->Echo->Checked;
Ignore_FF_Char = Setup->Ignore_FF->Checked;
// Radio WiFi Port
RadioIP = Setup->RadioIP->Text;
RadioPort = Setup->RadioPort->Text;
MultiCast = Setup->Radio_WiFi_Multicast->Checked;
// AIU Emulators WiFi Port
AIUIP = Setup->AIUIP->Text;
AIUPort = Setup->AIUPort->Text;
AIUMultiCast = Setup->AIU_WiFi_Multicast->Checked;
if (Setup->IPComboBox->Items->Count >  0 && Setup->IPComboBox->ItemIndex != -1) {
    FServerAddr.S_un.S_addr = WSocket_inet_addr(Setup->IPComboBox->Items->Strings[Setup->IPComboBox->ItemIndex].c_str());
    }
}

//---------------------------------------------------------------------------

void __fastcall TCommThread::CleanUp(void)     // called using Synchronize()
{
/*
I've learned not to intermix calls to MainForm->Delay() and Sleep(). Sleep() is ok
in the Execute() function but this function runs as the main thread which
makes a lot of calls to send commands to the TIU which uses MainForm->Delay().
*/
// not used
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioRecvChangeState(TObject *Sender,
      TSocketState OldState, TSocketState NewState)
{
// handle change of state
if (DEBUGGINGX(9)) PrintItem("Radio RX WiFi Change State = " + String(OldState) + " to " + String(NewState));
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioRecvSessionClosed(TObject *Sender,
      WORD ErrCode)
{
//ConnectedRadioLED->Checked = false;
//ConnectedRadioLED->Flashing = false;
SetLED(MainForm->ConnectedRadioLED, clLime, false, false);
bRadioWiFiConnected = false;
RadioWiFiRXConnected = false;
if (RadioRecv->MultiCastAddrStr.IsEmpty()) {
	if (DEBUGGING) PrintItem("Radio RX WiFi Unicast disconnected from " + RadioRecv->Addr + " - Port " + RadioRecv->Port);
	} else {
	if (DEBUGGING) PrintItem("Radio RX WiFi Multicast disconnected from " + RadioRecv->MultiCastAddrStr + " - Port " + RadioRecv->Port);
	}
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioRecvSessionConnected(TObject *Sender,
      WORD ErrCode)
{
if (ErrCode != 0) {
	PrintItem("Can't connect to Radio RX WiFi. Error #" + String(ErrCode));
	} else {
	//ConnectedRadioLED->Checked = true;
	//ConnectedRadioLED->Flashing = false;
	SetLED(MainForm->ConnectedRadioLED, clLime, true, false);
	bRadioWiFiConnected = true;
	RadioWiFiRXConnected = true;
	if (RadioRecv->MultiCastAddrStr.IsEmpty()) {
		if (DEBUGGING) PrintItem("Radio RX WiFi Unicast connected to " + RadioRecv->Addr + " - Port " + RadioRecv->Port);
        } else {
		if (DEBUGGING) PrintItem("Radio RX WiFi Multicast connected to " + RadioRecv->MultiCastAddrStr + " - Port " + RadioRecv->Port);
        }
	}
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioSendChangeState(TObject *Sender,
      TSocketState OldState, TSocketState NewState)
{
// handle change of state
if (DEBUGGINGX(9)) PrintItem("Radio TX WiFi Change State = " + String(OldState) + " to " + String(NewState));
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioSendSendData(TObject *Sender,
      int BytesSent)
{
//RadioDataSent = true;		// indicate that Winsock send buffer is empty, will be filled again by internal send buffer if data available to send.
if (DEBUGGINGX(7)) PrintItem("Radio TX WiFi Send Data");
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioSendDataSent(TObject *Sender, WORD ErrCode)
{
RadioDataSent = true;		// indicate that Internal sent buffer is completely delivered to winsock.
if (ErrCode > 0) {
	if (DEBUGGING) PrintItem("Radio TX WiFi Data Sent - ErrCode = " + String(ErrCode));
	} else {
	if (DEBUGGINGX(7)) PrintItem("Radio TX WiFi Send Data Sent");
	}
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioSendSessionClosed(TObject *Sender,
      WORD ErrCode)
{
//ConnectedRadioLED->Checked = false;
//ConnectedRadioLED->Flashing = false;
SetLED(MainForm->ConnectedRadioLED, clLime, false, false);
bRadioWiFiConnected = false;
RadioWiFiTXConnected = false;
if (DEBUGGING) PrintItem("Radio TX WiFi disconnected from " + RadioSend->Addr + " - Port " + RadioSend->Port);
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioSendSessionConnected(TObject *Sender,
      WORD ErrCode)
{
if (ErrCode != 0) {
	PrintItem("Can't connect to Radio TX WiFi. Error #" + String(ErrCode));
	} else {
	//ConnectedRadioLED->Checked = true;
	//ConnectedRadioLED->Flashing = false;
	SetLED(MainForm->ConnectedRadioLED, clLime, true, false);
	bRadioWiFiConnected = true;
	RadioWiFiTXConnected = true;
	if (DEBUGGING) PrintItem("Radio TX WiFi connected to " + RadioSend->Addr + " - Port " + RadioSend->Port);
	}
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::AIUSendChangeState(TObject *Sender,
      TSocketState OldState, TSocketState NewState)
{
// handle change of state
if (DEBUGGINGX(9)) PrintItem("AIU TX WiFi Change State = " + String(OldState) + " to " + String(NewState));
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::AIUSendSendData(TObject *Sender,
      int BytesSent)
{
//AIUDataSent = true;		// indicate that Winsock send buffer is empty, will be filled again by internal send buffer if data available to send.
if (DEBUGGINGX(7)) PrintItem("AIU TX WiFi Send Data");
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::AIUSendDataSent(TObject *Sender, WORD ErrCode)
{
AIUDataSent = true;		// indicate that Internal sent buffer is completely delivered to winsock.
if (ErrCode > 0) {
	if (DEBUGGING) PrintItem("AIU TX WiFi Data Sent - ErrCode = " + String(ErrCode));
	} else {
	if (DEBUGGINGX(7)) PrintItem("AIU TX WiFi Send Data Sent");
	}
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::AIUSendSessionClosed(TObject *Sender,
      WORD ErrCode)
{
bAIUWiFiTXConnected = false;
AIUWiFiTXConnected = false;
if (DEBUGGING) PrintItem("AIU TX WiFi disconnected from " + AIUSend->Addr + " - Port " + AIUSend->Port);
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::AIUSendSessionConnected(TObject *Sender,
      WORD ErrCode)
{
if (ErrCode != 0) {
	PrintItem("Can't connect to AIU TX WiFi. Error #" + String(ErrCode));
	} else {
	bAIUWiFiTXConnected = true;
	AIUWiFiTXConnected = true;
	if (DEBUGGING) PrintItem("AIU TX WiFi connected to " + AIUSend->Addr + " - Port " + AIUSend->Port);
	}
}
//---------------------------------------------------------------------------
/* From winsock2.h
//
struct in_addr {
  union {
    struct {
      u_char s_b1;
      u_char s_b2;
      u_char s_b3;
      u_char s_b4;
    } S_un_b;
    struct {
      u_short s_w1;
      u_short s_w2;
    } S_un_w;
    u_long S_addr;
  } S_un;
};
*/
//---------------------------------------------------------------------------

void __fastcall TCommThread::RadioRecvDataAvailable(TObject *Sender,
      WORD ErrCode)
{
if (ErrCode != 0) {
    if (DEBUGGING) PrintItem("Radio RX WiFi Connection Error " + String(ErrCode));
	return;
	}
int DetectorLen = sizeof(Detector);
iTIUbufCount = RadioRecv->ReceiveFrom((char *)cTIUbuf, sizeof(cTIUbuf), Detector, DetectorLen);
if (iTIUbufCount <  0) {		// ignore packets in error
    return;
    }
if (DEBUGGINGX(8)) {
	#ifdef DEBUGONLY
	PrintItem("Radio RX WiFi Dotted Host Address = " + String(WSocket_inet_ntoa(FServerAddr)));
    #endif
	PrintItem("Radio RX WiFi Sin Port = " + String(Detector.sin_port));
	PrintItem("Radio RX WiFi Dotted Sin Address = " + String(WSocket_inet_ntoa(Detector.sin_addr)));
    }
// Since we can use Multicast packet transmission from RTC (to IP 239.255.200.3), this means
// that RTC will receive its packets like an echo. We need to check for this and ignore those packets
// check all four octets
if (Detector.sin_addr.S_un.S_addr == FServerAddr.S_un.S_addr) {
	return;
    }
if (iTIUbufCount == 0) {
    if (DEBUGGING) PrintItem("Radio RX WiFi Connection Closed in OnDataAvaiable()");
    return;
    }
cTIUbuf[iTIUbufCount] = '\0';				// NULL Terminate
if (DEBUGGING) {
    SetTextValue(MainForm->TIU_Char_Count, String(iTIUbufCount));
    AddCntr(MainForm->TIU_Total_Byte_Count, iTIUbufCount);
    }
if (DEBUGGINGX(8)) PrintItem("Radio TX WiFi Raw Data : " + String((char *)cTIUbuf));
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::Execute()
{
//
// Calls to Sleep() are ok in this thread
// NO VCL Calls in this thread
//
if (DEBUGGINGX(1)) {
	// print the ThreadId of the Communications Thread
	PrintItem("TCommThread: ThreadID " + IntToHex((int)GetCurrentThreadId(), 8), (TColor) clGreen);
   	// print the Serial port ThreadIds if they are running --  NOTE It takes a few milliseconds for the serial threads to start
    // otherwise calls to getThreadId() return zero.
    Sleep(50);
    if (My_Radio_COM_Mode) {
    	PrintItem("TSerialEventThread: Serial Port ThreadID " + IntToHex((int)Serial_port->getThreadId(), 8), (TColor) clGreen);
        }
    if (My_RFID_COM_Mode) {
    	PrintItem("TSerialEventThread: RFID Port ThreadID " + IntToHex((int)RFID_serial_port->getThreadId(), 8), (TColor) clGreen);
        }
    }

//Synchronize(Initialize);

// this has no effect in BCB4.0 and Thread view not available in CB10.1
// it does work on BCB6.0
// http://docwiki.embarcadero.com/RADStudio/Rio/en/Converting_an_Unnamed_Thread_to_a_Named_Thread
THREADNAME_INFO info;
info.dwType = 0x1000;
info.szName = "Comm_Thread";
info.dwThreadID = -1;
info.dwFlags = 0;
__try {
	RaiseException(0x406D1388, 0, sizeof(info) / sizeof(DWORD), (DWORD*)&info);
	}
__except (EXCEPTION_CONTINUE_EXECUTION) {
	}

if (My_Radio_COM_Mode) {
    info.dwType = 0x1000;
    info.szName = "Radio_Serial_Thread";
    info.dwThreadID = Serial_port->getThreadId();
    info.dwFlags = 0;
    __try {
        RaiseException(0x406D1388, 0, sizeof(info) / sizeof(DWORD), (DWORD*)&info);
        }
    __except (EXCEPTION_CONTINUE_EXECUTION) {
        }
    }

if (My_RFID_COM_Mode) {
    info.dwType = 0x1000;
    info.szName = "RFID_Serial_Thread";
    info.dwThreadID = RFID_serial_port->getThreadId();
    info.dwFlags = 0;
    __try {
        RaiseException(0x406D1388, 0, sizeof(info) / sizeof(DWORD), (DWORD*)&info);
        }
    __except (EXCEPTION_CONTINUE_EXECUTION) {
        }
    }
// Moved from Constructor
// create Comm Request event
CommReqEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
assert(CommReqEvent != NULL);

if (My_Radio_WiFi_Mode || My_AIU_WiFi_Mode) {
    // Construct the Radio TX WiFi Port
    RadioSend = new TWSocket(NULL);
    assert(RadioSend);	// make sure constructor worked
    // Construct the Radio RX WiFi Port
    RadioRecv = new TWSocket(NULL);
    assert(RadioRecv);	// make sure constructor worked
    //
    //RadioSend->OnSendData = RadioSendSendData;
    RadioSend->OnDataSent = RadioSendDataSent;
    RadioSend->OnSessionClosed = RadioSendSessionClosed;
    RadioSend->OnSessionConnected = RadioSendSessionConnected;
    //RadioSend->OnChangeState = RadioSendChangeState;

	RadioSend->ReuseAddr 			= true;
    RadioSend->Proto 				= "udp";
	RadioSend->Addr  				= RadioIP;		// use Radio Modem IP from Setup
	RadioSend->Port  				= RadioPort;	// use Port from Setup
	// http://wiki.overbyte.eu/wiki/index.php/TWSocket.MultiCast
    // https://en.delphipraxis.net/topic/1429-multicast-configuration/
    // https://tldp.org/HOWTO/Multicast-HOWTO-6.html
    // https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/udp-examples.html
	if (MultiCast) {	// Use Multicasting when sending to one RTCModem device and one/many RTCAIU devices (they all reply unicast)		
		RadioSend->MultiCast 		= true;
        //RadioSend->MultiCastAddrStr = RadioIP;	// not used for TX, use Addr
		if (DEBUGGING) PrintItem("Radio TX WiFi Connect() (Multicast) to IP " + RadioSend->Addr + " - Port " + RadioSend->Port);
		} else {
		RadioSend->MultiCast 		= false;	// Use Unicasting when sending only to one RTCModem device
		if (DEBUGGING) PrintItem("Radio TX WiFi Connect() (Unicast) to IP " + RadioSend->Addr + " - Port " + RadioSend->Port);
        }
    //
	RadioSend->Connect();
	// Connect is asynchronous (non-blocking)-> We will wait while the
    // session is connecting or application terminated	
	while(RadioWiFiTXConnected == false) {
		Sleep(10);
		}
    //
#ifdef NOTNEEDED
	FServerAddr                = WSocketResolveHost("127.0.0.1");
	if (FServerAddr.s_addr == htonl(INADDR_LOOPBACK)) {
		// Replace loopback address by real localhost IP addr
		FServerAddr            = WSocketResolveHost(LocalHostName());
		}
#endif
	//RadioRecv->Close();
    RadioRecv->OnDataAvailable = RadioRecvDataAvailable;
	RadioRecv->OnSessionClosed = RadioRecvSessionClosed;
    RadioRecv->OnSessionConnected = RadioRecvSessionConnected;
    //RadioRecv->OnChangeState = RadioRecvChangeState;
    
	RadioRecv->ReuseAddr 			= true;
	RadioRecv->Proto   				= "udp";
	RadioRecv->Addr    				= "0.0.0.0";	// Listen on all available network connections simultaniously
    //RadioRecv->Addr    				= WSocket_inet_ntoa(FServerAddr);	// Listen to one network connection only
	RadioRecv->Port    				= RadioPort;	// use Port from Setup
	
	// http://wiki.overbyte.eu/wiki/index.php/TWSocket.MultiCast
    // https://en.delphipraxis.net/topic/1429-multicast-configuration/
	// https://tldp.org/HOWTO/Multicast-HOWTO-6.html
	// https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/udp-examples.html

	// RX is unicast to this computer's IP and Port
	try {
		RadioRecv->Listen();
		while(RadioWiFiRXConnected == false) {
			Sleep(10);
			}		
		if (DEBUGGING) PrintItem("Radio RX WiFi (Unicast) listening on Port " + RadioRecv->Port);
        } catch (const ESocketException &E) {
        // The socket is probably already in use
			if (DEBUGGING) PrintItem("Radio RX WiFi error : " + E.Message);
            throw;
        }
	}


if (My_AIU_WiFi_Mode) {
    // Construct the AIU TX WiFi Port
    AIUSend = new TWSocket(NULL);
    assert(AIUSend);	// make sure constructor worked
    //
    //AIUSend->OnSendData = AIUSendSendData;
    AIUSend->OnDataSent = AIUSendDataSent;
    AIUSend->OnSessionClosed = AIUSendSessionClosed;
    AIUSend->OnSessionConnected = AIUSendSessionConnected;
    //AIUSend->OnChangeState = AIUSendChangeState;

	AIUSend->ReuseAddr 			= true;
    AIUSend->Proto 				= "udp";
	AIUSend->Addr  				= AIUIP;		// use AIU IP from Setup
	AIUSend->Port  				= AIUPort;		// use AIU Port from Setup
	// http://wiki.overbyte.eu/wiki/index.php/TWSocket.MultiCast
    // https://en.delphipraxis.net/topic/1429-multicast-configuration/
    // https://tldp.org/HOWTO/Multicast-HOWTO-6.html
    // https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/udp-examples.html
	if (AIUMultiCast) {	// Use Multicasting when sending to one/many RTCAIUWiFi devices (they all reply unicast)
		AIUSend->MultiCast 		= true;
        //AIUSend->MultiCastAddrStr = AIUIP;	// not used for TX, use Addr
		if (DEBUGGING) PrintItem("AIU TX WiFi Connect() (Multicast) to IP " + AIUSend->Addr + " - Port " + AIUSend->Port);
		} else {
		AIUSend->MultiCast 		= false;	// Use Unicasting when sending only to one RTCAIUWiFi device - debug mode only
		if (DEBUGGING) PrintItem("AIU TX WiFi Connect() (Unicast) to IP " + AIUSend->Addr + " - Port " + AIUSend->Port);
        }
    //
	AIUSend->Connect();
    // Connect is asynchronous (non-blocking)-> We will wait while the
    // session is connecting or application terminated	
	while(AIUWiFiTXConnected == false) {
		Sleep(10);
		}
    //
    }
//
// FIFO - Thread safe
//
FIFO = new TThreadList();	// - Thread Safe
assert(FIFO != NULL);		// check for successful creation
TList *myList = FIFO->LockList();
try {
	myList->Capacity = 100;
    }
__finally {
	FIFO->UnlockList();
    }
    
String PrintLine;
//
// https://docs.microsoft.com/en-us/windows/desktop/sync/using-waitable-timer-objects
//
LARGE_INTEGER liDueTime;
liDueTime.QuadPart = 0L;

// Create an unnamed waitable timer.
// use this timer to periodically check for Terminated.
HANDLE MyTimer = CreateWaitableTimer(NULL, false, NULL);	// Autoreset --> the timer is a synchronization time
if (MyTimer == NULL) {
	assert(false);	// CreateWaitableTimer failed
    }

HANDLE handles[] = { MyTimer, CommReqEvent, MainForm->TerminationCTEvent };

// start the waitable Timer for 100 milliseconds and continue on if it starts up OK.
bool bretv = SetWaitableTimer(MyTimer, &liDueTime, 100, NULL, NULL, false);
if (bretv) {
    // Timer is running OK
    struct FIFOItem *MyFIFOItem;
    while(!Terminated) {
        // check the FIFO to see if anything is there - alerted by CommReqEvent
        DWORD dwretv = WaitForMultipleObjects(3, handles, FALSE, INFINITE);
        //------------------------  Termination Event -----------------------------------
        if (dwretv == WAIT_OBJECT_0 + 2 || dwretv == WAIT_FAILED) {	// TerminationCTEvent Event Signaled or unknown failure?
	        if (DEBUGGINGX(1)) PrintItem("TCommThread: TerminationPCEvent");
            Terminate();
            break;	// time to terminate
            }
        if (Terminated) {
            break;				// in case we are terminated
            }
        //------------------------  Timer Event -----------------------------------
        if (dwretv == WAIT_OBJECT_0 + 0) {	// MyTimer Event Signaled ?
            continue;		// continue to wait
            }
        if (dwretv != WAIT_OBJECT_0 + 1) continue;
        //------------------------  Communications Request Event -----------------------------------
        // The remaining event is the CommReqEvent (WAIT_OBJECT_0 + 1)
        while (!Terminated) {
            TList *myList = FIFO->LockList();
            //
            if (myList->Count > 0) {
                // FIFO has a command ready to send
                MyFIFOItem = reinterpret_cast<struct FIFOItem *>(myList->Items[0]);
                // remove the item pointer from the FIFO
                myList->Delete(0);
                //if (MyFIFOItem != NULL) FIFO->Remove(MyFIFOItem);
                FIFO->UnlockList();
				//
                // Send the command and wait for its completion
                //
                bool bretv = false;
                switch(MyFIFOItem->Type) {
                    case CMD_Type_SendString:
                        if (DEBUGGINGX(4)) {
                            // print the values in the struct pointed to by FIFOItem
                            PrintLine = "TCommThread: Type DL_Type_SendString Command " + String((char *)MyFIFOItem->Command) + " for Consist " + String(MyFIFOItem->ConsistNo);
                            PrintItem(PrintLine, clBlack);
                            }
                        // all commands execpt the 'V' command go out the port
                        switch(MyFIFOItem->Command[0]) {
                          case 'V':
                            if (DEBUGGINGX(2)) {
                            	PrintLine = "Sending Virtual Command: " + String(MyFIFOItem->Command);
                                PrintItem(PrintLine);
                                }
                            bretv = SendReadRAMCommand(MyFIFOItem);
                          	break;
                          default:
                          	//
	                        bretv = Send_String(MyFIFOItem->ConsistNo, MyFIFOItem->Command,
    	                        MyFIFOItem->TIUNo, MyFIFOItem->Remote, MyFIFOItem->ExpectedCount, MyFIFOItem->WaitTime, MyFIFOItem->Record_Flag,
        	                    MyFIFOItem->cLashUpEngineList,
            	                MyFIFOItem->ptrRespCode, MyFIFOItem->ResponseCount, MyFIFOItem->ResponseBuffer);
                            break;
                          }
                        if (!bretv) {	// the send string call failed
                        	if (MyFIFOItem->ptrRespCode) {	// return a failure code as a response - unless a code is already present
                                if (*MyFIFOItem->ptrRespCode < UNDETERMINED) *MyFIFOItem->ptrRespCode = UNDETERMINED;
                            	}
                            }
                        break;
                    case CMD_Type_WaitForResponse:
                        if (DEBUGGINGX(4)) {
                            // print the values in the struct pointed to by FIFOItem
                            PrintLine = "TCommThread: Type DL_Type_WaitForResponse Command " + byteHextoString(MyFIFOItem->Command) + " for Consist " + String(MyFIFOItem->ConsistNo);
                            PrintItem(PrintLine, clBlack);
                            }
                        bretv = TCommThread::TIUSendWaitforResponseRetry(MyFIFOItem->ConsistNo, MyFIFOItem->Command,
                            MyFIFOItem->ExpectedCount, MyFIFOItem->WaitTime, MyFIFOItem->Remote, MyFIFOItem->TIUNo, MyFIFOItem->ptrRespCode,
                            MyFIFOItem->ResponseCount, MyFIFOItem->ResponseBuffer);
                        if (!bretv) {	// the send string call failed
                        	if (MyFIFOItem->ptrRespCode) {	// return a failure code as a response - unless a code is already present
                                if (*MyFIFOItem->ptrRespCode < UNDETERMINED) *MyFIFOItem->ptrRespCode = UNDETERMINED;
                            	}
                            }
                        break;
                    case CMD_Type_SendHex:
                        if (DEBUGGINGX(4)) {
                            // print the values in the struct pointed to by FIFOItem
                            PrintLine = "TCommThread: Type DL_Type_SendHex Command " + byteHextoString(MyFIFOItem->Command) + " for Consist " + String(MyFIFOItem->ConsistNo);
                            PrintItem(PrintLine, clBlack);
                            }
                        bretv = TCommThread::SendHexStr(MyFIFOItem->Command, MyFIFOItem->Len, MyFIFOItem->ExpectedCount, MyFIFOItem->WaitTime, MyFIFOItem->TIUNo);  // Len is actual bytes to send
                        if (!bretv) {	// the send string call failed
                        	if (MyFIFOItem->ptrRespCode) {	// return a failure code as a response - unless a code is already present
                                if (*MyFIFOItem->ptrRespCode < UNDETERMINED) *MyFIFOItem->ptrRespCode = UNDETERMINED;
                            	}
                            }
                        break;
                   case CMD_Type_DelayedCommand:
                    	// This type of message must always converted to DL_Type_SendString
                        // before being put in the FIFO. Therefore, you should never see this type of
                        // command coming out of the FIFO.
                        assert(false);
                    	break;
                    default:
                    	// ignore bad commands but return an error code
                        if (MyFIFOItem->ptrRespCode) *MyFIFOItem->ptrRespCode = ILLEGALCOMMAND;
                        PrintLine = "TCommThread: Unknown Type Command " + byteHextoString(MyFIFOItem->Command) + " for Consist " + String(MyFIFOItem->ConsistNo);
                        PrintItem(PrintLine, clRed);
                    	break;
                    }

                HANDLE CommEvent = MyFIFOItem->CommEvent;   // save Event before deleting the request
                // Free the command's memory
                delete MyFIFOItem;
                if (CommEvent != NULL) SetEvent(CommEvent);		// signal complete
                // go and try to process another command
                } else {
                // FIFO is empty
                FIFO->UnlockList();
                break;
                }
            }
		}
	} else {
    // SetWaitableTimer failed
    assert(false);
    }

// Shut Down the WiFi
// http://wiki.overbyte.eu/wiki/index.php/TWSocket.HowTo.Close
if (RadioSend) {
	RadioSend->Close();
	delete RadioSend;
	RadioSend = NULL;
    }
//
if (RadioRecv) {
	RadioRecv->Close();
	delete RadioRecv;
	RadioRecv = NULL;
    }
//
if (AIUSend) {
	AIUSend->Close();
	delete AIUSend;
	AIUSend = NULL;
    }
//
if (MyTimer) {
	CancelWaitableTimer(MyTimer);
	CloseHandle(MyTimer);
    }

if (FIFO) {
	delete FIFO;
    FIFO = NULL;
    }
//Synchronize(CleanUp);	// not used
}
//---------------------------------------------------------------------------

bool __fastcall TCommThread::Send_String(unsigned char Consist, byte * Command_String,
	unsigned char TIU, unsigned char Remote, int Expected, int WaitTime, bool Record_Flag, byte * ptrLashUpEngineList,
    int * ptrRespCode, int * SSResponseCount, byte * SSResponseBuffer)
{
// return true on success, false on failure
assert (TIU >= 1 && TIU <= TIUMAX);
assert (Command_String != NULL);
assert (Remote >= 1 && Remote <= REMOTEMAX);
// if TIU Serial Port or WiFi port is not connected fail
if (!MainForm->IsRadioConnected()) return false;	// Radio must connected either by COM or WiFi
//
int LocalRespCode = UNDETERMINED;	// must have a response code even if the caller doesn't care so create a local spot
// *ptrRespCode is set from LocalRespCode just before the function exits
//
String TextValue;
String PrintLine;
//
bool bFlag;
//
// check pointers
if (Record_Flag == RECORD_OK) {
    assert(!SSResponseCount);	// recorded command cannot have a response
    assert(!SSResponseBuffer);
	}

Cmd Cmd_S1 = {"", 0, false};    // ASCII Command
Cmd Cmd_S2 = {"", 0, false};    // Encoded Command

// if the Consist is a LashUp, use #101 (DCS #102 (0x66) always)
if (Consist >= LASHUPMIN) Consist = LASHUPMIN;
//
Cmd_S1.idx = 0;  // Create the command in (unsigned char *)Cmd_S1
// Create command string for record function and do a Debug Print of it
// Some commands could be long, longest I know of is the "J" all engine mode command
String OSS = "";
// Consist
const unsigned char I_Consist = Consist + 1;  // In spinner, engine is 1 to 99, // in command it's 2 to 100
                                      // Engine number in command of 0 (DCS #1) is for factory reset engine
                                      // #100 (DCS #101) is used for "all engine operation"
                                      // Consist number for LashUps is 101 to 120 although #101 (DCS 102 (0x66))
                                      // is always sent to the TIU
Cmd_S1.Str[Cmd_S1.idx++] = I_Consist;
//
// insert Command_String (contains Command and Data)
// NOTE: Command_String is always ASCII so strlen() is OK.
const int StrLCS = strlen((char *)Command_String);
if (StrLCS > 0)	// If no string is supplied, skip this
	for (int i = 0; i < StrLCS ; i++) {
    	Cmd_S1.Str[Cmd_S1.idx++] = Command_String[i];
		}

	// DCS 5.00 : LashUpEngineList NOTE: LashUpEngineList will never
	// contain binary zero (because DCS engine 0 does not exist)
	// but could be any other value from 1 to 255 in DCS ver 5.00 or earlier.

    // DCS 6.00 : In ver 6.00, the engine numbers are sent as hex values but as ASCII characters.
    // LashUpEngineList starts with a ',' 0x2C, followed by the engine numbers.

    // When sent to the TIU the engine numbers are sent as ASCII Characters, for
    // example Engine 14 would show as DCS Engine 15 or 0x0F, which would be sent
    // as two ASCII characters '0' and 'F' or 0x30 0x46.
    // If it were reverse running, it would show with the high order bit
    // on or 0x8F and be sent as two ASCII characters '8' and 'F' or 0x38 0x46


if (ptrLashUpEngineList != NULL && (Consist >= LASHUPMIN && Consist <= LASHUPMAX)) // ignore ptrLashUpEngineList if not a Lashup
  if (ptrLashUpEngineList[0] != '\0') {
	assert(Consist >= LASHUPMIN && Consist <= LASHUPMAX);	// must be a LashUp
    // process lashup engine list
    const int StrLLU = strlen((char*) ptrLashUpEngineList);	// never contains 0x00 except the \0 at the end of the string
    // Find the end of the LashUpEngineList as denoted by 0xFF
    int StrLength;
    for(StrLength = 0 ; StrLength < LASHUP_ENGINE_LIST_LENGTH ; StrLength++) {
    	byte zz = ptrLashUpEngineList[StrLength];
        if (zz == 0xFF) break;
    	}
    StrLength++;
    assert((StrLLU == StrLength) && (StrLength != LASHUP_ENGINE_LIST_LENGTH));	// length from strlen() must be the same as determined programmatically
    //
    // NOTE NOTE: the TIUVersion fix that I implemented first in v0.3.24 is not bulletproof.
    // If you intermix TIU with different versions, RTC will construct commands
    // based on the last TIU that was connected via the [Connect] button.
    //  Here is what I added to the v3.24 readme:
    /*
    As of V0.3.24, I recommend that you upgrade all of your Remotes
	and TIUs to DCS v6.xx. MTH made some dramatic changes to the command
	format with v6.xx and TIU with that version will not accept all commands
	from a Remote running the older DCS version. At the very minumum,
	you can still use v5.xx but be certain that all TIUs and all Remotes
	are loaded with that version.
	*/

    if (MainForm->TIUVersion < 6.00) {	// TIUVersion is a float
	    // LashUpEngineList is a null terminated char string of binary engine numbers which are sent as binary number
    	for (int i = 0; i < StrLLU ; i++) {
        	Cmd_S1.Str[Cmd_S1.idx++] =  ptrLashUpEngineList[i]; 	// Build Command to TIU
            }
        }
    if (MainForm->TIUVersion >= 6.00) {
    	int CPtr = 0;
    	if (ptrLashUpEngineList[CPtr] == ',') {
	    	// comma sent as is
    	    Cmd_S1.Str[Cmd_S1.idx++] =  ptrLashUpEngineList[CPtr]; 	// Build Command to TIU
            CPtr++;
        	}
        // LashUpEngineList is a null terminated char string of binary engine numbers which are sent as ASCII
    	for (int i = CPtr; i < (StrLLU - 1) ; i++) {
        	char EngNum[2];
       		snprintf(EngNum, sizeof EngNum, "%02X", ptrLashUpEngineList[i] & 0xFF);
        	// first character of engine number
            Cmd_S1.Str[Cmd_S1.idx++] = EngNum[0];
            // second character of engine number
            Cmd_S1.Str[Cmd_S1.idx++] = EngNum[1];
        	}
        // terminating 0xFF sent as is
        Cmd_S1.Str[Cmd_S1.idx++] =  ptrLashUpEngineList[StrLLU - 1]; 	// Build Command to TIU
    	}
	}

//
// Sequence Number
Cmd_S1.Str[Cmd_S1.idx++] = Sequence;	// Sequence is an unsigned char

// TIU/remote
unsigned char I_TIU = TIU - 1;	// in spinner it's 1 to 5, in commands it's 0 to 4

// Super Mode 0x08 is added only for Engine commands not for AIU commands or a few other commands
if (Setup->SuperTIUMode) {	// SuperTIUMode is a bool (no VCL access)
	switch(Command_String[0]) {
    	// commands which do not run in Super Mode
		case 'A':	// Accy/Switch
        case 'B':	// Change TIU Number
        case 'E':	// Change Engine Number
        case 'F':	// Feature Reset
        case 'O':	// AC Input Frequency/DCS Setup
        case '!':	// TIU Version
        case 'o':	// Emergency Stop
        case 'x':	// Read number of AIU
        case 'q':   // Query
        case 'T':	// TMCC
	        // NOP
	        break;
        //
	    default:
	        // if TIU is in Super Mode, turn on the 0x08 bit in the TIU number
			if (Setup->TIUSuperMode[I_TIU]) I_TIU |= 0x08;	// TIUSuperMode is a bool- in commands it's 0x08 or'ed into the TIU number
	        break;
        }
    }
   
const unsigned char I_Remote = Remote;   // in spinner and in commands it is 0 to 15  (0x00 to 0x0F)
const unsigned char I_TIU_Remote = ((I_TIU << 4) | I_Remote);
Cmd_S1.Str[Cmd_S1.idx++] = I_TIU_Remote;
// NULL terminate
Cmd_S1.Str[Cmd_S1.idx]  = NULL;


if (DEBUGGING || MainForm->bRecording) {	// bRecording is a bool
	// create printable command string for debugging and for recording
	OSS = OSS + String(Consist) + " ";		// Consist Number
	//
	// insert Command_String (contains Command and Data)
	// NOTE: Command_String is always ASCII
	if (StrLCS > 0)	// If no string is supplied, skip this
		for (int i = 0; i < StrLCS ; i++) {
	        OSS = OSS + String((char)Command_String[i]); 	// ASCII Command String
			}
	OSS = OSS + " ";

	if (ptrLashUpEngineList != NULL) if (ptrLashUpEngineList[0] != '\0') {
	    OSS += "(";	// start with open parenthesis
	    // process lashup engine list
	    const int StrLLU = strlen((char*) ptrLashUpEngineList);
	    // Lashup is a null terminated char string of binary engine numbers
	    for (int i = 0; i < StrLLU ; i++) {
			// NOTE NOTE : special case of Lashup "U" command does not have the 0x2C in the LashUpEngineList String
	        // also special handling in Playback and Send_Str - I don't like this
	        if (i == 0 && Command_String[0] != 'U') { 	// 'U' Create Lashup is a special case
	            if (ptrLashUpEngineList[0] == 0x2C) continue;	// skip 0x2C if it's the first byte of non-'U' commands
	        	}                                   // 'U' command does not have the comma 0x2C as the first byte
            										// of the LashUpEngineList.
	        if ((ptrLashUpEngineList[i] == (unsigned char) 0xFF)) continue; 	// skip 0xFF bytes
	        //
	        unsigned char t1 = (unsigned char) ptrLashUpEngineList[i];
	        const bool SignChange = (t1 & 0x80) ? true : false;	// check for H.O. bit on (reverse running flag)
	        t1 = (t1 & 0x7F) - 1	; // convert from DCS Number to Engine Number
	        OSS += String(SignChange ? t1 * -1 : t1) + ",";
	        //
	    	}
	    // remove last comma
	    OSS.SetLength(OSS.Length()-1);
	    OSS += ") ";	// add close parenthesis
		}
	OSS = OSS + String(TIU) + " ";		// TIU Number
	OSS = OSS + String(Remote) + " ";	// Remote Number
	}

//  For this routine, we  put the information into the Display Boxes only for debugging
if (DEBUGGING && (StrLCS > 0)) {
    SetTextValue(MainForm->H_EngNo, IntToHex(I_Consist, 2));
    // command letter & first 4 data chars
    SetTextValue(MainForm->H_Command, IntToHex(Command_String[0], 2));
    if (StrLCS > 1) TextValue = IntToHex(Command_String[1], 2); else TextValue = "-";
    SetTextValue(MainForm->H_Data1, TextValue);
    if (StrLCS > 2) TextValue = IntToHex(Command_String[2], 2); else TextValue = "-";
    SetTextValue(MainForm->H_Data2, TextValue);
    if (StrLCS > 3) TextValue = IntToHex(Command_String[3], 2); else TextValue = "-";
    SetTextValue(MainForm->H_Data3, TextValue);
    if (StrLCS > 4) TextValue = IntToHex(Command_String[4], 2); else TextValue = "-";
    SetTextValue(MainForm->H_Data4, TextValue);
    SetTextValue(MainForm->H_Seq_Count, IntToHex(Sequence, 2));
    SetTextValue(MainForm->H_TIU_Remote, IntToHex(I_TIU_Remote, 2));
	}

if (DEBUGGINGX(2)) {
	PrintLine = "Sending String: (#" + String(Sequence) + ") "+ OSS;
    PrintItem(PrintLine);
    }

//
// Generate the CRC from Cmd_S1 and insert it into the correct spot in Cmd_S1
(void) CRCGen(&Cmd_S1, 0, Cmd_S1.idx - 1, true);	// from Cmd_S1 - generate CRC's - quietly
Cmd_S1.idx += 2;	// add two CRC bytes to the length for the CRC

// NOTE: Cmd_S1 may contain binary zero so don't take StrLen() of it

// encode command from Cmd_String into Encoded_CommandA
if (MainForm->bRecording && (Record_Flag == RECORD_OK)) {	// bRecording is a bool
    // TextValue is used in the "if" statement a few lines down and must be preserved until then
    TextValue = sEncodeAString(Cmd_S1);
    }

byteEncodeAString(&Cmd_S1, &Cmd_S2);	// result of encode of Cmd_S1 is put into Cmd_S2

// recording : time(NULL) = The number of seconds since January 1, 1970
if (MainForm->bRecording && (Record_Flag == RECORD_OK)) {	// bRecording is a bool
	if (MainForm->EncodedRecordFormat) {
		// Use Plain Text command format as a comment because it is easier to read
    	// 1439353496 5555595566A965696565666A656595A55A5A55A59A664D // 7 v4070 1 15
	    // 1439353517 5555669966A66565695956595955AAAA99A65A5A699556AA4D // 101 s0 (5,7) 1 15
        SetRecordingTextValue(String(time(NULL)) + " " + TextValue + " // " + OSS);
	    }
	if (!MainForm->EncodedRecordFormat) {		// DEFAULT format - Plain Text w/encoded comment
		// Use Encoded format as a comment because we can  - Plain Text is the default format
	    // 1439352676 7 v4070 1 15  // 5555595566A965696565666A656595A65A5A955A55594D
        // 1439352704 101 s0 (5,7) 1 15  // 5555669966A66565695956595955AAAA9AA55A5AA96599AA4D
        SetRecordingTextValue(String(time(NULL)) + " " + OSS + " // " + TextValue);
        }
    }

// TIU == 0 means don't send or receive anything. This is not the normal case and
// many parts of RTC will not function with a TIU number of 0.

if (TIU != 0) {
	// if running in Speed_mode, clear that mode temporarily for commands which require a response
	bool save_Speed_Mode = Speed_Mode;
	//
	if (Record_Flag == NO_RECORD) { 	// commands which are set to NO_RECORD cannot run in Speed_Mode
		Speed_Mode = false;	// turn off Speed_Mode
	    }
	// Send the encoded command
	// In All Engine Operation or if this is a LashUp, give the TIU three times as much time // AllEngineMode is a bool
	// Set bFlag with success(true)/failure(false), LocalRespCode with error code, SSresponseCount and SSResponseBuffer with response to Cmd_S2
    //
	bFlag = TIUSendWaitforResponseRetry(Consist, Cmd_S2.Str, Expected, ((MainForm->AllEngineMode || (Consist == LASHUPMIN)) ? (WaitTime * 3) : WaitTime),
    	Remote, TIU, &LocalRespCode, SSResponseCount, SSResponseBuffer);

    if (DEBUGGINGX(6)) {
	    PrintLine = "Sent String: (#" + String(Sequence) + ") "+ OSS + " Response Code = " + IntToHex(LocalRespCode,2);
    	if ((Expected == MINRESPONSE && LocalRespCode == 0x00) || LocalRespCode >= UNDETERMINED)
        	PrintItem(PrintLine, clRed);
            else
            PrintItem(PrintLine, clGreen);
    	}

    //
    /*   TODO Never retry outside of TIUSendWaitforResponseRetry() so this code is not needed anymore
    //
    // look at LocalRespCode and retry the send if necessary    
    //
    if (NumRetries > 0) { 	// only retry if NumRetries is > 0
        for (int RetryCount = 0 ; RetryCount < NumRetries ; RetryCount++) {
            // Check for certain error codes and if found, bump the sequence number and send the command again
            // This extra retry is only performed on RespCode of 0x00 when 11 bytes are expected or if a comm error occured
            // RespCode of 0x00 is normal when more than 11 bytes are expected (15, 39, 173, etc).
            if (((LocalRespCode & 0xF0) == 0x00 && Expected == 11) || (LocalRespCode >= UNDETERMINED)) {
                // Send again
                // Bump the sequence number on each command sent
                Sequence++;		// Sequence is an unsigned char

                if (DEBUGGINGX(2)) {
                	PrintLine = "Send String: Retry (#" + String(Sequence) + ") "+ OSS;
                    PrintItem(PrintLine);
                    }
                // Regenerate the CRC from Cmd_S1 and insert it into the correct spot in Cmd_S1
                Cmd_S1.idx -= 2;	// subtract out the old CRC bytes
                // NULL terminate
                Cmd_S1.Str[Cmd_S1.idx]  = NULL;
                // put in a new sequence number
                Cmd_S1.Str[Cmd_S1.idx - 2 ] = Sequence;	// Sequence is an unsigned char
                //  For this routine, we  put the information into the Display Boxes only for debugging
                if (DEBUGGING && (StrLCS > 0)) {
                    SetTextValue(MainForm->H_Seq_Count, IntToHex(Sequence, 2));
                    }

                (void) CRCGen(&Cmd_S1, 0, Cmd_S1.idx - 1, true);	// from Cmd_String - generate CRC's - quietly
                Cmd_S1.idx += 2;	// add two CRC bytes to the length for the CRC
                //
                byteEncodeAString(&Cmd_S1, &Cmd_S2);	// result of encode of Cmd_S1 is put into Cmd_S2
                // Send it
                bFlag = TCommThread::TIUSendWaitforResponseRetry(Consist, Cmd_S2.Str, Expected, ((MainForm->AllEngineMode || (Consist == LASHUPMIN)) ? (WaitTime * 3) : WaitTime),
                    Remote, TIU, &LocalRespCode, SSResponseCount, SSResponseBuffer);
                if (DEBUGGINGX(4)) {
                    PrintLine = "Sent String: Retry " + String(RetryCount + 1) + " (#" + String(Sequence) + ") "+ OSS + " 2nd Response Code = " + IntToHex(*ptrRespCode,2);
                    if ((Expected == MINRESPONSE && LocalRespCode == 0x00) || LocalRespCode >= UNDETERMINED)
                        PrintItem(PrintLine, clRed);
                        else
                        PrintItem(PrintLine, clGreen);
                    }
                // "for" loop to check the response code and try again if necessary
                } else {
                // retry was successful
                bFlag = true;
                break;
                }
            // TBD this message displays while retrying is going on
			// TBD if (RetryCount == 1) new TMessageDisplayThread(false, MyCaller, ParentHandle, String("Error - TIU ") + String(TIU), "Engine " + String(Consist) + " not on Track (NAK==0x00)", "", 2000);
            }
        } // if (NumRetries > 0)
        */
	Speed_Mode = save_Speed_Mode;	// restore Speed_Mode
    } else {
    // for TIU == 0 always return success, many parts of RTC will not work correctly with TIU == 0
    LocalRespCode = 0x00;
    bFlag = true;
    }	// if (TIU != 0) 

// Bump the sequence number on each command send
Sequence++;		// Sequence is an unsigned char
// *ptrRespCode is set from LocalRespCode just before the function exits
if (ptrRespCode) *ptrRespCode = LocalRespCode;
return bFlag;		// with flag indicating if the call to TIUSendWaitforResponseRetry() succeeded
}	// TCommThread::Send_String
//---------------------------------------------------------------------------

bool __fastcall TCommThread::TIUSendWaitforResponseRetry(unsigned char Consist, byte * Command,
	int ExpectedCount, int WaitTime, unsigned char Remote, unsigned char TIU, int * ptrRespCode,
    int * SWResponseCount, byte * SWResponseBuffer)
{
//
// Calls to Sleep() are ok in this thread
// NO VCL Calls in this thread
//
// This routine does not record, only Send_String() records
//
// Send Command string out TIU Port or the AIU Port and wait for a response of ExpectedCount bytes
// in WaitTime ms - NOTE: the ExpectedCount should include the 3 bytes of the "55554D" plus
// the number of bytes returned by the actual command. The TIU always responds with
// OKRESPONSE bytes when it receives the command, then more bytes up to MINRESPONSE bytes
// or more for some commands (ie 'x', '!', 'I', 'q', etc).
//
// Command is in binary and may contain 0x00.
//
// WaitTime in milliseconds and must be less than 20,000 (20 seconds).
//
// the encoded string received is decoded and placed into byteResp[]. Does not include the "55554D"
//
// This routine does NOT record. If needed, record must be performed before calling this routine.
//
// Return true on success, false on failure
//
// Examine value returned in *ptrRespCode for more details (see defines in Train_Control.h)
// ------
//
String PrintLine;
byte byteResp[CMDMAX]; 				// Array for decoded binary response
int LocalRespCode = UNDETERMINED;	// must have a response code even if the caller doesn't care so create a local spot

if (DEBUGGINGX(5)) {
	PrintLine =     "TIU: " + byteHextoString(Command) + " (Wait " + String(WaitTime) + "ms for " + String(ExpectedCount) + " bytes)";
    PrintItem(PrintLine, clBlack);
    // print decoded command
    if (DEBUGGINGX(7)) {
   		PrintLine = "                      : " + byteDecodeBinaryString(Command, strlen(Command) - 1, false);
    	PrintItem(PrintLine);
    	}
    }

// increment packets sent counter
if (DEBUGGING) Bump(MainForm->PacketsSent);
//
// Retries are always done ONLY in this function TIUSendWaitforResponseRetry()
// 0 = no retries (original send packet only)
// 1 = 1 retry
// etc.
//
// loop for NumRetries set on Setup Form (0-9) : 10 is "ignore serial ports" mode
//
for (int RetryCount = 0 ; RetryCount <= NumRetries ; RetryCount++) {	// always executes the loop once
	//
    SetTextValue(MainForm->Last_Response, "none");	// VCL Reference
    //
    if (DEBUGGING) SetTextValue(MainForm->TIU_Char_Count, "0");
	// send the command, SendHexStr() does not record
    // Command may contain 0x00 so use strcspn() to find the end of the command
    int Len = strcspn(Command, "M") + 1;	// locate the 0x4D at the end of the command
    //---------------------------------------------------------
    RadioDataSent = false;
    AIUDataSent = false;
    memset((byte *) cTIUbuf, NULL, CMDMAX);	// clear the receive buffer
    iTIUbufCount = 0;						// clear the receive byte counter
    bool SendOK = SendHexStr(Command, Len, ExpectedCount, WaitTime, TIU);
    //
    while(!RadioDataSent && !AIUDataSent) {
    	Sleep(10);
        Application->ProcessMessages();
        }	// wait until the command is sent out the COM port or the WiFi port
    //---------------------------------------------------------
    //
    // no data response expected or don't wait at all for response
    //
	if (ExpectedCount == 0 || WaitTime == 0) {
        LocalRespCode = 0x10 | TIU;						// no error (see note above, this byte is 0x1T and could be 0x10 to 0x17)
    	if (DEBUGGINGX(8)) {
        	PrintLine = "TIU: Not waiting for a response from TIU";
            PrintItem(PrintLine, clGreen);
            }
		// Copy local response to global response
        if (SWResponseCount) *SWResponseCount = 0;
        if (SWResponseBuffer) SWResponseBuffer[0] = '\0';
        if (ptrRespCode) *ptrRespCode = LocalRespCode;
		return true;	// a Debugging mode is active - Success
        }
 
    // Debugging - don't wait for response:
    // 1. in Speed_Mode
    // 2. in ignore serial ports mode
    // 3. if Emergency Stop was pressed
    // 4. in TX Only Mode
    //
    // in all Debugging cases, fake a response so we might get fewer error messages.
	if (Speed_Mode || (NumRetries == 10) || bEmergencyStop || TXOnlyComm) {	// bEmergency is a bool
        // NumRetries == 10 is a debug mode which disables a lot of error checking
        iTIUbufCount = 173;
        //
    	if (DEBUGGING) SetTextValue(MainForm->TIU_Char_Count, "173");
        // set bTIUbuf byteResp to some default
        memset((byte *) cTIUbuf, '5', CMDMAX);
		memset(byteResp, 0x55, sizeof(byteResp));
		if (!TXOnlyComm)  {
            // set TIU box color to yellow indicating test mode
            SetColor(MainForm->TIUSpin, clYellow);
            } else {
            // otherwise set to green indicating good communications
            SetColor(MainForm->TIUSpin, clLime);
            }
        if (DEBUGGINGX(3)) {
        	PrintLine = "TIU: Speed or debug mode, Not waiting for response from TIU";
            PrintItem(PrintLine, clGreen);
            }
        LocalRespCode = 0x10 | TIU;			// no error (see note above, this byte is 0x1T and could be 0x10 to 0x17)
		// Copy local response to global response
        if (SWResponseCount) *SWResponseCount = iTIUbufCount;
        if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
		if (ptrRespCode) *ptrRespCode = LocalRespCode;
    	return true;	// a Debugging mode is active - Success
        }
    //
    if (!SendOK) {	// sendData failed with busy indication (the only error possible to make it to here)
    	Sleep(100);	// wait 100 ms
    	if (DEBUGGING) {
        	Bump(MainForm->sendDataBusy);	//  increment busy counter
			Bump(MainForm->Retries);
            PrintLine = "TIU: Busy Retry";
            PrintItem(PrintLine, clRed);
            }
        continue; // the "for (int RetryCount = 0 ; RetryCount <= NumRetries ; RetryCount++)" loop
        }

    // -----------------------------------------------------------------------------------------
    //
	// wait for <WaitTime> for <ExpectedCount> bytes (eg: "55554D" + "A595696655555565")
	// if we don't get them within <WaitTime>, retry <NumRetries> times
    //
    int step;
	for (step = WaitTime/DELAYSTEP; step > 0 ; step--) {
        //
        // Here we wait for the serial port thread to send the packet and receive the response
        // Debug PrintItem("Delay " + String(DELAYSTEP) + " - " + String(step), clBlack);
    	Sleep(DELAYSTEP);	// delay for DELAYSTEP ms
        Application->ProcessMessages();
        // see if we got the expected number of bytes from the TIU
		if (iTIUbufCount == ExpectedCount) {   			// Number of characters returned depends on command - given as "ExpectedCount"
    		// Got correct number of bytes in response 	// got "55554D" followed by rest of response
            //
            if (bRadioWiFiConnected && DEBUGGING) {
                String TextValue = "";
                // convert byte string to a character String type for display in Buffer
                for (int i = 0 ; i < iTIUbufCount ; i++) {
                    TextValue += IntToHex(cTIUbuf[i] & 0xFF,2);
                    }
                // display it in the TIU buffer (in Red)
                ShowinTIUBuffer(TextValue, clRed);
                }
            //
            // use cTIUbuf loaded by SerialEventManager
            LocalRespCode = RXLookUp(cTIUbuf[9] << 8 | cTIUbuf[10]); // set from 10th and 11th bytes of response
            //
        	if (DEBUGGING) {
	        	// VCL Reference
    			SetTextValue(MainForm->Last_Response, byteHextoString((byte *) &cTIUbuf[3]));
                }
            if (DEBUGGINGX(4)) {
			    PrintLine = "TIU: Response after " + String(((WaitTime/DELAYSTEP) - step) * DELAYSTEP) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
    	        PrintItem(PrintLine, clGreen);
            	}
            if (DEBUGGINGX(7)) {
                if (iTIUbufCount > 3) {
	    			SetTextValue(MainForm->Last_Response_Decoded, byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false));
                    }
                if (iTIUbufCount > 3) {
				    PrintLine = "     Last Response  : " + byteHextoString((byte *) &cTIUbuf[3]) + " (0x" + IntToHex(LocalRespCode, 2) + ")";
    		        PrintItem(PrintLine);
				    PrintLine = "                    : " + byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false);
            		PrintItem(PrintLine);
                    }
	            }
            //
            // -----------------------------------------------------------------
            //
			// go through each 4 characters in bTIUbuf and build byteResp in binary
            //
            // response is encoded binary
			// decode it. NOTE: byteResp may contain 0x00 so you can't do a strlen() call on it
            memset(byteResp, NULL, sizeof(byteResp));
 		 	// convert 4 encoded ASCII characters into a binary value for the response starting at the 3rd byte
            int iLen = (iTIUbufCount - 3) / 2;
			for (int i = 0 ; i < iLen; i++) {
				// Note: "55554D" part of response is not converted
		    	byteResp[i] = RXLookUp(cTIUbuf[(i * 2) + 3] << 8 | cTIUbuf[(i * 2) + 4]);
				}
            // -----------------------------------------------------------------
            //
            // check the remote number in the response and make sure the response packet is for us.
            //
            if ((byteResp[2] & 0x0F) != Remote) {
                // packet was not for us, so just ignore it
				if (DEBUGGING) {
	                // Bump(MainForm->ForeignPacket);
				    SetTextValue(MainForm->TIU_Char_Count, "0");
                	PrintLine = "TIU: Foreign Packet - for Remote " + String(byteResp[2] & 0x0F);
                    PrintItem(PrintLine, clRed);
                    }
                //
                //clear packet area
                LocalRespCode = UNDETERMINED;	// undetermined response
    			memset((byte *) cTIUbuf, NULL, CMDMAX);
                iTIUbufCount = 0;
                // Response packet was for some other Remote
                continue; // "for (step = WaitTime/DELAYSTEP; step > 0 ; step--)" loop and wait for another response
            	}
            // -----------------------------------------------------------------
			/*
			 * Check CRC
			 *
			 * A9A995955666555555595555555556555955
			 *   FC  C0  13  00  04  00  00  02  08
			 * reveng -c -m CRC-16/MCRF4XX 13000400000208c0fc
			 *			returns 0000
			 */
			// evaulate the received CRC - it should be 0x0000 (evaluates case where iTIUbufCount == ExpectedCount)
            unsigned int crc_result = CRCEvaluation((Cmd *) &byteResp[0], 0, iLen - 1, true);	// from bTIUBuf - Check CRC as received from TIU - Quietly
			if (crc_result != 0x0000) {
				// check the CRC and increment the error count
               	if (DEBUGGING) {
                	Bump(MainForm->NAKCRC);		// increment NAKCRC (NAK4) counter
					Bump(MainForm->Retries);    // increment retries counter
    			    PrintLine = "TIU: (0x84) Command Retrying - CRC Error";	// Note: post CRC popup message only if CRC error persists
        	        PrintItem(PrintLine, clRed);
					(void) new TMessageDisplayThread(false, MyCaller, ParentHandle, String("Error - TIU ") + String(TIU), " CRC Error", "", 2000);
                    }
                LocalRespCode = CRCERROR;	// CRC Error
                break; // out of "for (step = WaitTime/DELAYSTEP; step > 0 ; step--)" loop and retry
				}
            // -----------------------------------------------------------------
            if ((ExpectedCount == MINRESPONSE) && (iTIUbufCount == MINRESPONSE)) {
	            //
	            // If we are waiting for 11 bytes, and we got 11 bytes, check the response code
                //
    	        // A 0x0T is a error response,  0x1T is some kind of good response.
                // "T" is the TIU number 0-4
                // DCS v5.0 note: The TIU now sends 0x1T as the response byte. Seems to indicate success. T is the TIU number
                // DCS v5.0 note: The TIU sends the TIU number in the response only for commands that return a value.
                //		Simple ACK or NAK responses don't include the TIU number, they just return the remote number, TIU nybble set to zero
				//
				//  A995699555565566
				//    E8  68  01  1T
                //            ^             TIU nybble is 0
                //             ^			Remote nybble here is 1
                //                ^^		Status byte is 0x1T indicating ACK where T is the TIU Number 0-4
				//  66A665A55A5A5555
				//    73  70  0F  0T
                //            ^             TIU nybble is 0
                //             ^			Remote nybble here is 15
                //                ^^		Status byte is 0x0T indicating NAK where T is the TIU Number 0-4
				//
         	    if ((LocalRespCode & 0xF0) == 0x00) {	// did we get a 0x0T for the status byte?
            		// yes, something went wrong - but I don't know what
                    // NOTE: 0x0T is a typical good response for a "non Engine specific" command - x, !, I0, Ix, Dq, etc.
                    //			Handled elsewhere.
    				// increment NAK0T counter
                    // Translate to an error code
                    LocalRespCode = NAKRESPONSE;
                    if (DEBUGGING) {
	                   	Bump(MainForm->NAK0T);	// Count NAK0T errors and display
    	               	// I believe that a response code of 0x00 for an 11 expected count means engine was not found
						Bump(MainForm->EngNotFound);
			        	PrintLine = "TIU: (0x0T) NAK after " + String(((WaitTime/DELAYSTEP) - step) * DELAYSTEP) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
			        	PrintItem(PrintLine, clRed);
                    	}
            		// Message and then retry - no "not on track" message until the retries fail
					//if (NumRetries >= 1) (void) new TMessageDisplayThread(false, MyCaller, ParentHandle, "Warning", "Engine " + String(Consist) + " not on Track (NAK==0x00)");

        			if (DEBUGGINGX(7)) {
            			// SetTextValue is a VCL Reference using messages to the main thread
   	        			if (iTIUbufCount > 3) {
						    SetTextValue(MainForm->Last_Response_Decoded, byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false));
							}
				        if (iTIUbufCount > 3) {
					    	PrintLine = "     Last Response  : " + byteHextoString((byte *) &cTIUbuf[3]) + " (0x" + IntToHex(LocalRespCode, 2) + ")";
 			       			PrintItem(PrintLine);
				    		PrintLine = "                    : " + byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false);
				        	PrintItem(PrintLine);
 			               }
                       }
			        // -----------------------------------------------------------------
			        // Packet analysis complete
			        // set TIU box color to green indicating good comm with the TIU (but this is a NAK)
        		    SetColor(MainForm->TIUSpin, clLime);
                    //
                    if (SWResponseCount) *SWResponseCount = iTIUbufCount;
                    if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
                    #ifdef CASE1
                    // TODO Failed with a 0x0T NAK - just retry on a 0x0T NAK
                    break; // out of "for (step = WaitTime/DELAYSTEP; step > 0 ; step--)"  loop and retry
                    #endif
                    #ifdef CASE2
					// TODO consider setting response and exiting with false - that is, don't retry on a 0x0T NAK
		            if (ptrRespCode) *ptrRespCode = LocalRespCode;
					return false;    // Failed with a 0x0T NAK
                    #endif
                    #ifndef CASE3
					// TODO consider setting response and exiting with true - that is, don't retry on a 0x0T NAK and pretend it worked
		            if (ptrRespCode) *ptrRespCode = 0x00;
					return true;    // pretend command worked with a 0x0T NAK response
                    #endif

   	                } else {
                    //
                    // status byte is 0x1T which is an ACK (good) response
                    // 
                    }
                }   // if ((ExpectedCount == MINRESPONSE) && (iTIUbufCount == MINRESPONSE))
            // -----------------------------------------------------------------
            //
            // Packet analysis complete - Got Expected Count, good response code and good CRC
            //
            // Set TIU box color to green indicating good comm with the TIU
            SetColor(MainForm->TIUSpin, clLime);
            // This is the normal (good response) exit
			// Copy local response to global response
    	    if (SWResponseCount) *SWResponseCount = iTIUbufCount;
        	if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
            if (DEBUGGINGX(2)) {
    			if ((ExpectedCount == MINRESPONSE && (LocalRespCode & 0xF0) == 0x00) || LocalRespCode >= UNDETERMINED)
        			PrintItem("TIU: ACK   Response Code = 0x" + IntToHex(LocalRespCode,2), clRed);
            		else
            		PrintItem("TIU: ACK   Response Code = 0x" + IntToHex(LocalRespCode,2), clGreen);
                }

            if (ptrRespCode) *ptrRespCode = LocalRespCode;
            // This is the primary success exit from this function
			return true;    // Success - Got ExpectedCount bytes, LocalRespCode != 0x00 and the CRC was good
            //
        	}  // if (TIUbufCount == ExpectedCount)

        // "break" statements come out here
        // Go back and keep on waiting for complete response (until WaitTime) since
        // we did not get the ExpectedCount number of bytes yet
		}	// for (int step = WaitTime/DELAYSTEP; step > 0 ; step--)
    //
    // ----------------------- Response Timed Out ------------------------------------------
    //
	// At this point, we have waited for one "WaitTime" in steps of WaitTime/DELAYSTEP
	//
	// If we expected more than 11 characters but we got exactly 11 (MINRESPONSE) characters
	// that probably means we got a NAK for a "non Engine" command - x, !, I0, Ix, etc.
	//
    // A "0x00" is a error response, 0x1T is some kind of good response
    //
    // DCS v5.0 note: The TIU now sends 0x11 as the response byte. Seems to indicate success.
    // DCS v5.0 note: The TIU sends the TIU number in the response only for commands that return a value.
    //		Simple ACK or NAK responses don't include the TIU number, they just return the remote number, TIU nybble set to zero
	//
	//  A995699555565566
	//    E8  68  01  1T
    //            ^             TIU nybble is always == 0
    //             ^			Remote nybble here is 1
    //                ^^		Status byte is 0x1T indicating ACK
	//  66A665A55A5A5555
	//    73  70  0F  0T
    //            ^             TIU nybble is always == 0
    //             ^			Remote nybble here is 15
    //                ^^		Status byte is 0x0T indicating NAK
	//
    // Display the packet
	//
    if (bRadioWiFiConnected && DEBUGGING) {
        String TextValue = "";
        // convert byte string to a character String type for display in Buffer
        for (int i = 0 ; i < iTIUbufCount ; i++) {
            TextValue += IntToHex(cTIUbuf[i] & 0xFF,2);
            }
        // display it in the TIU buffer (in Red)
        ShowinTIUBuffer(TextValue, clRed);
        }
	//
	if ((ExpectedCount > MINRESPONSE) && (iTIUbufCount == MINRESPONSE)) {
	    //
   		// Got a short but possibly valid response 	// got "55554D" followed by rest of response, probably a NAK
        //
        // I think that when this occurs, the response is ALWAYS a NAK
        //
        if (DEBUGGING && (iTIUbufCount > 3)) {		// VCL Response
		    SetTextValue(MainForm->Last_Response, byteHextoString((byte *) &cTIUbuf[3]));
            }
        //
        // -----------------------------------------------------------------
		// go through each 4 characters in bTIUbuf and build byteResp in binary
        //
	 	// convert 4 encoded ASCII characters into a binary value for the entire response
        memset(byteResp, NULL, sizeof(byteResp));
		// convert 4 encoded ASCII characters into a binary value for the response starting at the 3rd byte
        // NOTE: byteResp may contain 0x00 so you can't do a strlen() call on it
        int iLen = (iTIUbufCount - 3) / 2;
		for (int i = 0 ; i < iLen; i++) {
			// Note: "5554D" part of response is not converted
	    	byteResp[i] = RXLookUp(cTIUbuf[(i * 2) + 3] << 8 | cTIUbuf[(i * 2) + 4]);
			}

        // -----------------------------------------------------------------
		/*
		 * Check CRC
		 *
		 * A9A995955666555555595555555556555955
		 *   FC  C0  13  00  04  00  00  02  08
		 * reveng -c -m CRC-16/MCRF4XX 13000400000208c0fc
		 *			returns 0000
		 */
		// evaulate the received CRC - it should be 0x0000 (evaluates when a timeout occured)
        // This determines if we have, at lease, received a complete response from the TIU
		unsigned int crc_result =  CRCEvaluation((Cmd *)&byteResp[0], 0, iLen - 1, true);	// from bTIUbuf - Check CRC as received from TIU - Quietly
		if (crc_result != 0x0000) {
			// check the CRC and increment the error count
			if (DEBUGGING) {
				Bump(MainForm->NAKCRC);	// increment NAKCRC counter
				Bump(MainForm->Retries);// increment retries counter
			    PrintLine = "TIU: (0x84) Command Retrying - CRC Error";	// Note: post CRC popup message only if CRC error persists
    	        PrintItem(PrintLine, clRed);
				new TMessageDisplayThread(false, MyCaller, ParentHandle, String("Error - TIU ") + String(TIU), " CRC Error", "", 2000);
         		}
            LocalRespCode = CRCERROR;	// CRC Error
            continue; // the "for (int RetryCount = 0 ; RetryCount <= NumRetries ; RetryCount++)" loop
		 	}
        // -----------------------------------------------------------------
        //
        // check the remote number in the response and make sure the response packet is for us.
        //
        if ((byteResp[2] & 0x0F) != Remote) {
            // packet was not for us, so just ignore it
			if (DEBUGGING) {
	            //Bump(MainForm->ForeignPacket);
	            SetTextValue(MainForm->TIU_Char_Count, "0");
            	PrintLine = "TIU: Foreign Packet - Remote " + String(Remote);
                PrintItem(PrintLine, clBlack);
                }
			//
            // clear packet area
            LocalRespCode = UNDETERMINED;	// undetermined response
   			memset((byte *) cTIUbuf, NULL, CMDMAX);
            RetryCount--;	// this packet does not count as a retry
            iTIUbufCount = 0;
            continue; // the "for (int RetryCount = 0 ; RetryCount <= NumRetries ; RetryCount++)" loop
           	}
		// -----------------------------------------------------------------
        //
        // use cTIUbuf loaded from ISR
        LocalRespCode = RXLookUp(cTIUbuf[9] << 8 | cTIUbuf[10]); // set from 10th and 11th bytes of response
        //
        if (DEBUGGING) {
        	PrintLine = "TIU: Short NAK after " + String(((WaitTime/DELAYSTEP) - step) * DELAYSTEP) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
        	PrintItem(PrintLine, clRed);
        	}
        if (DEBUGGINGX(7)) {
            // VCL Reference
   	        if (iTIUbufCount > 3) {
			    SetTextValue(MainForm->Last_Response_Decoded, byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false));
                }
	        if (iTIUbufCount > 3) {
		    	PrintLine = "     Last Response  : " + byteHextoString((byte *) &cTIUbuf[3]) + " (0x" + IntToHex(LocalRespCode, 2) + ")";
        		PrintItem(PrintLine);
	    		PrintLine = "                    : " + byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false);
	        	PrintItem(PrintLine);
                }
            }
        // -----------------------------------------------------------------
		//
		// A Short NAK will never recover no matter how many retries you do until the sequence number is incremented
        //
		#ifndef CASE20
        Bump(MainForm->NAK0T);	// increment NAK0T counter
        // don't increment EngNotFound or Retries counters
        break;
		#endif
		#ifdef CASE21
        //
        if ((LocalRespCode == 0x00) && (ExpectedCount > MINRESPONSE) && (iTIUbufCount == MINRESPONSE)) {
			// did we get a 0x0T for the status byte?
        	// this is case where a NAK was returned for a command which expected more than MINRESPONSE (11 bytes)
        	// but actually got MINRESPONSE (11 bytes)
        	// yes, something went wrong - but I don't know what
			if (DEBUGGING) {
    	       	Bump(MainForm->NAK0T);	// increment NAK0T counter
        		// I believe that a response code of 0x00 for an 11 expected count means engine was not found
				Bump(MainForm->EngNotFound);
				Bump(MainForm->Retries);	// increment retries counter
                }
       		// Show Message and then retry
			if (DEBUGGING && NumRetries > 0) {
            	PrintLine = "TIU: (0x00) Engine " + String(Consist) + " not on track (NAK==0x00 expected long response)";
                PrintItem(PrintLine, clRed);
                }
            continue; // the "for (int RetryCount = 0 ; RetryCount <= NumRetries ; RetryCount++)" loop
			}
		#endif
        // -----------------------------------------------------------------
        //
        #ifdef NOTNEEDED
        // Packet analysis complete - ACK response to what was probably a duplicate packet number
        // In this case, we don't want to retry because it will just fail again.
        //
        // set TIU box color to green indicating good comm with the TIU
        SetColor(MainForm->TIUSpin, clLime);
		// Copy local response to global response
        if (SWResponseCount) *SWResponseCount = iTIUbufCount;
        if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
		if (DEBUGGING) {
        	PrintLine = "TIU: (0x00) NAK (no retry)";
            PrintItem(PrintLine, clRed);
            }
        if (ptrRespCode) *ptrRespCode = LocalRespCode;
		return false;    // Got MINRESPONSE and the CRC was good - but expected more bytes, comm to TIU is good
        #endif
       	}	// 	if (ExpectedCount > MINRESPONSE && TIUbufCount == MINRESPONSE)
	//
    // Finished processing packet with expected Byte Count or with shorter (MINRESPONSE) Byte Count
    //
    // Now its an unknown problem, display error and retry
    //
    // increment retries counter
	if (DEBUGGING) Bump(MainForm->Retries);

    if (DEBUGGING && iTIUbufCount > 3) {		// get whatever response bytes are available above the "55554D"
	    // VCL Reference
	    SetTextValue(MainForm->Last_Response, byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false));
		}
	//if (NumRetries > 0) new TMessageDisplayThread(false, MyCaller, ParentHandle, "Error", "Engine " + String(Consist) + " Timeout", "", 2000);

    if (DEBUGGING && NumRetries > 0) {
    	PrintLine = "TIU: Retrying after " + String(((WaitTime/DELAYSTEP) - step) * DELAYSTEP) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
    	PrintItem(PrintLine, clRed);
    	}

    // Don't try to decode the response because it is most likely garbage
    // go back and try again until "NumRetries" retry count is reached
	}	// // the "for (int RetryCount = 0 ; RetryCount <= NumRetries ; RetryCount++)" loop

//
// If we didn't get the expected number of bytes in return (after "WaitTime" timeout), give
// an error depending on how many bytes were actually returned. If it was some other error, display
// the appropriate error message. All requested retries also failed
//
// ALL ERRORS COME THROUGH HERE
//
//// set TIU box color to red indicating bad comm with the TIU but we are still connected
//SetColor(MainForm->TIUSpin, clRed);
//
// Every command to the TIU should respond with an OK - "55554D"
if (iTIUbufCount < OKRESPONSE) {
	// we didn't get the OK from the TIU
	// set TIU box color to red indicating bad comm with the TIU but we are still connected
	SetColor(MainForm->TIUSpin, clRed);
	if (DEBUGGING) Bump(MainForm->NAKShort);	// increment NAK counter #1
    // post message "Error - TIUx not responding"
	if (NumRetries >  0) (void) new TMessageDisplayThread(false, MyCaller, ParentHandle, "", String("Error - TIU ") + String(TIU) + " not responding", "", 2000);
    //
    if (iTIUbufCount > 0) {
	    LocalRespCode = OKMISSED;	// OK response not received
	    if (DEBUGGING) {
    		PrintLine = "TIU: (0x82) Check Track - NAK - OK response not received after " + String(WaitTime) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
    		PrintItem(PrintLine, clRed);
	        if (iTIUbufCount > 3) {
		    	PrintLine = "     Last Response  : " + byteHextoString((byte *) &cTIUbuf[3]) + " (0x82)";
    			PrintItem(PrintLine);
	    		PrintLine = "                    : " + byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false);
		    	PrintItem(PrintLine);
        	    }
	    	}
        } else {
	    LocalRespCode = TIMEDOUT;	// response not received
	    if (DEBUGGING) {
    		PrintLine = "TIU: (0x85) Check Track - No Response after " + String(WaitTime) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
    		PrintItem(PrintLine, clRed);
	        if (iTIUbufCount > 3) {
		    	PrintLine = "     Last Response  : " + byteHextoString((byte *) &cTIUbuf[3]) + " (0x85)";
    			PrintItem(PrintLine);
	    		PrintLine = "                    : " + byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false);
		    	PrintItem(PrintLine);
        	    }
	    	}
        }
	// Copy local response to global response
    if (SWResponseCount) *SWResponseCount = iTIUbufCount;
    if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
	if (ptrRespCode) *ptrRespCode = LocalRespCode;
    // come through here if the engine is not on track
    return false;	// failure - OK missed or timed-out
	}
//
// see if we got more or fewer than the expected number of bytes back from the TIU
//
if (iTIUbufCount != ExpectedCount) {
	// More or Fewer than "ExpectedCount" bytes returned
	// set TIU box color to yellow indicating unknown comm with the TIU but we are still connected
	SetColor(MainForm->TIUSpin, clYellow);
    if (DEBUGGING) Bump(MainForm->NAKLen);	// increment NAK counter #2
    // post message for 1 second as long as NumRetries is > 0
	if (NumRetries >  0 && iTIUbufCount != MINRESPONSE) new TMessageDisplayThread(false, MyCaller, ParentHandle, String("Error - TIU ") + String(TIU), String("Check Track") + String(" - Engine ") + String(Consist), "", 2000);
    //
    if (DEBUGGING) {
    	PrintLine = "TIU: (0x81/0x83) Wrong # of Bytes Received - NAK Length  after " + String(WaitTime) + " milliseconds - Expected: " + String(ExpectedCount) + " Got: " + String(iTIUbufCount);
    	PrintItem(PrintLine, clRed);
        if (iTIUbufCount > 3) {
	    	PrintLine = "     Last Response  : " + byteHextoString((byte *) &cTIUbuf[3]) + " (0x81/0x83)";
    		PrintItem(PrintLine);
    		PrintLine = "                    : " + byteDecodeBinaryString((byte *) &cTIUbuf[3], iTIUbufCount - 3, false);
	    	PrintItem(PrintLine);
            }
    	}
    if (iTIUbufCount > ExpectedCount)
	    LocalRespCode = TOMANYBYTES;	// too many bytes returned
        else
		LocalRespCode = TOOFEWBYTES;	// too few bytes received
	// Copy local response to global response
    if (SWResponseCount) *SWResponseCount = iTIUbufCount;
	if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
	if (ptrRespCode) *ptrRespCode = LocalRespCode;
    return false;	// failure - either too many bytes or too few bytes
	}
// if the code gets here, some other error has occured - LocalRespCode is set with the error
// show error popup and exit
if (NumRetries > 0) new TMessageDisplayThread(false, MyCaller, ParentHandle, String("Error - TIU ") + String(TIU), String(" Unknown Error") + String(" - Engine ") + String(Consist), "", 2000);
// Copy local response to global response
if (SWResponseCount) *SWResponseCount = iTIUbufCount;
if (SWResponseBuffer) memcpy(SWResponseBuffer, byteResp, CMDMAX);
if (ptrRespCode) *ptrRespCode = LocalRespCode;
return false;	// failure - unknown error
}	// TIUSendWaitforResponseRetry
//---------------------------------------------------------------------------

bool __fastcall TCommThread::SendHexStr(byte * Cmd_In, int Len, int ExpectedByteCount, int Timeout, char cTIULocal)  // Len is actual bytes to send
{
//
// Calls to Sleep() are ok in this thread
// NO VCL Calls in this thread
//
assert(Cmd_In != NULL);
assert(Len > 0);
//
// Send to TIU Port or AIU port, don't wait for response, simple error handling
// This is the only routine that actually sends to the TIU and AIU ports
//
// Cmd_In is in binary and might contain 0x00 - use Len
//
if (!MainForm->IsRadioConnected()) return false;	// Radio must be connected either by COM or WiFi
//
String PrintLine;
String TextValue;

// No recording, if needed it must be done prior to calling this routine

if (DEBUGGING) {
	TextValue = "";
    // convert byte string to a character String type for display in Buffer
	for (int i = 0 ; i < Len ; i++) {
        TextValue += IntToHex(Cmd_In[i] & 0xFF,2);
    	}
	// display it in the TIU buffer
    ShowinTIUBuffer(TextValue);
    }

// ignore serial port mode (don't send and don't receive)
if (NumRetries == 10) {
    // NumRetries == 10 is a debug mode which disables a lot of error checking
	return true;
	}
//
// TODO - Should I do some kind of retry if busy here?
// Return true on sucess, false on failure
// Right now, busy is handled at a higher level with a retry
//
// either bRadioCOMConnected or bRadioWiFiConnected must be true to send a packet
//
/*
 Truth Table

	USB			WiFi      	AIU       	AIU			Send
    Radio       Radio       WiFi TX     Emulator    Using
    Connected   Connection  Connection  [TIU -1 ]
                                                 ||
    0			0			0 			0           X (can't occur)
    0           0           0 			1           X (can't occur)
    0           0           1 			0           X (can't occur)
    0           0           1 			1           X (can't occur)
    0           1           0 			0			WiFi Radio
    0           1           0 			1			WiFi Radio
    0           1           1 			0			WiFi Radio
    0           1           1 			1			AIU WiFi
    1           0           0 			0			USB
    1           0           0 			1			USB
    1           0           1 			0			USB
    1           0           1 			1			AIU WiFi
    1           1           0 			0           X (can't occur)
    1           1           0 			1           X (can't occur)
    1           1           1 			0           X (can't occur)
    1           1           1 			1           X (can't occur)

 */

if (bRadioCOMConnected && !(bAIUWiFiTXConnected  && (AIUEmulator[cTIULocal-1].Found == true))) {
	//
    // Send a packet to a Radio over the USB Serial port -------------------
    //
    if (DEBUGGINGX(9)) {
        PrintItem("SendHexStr: Sending over Radio TX USB Serial");
        }
	// Returns 0 on sucess, 1 on bad parameters, 2 on busy
	int retv = Serial_port->sendData(Cmd_In, Len);  	// ******* THIS IS THE ONLY CALL TO Serial_port.sendData() *************

	// error trap
	assert(retv != 1);	// bad parameters
	assert(retv != 2);	// port busy
    if (retv) {	// failure
        if (DEBUGGING) {
            PrintLine = "SendHexStr: Error: " + MainForm->ConnectedRadioLED->Caption + " sendData failed : " + String(retv);
            PrintItem(PrintLine, clRed);
            }
        // Error retv == 1 - Bad Parameters
        // Error retv == 2 - Port Busy
        return false;	// indicate failure
        }
    }
if (bRadioWiFiConnected || bAIUWiFiTXConnected) {
	//
    // Send a packet to a TIU or AIU Emulator over WiFi -------------------
    //
    if (DEBUGGINGX(9)) {
		PrintLine = "SendHexStr: WiFi Sending";
    	PrintItem(PrintLine);
        }
    // Build up the message by prepending the Expected Byte Count, Timeout (in tenths of a second) and Checksum to the command
    // The Expected Byte Count and Timeout can both be zero. The Checksum always has the 0x80 bit set.
    // I do a 1's complement ('~') to the byte count and timeout before they are inserted in the message. Then the
    // code in the Modem has to do the same 1's complement to get the original bytes. This does mean that Expected Byte Count
    // and Timeout can never have a value of 255. A 0x80 is or'ed into the Checksum to be sure that it is never zero.
    #define PACKETSTART 3
    Cmd CommandOut;
    CommandOut.Str[0] = ~ExpectedByteCount;
    CommandOut.Str[1] = ~(Timeout/100);	// convert to tenths of a second from ms
    CommandOut.Str[2] =  0x80 || (ExpectedByteCount + Timeout);
    strncpy(&CommandOut.Str[PACKETSTART], Cmd_In, Len + 1);	// plus one to copy the NULL terminator

    String LocalStr = String((char *)CommandOut.Str);
    // if an AIU Emulator is running for this TIU, send the packet using AIUSend
    // otherwise use RadioSend
    if (bAIUWiFiTXConnected && (AIUEmulator[cTIULocal-1].Found == true)) {
    	AIUSend->SendStr(LocalStr);
	    if (DEBUGGINGX(7)) PrintItem("SendHexStr: Sending over AIU TX WiFi");
    	} else {
		RadioSend->SendStr(LocalStr);
	    if (DEBUGGINGX(7)) PrintItem("SendHexStr: Sending over Radio TX WiFi");
        }
/*
    // TBD DEBUG ONLY   - Cannot Change Port if not closed
    if (bAIUWiFiTXConnected) {
        AIUSend->Port = 8891;
        CommandOut.Str[3] = '#';
        strncpy(&CommandOut.Str[PACKETSTART + 1], Cmd_In, Len + 1);	// plus one to copy the NULL terminator
        String LocalStr = String((char *)CommandOut.Str);
        AIUSend->SendStr(LocalStr);
        AIUSend->Port = 8890;
        }
*/
	}
//
if (DEBUGGINGX(8)) {
	PrintLine = "SendHexStr: Sent";
    PrintItem(PrintLine);
    }

// in Speed_Mode, ignore the response completely (Speed_Mode is turned off for commands which REQUIRE a response.)
if (Speed_Mode) {
	Sleep(300);		// originally this was too speedy for the TIU so I had to add the delay
    return true;	// always indicate success
    }
if (TXOnlyComm) {
	Sleep(500);		// originally this was too speedy for the TIU so I had to add the delay
    return true;	// always indicate success
    }
//
return true; // indicate success
}	// SendHexStr(byte *)
//---------------------------------------------------------------------------
void __fastcall TCommThread::RunOnTerminate(TObject * Sender)     // called Automatically using Synchronize()
{
// not used
}
//---------------------------------------------------------------------------
// Beginning of Interface Routines
// These routines are the way that threads other than the main thread
// communicate with the VCL
//---------------------------------------------------------------------------
void __fastcall TCommThread::PrintItem(String PrintLine, TColor PrintColor) {
UI_Struct * Message1;
// This routine does the equivalent of the next line but without calling the VCL directly
//if (DEBUGGING) DB->Debug_Print(PrintLine, PrintColor);
if (!DEBUGGING) return;	// only print the PrintLine if DEBUGGING is true
						// because the output goes to the Debug_Messages buffer
// type 1 message
Message1 = new UI_Struct;
Message1->PrintColor = PrintColor;
if (DB->TimeStampGroup->ItemIndex == 0 || (DB->TimeStampGroup->ItemIndex == 1 && PC_Tenths == 0)) {
    // https://www.drdobbs.com/windows/win32-performance-measurement-options/184416651
    // https://stackoverflow.com/questions/1695288/getting-the-current-time-in-milliseconds-from-the-system-clock-in-windows
    SYSTEMTIME st;
    GetLocalTime(&st);
    char currentTime[24];
    snprintf(currentTime, sizeof currentTime, "%02d:%02d:%02d.%03d", st.wHour, st.wMinute, st.wSecond , st.wMilliseconds);
    Message1->Line = String(currentTime) + " T " + PrintLine;
    } else {
    if (DB->TimeStampGroup->ItemIndex == 1) {
		Message1->Line = Format(String("%9.1f"), ARRAYOFCONST((PC_Tenths/10.0))) + " T " + PrintLine;
        } else {
        Message1->Line = Format(String("%7d"), ARRAYOFCONST((MainForm->MsgNum++))) + " T " + PrintLine;
        }
    }
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_PrintItem, (LPARAM) Message1);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::Bump(TAbNumEdit * Countr) {
UI_Struct * Message2;
// Countr->Value++;
// type 2 message
Message2 = new UI_Struct;
Message2->Countr = Countr;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_Bump, (LPARAM) Message2);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetColor(TAbNumSpin * Spinner, TColor ColorDefault) {
UI_Struct * Message3;
// Spinner->ColorDefault = ColorDefault;
// type 3 message
Message3 = new UI_Struct;
Message3->Spinner = Spinner;
Message3->ColorDefault = ColorDefault;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetColor, (LPARAM) Message3);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetTextValue(TEdit * EditBox, String TextValue) {
UI_Struct * Message4;
// EditBox->Text = TextValue;
// type 4 message
Message4 = new UI_Struct;
Message4->EditBox = EditBox;
Message4->Line = TextValue;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetTextValue, (LPARAM) Message4);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetTextValue(TEdit * EditBox, String &TextValue) {
UI_Struct * Message5;
// TextValue = EditBox->Text;
// type 5 message
Message5 = new UI_Struct;
Message5->EditBox = EditBox;
Message5->Line = "";

// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetTextValue, (LPARAM) Message5);	// wait for response

TextValue = Message5->Line;
delete Message5;	// delete the message after processing the return value
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetRecordingTextValue(String TextValue) {
UI_Struct * Message6;
// RecordingBuffer->SelStart = RecordingBuffer->GetTextLen();
// RecordingBuffer->SelText = TextValue;
// type 6 message
Message6 = new UI_Struct;
Message6->Line = TextValue;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetRecordingTextValue, (LPARAM) Message6);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::ShowinTIUBuffer(String TextValue, TColor Color) {
UI_Struct * Message7;
// TIUBuffer->SelAttributes->Color = Color;
// TIUBuffer->SelStart = TIUBuffer->GetTextLen();
// TIUBuffer->SelText = String(MainForm->MsgNum++) + " M " + TextValue + TEXT("\n");
// type 7 message
Message7 = new UI_Struct;
Message7->Line = TextValue;
Message7->ColorDefault = Color;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_ShowinTIUBuffer, (LPARAM) Message7);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::PopupItem(String PrintLine, int Duration) {
// duration in milliseconds (optional - defaults to 1000)
UI_Struct * Message8;
// DebugMessagePopup("Information", PrintLine, "", Duration);
// type 8 message
Message8 = new UI_Struct;
Message8->Line = PrintLine;
Message8->iValue = Duration;		// Duration in milliseconds
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_PopupItem, (LPARAM) Message8);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::ShowinBuffer(TRichEdit * Buf, String PrintLine, TColor PrintColor, bool Scroll) {
UI_Struct * Message9;
//Buf->SelAttributes->Color = PrintColor;
//Buf->SelStart = Buf->GetTextLen();
//Buf->SelText = String(PrintLine) + TEXT("\n");
// type 9 message
Message9 = new UI_Struct;
Message9->Buffer = Buf;
Message9->Line = PrintLine;
Message9->PrintColor = PrintColor;
Message9->bValue = Scroll;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_ShowinBuffer, (LPARAM) Message9);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetNumSpinValue(TAbNumSpin * Spinner, int Value) {
UI_Struct * Message10;
// Spinner->Value = Value;
// type 10 message
Message10 = new UI_Struct;
Message10->Spinner = Spinner;
Message10->iValue = Value;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetNumSpinValue, (LPARAM) Message10);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetNumEditValue(TAbNumEdit * Countr, int Value) {
UI_Struct * Message11;
// Countr->ValueAsInt = Value;
// type 11 message
Message11 = new UI_Struct;
Message11->Countr = Countr;
Message11->iValue = Value;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetNumEditValue, (LPARAM) Message11);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::Dec(TAbNumEdit * Countr) {
UI_Struct * Message12;
// Countr->Value--;
// type 12 message
Message12 = new UI_Struct;
Message12->Countr = Countr;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_Dec, (LPARAM) Message12);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetCounterValue(TAbNumEdit * Countr, int * CntrVal) {
UI_Struct * Message13;
// CntrValue = Countr->Value;
// type 13 message
Message13 = new UI_Struct;
Message13->Countr = Countr;
Message13->Line = "";

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetCntrValue, (LPARAM) Message13);	// wait for response

*CntrVal = Message13->Line.ToIntDef(0);
delete Message13;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetPCLED(TAbLED * LED, int iState, String Line) {
UI_Struct * Message14;
// LED->Visible = true
// LED->Checked = bState
// type 14 message
Message14 = new UI_Struct;
Message14->LED = LED;
Message14->iState = iState;
Message14->Line = Line;
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetPCLED, (LPARAM) Message14);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetNumEditColor(TAbNumEdit * Countr, TColor Color) {
UI_Struct * Message15;
// Countr->ValueAsInt = Value;
// type 15 message
Message15 = new UI_Struct;
Message15->Countr = Countr;
Message15->ColorDefault = Color;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetNumEditColor, (LPARAM) Message15);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetComboBoxText(TComboBox * ComboBox, AnsiString &TextValue, int Index) {
UI_Struct * Message16;
// TextValue = ComboBox->Items[Index] ->Text;
// type 16 message
Message16 = new UI_Struct;
Message16->ComboBox = ComboBox;
Message16->iValue = Index;
Message16->Line = "";

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetComboBoxText, (LPARAM) Message16);	// wait for response

TextValue = Message16->Line;
delete Message16;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::ClearCounter(TAbNumEdit * Countr) {
UI_Struct * Message17;
// Countr->Value = 0;
// type 17 message
Message17 = new UI_Struct;
Message17->Countr = Countr;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_ClearCntr, (LPARAM) Message17);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetCounter(TAbNumEdit * Countr, float value) {
UI_Struct * Message18;
// Countr->Value = value;
// type 18 message
Message18 = new UI_Struct;
Message18->Countr = Countr;
Message18->fValue = value;
//
// Send CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetCntr, (LPARAM) Message18);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::OpenAIUWindow(int TIU, int AIU) {
UI_Struct * Message19;
// type 19 message
Message19 = new UI_Struct;
Message19->iTIU = TIU;
Message19->iAIU = AIU;
//
// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the window construction before proceeding
(void) SendMessage(MainForm->Handle, CM_UI, M_OpenAIUWindow, (LPARAM) Message19);	// wait for response
// MessageX deleted in Train_Control.cpp
}

//---------------------------------------------------------------------------
void __fastcall TCommThread::GetAIUSwitchState(int iChan, int iAIU, int iTIU, bool &bState, int MyPCNo) {
UI_Struct * Message20;
// type 20 message
Message20 = new UI_Struct;
Message20->iTIU = iTIU;
Message20->iAIU = iAIU;
Message20->iChan = iChan;
Message20->PCNo = MyPCNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetAIUSwitchState, (LPARAM) Message20);	// wait for response

bState = Message20->bValue;
delete Message20;	// delete the message after processing it
}

//---------------------------------------------------------------------------
void __fastcall TCommThread::GetAIUAccState(int iChan, int iAIU, int iTIU, bool &bState, int MyPCNo) {
UI_Struct * Message21;
// type 21 message
Message21 = new UI_Struct;
Message21->iTIU = iTIU;
Message21->iAIU = iAIU;
Message21->iChan = iChan;
Message21->PCNo = MyPCNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetAIUAccState, (LPARAM) Message21);	// wait for response

bState = Message21->bValue;
delete Message21;	// delete the message after processing it
}

//---------------------------------------------------------------------------
void __fastcall TCommThread::GetAIUSwitchName(int iChan, int iAIU, int iTIU, AnsiString &TextValue, int MyPCNo) {
UI_Struct * Message22;
// type 22 message
Message22 = new UI_Struct;
Message22->iTIU = iTIU;
Message22->iAIU = iAIU;
Message22->iChan = iChan;
Message22->PCNo = MyPCNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetAIUSwitchName, (LPARAM) Message22);	// wait for response

TextValue = Message22->Line;
delete Message22;	// delete the message after processing it
}

//---------------------------------------------------------------------------
void __fastcall TCommThread::GetAIUAccName(int iChan, int iAIU, int iTIU, AnsiString &TextValue, int MyPCNo) {
UI_Struct * Message23;
// type 23 message
Message23 = new UI_Struct;
Message23->iTIU = iTIU;
Message23->iAIU = iAIU;
Message23->iChan = iChan;
Message23->PCNo = MyPCNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetAIUAccName, (LPARAM) Message23);	// wait for response

TextValue = Message23->Line;
delete Message23;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::MakeItSo(int iAIU, int iTIU, int MyPCNo) {
UI_Struct * Message24;
// type 24 message
Message24 = new UI_Struct;
Message24->iTIU = iTIU;
Message24->iAIU = iAIU;
Message24->PCNo = MyPCNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) PostMessage(MainForm->Handle, CM_UI, M_MakeItSo, (LPARAM) Message24);	// wait for response
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetVirtualAIUAccState(bool bState, int iChan, int iAIU, int iTIU, int MyPCNo) {
UI_Struct * Message25;
// type 25 message
Message25 = new UI_Struct;
Message25->iTIU = iTIU;
Message25->iAIU = iAIU;
Message25->iChan = iChan;
Message25->PCNo = MyPCNo;
Message25->bValue = bState;

// Send CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetVirtualAIUAccState, (LPARAM) Message25);	// wait for response
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetStringGridText(TStringGrid * StringGrid, AnsiString &TextValue, int &PSLevel, int Index) {
UI_Struct * Message26;
// TextValue = ComboBox->Items[Index] ->Text;
// type 26 message
Message26 = new UI_Struct;
Message26->StringGrid = StringGrid;
Message26->iValue = Index;
Message26->Line = "";

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetStringGridText, (LPARAM) Message26);	// wait for response

TextValue = Message26->Line;
PSLevel = Message26->iValue;
delete Message26;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetLabelText(TLabel * Label, String TextValue) {
UI_Struct * Message27;
// Label->Caption = TextValue;
// type 27 message
Message27 = new UI_Struct;
Message27->Label = Label;
Message27->Line = TextValue;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetLabelText, (LPARAM) Message27);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::PopupMessage(String Message, String Caption, String Status, int dly) {
UI_Struct * Message28;
// Message displayed in a popup window
// type 28 message
Message28 = new UI_Struct;
Message28->Line = Message;
Message28->Caption = Caption;
Message28->Status = Status;
Message28->iValue = dly;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_PopupMessage, (LPARAM) Message28);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetFlag(int FlagNo, int &iState) {
// each thread that calls this must protect with Global_Flag_CS critical section
UI_Struct * Message29;
// bValue = FLag->Checked
// type 29 message
Message29 = new UI_Struct;
Message29->iValue = FlagNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetFlag, (LPARAM) Message29);	// wait for response
// SendMessage used in SetFlag, GetFlag, GetFlagName and SetFlagName
iState = Message29->iState;
delete Message29;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetFlag(int FlagNo, int &iState) {
// each thread that calls this must protect with Global_Flag_CS critical section
UI_Struct * Message30;
// Label->Caption = TextValue;
// type 30 message
Message30 = new UI_Struct;
Message30->iValue = FlagNo;
Message30->iState = iState;
//
// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_SetFlag, (LPARAM) Message30);
// SendMessage used in SetFlag, GetFlag, GetFlagName and SetFlagName
iState = Message30->iState;
delete Message30;    // delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetFlagName(int FlagNo, String &FlagName) {
// each thread that calls this must protect with Global_Flag_CS critical section
UI_Struct * Message31;
// FlagName = FLag->Caption
// type 31 message
Message31 = new UI_Struct;
Message31->iValue = FlagNo;
Message31->Line = "";

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetFlagName, (LPARAM) Message31);	// wait for response
// SendMessage used in SetFlag, GetFlag, GetFlagName and SetFlagName
FlagName = Message31->Line;
delete Message31;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetFlagName(int FlagNo, String FlagName, int &iState) {
// each thread that calls this must protect with Global_Flag_CS critical section
UI_Struct * Message32;
// Label->Caption = TextValue;
// type 32 message
Message32 = new UI_Struct;
Message32->iValue = FlagNo;
Message32->Line = FlagName;
//
// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_SetFlagName, (LPARAM) Message32);
// SendMessage used in SetFlag, GetFlag, GetFlagName and SetFlagName
if (iState != NULL) iState = Message32->iState;
delete Message32;	// delete the message after processing it
}
//---------------------------------------------------------------------------
// Message 33 deleted - it was a duplicate of Message 38
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetAspect(int SignalNo, int &iAspect, int &iNumLights, int &iType) {
// each thread that calls this must protect with Signals_CS critical section
UI_Struct * Message34;
// bValue = FLag->Checked
// type 34 message
Message34 = new UI_Struct;
Message34->iSignalNo = SignalNo;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetAspect, (LPARAM) Message34);	// wait for response
// SendMessage used in SetAspect, GetAspect, GetSignalName and CreateSignal
iAspect = Message34->iAspect;		// return Aspect
iNumLights = Message34->iNumLights;	// return number of lights on the signal head
iType = Message34->iType;
delete Message34;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetAspect(int SignalNo, int &iAspect) {
// each thread that calls this must protect with Signals_CS critical section
UI_Struct * Message35;
// type 35 message
Message35 = new UI_Struct;
Message35->iSignalNo = SignalNo;
Message35->iAspect = iAspect;
//
// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_SetAspect, (LPARAM) Message35);
// SendMessage used in SetAspect, GetAspect, GetSignalName and CreateSignal
iAspect = Message35->iAspect;	// return Aspect
delete Message35;    // delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetSignalName(int SignalNo, String &SignalName, int &NumberPlate, int &iType) {
// each thread that calls this must protect with Signals_CS critical section
UI_Struct * Message36;
// SignalName = FLag->Caption
// type 36 message
Message36 = new UI_Struct;
Message36->iSignalNo = SignalNo;
Message36->Line = "";
Message36->iNumberPlate = 0;
Message36->iType = 0;

// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetSignalName, (LPARAM) Message36);	// wait for response
// SendMessage used in SetAspect, GetAspect, GetSignalName and CreateSignal
SignalName = Message36->Line;			// return signal name
NumberPlate = Message36->iNumberPlate;	// return Number Plate value
iType = Message36->iType;				// return Signal type
delete Message36;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::CreateSignal(int iSignalNo, int iDir, int iType, int iCol, int iRow, String SignalName, int &iAspect, int iNumLights, int NumberPlate) {
// each thread that calls this must protect with Signals_CS critical section
UI_Struct * Message37;;
// type 37 message
Message37 = new UI_Struct;
Message37->iSignalNo = iSignalNo;
Message37->iDir = iDir;
Message37->Line = SignalName;
Message37->iAspect = iAspect;
Message37->iNumLights = iNumLights;
Message37->iCol = iCol;
Message37->iRow= iRow;
Message37->iNumberPlate = NumberPlate;
Message37->iType = iType;
//
// Send CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_CreateSignal, (LPARAM) Message37);
// SendMessage used in SetAspectl, GetAspect, GetSignalName and CreateSignal
iAspect = Message37->iAspect;	// return Aspect
delete Message37;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::DeleteSignal(int iSignalNo) {
// Called with a signal number
// each thread that calls this must protect with Signals_CS critical section
int iDeleted;
CreateSignal(-iSignalNo, 0, 0, 0, 0, "", iDeleted, 0, 0);	// negative signal number tells CreateSignal() to delete it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::ShowLayout(void) {
UI_Struct * Message38;
// type 38 message
Message38 = new UI_Struct;
// no operands
// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_ShowLayout, (LPARAM) Message38);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::DeleteAllSignals(void) {
UI_Struct * Message39;
// type 39 message
Message39 = new UI_Struct;
// no operands
// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_DeleteAllSignals, (LPARAM) Message39);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetCabAspect(int iSignal, int Aspect, int MyPCNo) {
UI_Struct * Message40;
// type 40 message

Message40 = new UI_Struct;
Message40->PCNo = MyPCNo;
// head
Message40->iSignalNo = iSignal;
Message40->iAspect = Aspect;
// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_SetCabAspect, (LPARAM) Message40);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetCabLabels(int iSignal, String sName, int iNumberPlate, int iType, int MyPCNo) {
UI_Struct * Message41;
// type 41 message

Message41 = new UI_Struct;
Message41->PCNo = MyPCNo;
// head
Message41->iSignalNo = iSignal;
Message41->Line = sName;
Message41->iNumberPlate = iNumberPlate;
Message41->iAspect = -1;	// -1 means don't set the Aspect (use SetCabAspect())
Message41->iType = iType;
// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_SetCabLabels, (LPARAM) Message41);
// MessageX deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::CreateCabSignals(int iSignal, int iDistant, int MyPCNo) {
UI_Struct * Message42;
// type 42 message

Message42 = new UI_Struct;
Message42->PCNo = MyPCNo;
// head
Message42->iSignalNo = iSignal;
Message42->iDistant = iDistant;	// can be zero to skip creating Distant signal
// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_CreateCab, (LPARAM) Message42);
// Message42 deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::GetCabAspect(int iSignal, int &iAspect, int &iNumLights, int &iType, int MyPCNo) {	// and Type
UI_Struct * Message43;

// type 43 message
Message43 = new UI_Struct;
Message43->PCNo = MyPCNo;
// head
Message43->iValue = iSignal;
// Post CM_UI message to MainForm - Use SendMessage() so that we wait for the message processing to complete
(void) SendMessage(MainForm->Handle, CM_UI, M_GetCabAspect, (LPARAM) Message43);
iAspect = Message43->iAspect;
iNumLights = Message43->iNumLights;
iType = Message43->iType;
delete Message43;	// delete the message after processing it
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::SetLED(TAbLED * LED, TColor ColorOn, bool Checked, bool Flashing, String Line) {
UI_Struct * Message44;
//
// type 44 message
Message44 = new UI_Struct;
Message44->LED = LED;
Message44->iState = (int) ColorOn;
Message44->iValue = Checked;
Message44->iType = Flashing;
Message44->Line = Line;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_SetLED, (LPARAM) Message44);
// Message44 deleted in Train_Control.cpp
}
//---------------------------------------------------------------------------
void __fastcall TCommThread::AddCntr(TAbNumEdit * Countr, int Val) {
UI_Struct * Message45;
// Countr->Value += Val;
// type 45 message
Message45 = new UI_Struct;
Message45->Countr = Countr;
Message45->iValue = Val;
//
// Post CM_UI message to MainForm
(void) PostMessage(MainForm->Handle, CM_UI, M_AddCntr, (LPARAM) Message45);
// MessageX deleted in Train_Control.cpp
}

//---------------------------------------------------------------------------
// End of Interface Routines
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------

unsigned int __fastcall TCommThread::CRCGen(Cmd * Command, int StartChar, int EndChar, bool Quiet)
{
//
// This function generates CRC for packets to sent to the TIU for CRC-16/MCRF4XX
//
// NOTE: Cmd_String may contain binary zero
//
cm_t var;
p_cm_t p_cm = &var;
p_cm->cm_width = 16;
p_cm->cm_poly  = 0x1021L;
p_cm->cm_init  = 0x0000FFFFL;
p_cm->cm_refin = TRUE;
p_cm->cm_refot = TRUE;
p_cm->cm_xorot = 0L;

cm_ini(p_cm);

// feed in the data bytes in reverse order
for (int ii = EndChar ; ii >= StartChar ; ii--) {
	cm_nxt(p_cm, Command->Str[ii]);
    }
//
ulong const crc_result = cm_crc(p_cm) & 0xFFFF;

// Insert the CRC byte in the command string - LSB then MSB
Command->Str[EndChar+1] = (crc_result & 0x00FF);
Command->Str[EndChar+2] = (crc_result & 0xFF00) >> 8;
// Insert the NULL
Command->Str[EndChar+3] = NULL;	    // NULL Terminated

// then put it in Command Decode area in preparation for sending
if (DEBUGGING) {
    SetTextValue(MainForm->H_CRC_LSB, IntToHex(Command->Str[EndChar+1], 2));
    SetTextValue(MainForm->H_CRC_MSB, IntToHex(Command->Str[EndChar+2], 2));
    }


return crc_result;
}
//---------------------------------------------------------------------------

unsigned int __fastcall TCommThread::CRCEvaluation(Cmd * Cmd_String, int StartChar, int EndChar, bool Quiet)
{
//
// This function checks packets received from the TIU for CRC for CRC-16/MCRF4XX
//
// NOTE: Cmd_String may contain binary zero
//
//   A9A995955666555555595555555556555955
//     FC  C0  13  00  04  00  00  02  08
// reveng -c -m CRC-16/MCRF4XX 13000400000208c0fc
//
cm_t var;
p_cm_t p_cm = &var;
p_cm->cm_width = 16;
p_cm->cm_poly  = 0x1021L;
p_cm->cm_init  = 0x0000FFFFL;
p_cm->cm_refin = TRUE;
p_cm->cm_refot = TRUE;
p_cm->cm_xorot = 0L;

cm_ini(p_cm);

// feed in the data bytes in forward order    *******
for (int ii = StartChar + 2 ; ii <= EndChar ; ii++) {
	cm_nxt(p_cm, Cmd_String->Str[ii]);
    }
// feed in the CRC bytes - LSB then MSB     ******
cm_nxt(p_cm, Cmd_String->Str[StartChar+1]);
cm_nxt(p_cm, Cmd_String->Str[StartChar]);
//
ulong const crc_result = cm_crc(p_cm) & 0xFFFF;

//  ******   crc_result must be zero

return crc_result;
}
//---------------------------------------------------------------------------

void __fastcall TCommThread::charstringtobinary(char * Command, char * Packet, int Len)
{
//
// Input is Command, a (char *) to an array of hexidecimal digits in ASCII. Null Terminated. Must
//	always be an even number of digits and be equal to (Len * 2).
//	eg: "55555A566A99656966669996696A5A55695A4D"
// Input is Len, an integer count of the number of bytes in the Packet array.
//	eg:	Len = strlen(Command)/2;
// Output is Packet, a (char *) to an array of the binary representation of Command, converting/
//	two hexidecimal digits into one 8 bit byte. It will contain Len bytes. It may contain
//	a 0x00 byte.
//	eg: "UUZVjeiffijZuiZM" where the 'M' is the 0x4D End of Packet
//
assert(Command);
assert(Packet);
assert(Len < 254);

static const char* const lut = "0123456789ABCDEF";
// Convert Hex character string to Hex
for (int i = 0; i < (Len * 2); i += 2) {
    char a = Command[i];
   	const char* p = std::lower_bound(lut, lut + 16, a);
    if (*p != a) {
        return;
        }
    char b = Command[i + 1];
    const char* q = std::lower_bound(lut, lut + 16, b);
    if (*q != b) {
        return;
        }
	Packet[i/2] = (((p - lut) << 4) | (q - lut));
	}
}
//---------------------------------------------------------------------------
// Set/Reset RTS on TIU Port
void __fastcall TCommThread::RTSSwitch(bool Checked)
{
if (bRadioCOMConnected)
	Serial_port->RTS(Checked);
}
//---------------------------------------------------------------------------
// Set/Reset DTR on TIU Port
void __fastcall TCommThread::DTRSwitch(bool Checked)
{
if (bRadioCOMConnected)
	Serial_port->DTR(Checked);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Beginning of FIFO Routines
// These routines run in the context of the calling thread, not TCommThread
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Send_String_FIFO()
// Enqueue the Command_String and associated data in the FIFO
// Wait for CommDoneEvent
// Runs in the context of the calling thread, not TCommThread
// Called by almost all UI windows, PC_Thread.cpp, DL_Thread.cpp and Engine_Status_Thread.cpp
// Typical usage:
/* from TProg_Control_Thread::FetchByte()
bool bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, CommandString, TIUNo, MyRemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
if (!bretv) {	// retry once
   	if (DEBUGGING) ptrTIUCommThread->PrintItem("Send_String_FIFO() Retrying " + String(CommandString));
	bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, CommandString, TIUNo, MyRemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
	if (!bretv) return false;	// return false with value == 0 if we can't read RAM
	}
*/
////---------------------------------------------------------------------------
bool __fastcall TCommThread::Send_String_FIFO(unsigned char Consist, byte * Command_String,
	unsigned char TIU, unsigned char Remote, int Expected, int WaitTime, bool Record_Flag, byte * ptrLashUpEngineList,
    int * ptrRespCode, int * SSResponseCount, byte * SSResponseBuffer)
{
assert (TIU >= 1 && TIU <= TIUMAX);
assert (Command_String != NULL);
assert (Remote >= 1 && Remote <= REMOTEMAX);
//
String PrintLine;
bool bretv = false;
int LocalRespCode = UNDETERMINED;		// must have a response code even if the caller doesn't care so create a local spot

HANDLE LocalCommDoneEvent;		// for event communications with CommThread
//
// runs in the context of the thread
//
// if CommThread is created, use it
if (ptrTIUCommThread) {	// make sure the comm thread is running
    //  Wait for response using signal event
	LocalCommDoneEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
	assert(LocalCommDoneEvent != NULL);

	// allocate the FIFO list structure for one command
	struct FIFOItem * pFItem;
	assert (sizeof (struct FIFOItem) > 8);
    pFItem = new struct FIFOItem;				// deleted in TCommThread->Execute()
    pFItem->Type = CMD_Type_SendString;			// type CMD_Type_SendString command into the list
	pFItem->TVal = 0;							// execute now when calling any of the *_FIFO routines
    pFItem->ConsistNo = Consist;
    pFItem->TIUNo = TIU;
    (void) strncpy(pFItem->Command, Command_String, 128);
    pFItem->Remote = Remote;
    //pFItem->Flag = (int *) NULL;				// not used
    //pFItem->FlagValue = 0;					// not used
    pFItem->ExpectedCount = Expected;
    pFItem->WaitTime = WaitTime;
    pFItem->Record_Flag = Record_Flag;
    pFItem->ptrRespCode = &LocalRespCode;

    pFItem->ResponseCount = SSResponseCount;
	pFItem->ResponseBuffer = SSResponseBuffer;
    // use the Lashup Engine List only if the Consist number is > 101
    if ((Consist>= 101) && (ptrLashUpEngineList != NULL)) {
    	(void) strncpy(pFItem->cLashUpEngineList, ptrLashUpEngineList, LASHUP_ENGINE_LIST_LENGTH);
        } else {
		pFItem->cLashUpEngineList[0] = '\0';
        }

    pFItem->CommEvent = LocalCommDoneEvent;

    if (!FIFO) {
    	// if the FIFO does not exists, toss the command
        PrintLine = "Send_String_FIFO: FIFO does not exist";
        PrintItem(PrintLine);
    	return false;
        }

    // add to the queue
    TList *myList = ptrTIUCommThread->FIFO->LockList();
    assert(myList);
    ptrTIUCommThread->FIFO->Add(pFItem);
    // Signal Comm_Thread that a command is ready
    SetEvent(CommReqEvent);
    ptrTIUCommThread->FIFO->UnlockList();
    if (DEBUGGING) Bump(MainForm->FIFO_Count);
    //  Wait for response using signal event
    if (DEBUGGINGX(9)) {
        PrintLine = "Send_String_FIFO: WaitForObject(LocalCommDoneEvent)";
        PrintItem(PrintLine);
        }
    DWORD dwWait;
    //
    while(true) {
        // Wait for the event to show up
        dwWait = WaitForSingleObject(LocalCommDoneEvent, DWDELAY);
        if (bEmergencyStop) {
            bretv = false;	// failure
            break;	// get out in a hurry
			}
        //
        if (dwWait == WAIT_TIMEOUT || dwWait == WAIT_FAILED) {	// DWDELAY timed out or other failure?
            // terminate Comm_Thread because of unknown failure
            // assert(false);	// this may execute if single stepping through the communications code
            // This failure mode does occur under some yet unknown circumstances.
            // With the assert() call, the program terminates. It seems to me to be
            // better to just return a failure than to terminate.
            // Caller should evaluate the failure and do something.
            bretv = false;	// failure
            break;
            }
        if (dwWait == WAIT_OBJECT_0) {	// Event Signaled ?
        	bretv = true;		// possible success
            if (DEBUGGINGX(7)) {
                PrintLine = "Send_String_FIFO() : CommDoneEvent Signaled";
                PrintItem(PrintLine, clBlack);
                }
            if (LocalRespCode >= UNDETERMINED) bretv = false;	// failed
            if (ptrRespCode) *ptrRespCode = LocalRespCode;
            break;
            }
        if (bEmergencyStop) break;	// get out in a hurry
        // keep on waiting
        }
    CloseHandle(LocalCommDoneEvent);
	return bretv;
    }
// may reach this point if the port closes before all commands are sent
CloseHandle(LocalCommDoneEvent);
assert(false);
return false;
}

//---------------------------------------------------------------------------
// TIUSendWaitforResponse_FIFO()
// Enqueue the Command and associated data in the FIFO
// Wait for CommDoneEvent
// Runs in the context of the calling thread, not TCommThread
// Called only by PBFileThread.cpp
//---------------------------------------------------------------------------
bool __fastcall TCommThread::TIUSendWaitforResponse_FIFO(unsigned char Consist, byte * Command,
	int ExpectedCount, int WaitTime, unsigned char Remote, unsigned char TIU, int * ptrRespCode,
    int * SWResponseCount, byte * SWResponseBuffer)
{
assert(ExpectedCount == 0 || ExpectedCount >= MINRESPONSE); 	// ExpectedCount must be 0 or 11 or more
assert(WaitTime >= 0 && WaitTime <= 20000);	// WaitTime must be positive and less than 20 seconds
assert(Remote <= REMOTEMAX);			// Remote must be 15 or less
assert(Command != NULL);				// Command must be present
//
String PrintLine;
bool bretv = false;
HANDLE LocalCommDoneEvent;		// for event communications with CommThread
int LocalRespCode = UNDETERMINED;	// must have a response code even if the caller doesn't care so create a local spot

// if CommThread is created, use it
if (ptrTIUCommThread) {	// right now, pressing the debug tab's "Connect" button creates this thread

	//  Wait for response using signal event
    LocalCommDoneEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
    assert(LocalCommDoneEvent != NULL);

	// allocate the FIFO list structure for one command
	struct FIFOItem * pFItem;
    assert (sizeof (struct FIFOItem) > 8);
    pFItem = new struct FIFOItem;					// deleted in TCommThread->Execute()
    pFItem->Type = CMD_Type_WaitForResponse;		// type CMD_Type_WaitForResponse command into the list
	pFItem->TVal = 0;	// execute now
    pFItem->ConsistNo = Consist;
    pFItem->TIUNo = TIU;
    (void) strncpy(pFItem->Command, Command, 128);
    pFItem->Remote = Remote;
    pFItem->ExpectedCount = ExpectedCount;
    pFItem->WaitTime = WaitTime;
	pFItem->ptrRespCode = &LocalRespCode;
    pFItem->ResponseCount = SWResponseCount;
    pFItem->ResponseBuffer = SWResponseBuffer;

    pFItem->CommEvent = LocalCommDoneEvent;

    if (!FIFO) {
    	// if the FIFO does not exists, toss the command
        PrintLine = "TIUSendWaitforResponse_FIFO: FIFO does not exist";
        PrintItem(PrintLine);
    	return false;
        }

    // add to the queue
    TList *myList = ptrTIUCommThread->FIFO->LockList();
    assert(myList);
    ptrTIUCommThread->FIFO->Add(pFItem);
    // Signal Comm_Thread that a command is ready
    SetEvent(CommReqEvent);
    ptrTIUCommThread->FIFO->UnlockList();
    if (DEBUGGING) Bump(MainForm->FIFO_Count);
    //  Wait for response using signal event
    if (DEBUGGINGX(7)) {
        PrintLine = "TIUSendWaitforResponse_FIFO: WaitForObject(CommDoneEvent)";
        PrintItem(PrintLine, clBlack);
        }
    DWORD dwWait;
    //
    for (;;) {
        // Wait for the CommDoneEvent
        dwWait = WaitForSingleObject(LocalCommDoneEvent, DWDELAY);
        if (bEmergencyStop) break;	// get out in a hurry
        //
        if (dwWait == WAIT_TIMEOUT || dwWait == WAIT_FAILED) {	// DWDELAY timed out or other failure?
            // terminate Comm_Thread because of unknown failure
            assert(false);
            bretv = false;	// failure
			break;
            }
        if (dwWait == WAIT_OBJECT_0) {	// Event Signaled ?
	        bretv = true;	// possible success
            if (DEBUGGINGX(7)) {
                PrintLine = "TIUSendWaitforResponse_FIFO() : CommDoneEvent Signaled";
                PrintItem(PrintLine, clBlack);
                }
			// *ptrRespCode is set from LocalRespCode just before the function exits
            if (LocalRespCode >= UNDETERMINED) bretv = false;	// failed
			if (ptrRespCode) *ptrRespCode = LocalRespCode;
            break;
            }
        // process all messages
        bool MsgAvailable;
        MSG msg;
        while (true) {
            MsgAvailable = PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);	// message available ?
            if (!MsgAvailable) break;	// no messages available
            // translate and dispatch the message
            TranslateMessage(&msg);
            DispatchMessage(&msg);
            // check for another message
            }
        if (bEmergencyStop) break;	// get out in a hurry
        // keep on waiting
        }
    CloseHandle(LocalCommDoneEvent);
    return bretv;
    }
// may reach this point if the port closes before all commands are sent
CloseHandle(LocalCommDoneEvent);
assert(false);					// but its a hard failure if it occurs
return false;
}

//---------------------------------------------------------------------------
// SendHexStrStr_FIFO()
// FiFO version of SendHexStr() which uses a String as input
// Runs in the context of the calling thread, not TCommThread
// Called by Train_Control.cpp
//---------------------------------------------------------------------------

bool __fastcall TCommThread::SendHexStrStr_FIFO(int Port, String Cmd_In, int Len)  // Len is actual bytes to send
{
assert(Port == TIU_PORT);
assert(Cmd_In.Length() > 0);
assert(Len > 0);
//
// convert Cmd_In from a String to unsigned char[] and pass it to SendHexStr_FIFO()
unsigned char Packet[CMDMAX];
// this routine is used the CB10.1 compile where a wchar_t needs to be converted to char
AnsiString aCmd_In = Cmd_In;	// must be converted to AnsiString as SendHexStr_FIFO() requires a char string
charstringtobinary(aCmd_In.c_str(), Packet, Len);	// convert ASCII char string to binary
return SendHexStr_FIFO(Port, Packet, Len);
}	// SendHexStrStr(String)

//---------------------------------------------------------------------------
// SendHexStr_FIFO()
// FIFO version of SendHexStr() which uses a binary hex string as input
// Runs in the context of the calling thread, not TCommThread
//---------------------------------------------------------------------------

bool __fastcall TCommThread::SendHexStr_FIFO(int Port, byte * Cmd_In, int Len)  // Len is actual bytes to send
{
assert(Port == TIU_PORT);
assert(Cmd_In != NULL);
assert(Len > 0);
//
// Enqueue the Cmd_In and associated data in the FIFO
// Wait for CommDoneEvent
//
String PrintLine;
bool retv = false;
HANDLE LocalCommDoneEvent;		// for event communications with CommThread

// Send to TIU Port only, don't wait for packet response, simple error handling
// This routine sends the command to the FIFO and waits for a CommDoneEvent

// Cmd_In is in binary and might contain 0x00 so don't take length() but use Len

// if CommThread is created, use it
if (ptrTIUCommThread && (Port == TIU_PORT)) {	// right now, pressing the debug tab's "Connect" button creates this thread

    //  Wait for response using signal eventt
    LocalCommDoneEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
    assert(LocalCommDoneEvent != NULL);

	// allocate the FIFO list structure for one command
	struct FIFOItem * pFItem;
	assert (sizeof (struct FIFOItem) > 8);
	pFItem = new struct FIFOItem;				// deleted in TCommThread->Execute()
    pFItem->Type = CMD_Type_SendHex;			// type CMD_Type_SendHex command into the list
	pFItem->TVal = 0;							// execute now
    pFItem->ExpectedCount = 0;
    pFItem->WaitTime = 0;
    (void) strncpy(pFItem->Command, Cmd_In, Len);
    pFItem->Len = Len;

    pFItem->CommEvent = LocalCommDoneEvent;

    // other variables in the FIFOItem structure are not used

    if (!FIFO) {
    	// if the FIFO does not exists, toss the command
        PrintLine = "SendHexStr_FIFO: FIFO does not exist";
        PrintItem(PrintLine);
    	return false;
        }

    // add to the queue
    TList *myList = ptrTIUCommThread->FIFO->LockList();
    assert(myList);
    ptrTIUCommThread->FIFO->Add(pFItem);
    // Signal Comm_Thread that a command is ready
    SetEvent(CommReqEvent);
    ptrTIUCommThread->FIFO->UnlockList();
    if (DEBUGGING) Bump(MainForm->FIFO_Count);
    //  Wait for response using signal event
    if (DEBUGGINGX(7)) {
        PrintLine = "SendHexStr_FIFO: WaitForObject(CommDoneEvent)";
        PrintItem(PrintLine, clBlack);
        }

    DWORD dwDelay = 1000; 	// absolute time out value of 1 second (not the same as WaitTime)
    DWORD dwWait;
    //
    for (;;) {
        // Wait for our event to show up
        dwWait = WaitForSingleObject(LocalCommDoneEvent, dwDelay);
        if (bEmergencyStop) break;	// get out in a hurry
        //
        if (dwWait == WAIT_TIMEOUT || dwWait == WAIT_FAILED) {	// dwDelay timed out or other failure?
            // terminate Comm_Thread because of unknown failure
            assert(false);
            retv = false;	// failure
            break;
            }
        if (dwWait == WAIT_OBJECT_0) {	// Event Signaled ?
			if (DEBUGGINGX(7)) {
                PrintLine = "SendHexStr_FIFO: CommDoneEvent Signaled";
                PrintItem(PrintLine, clBlack);
                }
            retv = true; 	// success
            break;
            }
        if (bEmergencyStop) break;	// get out in a hurry
        // keep on waiting
        }
    CloseHandle(LocalCommDoneEvent);
    return retv;
    }
// may reach this point if the port closes before all commands are sent
CloseHandle(LocalCommDoneEvent);
assert(false);
return false;
}
//---------------------------------------------------------------------------

bool __fastcall TCommThread::SendReadRAMCommand(struct FIFOItem * PItem) {
//
// The 'V' command gets broken into 2-5 'q' commands to be sent as a group to the TIU
// It is very important that the group go to the TIU one after the other.
//
// The 'V' command is a special command made up for RTC
// Vbaaa
// where:
//    b = number of bytes to read 1,2,4
//    aaa = lowest byte address (we read up big endian so only 1, 2, or 4 big endian bytes are valid)
//
//#define RAMIDLE 50	// Did not improve the NAK rate
int NumBytes;
unsigned int addr, TranslatedAddr;
char cmd[15];
(void) sscanf(PItem->Command, QUERYATOM, &NumBytes, &addr);
//assert(NumBytes == 1 || NumBytes == 2 || NumBytes == 4); // this is tested in the below switch statement
assert(addr <= 0x3FF);
bool bretv = true;	// starts at true and set to false if any of the Send_String() calls fails
switch(NumBytes) {
	case 1:
        TranslatedAddr = RAMTOQ(addr, 0);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
        bretv &=  Send_String(PItem->ConsistNo, cmd,
            PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
            PItem->cLashUpEngineList,
            NULL, NULL, NULL);	// we don't care about the response until the 'q' command

		break;
	case 2:
		TranslatedAddr = RAMTOQ(addr, 0);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
		bretv &=  Send_String(PItem->ConsistNo, cmd,
			PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
			PItem->cLashUpEngineList,
			NULL, NULL, NULL);	// we don't care about the response until the 'q' command
		if (!bretv) break;	// don't continue on error as the RAM value will be wrong
        //
		TranslatedAddr = RAMTOQ(addr+1, 1);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
		bretv &=  Send_String(PItem->ConsistNo, cmd,
			PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
			PItem->cLashUpEngineList,
			NULL, NULL, NULL);	// we don't care about the response until the 'q' command
		//Sleep(RAMIDLE);	// delay in case we are hitting the TIU too hard with 'q' commands

		break;
	case 4:
		TranslatedAddr = RAMTOQ(addr, 0);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
		bretv &=  Send_String(PItem->ConsistNo, cmd,
			PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
			PItem->cLashUpEngineList,
			NULL, NULL, NULL);	// we don't care about the response until the 'q' command
		if (!bretv) break;	// don't continue on error as the RAM value will be wrong
        //
		TranslatedAddr = RAMTOQ(addr+1, 1);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
		bretv &=  Send_String(PItem->ConsistNo, cmd,
			PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
			PItem->cLashUpEngineList,
			NULL, NULL, NULL);	// we don't care about the response until the 'q' command
		if (!bretv) break;	// don't continue on error as the RAM value will be wrong
        //
		TranslatedAddr = RAMTOQ(addr+2, 2);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
		bretv &=  Send_String(PItem->ConsistNo, cmd,
			PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
			PItem->cLashUpEngineList,
			NULL, NULL, NULL);	// we don't care about the response until the 'q' command
		if (!bretv) break;	// don't continue on error as the RAM value will be wrong
        //
		TranslatedAddr = RAMTOQ(addr+3, 3);
		snprintf(cmd, sizeof cmd, QUERY, TranslatedAddr);
		bretv &=  Send_String(PItem->ConsistNo, cmd,
			PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
			PItem->cLashUpEngineList,
			NULL, NULL, NULL);	// we don't care about the response until the 'q' command

		break;
	default:
		assert(false);	// NumBytes can only be 1, 2, or 4
		return false;
	}

if (!bretv) return false;	// don't continue on error as the RAM value will be wrong
//
// Now send a 'q' command to actually get the data
bretv &=  Send_String(PItem->ConsistNo, "q",
	PItem->TIUNo, PItem->Remote, PItem->ExpectedCount, PItem->WaitTime, PItem->Record_Flag,
	PItem->cLashUpEngineList,
	PItem->ptrRespCode, PItem->ResponseCount, PItem->ResponseBuffer);
//
return bretv;	// all calls to Sent_String must return true for this function to return true
}
//---------------------------------------------------------------------------

