//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2019 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
//---------------------------------------------------------------------------
#include <vcl.h>
#include <tchar.h>
#pragma hdrstop

#include "Train_Control.h"
#include "Comm_Thread.h"
#include "Debug_Controls.h"
#include "Setup_Control.h"
#include "Clipbrd.hpp"

#include <stdio.h>
#include <time.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "HideButton"
#pragma link "AbNumEdit"
#pragma resource "*.dfm"

TDB *DB;

extern TCommThread * ptrTIUCommThread;				// pointer to TIU communications thread
extern int PC_Tenths;								// RunTimeinTenths counter (system wide)


//---------------------------------------------------------------------------

__fastcall TDB::TDB(TComponent* Owner)
    : TForm(Owner)
{
MainForm->bDebug = false;	// turn off debugging until form is shown.
//https://stackoverflow.com/questions/6049053/increase-size-limit-of-trichedit
// expand the size of RichEdit control
Debug_Messages->MaxLength = System::MaxInt-2;

// Experimental - use Drop Down Menu style in CB10.1
#if __BORLANDC__ >= 0x720
enum DECLSPEC_DENUM TButtonStyle : unsigned char { bsPushButton, bsCommandLink, bsSplitButton };
//Prog_Control_Button->Style = Prog_Control_Button->Style + 2;    // bsSplitButton defined in Vcl.StdCtrls.hpp is really not defined
ClearDebugButton->Style = bsSplitButton;
ClearDebugButton->DropDownMenu = DebugPopup;
#endif

}
//---------------------------------------------------------------------------

__fastcall TDB::~TDB(void)
{
FindDialogDebugMessages->CloseDialog();
ReplaceDialogDebugMessages->CloseDialog();
if (MainForm->bSaveRegistryInfo) {
    // write Preferences from registry if they exist
    TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
    assert(RegFile != NULL);
    __try {
		RegFile->WriteInteger("Debug", "Top", Top > 0 ? Top : 0);
		RegFile->WriteInteger("Debug", "Height", Height);
        RegFile->WriteInteger("Debug", "Left", Left > 0 ? Left : 0);
        RegFile->WriteInteger("Debug", "Width", Width);
        RegFile->WriteInteger("Debug", "Level", DBLevel->ValueAsInt);
        RegFile->WriteInteger("Debug", "AutoScroll", AutoScroll->Checked);
        RegFile->WriteInteger("Debug", "TimeStamp", TimeStampGroup->ItemIndex);
        }
    __finally {
        // Deallocate RegFile
        delete RegFile;
        }
    }
// save Debug Messages window always as rich text file
if (DEBUGGING && Debug_Messages->Lines->Count > 0) {
    Debug_Messages->PlainText = false;
    const String FileName = String(MainForm->cStartDir) + "\\" + "Debug_Messages.rtf";
    Debug_Messages->Lines->SaveToFile(FileName);
    }

// save CRC Buffer window as either rich text or plain text file
if (DEBUGGING && MainForm->CRCBuffer->Lines->Count > 0) {
	MainForm->CRCBuffer->PlainText = MainForm->PlainTextFile->Checked;
	if (MainForm->CRCBuffer->PlainText) {
    	const String FileName = String(MainForm->cStartDir) + "\\" + "CRC_Messages.txt";
		MainForm->CRCBuffer->Lines->SaveToFile(FileName);
		} else {
    	const String FileName = String(MainForm->cStartDir) + "\\" + "CRC_Messages.rtf";
		MainForm->CRCBuffer->Lines->SaveToFile(FileName);
		}
    
    }
// save TIU Buffer window as either rich text or plain text file
if (DEBUGGING && MainForm->TIUBuffer->Lines->Count > 0) {
	MainForm->TIUBuffer->PlainText = MainForm->PlainTextFile->Checked;
	if (MainForm->TIUBuffer->PlainText) {
    	const String FileName = String(MainForm->cStartDir) + "\\" + "TIU_Messages.txt";
		MainForm->TIUBuffer->Lines->SaveToFile(FileName);
		} else {
    	const String FileName = String(MainForm->cStartDir) + "\\" + "TIU_Messages.rtf";
		MainForm->TIUBuffer->Lines->SaveToFile(FileName);
		}
    
    }
// save RFID Buffer window as either rich text or plain text file
if (DEBUGGING && MainForm->RFIDBuffer->Lines->Count > 0) {
	MainForm->RFIDBuffer->PlainText = MainForm->PlainTextFile->Checked;
	if (MainForm->RFIDBuffer->PlainText) {
    	const String FileName = String(MainForm->cStartDir) + "\\" + "RFID_Messages.txt";
		MainForm->RFIDBuffer->Lines->SaveToFile(FileName);
		} else {
    	const String FileName = String(MainForm->cStartDir) + "\\" + "RFID_Messages.rtf";
		MainForm->RFIDBuffer->Lines->SaveToFile(FileName);
		}
    
    }

MainForm->bDebug = false;	// clear flag on exit
//
// Debug window is always opaque
//
}

//---------------------------------------------------------------------------

void __fastcall TDB::ShowIt(int Remote, int Engine, int TIU)
{
// grab the Remote, Engine and TIU numbers
RemoteNo = Remote;
EngineNo =  Engine;
TIUNo = TIU;
DebugTopPanel->Caption = "Debug Messages";
Caption = DebugTopPanel->Caption;
MainForm->bDebug = true;	// turn on debugging

// turn on debug check box panel
Setup->SerialDebugPanel->Visible = true;

//Setup->SuperTIUMode->Visible = true;
Setup->AllowZero->Visible = true;
Setup->CRC_Trace->Visible = true; 	// not saved in Registry

Setup->EncodedRecordFormat->Visible = true;

Setup->BaudRateSel->Enabled = true;
Setup->ParitySel->Enabled = true;
Setup->BitsSel->Enabled = true;
Setup->StopSel->Enabled = true;

//Setup->COMSelGroup->Top = 50;	// Serial port box position
//Setup->SerialPortListGroup->Top = 50;

// turn on debug version of port control
Setup->SerialPortListGroup->Visible = false;
Setup->BaudRateSel->Visible = true;
Setup->ParitySel->Visible = true;
Setup->StopSel->Visible = true;
Setup->ParitySel->Visible = true;
Setup->BitsSel->Visible = true;

MainForm->Hint = "RTC Debug Screen - Use at your own risk";

// horizontal scroll bar on EngineData StringGrid
MainForm->EngineData->ScrollBars = ssBoth;

// minumum width
Constraints->MinWidth = Width;

Show();  	      // Same as setting Visible to true

// Note: I want to have the window come up at the postion given by
// the saved coordinates. If I put the Show() call after this registry
// code, the Show() puts its own coordinates in.
// With the Show() before the registry code, the window flashes in
// the windows default position and then, at least, moves to its
// new position. I would like to get rid of the flashing but I don't know
// how

// read preferences from the Registry
RegFileRootKey = MainForm->RegFileRootKey;
// read Preferences from registry if they exist
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);
__try {
	Top = RegFile->ReadInteger("Debug", "Top", Top);
	Left = RegFile->ReadInteger("Debug", "Left", Left);
    // if using multiple monitors and one was removed, pull the window
    // back onto the remaining monitor
    int MC = Screen->MonitorCount;
    if (MC == 1) {
        int Wth = Screen->Monitors[0]->Width;
		if (Left > Wth) Left -= Wth;
        int Hth = Screen->Monitors[0]->Height;
        if (Top > Hth) Top -= Hth;
        }
	Height = RegFile->ReadInteger("Debug", "Height", Height);
	Width = RegFile->ReadInteger("Debug", "Width", Width);
	DBLevel->ValueAsInt = RegFile->ReadInteger("Debug", "Level", DBLevel->ValueAsInt);
    DBLevel->Options = DBLevel->Options << eoLimitMin;	// necessary to repair some subtle bug in the AbNumSpin
	AutoScroll->Checked = RegFile->ReadInteger("Debug", "AutoScroll", AutoScroll->Checked);
	TimeStampGroup->ItemIndex = RegFile->ReadInteger("Debug", "TimeStamp", TimeStampGroup->ItemIndex);
    }
__finally {
	// Deallocate RegFile
	delete RegFile;
    }
//
TimeStampButtonClick(NULL);	// place a time stamp at the beginning of the debug messages window
MainForm->GlobalDBLevel = DBLevel->ValueAsInt;	// use this instead of the DBLevel->Value VCL property in threads
}
//---------------------------------------------------------------------------

bool __fastcall TDB::Debug_Print(String Msg, TColor Color)
{
if (!MainForm->bDebug && !Debug_Messages) return false;		// just in case the TRichEdit was destroyed
    Debug_Messages->SelAttributes->Color = Color;
    if (TimeStampGroup->ItemIndex == 0 || (TimeStampGroup->ItemIndex == 1 && PC_Tenths == 0)) {
	    // https://www.drdobbs.com/windows/win32-performance-measurement-options/184416651
        // https://stackoverflow.com/questions/1695288/getting-the-current-time-in-milliseconds-from-the-system-clock-in-windows
       	SYSTEMTIME st;
        GetLocalTime(&st);
        char currentTime[24];
        snprintf(currentTime, sizeof currentTime, "%02d:%02d:%02d.%03d", st.wHour, st.wMinute, st.wSecond , st.wMilliseconds);
        DB->Debug_Messages->SelStart = DB->Debug_Messages->GetTextLen();
        DB->Debug_Messages->SelText = String(currentTime) + " D " + Msg + TEXT("\n");
        } else {
        if (TimeStampGroup->ItemIndex == 1) {
            DB->Debug_Messages->SelStart = DB->Debug_Messages->GetTextLen();
			DB->Debug_Messages->SelText = Format(String("%9.1f"), ARRAYOFCONST((PC_Tenths/10.0))) + " D " + Msg + TEXT("\n");
			} else {
			DB->Debug_Messages->SelStart = DB->Debug_Messages->GetTextLen();
			DB->Debug_Messages->SelText = Format(String("%7d"), ARRAYOFCONST((MainForm->MsgNum++))) + " D " + Msg + TEXT("\n");
            }
        }
    Debug_Messages->SelAttributes->Color = clBlack;
    //
	// remain at end of buffer
	//
	// Method 1 does not work under CB10.1
	//Debug_Messages->SelStart = Debug_Messages->GetTextLen();
	//Debug_Messages->Perform(EM_SCROLLCARET, 0, 0);
	// Method 2 does not work under CB10.1
	//Debug_Messages->SelStart = Debug_Messages->Text.Length();
	//SendMessage(Debug_Messages->Handle, EM_SCROLLCARET,0,0);
	// Method 3 works under both BCB 4.0 and CB10.1 But sometimes line down doesn't get to end of buffer
	//Debug_Messages->Perform(EM_SCROLL, SB_LINEDOWN, 0);	// remain at end of buffer
	// Method 4 works under CB10.1 but not BCB 4.0
	//http://codeverge.com/embarcadero.delphi.general/d2009-richedit/1040993
	//
   	if (AutoScroll->Checked) {
   		Debug_Messages->HideSelection = false;
   		Debug_Messages->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
   		Debug_Messages->HideSelection = true;
        }
	//
return true;
}
//---------------------------------------------------------------------------

void __fastcall TDB::HideButtonClick(TObject *Sender)
{
FindDialogDebugMessages->CloseDialog();
ReplaceDialogDebugMessages->CloseDialog();

MainForm->bDebug = false;	// turn off debugging
MainForm->GlobalDBLevel = 0;
Hide();
// close up mainform
MainForm->Width = MainForm->iSavedWidth;
MainForm->Height = MainForm->iSavedHeight;
// turn off horizontal scroll bar on EngineData StringGrid
MainForm->EngineData->ScrollBars = ssVertical;

// turn off debug check box panel
Setup->SerialDebugPanel->Visible = false;

Setup->AllowZero->Visible = false;
Setup->CRC_Trace->Visible = false; 	// not saved in Registry
Setup->EncodedRecordFormat->Visible = false;
//Setup->SuperTIUMode->Visible = false;

Setup->BaudRateSel->Enabled = false;
Setup->ParitySel->Enabled = false;
Setup->BitsSel->Enabled = false;
Setup->StopSel->Enabled = false;

//Setup->COMSelGroup->Top = 18;	// Serial port box position
//Setup->SerialPortListGroup->Top = 18;

// turn off debug version of port control
Setup->SerialPortListGroup->Visible = true;
Setup->BaudRateSel->Visible = false;
Setup->ParitySel->Visible = false;
Setup->StopSel->Visible = false;
Setup->ParitySel->Visible = false;
Setup->BitsSel->Visible = false;

MainForm->Hint = "Remote Control Main Window";

// remove horizontal scroll bar on EngineData StringGrid
MainForm->EngineData->ScrollBars = ssVertical;
}
//---------------------------------------------------------------------------

void __fastcall TDB::CutDebugClick(TObject *Sender)
{
// Debug Buffer popup menu item
Debug_Messages->CutToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TDB::CopyDebugClick(TObject *Sender)
{
// Debug Buffer popup menu item
Debug_Messages->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TDB::PasteDebugClick(TObject *Sender)
{
// Debug Buffer popup menu item
Debug_Messages->PasteFromClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TDB::SelectAllDebugClick(TObject *Sender)
{
// Debug Buffer popup menu item
Debug_Messages->SelectAll();
}
//---------------------------------------------------------------------------

void __fastcall TDB::SaveDebugClick(TObject *Sender)
{
MainForm->WritetoDiskClick(Sender);	
}
//---------------------------------------------------------------------------

void __fastcall TDB::ClearDebugButtonClick(TObject *Sender)
{
// clear Debug Messages and line count
Debug_Messages->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TDB::FindDialogClick(TObject *Sender)
{
// http://www.bcbjournal.org/articles/vol1/9710/Using_the_Find_and_Replace_dialog_boxes.htm
//FindDialog1->Position = Point(Debug_Messages->Left + Debug_Messages->Width, Debug_Messages->Top);
FindDialogDebugMessages->Execute();
}
//---------------------------------------------------------------------------

void __fastcall TDB::FindDialogDebugMessagesFind(TObject *Sender)
{
  // Dialog that generated the event could be either
  // Find or Replace dialog so just cast
  // Sender to a TFindDialog.
  TFindDialog* Dlg = 
     dynamic_cast<TFindDialog *>(Sender);

  // Set up the options.
  TSearchTypes Options;
  if (Dlg->Options.Contains(frWholeWord))
    Options << stWholeWord;
  if (Dlg->Options.Contains(frMatchCase))
    Options << stMatchCase;

  // Deselect any selected text and advance the cursor
  // position. Prevents us from selecting the same
  // text over and over again.
  if (Debug_Messages->SelLength) {
    Debug_Messages->SelStart += Debug_Messages->SelLength;
    Debug_Messages->SelLength = 0;
  }

  // Find the text.
  int Pos = Debug_Messages->FindText(Dlg->FindText,
    Debug_Messages->SelStart, Debug_Messages->GetTextLen(),
    Options);

  // If text is found, highlight it, scroll window to   
  // found text, and set focus to RichEdit control. 
  // If no text is found, let user know.
  if (Pos != -1) {
    Debug_Messages->SelStart = Pos;
    Debug_Messages->SelLength = Dlg->FindText.Length();
    // Method 4 works under CB10.1 but not BCB 4.0
    Debug_Messages->HideSelection = false;
    Debug_Messages->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    Debug_Messages->HideSelection = true;
    //
    Debug_Messages->SetFocus();
  }
  else
    Application->MessageBox(TEXT("Search text not found."), TEXT("Find Text"), MB_OK);

}
//---------------------------------------------------------------------------

void __fastcall TDB::ReplaceDialogDebugMessagesReplace(TObject *Sender)
{
// If user clicked Replace All, then do find/replace
// operation until no more matches are found.
if (ReplaceDialogDebugMessages->Options.Contains(frReplaceAll)) {
    static int Count;
	Count = 0;
    TSearchTypes Options;
    if (ReplaceDialogDebugMessages->Options.Contains(frWholeWord))
      Options << stWholeWord;
    if (ReplaceDialogDebugMessages->Options.Contains(frMatchCase))
      Options << stMatchCase;

    // Find the first match.
    int Pos = Debug_Messages->FindText(
      ReplaceDialogDebugMessages->FindText, Debug_Messages->SelStart,
      Debug_Messages->GetTextLen(), Options);

    // Loop while we are still finding text.
    while (Pos != -1) {
      // Same code as in the OnFind event handler.
      Debug_Messages->SelStart = Pos;
      Debug_Messages->SelLength =
        ReplaceDialogDebugMessages->FindText.Length();

      // Replace the found text with the new text.
      Debug_Messages->SetSelTextBuf(
        ReplaceDialogDebugMessages->ReplaceText.c_str());
      Debug_Messages->Perform(EM_SCROLLCARET, 0, 0);

      // Increment the counter so we can tell the user
      // how many replacements we made.
      Count++;

      // Do another search.
      Pos = Debug_Messages->FindText(
        ReplaceDialogDebugMessages->FindText, Debug_Messages->SelStart,
        Debug_Messages->GetTextLen(), Options);
      }

    // Dismiss the Replace dialog.
    ReplaceDialogDebugMessages->CloseDialog();

    // Tell the user what we did.
    TCHAR buff[128];
    // NOTE: under some circumstances, the next line gets a compile error.
    // E2268 Call to undefined function '_sntprintf'. But that function is used in other source files.
    // Then if I just recompile, I do not get the error    
    _sntprintf(buff, sizeof(buff),TEXT("Finished searching text.\r\n%d replacements made."), Count);
	Application->MessageBox(buff, TEXT("Replace Text"), MB_OK);
	} else {
    // If not doing a Replace All, then just replace the
    // found text with the new text.
    // No text selected? Then return.
    if (Debug_Messages->SelLength == 0) return;
    Debug_Messages->SetSelTextBuf(
      ReplaceDialogDebugMessages->ReplaceText.c_str());
    Debug_Messages->SelLength = 0;
    }

// Set focus to the Debug_Messages.
Debug_Messages->SetFocus();
}
//---------------------------------------------------------------------------

void __fastcall TDB::CopyButtonClick(TObject *Sender)
{
Debug_Messages->SelectAll();
Debug_Messages->CopyToClipboard();
Debug_Messages->SelLength = 0;	// clear selection highlight
}
//---------------------------------------------------------------------------

void __fastcall TDB::Font1Click(TObject *Sender)
{
FontDialog1->Options << fdApplyButton;
FontDialog1->Execute();
}
//---------------------------------------------------------------------------

void __fastcall TDB::FontDialog1Apply(TObject *Sender, HWND Wnd)
{
if (ActiveControl->ClassNameIs("TEdit"))
   ((TEdit *)ActiveControl)->Font->Assign(FontDialog1->Font);
  else if (ActiveControl->ClassNameIs("TRichEdit"))
    ((TRichEdit *)ActiveControl)->SelAttributes->Assign(FontDialog1->Font);
    else
    MessageBeep(0);
}
//---------------------------------------------------------------------------

void __fastcall TDB::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
// Make sure KeyPreview is true for this form
assert(KeyPreview);
// Pass function keys to the MainForm, handle rest of the keys here.
switch (Key) {
    case VK_F1:
    case VK_F2:
    case VK_F3:
    case VK_F4:
    case VK_F5:
    case VK_F6:
    case VK_F7:
    case VK_F8:
    case VK_F9:
    case VK_F10:
    case VK_F11:
    case VK_F12:
		MainForm->FormKeyDown(Sender, Key, Shift);
        Key = 0;	// handled
		break;
	}	
}
//---------------------------------------------------------------------------

void __fastcall TDB::CopyrightClick(TObject *Sender)
{
// Copy the Form to the clipboard.
// In BCB4.0, there is no SetSize method associated with TBitmap
// so we have to use Width and Height separately.
//
// see:
// https://stackoverflow.com/questions/3190777/how-do-i-copy-a-form-as-an-image-to-the-clipboard
//
Graphics::TBitmap * bm;
bm = new Graphics::TBitmap;
try
	{
    #if __BORLANDC__ >= 0x720
	// Defined for CB10.1 C++ programs that use the VCL.
	bm->SetSize(ClientWidth, ClientHeight);
    #else
    bm->Width = ClientWidth;
    bm->Height = ClientHeight;
    #endif
	BitBlt(bm->Canvas->Handle, 0, 0, ClientWidth, ClientHeight, Canvas->Handle, 0, 0, SRCCOPY);
    Clipboard()->Assign(bm);
    }
__finally
	{
	delete bm;
    }
// close up mainform
//MainForm->Width = MainForm->iSavedWidth;
//MainForm->Height = MainForm->iSavedHeight;
}
//---------------------------------------------------------------------------

void __fastcall TDB::TimeStampButtonClick(TObject *Sender)
{
// place a time stamp in the Debug Buffer in the form "Thu Jan 20 15:19:21 2022"
  time_t rawtime;
  struct tm * timeinfo;

  time ( &rawtime );
  timeinfo = localtime ( &rawtime );
  Debug_Print("Timestamp : " + String(asctime (timeinfo)));
}
//---------------------------------------------------------------------------

void __fastcall TDB::DBLevelValueChanged(TObject *Sender)
{
Debug_Print("Debug Level changed to " + DBLevel->Text);
MainForm->GlobalDBLevel = DBLevel->ValueAsInt;	// keep global value in sync with VCL property
}
//---------------------------------------------------------------------------

void __fastcall TDB::FormCloseQuery(TObject *Sender, bool &CanClose)
{
// ask user to use the Hide button instead of the X button
MainForm->DebugMessagePopup("Warning", "Use the [Hide] button to close the Debug window", "");
CanClose = false;	// disallow close
}
//---------------------------------------------------------------------------

void __fastcall TDB::ToggleWordWrapClick(TObject *Sender)
{
Debug_Messages->WordWrap = !Debug_Messages->WordWrap;
ToggleWordWrap->Checked = Debug_Messages->WordWrap;
}
//---------------------------------------------------------------------------

void __fastcall TDB::WMEnable(TWMEnable &Message)
// https://community.idera.com/developer-tools/b/blog/posts/keep-a-non-modal-window-active-when-a-modal-window-is-being-shown
{
// Handle Message : WM_ENABLE
if (!Message.Enabled) {
	PostMessage(Handle, CM_REENABLE, 1, 0);
    }
//
}
//---------------------------------------------------------------------------

void __fastcall TDB::CMReenable(TMessage &Message)
{
// Handle Message : Reenable window CM_REENABLE
//
EnableWindow(Handle, true);
//
}
//---------------------------------------------------------------------------



