//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2022 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page : http://www.silogic.com/trains/RTC_Running.html
//---------------------------------------------------------------------------
/*
Forms which are auto created:

MainForm
DB
NumberQuery
Command_Query_Form
DetectedRFIDTags
BlockNameQuery

All of the rest of the forms are manually created.

Linker
	[x] Create debug information
	[ ] Use dynamic RTL
	[x] Use debug Libraries

Packages
	[ ] Build with runtime packages
//---------------------------------------------------------------------------
There may be a bug in Borland C++ Builder which causes a stack overflow
if the overridden constructor only has one additional parameter, an int.

All overrider constructors should have at least two additional int paramters.
//---------------------------------------------------------------------------
A bug in BCB4 and 5 (fixed in BCB 6) caused trouble in starting of threads.

According to Remy Lebeau(TeamB) in this post:

"If you set CreateSuspended to False, the thread will begin running automatically,
AFTER all constructors have been called and exited first (early versions
had a bug in this area, which was finally fixed in D/CB v6."

Then he followed up in this post:

"That is the bug I referred to which was fixed in CB/Delphi v6. In earlier versions,
TThread's constructor did not take descendant constructors into account, so it
allowed the thread to begin running immediately, before descendant constructors
did their work. But in v6 onwards, that is no longer the case. TThread correctly
waits until all constructors have finished their work before then allowing the
thread to run."

This means that in the BCB4 version of RTC, I must be careful about constructing
objects in the TThread's constructor. Construction of objects could probably
be moved to the beginning of the Execute() function.

*/
//---------------------------------------------------------------------------
// Writing C++ Code for BCB 4.0 and CB10.1
// Requires option "Use Classic Borland Compiler to set to true in CB10.1
// http://docwiki.embarcadero.com/RADStudio/Tokyo/en/Predefined_Macros#C.2B.2B_Compiler_Versions_in_Predefined_Macros
#if __BORLANDC__ >= 0x720
// Defined for CB10.1 C++ programs that use the VCL.
#endif
//---------------------------------------------------------------------------
/*

Use the DLL versions of LUA, SOCKET and MIME. Do not use the LIB versions.

That way only one copy of the LUA code has to be loaded.

--> luadll.lib has to be included in the BCB 4.0 and 6.0 projects
--> lua535dll.lib has to be included in the CB10.1 and CB10.3 projects

--> luadll.dll has to be present with the RTC executable for BCB 4.0 and 6.0 projects
--> lua535dll.dll has to be present with the RTC executable for CB10.1 and CB10.3 project

--> BCB4.0 version of mimecore.dll renamed to core.dll has to be in the socket folder for BCB 4.0 and 6.0
--> BCB4.0 version socketcore.dll renamed to core.dll has to be in the socket folder for BCB 4.0and BCB 6.0

--> CB10.1 version of mimecore.dll renamed to core.dll has to be in the socket folder for CB 10.1 and CB 10.3
--> CB10.1 version socketcore.dll renamed to core.dll has to be in the socket folder for CB 10.1 and CB 10.3

--> BCB 4.0 version of ssl.dll has to be present with the RTC executable for BCB 4.0 and 6.0 projects
--> CB 10.1 version of ssl.dll has to be present with the RTC executable for CB10.1 and CB10.3 projects

--> luaSec renamed to ssl.dll version 0.4.1 has to be present with the RTC executable for BCB 4.0 and 6.0 projects
--> luaSec renamed to ssl.dll version 0.8 has to be present with the RTC executable for CB10.1 and CB10.3 projects

--> for all compilers libeay32.dll and ssleay32.dll version 1.0.2s (Win 32 version) in the folder with the RTC executable.
*/
//---------------------------------------------------------------------------

#include <vcl.h>
#include <tchar.h>
#pragma hdrstop

// RTC Near Field Communications
#include "RTCMegaMaster.h"
#include "RTC_NFC.h"

#include "Engine_Commands.h"
#include "Train_Control.h"
#include "AIU_Control.h"
#include "Op_Control.h"
#include "Volume_Control.h"
#include "Tserial_event.h"
#include "Utility.h"
#include "Setup_Control.h"
#include "Debug_Controls.h"
#include "Num_Query.h"
#include "Message_Popup.h"
#include "Conv_Mode.h"
#include "DCS_Setup_Controls.h"
#include "LashUp_Selection.h"
#include "RTC_Info.h"
#include "TMCC_Control.h"
#include "Quick_Controls.h"
#include "RTC_Help.h"
#include "Route_Scene.h"
#include "Hot_Com.h"
#include "Super_TIU_Mode.h"
#include "Comm_Thread.h"
#include "Track_Signal_Thread.h"
#include "Track_Signal.h"
#include "Engine_Information_Thread.h"
#include "Engine_Info.h"
#include "Engine_Status_Thread.h"
#include "Engine_Status.h"
#include "Timer_Thread.h"
#include "PC_Thread.h"
#include "Prog_Control.h"
#include "Message_Thread.h"
#include "DL_Thread.h"
#include "DetectedTags.h"
#include "Layout.h"
#include "Clipbrd.hpp"

#include <algorithm>	// STL - needed for lower_bound()
#include <dir.h>        // needed for getcwd() in CB10.1
#include <math.h>		// needed for pow() in BCB4.0
#include <stdio.h>
#include <time.h>		// for printout timestamp
#include <mmsystem.h>	// for PlaySound

// Rocksoft Model CRC Algorithm
#include "crcmodel.h"

// RTC Near Field Communications
#include "RTC_NFC.h"
// Differences between BCB4.0/BCB6.0 and CB10.1
//
//

#if __BORLANDC__ >= 0x720
// used by TSocketServer's SSL routines
// These should have been linked in automatically but I couldn't get that to work
#pragma link "cryptui.lib"
#pragma link "crypt32.lib"
#pragma link "ncrypt.lib"
#endif

#pragma resource "*.dfm"
TMainForm *MainForm;

//---------------------------------------------------------------------------
// Copyright comment into the executable file
#pragma comment(exestr, " Copyright 2022 by Mark DiVecchio")
//---------------------------------------------------------------------------

// pointers to Radio and RFID serial ports (actual thread created with "new")
Tserial_event * RFID_serial_port = NULL;
Tserial_event * Serial_port = NULL;
//static void SerialEventManager2(uint32 object, uint32 event);

bool Ignore_FF_Char;

// Serial port settings
const static int Baud[] = {2400, 4800, 9600, 19200, 38400, 57600, 115200};
const static int Parity[] = {SERIAL_PARITY_NONE,  SERIAL_PARITY_ODD, SERIAL_PARITY_EVEN};
const static char Bits[] = {7, 8};
const static int Stop[] = {SERIAL_STOP_1, SERIAL_STOP_2};

// global vars that might be used in an ISR should be defined as volatile
volatile bool bStop;
volatile int CTS_Count, DSR_Count, CD_Count;
volatile bool bRFIDSendComplete;
// RFID Serial Port Status
volatile bool bRFIDCOMConnected = false;	    // RFID Serial Port Connection Complete Flag
//volatile int iRFIDbufCount;					// RFID Serial Port Byte Counter
// Radio Serial Port Status
volatile bool bRadioCOMConnected = false;	    // Radio Serial Port Connection Complete Flag
// RFID UDP WiFi Port Status
volatile bool bRFIDUDPWiFiConnected = false;	// RFID UDP WiFi Port Connection Complete Flag
volatile int iRFIDUDPWiFibufCount = 0;			// RFID UDP WiFi Port Byte Counter
volatile bool bRFIDUDPWiFiRXConnected = false;	// RFID UDP WiFi Listen() Success Flag
// RFID TCP WiFi Port Status
volatile bool bRFIDTCPWiFiConnected = false;	// RFID TCP WiFi Port Connection Complete Flag
volatile int iRFIDTCPWiFibufCount = 0;			// RFID TCP WiFi Port Byte Counter
//volatile bool bRFIDTCPWiFiRXConnected = false;	// RFID TCP WiFi Listen() Success Flag
// Radio WiFi Port Status
volatile bool bRadioWiFiConnected = false;	// Radio WiFi Port Connection Complete Flag
volatile int iRadioWiFibufCount = 0;		// Radio WiFi Port Byte Counter
// AIU WiFi Port Status
volatile bool bAIUWiFiTXConnected;			// AIU WiFi TX Port Connection Complete Flag

// TIU Connection Status
bool bTIUConnected = false;				// set to true when the TIU acknowledged a Get AIU Count 'x' Command

// For threaded TIU Communications
volatile bool DisconnectRequest = false;		// Indicates [Disconnect] button was pressed
extern volatile bool RadioDataSent;				// Indicates that the Radio packet was sent either via COM or WiFi

bool bEmergencyStop;		// emergency stop in effect

// global vars that might be used in an ISR should be tagged volatile
extern volatile int iTIUbufCount;
extern volatile byte cTIUbuf[CMDMAX];			// data from the TIU on Port A

// Forward Declarations
unsigned char ValidDigit(unsigned char d);
unsigned char ValidByte(unsigned char c);
byte charToHex(byte ch);
unsigned char Hex2Digit_to_Decimal(unsigned char * str);
char * Amp(const char * a);
//---------------------------------------------------------------------------
// Workaround for SetLayeredWindowAttributes() to make transparent windows
lpfnSetLayeredWindowAttributes m_pSetLayeredWindowAttributes;
//---------------------------------------------------------------------------
// Global variables used for v2 routines that keep the data
// as binary rather than in String data types (can't be static - referenced in Comm_Thread.cpp)
//byte cRFIDbuf[CMDMAX];	// data from the RFID Tag Readers on RFID Port
//
byte Sequence;			// Sequence number for TX packets
//---------------------------------------------------------------------------
// Pointers to Forms for 121 Engines/LashUps, 5 TIU each with 5 AIU:
//		Factory Reset Engine 0 (DCS 1)	 	(1)
//		Real Engines 1-99 (DCS 2-100)    	(99)
//		All Engine Operation 100 (DCS 101)	(1)
//		LashUps 101-120 (always DCS #102)	(20)  -- total of 121 entries indexed as 0-120
// initialized to NULL in the MainForm Constructor

TVolume * ptrVolume[TIUMAX][ABSOLUTEMAX];	// Volume Control Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
TOp * ptrOp[TIUMAX][ABSOLUTEMAX];   		// Operation Control Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
char * ptrLUEngineList[TIUMAX][ABSOLUTEMAX];	// LashUp Engine List Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
char * ptrLUNameList[TIUMAX][ABSOLUTEMAX];		// LashUp Engine Name Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
TAIUControl * ptrAIUControl[TIUMAX][AIUMAX];	// AIU Control Pointers for 25 AIU - indexed by TIUNo-1 and AIUNo-1
struct AIU_EMU AIUEmulator[TIUMAX];				// Flags to indicate an AIU Emulator is attached to an emulated TIU
struct MSG_DISPLAY MessageDisplay[100];          // Flags to indicate a message display is available
int AIUCount[TIUMAX];							// Number of AIU connected to each TIU
TConvMode * ptrConvMode[TIUMAX][CONVMAX];		// Conventional Mode Pointers, 5 TIU, 2 channels - indexed by TIUNo-1 and ChannelNo-1
TDCS_Setup * ptrDCS[TIUMAX];			// DCS Setup Pointers, 5 TIU - indexed by TIUNo-1
TLashUp * ptrLashUp[TIUMAX];			// LashUp Pointers, 5 TIU - indexed by TIUNo-1
TQuickCtrl * ptrQuick[TIUMAX];			// Quick Control Pointers, 5 TIU, indexed by TIUNo-1
TStringList * ptrEngDat[TIUMAX];		// 5 pointers for 5 TIU's worth of Engine data
TStringList * ptrLashupDat[TIUMAX];		// 5 pointers for 5 TIU's worth of Lashup data
// TODO: TMCC code needs to be completed
// #define TMCCENGINEMAX 100
//TTMCCOp * ptrTMCCOp[TIUMAX][TMCCENGINEMAX];	// TMCC Operation Control Pointers - indexed by actual engine number, 1-99 index 0 is not used
THot_Command * ptrHot_Command[HOTPAGEMAX];		// Hot Button Pointers, up to 5 windows for 100 buttons
TEngineInfo * ptrEngineInfo[TIUMAX][ABSOLUTEMAX];		// Engine Information Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
TEngineStatus * ptrEngineStatus[TIUMAX][ABSOLUTEMAX];	// Engine Status Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
TTrackSignal * ptrTrackSignal[TIUMAX][ABSOLUTEMAX];	// Track Signal Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine
TCommThread * ptrTIUCommThread;				// pointer to TIU communications thread
TDispatch_List_Thread * ptrDL_Thread;		// pointer to Dispatch List Thread
TThread * PBThread;							// pointer to Setup's playback thread
TThread * ptrPBThread[20 * HOTPAGEMAX];		// pointers to Hot Com's playback threads
TProgram_Control *ptrProgram_Control[PCMAX];	// pointer(s) to Program Control Windows
TThreadList * Event_FIFO[PCMAX];	    	//  FIFO Event List for Program Control Threads - Thread Safe
TAbNumEdit *ptrPC_ETime[PCMAX];				// pointer(s) to ETime control in each Program Control window
int PC_Tenths;								// RunTimeinTenths counter (system wide)
int ptrPC_Tenths[PCMAX];					// RunTimeinTenths counter in each Program Control window
TMyTimerThread * ptrTimer_Thread;			// pointer to single Timer Thread
HANDLE hRFIDReqEvent[PCMAX];				// pointer(s) to RFID Events created by PC threads
//---------------------------------------------------------------------------
// Signal number of each of two cab signals
int iCabSignal[PCMAX][2];					// Signal number of each of two cab signals in PC Control window
											// [0] is HOMESIGNAL, [1] is DISTANTSIGNAL
//---------------------------------------------------------------------------
// ComboBox filename for saved items
String TagLocationsFileName = "TagLocations.txt";
String EngineTypesFileName = "EngineTypes.txt";
String CarModelsFileName = "CarModels.txt";
String EngineModelsFileName = "EngineModels.txt";
String RailroadNamesFileName = "RailroadNames.txt";


//-------------------------   Constructor -----------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
	: TForm(Owner)
{
int DebugCommandLine = 0;
MessagePopup = NULL;	// make sure the pointer is NULL
iSS = 1;		// Splash Window if == 1
// shrink MainForm
// First save the larger debug window size
SavedDebugWidth = Width;
SavedDebugHeight = Height;
// Calculate the smaller normal operation window size
iSavedWidth = ESTOP->Left + ESTOP->Width + 40; // 596;
iSavedHeight = ESTOP->Top + ESTOP->Height + StatusBar->Height + 42; // 324;
// Set the window to the smaller size
Width = iSavedWidth;
Height = iSavedHeight;
// Status Bar widths
StatusBar->Panels->Items[RADIO_PORT_STATUS]->Width = Width / 3;
StatusBar->Panels->Items[TIU_STATUS]->Width = Width / 3;
// make sure Debug flag is cleared
bDebug = false;
// Insert compile date and time into *.exe for information
String MyDT = __DATE__ ;
MyDT += " - ";
MyDT += __TIME__;
//
#if __BORLANDC__ >= 0x740
Date_Time->Caption = MyDT + " (CB 10.3 0x" + IntToHex(__TCPLUSPLUS__, 4) + ")";	// CB10.3 Rio
#elif __BORLANDC__ >= 0x720
Date_Time->Caption = MyDT + " (CB 10.1 0x" + IntToHex(__TCPLUSPLUS__, 4) + ")";	// CB10.1 Berlin
#elif __BORLANDC__ >= 0x560
Date_Time->Caption = MyDT + " (BCB 6.0 0x" + IntToHex(__TCPLUSPLUS__, 4) + ")";	// BCB 6.0
#else
Date_Time->Caption = MyDT + " (BCB 4.0 0x" + IntToHex(__TCPLUSPLUS__, 4) + ")";	// BCB 4.0
#endif

Application->ShowHint = true;			// global hint enable but does not work for menus
LocalTransparencyOverride = false;
MsgNum = 0;								// initialize message number counter

// This has no effect in BCB4.0 and Thread view not available in CB10.1
// It does work on BCB6.0 and shows the thread name in the Thread View window
// http://docwiki.embarcadero.com/RADStudio/Rio/en/Converting_an_Unnamed_Thread_to_a_Named_Thread
THREADNAME_INFO info;
info.dwType = 0x1000;
info.szName = "VCL_Thread";
info.dwThreadID = -1;
info.dwFlags = 0;
__try {
	RaiseException(0x406D1388, 0, sizeof(info) / sizeof(DWORD), (DWORD*)&info);
	}
__except (EXCEPTION_CONTINUE_EXECUTION) {
	}

// NameThreadForDebugging is not available in BCB4.0 or BCB6.0
// http://docwiki.embarcadero.com/Libraries/Rio/en/System.Classes.TThread.NameThreadForDebugging

#ifdef DEBUGTESTCODE
// https://docs.microsoft.com/en-us/windows/win32/api/debugapi/nf-debugapi-isdebuggerpresent
if (IsDebuggerPresent()) {
	// https://docs.microsoft.com/en-us/windows/win32/api/debugapi/nf-debugapi-outputdebugstringa
    OutputDebugString(TEXT("Break in MainForm constructor"));
	// https://docs.microsoft.com/en-us/windows/win32/api/debugapi/nf-debugapi-debugbreak
    DebugBreak();
    }
#endif

//
// generate the version number and put it in the Status Bar & popup menu
//
DWORD  verHandle = NULL;
UINT   size      = 0;
LPBYTE lpBuffer  = NULL;
DWORD  verSize   = GetFileVersionInfoSize( Application->ExeName.c_str(), &verHandle);

if (verSize != NULL) {
    LPSTR verData = new char[verSize];	// deleted below
    assert (verData != NULL);
	if (GetFileVersionInfo( Application->ExeName.c_str(), verHandle, verSize, verData)) {
        if (VerQueryValue(verData,TEXT("\\"),(VOID FAR* FAR*)&lpBuffer,&size)) {
            if (size) {
                VS_FIXEDFILEINFO *verInfo = (VS_FIXEDFILEINFO *)lpBuffer;
                if (verInfo->dwSignature == 0xfeef04bd) {
                    // Doesn't matter if you are on 32 bit or 64 bit,
                    // DWORD is always 32 bits, so first two revision numbers
                    // come from dwFileVersionMS, last two come from dwFileVersionLS
                    _sntprintf(Version, sizeof Version, TEXT("V%d.%d.%d.%d (%s)"),
                    ( verInfo->dwFileVersionMS >> 16 ) & 0xffff,
                    ( verInfo->dwFileVersionMS >>  0 ) & 0xffff,
                    ( verInfo->dwFileVersionLS >> 16 ) & 0xffff,
                    ( verInfo->dwFileVersionLS >>  0 ) & 0xffff,
                    TEXT(__DATE__));
                    StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = Version;
                    AboutRTC->Caption = "About RTC " + StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text;
                    Caption = Caption + " - " + StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text;
                	}
            	}
        	}
    	}
    delete[] verData;	// this is the only place where verData is deleted
	} else {
    // Retrieve the system error message for the last-error code
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError();

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );
    // Display the error message
    MessageBox(NULL, (LPCTSTR)lpMsgBuf, TEXT("Error"), MB_OK);
	LocalFree(lpMsgBuf);
    }

getcwd(cStartDir, MAX_PATH);	// Directory for *.ini and Car/Engine/Railroad files
Last_Response->Text = cStartDir;	// cStartDir never changes

WorkingDir =  cStartDir;		// Directory for save/load file diaglogs
								// WorkingDir changes depending on the user changing folders in the dialog boxes
                                // WorkingDir is saved in the registry on exit
CTS_Count = 0;		// TIU Port Signal counters
DSR_Count = 0;
CD_Count = 0;
LastHotComPage = 1;
// User Registry Data
RegFileRootKey = "Software\\Remote Control\\Serial";
// Counters
CTS_Counter->Text = String(CTS_Count);
DSR_Counter->Text = String(DSR_Count);
CD_Counter->Text = String(CD_Count);
// Buttons
DisconnectButton->Enabled = false;
ConnectButton->Enabled = true;
ConnectButton->Caption = "&Connect";
ExitButton->Enabled = true;
Send_ASCII_TIU->Enabled = false;
Send_ASCII_RFID->Enabled = false;
Send_Hex_TIU->Enabled = false;
Send_Hex_RFID->Enabled = false;

// All controls disabled on start
SoundCtrlsButton->Enabled = false;
OpButton->Enabled = false;
StartupButton->Enabled = false;
QuickButton->Enabled = false;
Var1ModeButton->Enabled = false;
Var2ModeButton->Enabled = false;
ShutdownButton->Enabled = false;
StopAllButton->Enabled = false;
ReadEnginesButton->Enabled = false;

// AIU Button enabled at start
AIUButton->Enabled = true;

// PC Button enabled on start
Prog_Control_Button->Enabled = true;

// DTR and RTS disabled on start
RTSSwitch->Enabled = false;
DTRSwitch->Enabled = false;

STOPem->Enabled = false;
StartedConsists->Value = 0;
CommandDecoded = false;

AltKeyState = ShiftKeyState = CtrlKeyState = 0;

// RFID Port
//iRFIDbufCount = 0;	// RFID Port Byte Counter

// set max TIU and AIU Spinners
TIUSpin->MaxValue = TIUMAX;
AIUSpin->MaxValue = AIUMAX;
RemoteSpin->MaxValue = REMOTEMAX;

// check for command line parameters <TIUNo> <RemoteNo>
ProgName->Caption = ParamStr(0);
int PCount = ParamCount();
TIUNo = RemoteNo = EngineNo = -1;
try {
	if (PCount > 0) for (int i = 1 ; i <= PCount ; i++) {
        if (i == 1 && ParamStr(i)[1] != '/') {
        	TIUNo = ParamStr(1).ToInt(); 		// TIU Number
            if (TIUNo < 1 || TIUNo > TIUMAX) throw(10);	// TIU 1 to 5
            TIUSpin->Value = TIUNo;
            continue;
            }
        if (i == 2 && ParamStr(i)[1] != '/') {
        	RemoteNo = ParamStr(2).ToInt();		// Remote Number
            if (RemoteNo < 0 || RemoteNo > REMOTEMAX) throw(11);	// Remote 0 to 15
            RemoteSpin->Value = RemoteNo;
            continue;
            }
        // See if it's a parameter starting with /
        if (ParamStr(i)[1] == '/') {
        	// "/d" Debug
            if (ParamStr(i)[2] == 'd') {
            	DebugCommandLine++;
                }
            //
            // "/iZZZZZ" Identity used to access Registry for Program settings
            if (ParamStr(i)[2] == 'i') {
            	RegFileRootKey = RegFileRootKey + "-" + ParamStr(i).SubString(3,10);
                }
            //
            continue;
            }
		}
    }
catch(int) {
    ShowMessage("Invalid Command Line Parameter(s)" + String("\n") + "Should be two numbers - TIU # (1 to 5) and Remote # (0 to 15)");
    Application->Terminate();
	}
catch(...) {
    ShowMessage("Invalid Command Line Parameter(s)" + String("\n") + "Should be TIU # and Remote # separated by a space");
    Application->Terminate();
	}

// read Preferences from registry if they exist
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);			// make sure that construction worked
// display for debug
Last_Message->Text = RegFileRootKey;
__try {
    if (TIUNo < 0) { 	// not given on command line so read it from the Registry
		TIUSpin->Value = RegFile->ReadInteger("Options", "TIU", 1);	// causes TIUNo to change (does not cause OnChange event to fire)
		TIUNo = TIUSpin->Value;
        assert(TIUNo >= 1 && TIUNo <= TIUMAX);
        }
    if (RemoteNo < 0) { 	// not given on command line so read it from the Registry
	    RemoteSpin->Value = RegFile->ReadInteger("Options", "Remote", REMOTEMAX);
	    RemoteNo = RemoteSpin->Value;
        assert(RemoteNo >= 0 && RemoteNo <= REMOTEMAX);
        }
	//
    // Engine No and AIUNo loaded by LoadTIUData() on a TIU by TIU basis
    //
    // Options common over all TIU
	PlainTextFile->Checked = RegFile->ReadInteger("Options", "PlainTextFile", PlainTextFile->Checked);
	TIULog->Checked = RegFile->ReadInteger("Options", "TIULog", TIULog->Checked);
	RFIDLog->Checked = RegFile->ReadInteger("Options", "RFIDLog", RFIDLog->Checked);
	CRCSel->ItemIndex = RegFile->ReadInteger("Options", "CRCSel", CRCSel->ItemIndex);
    WorkingDir = RegFile->ReadString("Options","WorkingDir" + String(IntToHex(__TCPLUSPLUS__, 4)), WorkingDir);
    iSS = RegFile->ReadInteger("Options", "SS", iSS);	// this gets set to 0 when you click on RTCInfo Button
    if (iSS == 0) AboutRTC->Checked = true;
	// Engine RAM Dump
    LastLine->ValueAsInt = RegFile->ReadInteger("Debug", "LastLine", LastLine->ValueAsInt);
    FirstLine->ValueAsInt = RegFile->ReadInteger("Debug", "FirstLine", FirstLine->ValueAsInt);
    LastHotComPage = RegFile->ReadInteger("Options", "LastHotComPage", LastHotComPage);
    }
__finally {
	// Deallocate RegFile
	delete RegFile;
    }
//
if (iSS) {
	// Splash Window
    Splash = new TSplash(Application);
    Splash->Memo1->Lines->Strings[0] = "Remote Train Control - " + String(Version);
    Splash->Show();			// Show Splash Window
    Splash->Update();		// Paint Splash Window
    }


// Load strings into Combo Boxes if file exists
if (FileExists(String(cStartDir) + "\\" + CarModelsFileName)) {
	CarModelComboBox->Items->LoadFromFile(String(cStartDir) + "\\" + CarModelsFileName);
	CarModelComboBox->ItemIndex = 0;
	}
if (FileExists(String(cStartDir) + "\\" + EngineModelsFileName)) {
	EngineModelComboBox->Items->LoadFromFile(String(cStartDir) + "\\" + EngineModelsFileName);
	EngineModelComboBox->ItemIndex = 0;
	}
if (FileExists(String(cStartDir) + "\\" + RailroadNamesFileName)) {
	RailroadNameComboBox->Items->LoadFromFile(String(cStartDir) + "\\" + RailroadNamesFileName);
	RailroadNameComboBox->ItemIndex = 0;
	}
if (FileExists(String(cStartDir) + "\\" + TagLocationsFileName)) {
	TagLocationComboBox->Items->LoadFromFile(String(cStartDir) + "\\" + TagLocationsFileName);
	TagLocationComboBox->ItemIndex = 0;
	}
if (FileExists(String(cStartDir) + "\\" + EngineTypesFileName)) {
	EngineTypeComboBox->Items->LoadFromFile(String(cStartDir) + "\\" + EngineTypesFileName);
	EngineTypeComboBox->ItemIndex = 0;
	}

ConnectMsg = "Press the [Connect] button before performing this operation";
DisconnectMsg = "Press the [Disconnect] button before performing this operation";
TXMsg = "Two Way Communications is required before performing this operation";
bRecording = bPlayingBack = bPlaybackCommError = AllEngineMode = false;
iPlayingBackCounter = 0;		// indicate no playback is taking place
bEmergencyStop = false;


// clear pointer arrays and data
for (int iTIU = 1; iTIU <= TIUMAX ; iTIU++) { // TIU 1-5
	bEngineDataCorrupt[iTIU-1] = bTIUDataLoaded[iTIU-1] = false;
    AIUCount[iTIU-1] = AIUMAX;
    ptrEngDat[iTIU-1] = NULL;
    ptrLashupDat[iTIU-1] = NULL;
    AIUEmulator[iTIU-1].Found = false;
	for (int iEng = 0; iEng < ABSOLUTEMAX ; iEng++) {  	// Engine 0 to 120 - factory reset to end of LashUps
		ptrVolume[iTIU-1][iEng] = NULL;
		ptrOp[iTIU-1][iEng] = NULL;
        ptrEngineInfo[iTIU-1][iEng] = NULL;
        ptrTrackSignal[iTIU-1][iEng] = NULL;
        ptrEngineStatus[iTIU-1][iEng] = NULL;
    	ptrLUEngineList[iTIU-1][iEng] = NULL;
       	ptrLUNameList[iTIU-1][iEng] = NULL;
    	}
    }

// clear pointer arrays for each TIU/AIU pair & each Conventional channel, DCS and Lashup
for (int iTIU = 1; iTIU <= TIUMAX ; iTIU++) { // TIU 1-5
	ptrConvMode[iTIU-1][0] = NULL;	// Var Channel 1
	ptrConvMode[iTIU-1][1] = NULL;	// Var Channel 2
    ptrDCS[iTIU-1] = NULL;
    ptrLashUp[iTIU-1] = NULL;
    ptrQuick[iTIU-1] = NULL;
   	for (int iAIU=1 ; iAIU <= AIUMAX ; iAIU++) {   // AIU 1-5
       	ptrAIUControl[iTIU-1][iAIU-1] = NULL;
        }
    }
// clear pointer arrays for Program Control windows
for (int iPCNo = 1; iPCNo <= PCMAX ; iPCNo++) { // windows 1-5
    ptrProgram_Control[iPCNo - 1] = NULL;
    ptrPC_ETime[iPCNo - 1] = NULL;
    ptrPC_Tenths[iPCNo - 1] = 0;	// clear the Timer Thread's local elapsed times
    hRFIDReqEvent[iPCNo - 1] = NULL;
    Event_FIFO[iPCNo - 1] = NULL;
    iCabSignal[iPCNo - 1][HOMESIGNAL] = iCabSignal[iPCNo - 1][DISTANTSIGNAL] = 0;	// Home and Distant Cab signal numbers
	}

PC_Tenths = 0;	// clear the Timer Thread's global time

for (int iHot = 1; iHot <= HOTPAGEMAX ; iHot++) { // Hot Button pages
    ptrHot_Command[iHot-1] = NULL;
	}

ptrTIUCommThread = NULL;
ptrDL_Thread = NULL;

// Fill the TIU specific data from a disk file
LoadTIUData(EngineData, TIUNo);	// First Time load of EngineData from disk EngineDataX.ini file

EngineData->DefaultColWidth = 30;		// default
EngineData->ColWidths[ENG_NAME] = EngineData->Width - EngineData->DefaultColWidth;	// lots of space for the engine name

// create the Setup window
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Setup Form should be manually created
if (!Setup) {	// since I used the C++ pointer, it will get automatically deleted at close
	Setup = new TSetup(this, TIUNo, RemoteNo);     // Pass "this" as the owner
	assert(Setup != NULL);			// make sure that construction worked
    //Setup->COMSelGroup->Top = 16;	// Serial port box position for non-debug mode
	//Setup->SerialPortListGroup->Top = 16;
   	}
if (Setup->DisableEmergencyStopButton->Checked) {
	ESTOP->Enabled = false;
    ESTOP->Hint = "Disabled in Setup";
    ESTOP->Cursor = crNo;
    }

bSaveRegistryInfo = true;	// Normally we save Registry info in the Registry - except
							// when user clicked Reset Preferences button in Debug Mode
// in RFID Com mode, enable the serial port select
Setup->RFIDNum->Enabled = Setup->RFID_COM_Mode->Checked;
Setup->RFIDLabel->Enabled = Setup->RFID_COM_Mode->Checked;
// in Radio Com mode, enable the serial port select
Setup->PortNum->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->COMLabel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->COMCheckListBox->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->BaudRateSel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->ParitySel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->BitsSel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->StopSel->Enabled = Setup->Radio_COM_Mode->Checked;
// in Radio WiFi mode, enable the PortIP edit box
Setup->RadioIP->Enabled = Setup->Radio_WiFi_Mode->Checked;
Setup->RadioPort->Enabled = Setup->Radio_WiFi_Mode->Checked;

// Process what was found on the command line
if (DebugCommandLine) DebugButton->OnClick(DebugButton);

//I don't know if this is needed or not
/*
// Enable the Undo menu item if the RichEdit can handle it
if (SendMessage(RFIDBuffer->Handle, EM_CANUNDO, 0, 0))
	UndoRFIDBuffer->Enabled = true;
    else
    UndoRFIDBuffer->Enabled = false;
// Enable the Undo menu item if the RichEdit can handle it
if (SendMessage(TIUBuffer->Handle, EM_CANUNDO, 0, 0))
	UndoTIUBuffer->Enabled = true;
    else
    UndoTIUBuffer->Enabled = false;
// Enable the Undo menu item if the RichEdit can handle it
if (SendMessage(CRCBuffer->Handle, EM_CANUNDO, 0, 0))
	UndoCRCBuffer->Enabled = true;
    else
    UndoCRCBuffer->Enabled = false;
*/

// expand the size of all RichEdit controls
//https://stackoverflow.com/questions/6049053/increase-size-limit-of-trichedit
RFIDBuffer->MaxLength = System::MaxInt-2;
TIUBuffer->MaxLength = System::MaxInt-2;
CRCBuffer->MaxLength = System::MaxInt-2;
RecordingBuffer->MaxLength = System::MaxInt-2;

//
#if __BORLANDC__ < 0x720
// Workaround BCB4.0 for SetLayeredWindowAttributes() which enables transparent windows
/*Load dll that holds the API func*/
HMODULE hUser32 = GetModuleHandle("USER32.DLL");
/*Grab pointer to API*/
m_pSetLayeredWindowAttributes = (lpfnSetLayeredWindowAttributes)GetProcAddress(hUser32, "SetLayeredWindowAttributes");
if (NULL == m_pSetLayeredWindowAttributes) {
	MessageBox(NULL,"Error!","SetLayeredWindowAttributes Not Found",MB_OK);
	return;
	}
#endif
//
EngineData->ScrollBars = ssVertical;	// sometimes we lose this setting
//
//
// Experimental - use Drop Down Menu style in CB10.1
#if __BORLANDC__ >= 0x720
// bsSplitButton defined in Vcl.StdCtrls.hpp is really not defined
enum DECLSPEC_DENUM TButtonStyle : unsigned char { bsPushButton, bsCommandLink, bsSplitButton };
Prog_Control_Button->Style = bsSplitButton;
Prog_Control_Button->DropDownMenu = PCPopupMenu;
AIUButton->Style = bsSplitButton;
AIUButton->DropDownMenu = TIUPopupMenu;
StartupButton->Style = bsSplitButton;
StartupButton->DropDownMenu = StartupPopupMenu;
HotCommandButton->Style = bsSplitButton;
HotCommandButton->DropDownMenu = HotButtonPopup;
ReadEnginesButton->Style = bsSplitButton;
ReadEnginesButton->DropDownMenu = ReadPopupMenu;
#endif
// -- create Dispatch List request event
DispatchListReqEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
assert(DispatchListReqEvent != NULL);
// -- Create terminate events
// -- create Terminate Comm Thread Event
TerminationCTEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
assert(TerminationCTEvent != NULL);
// -- create Terminate Dispatch List Thread Event
TerminationDLEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
assert(TerminationDLEvent != NULL);
// -- create Terminate Playback Thread Event - this event will terminate any (and all) Playback threads
TerminationPBEvent = CreateEvent(NULL, false, false, NULL);	// Auto Reset - nonsignaled initial state
assert(TerminationPBEvent != NULL);
//---------------------------------------------------------------------------
// Critical Section to protect Program Control window's Flags and Global integers & strings
InitializeCriticalSection(&Global_Flag_CS);
// Critical Section to protect the Fixed Signals
InitializeCriticalSection(&Signals_CS);
//---------------------------------------------------------------------------
//
// Process the Configuration Script
//
// Lua State Pointer
M = NULL;
if (M != NULL) {
	lua_close(M);
    M = NULL;
    }

// initialization
AnsiString ConfigFileName = String(cStartDir) + "\\" + "config.lua";

if (FileExists(ConfigFileName)) {
    M = luaL_newstate();
    if (M == NULL) {
        if (DEBUGGING) DB->Debug_Print("Cannot create state: not enough memory");
        return;
        }
    luaL_openlibs(M);
    // Add the WorkingDir at the end of the Lua package path
    {
    AnsiString PathString = "package.path = package.path .. [[;" + WorkingDir + "?.lua]]";
    boolean bretv = luaL_dostring(M, PathString.c_str());
    assert(!bretv);	// fail on error
    }
    // compile and execute file
    ConfigurationSuccess = false;

    int load_stat = luaL_loadfile(M, ConfigFileName.c_str());	// load the chunk
    // Process the Configuration file if it is there
    if (load_stat == LUA_OK) {
        // Config file found, now run it
        int retv = lua_pcall(M, 0, 0, 0);	// run the chunk
        if (retv != LUA_OK) {	// LUA_ERRRUN, LUA_ERRMEM, LUA_ERRERR, LUA_ERRGCMM
            // Failed to run
            ShowMessage("Error " + String(retv) + " : " + String(lua_tostring(M, -1)));
            if (M != NULL) {
                lua_close(M);
                M = NULL;
                }
            }
    //
    // Load Configuration values
    //
        if (retv == LUA_OK) {
            ConfigurationSuccess = true;
            if (lua_getglobal(MainForm->M, "Engine_List_Title") == LUA_TSTRING) {
                Engine_List_Label->Caption = lua_tolstring(MainForm->M, -1, NULL);	// Engine List Title
                }
            lua_pop(M,1);	// remove global value
            if (lua_getglobal(MainForm->M, "Working_Directory") == LUA_TSTRING) {
                WorkingDir = lua_tolstring(MainForm->M, -1, NULL);					// Working Directory
                }
            lua_pop(M,1);	// remove global value
            if (lua_getglobal(MainForm->M, "ConnectMsg") == LUA_TSTRING) {
                ConnectMsg = lua_tolstring(MainForm->M, -1, NULL);					// ConnectMsg
                }
            lua_pop(M,1);	// remove global value
            if (lua_getglobal(MainForm->M, "DisconnectMsg") == LUA_TSTRING) {
                DisconnectMsg = lua_tolstring(MainForm->M, -1, NULL);				// DisconnectMsg
                }
            lua_pop(M,1);	// remove global value

            assert(Setup);	// Setup window was constructed above
            //
            // Load Configuration Variables into Setup Variables
            //
            if (lua_getglobal(MainForm->M, "TIU") == LUA_TNUMBER) {
                TIUSpin->Value = lua_tonumber(MainForm->M, -1);			// TIU Number
                }
            lua_pop(M,1);	// remove global

            if (lua_getglobal(MainForm->M, "Remote") == LUA_TNUMBER) {
                RemoteSpin->Value = lua_tonumber(MainForm->M, -1);		// Remote Number
                }
            lua_pop(M,1);	// remove global

            if (lua_getglobal(MainForm->M, "Engine") == LUA_TNUMBER) {
                EngineSpin->Value = lua_tonumber(MainForm->M, -1);		// Engine Number
                }
            lua_pop(M,1);	// remove global

            #define CONFIG_SETUP_VALUE(VAR) \
                if (lua_getglobal(MainForm->M, #VAR) == LUA_TNUMBER) { \
                    Setup->VAR->Value = lua_tonumber(MainForm->M, -1); \
                    } \
                lua_pop(M,1);	// remove global

            CONFIG_SETUP_VALUE(PortNum);		// Radio Comm Port
            CONFIG_SETUP_VALUE(RFIDNum);		// RFID Comm Port
            CONFIG_SETUP_VALUE(MaxSpeed);		// Maximum Speed
            CONFIG_SETUP_VALUE(Quick_Ctrl_Scaling);	// Quick Ctrl Scaling
            CONFIG_SETUP_VALUE(TransparentSpin);	// Transparency
            CONFIG_SETUP_VALUE(MPHScaling);			// Speed Scaling
            CONFIG_SETUP_VALUE(NumRetries);			// Number of Retries

            #define CONFIG_SETUP_CHECKBOX(VAR) \
                if (lua_getglobal(MainForm->M, #VAR) == LUA_TBOOLEAN) { \
                    Setup->VAR->Checked = lua_toboolean(MainForm->M, -1);  \
                    } \
                lua_pop(M,1);	// remove global

            CONFIG_SETUP_CHECKBOX(AutoConnect); 	// Auto Connect to Comm Ports and TIU - true or false
            CONFIG_SETUP_CHECKBOX(RFID_COM_Mode);	// RFID Serial COM enabled - true or false
            CONFIG_SETUP_CHECKBOX(RFID_WiFi_Mode);	// RFID WiFi enabled - true or false
            CONFIG_SETUP_CHECKBOX(Radio_COM_Mode);	// Radio Serial COM enabled - true or false
            CONFIG_SETUP_CHECKBOX(Radio_WiFi_Mode);	// Radio WiFi enabled - true or false
            CONFIG_SETUP_CHECKBOX(AutoCoupler);		// Auto Coupler Slack Sound
            CONFIG_SETUP_CHECKBOX(AutoBell);	// Auto Bell Sounds
            CONFIG_SETUP_CHECKBOX(AutoWhistle);	// Auto Whistle Sounds
            CONFIG_SETUP_CHECKBOX(ShortBellWhistleSequence);		//	Short Bell/Whistle Sequence
            CONFIG_SETUP_CHECKBOX(ExtendedSUSD);	// Extended Startup and Shutdown Sounds
            CONFIG_SETUP_CHECKBOX(StandardAccDec);	// Use Acc/Dec = 1
            CONFIG_SETUP_CHECKBOX(DisableEmergencyStopButton);		// Disable Emergency Stop
            CONFIG_SETUP_CHECKBOX(AutoHotButtons);		// Auto Hot Buttons
            CONFIG_SETUP_CHECKBOX(AutoQuickControls);	// Auto Quick Controls
            CONFIG_SETUP_CHECKBOX(AutoSoundCtrls);		// Auto Sound Ctrls
            CONFIG_SETUP_CHECKBOX(SmokeOFFatStartup);	// Smoke off at Engine Startup
            CONFIG_SETUP_CHECKBOX(EncodedRecordFormat);	// Encoded Record Format
            CONFIG_SETUP_CHECKBOX(RepeatPlayback);		// Repeat Playback
            CONFIG_SETUP_CHECKBOX(Safety_Mode);			// enables double switch throws, whistle off and bell off in safety mode
            CONFIG_SETUP_CHECKBOX(GeneralSounds);		// Allow RTC to generate sound effects
            CONFIG_SETUP_CHECKBOX(NoSleep);		        // Prevent PC from going to Sleep
            CONFIG_SETUP_CHECKBOX(Radio_WiFi_Multicast);	// enables TX WiFi Multicast

            #define CONFIG_SETUP_BOOLEAN(VAR) \
                if (lua_getglobal(MainForm->M, #VAR) == LUA_TBOOLEAN) { \
                    Setup->VAR = lua_toboolean(MainForm->M, -1);  \
                    } \
                lua_pop(M,1);	// remove global

            CONFIG_SETUP_BOOLEAN(Show_Pause_Button);	// Show Pause Button on Program Control window
            CONFIG_SETUP_BOOLEAN(Comm_Not_Required);	// enables some windows even without communications - use with care

            #define CONFIG_SETUP_STRING(VAR) \
                if (lua_getglobal(MainForm->M, #VAR) == LUA_TSTRING) { \
                    Setup->VAR = lua_tostring(MainForm->M, -1);  \
                    } \
                lua_pop(M,1);	// remove global

            Setup->SwitchStraight = "Normal";
            Setup->SwitchDiverge = "Reverse";
            CONFIG_SETUP_STRING(SwitchStraight);	// word used for Switch when in the Normal position
            CONFIG_SETUP_STRING(SwitchDiverge);		// word used for Switch when in the Reverse position

             #define CONFIG_SETUP_TEXT(VAR) \
                if (lua_getglobal(MainForm->M, #VAR) == LUA_TBOOLEAN) { \
                    Setup->VAR->Text = lua_tostring(MainForm->M, -1);  \
                    } \
                lua_pop(M,1);	// remove global

            CONFIG_SETUP_TEXT(RadioIP);			// IP of the Modem WiFi/AIU Radios (Unicast or Multicast)
            CONFIG_SETUP_TEXT(RadioPort);		// Port of the Modem WiFi/AIU Radios
            }
        } else {
        // Config file not found - if we are debugging, then show a message otherwise continue on without one
        if (DEBUGGING) DB->Debug_Print("Error " + String(load_stat) + " : " + String(lua_tostring(M, -1)));
        if (M != NULL) {
            lua_close(M);
            M = NULL;
            }
        }
    }
//---------------------------------------------------------------------------
// TODO Command Line Processor - not debugged
//
//#define ENABLE_COMMAND_LINE
#ifdef ENABLE_COMMAND_LINE
/*
 ---- NOT DEBUGGED
assert(falsex);	// Command Line code not used - cause a compile error
// Check for command line *.rtc files
const int NumDrop = ParamCount();
if (NumDrop > 0) {
	// Connect
    ConnectButton->OnClick(ConnectButton);   // note: this calls Debug which is not constructed yet
	for (int i = 1; i <= NumDrop; i++) {
       	Application->ProcessMessages();
        Repaint();
        String Parameter = ParamStr(i);
        String DisplayPathName = Parameter;  // ShellGetDisplayPathName(Parameter);
        String FilePath = ExtractFilePath(DisplayPathName);
		// TODO needs to be completly rewritten after moving playback into a thread
        if (FileExists(DisplayPathName)) {
			// it's a filename
            if (ExtractFileExt(DisplayPathName) == ".rtc") { // is it a RTC recorded file?
				// queue it up for playing
                RecordingBuffer->Clear();
                RecordingBuffer->Lines->LoadFromFile(DisplayPathName);
                // start timer and playback
				PlaybackTimer->Interval = 1000;	// 1 second
				PlaybackTimer->Tag = -1;	// indicated just started
				PlaybackTimer->Enabled = true;
				PlayingBack = true;
                PlayingBackCounter++;
                PlaybackCommError = false;
				Setup->PlaybackStartButton->Enabled = false;
				Setup->PlaybackStopButton->Enabled = true;
				Setup->RecordStartButton->Enabled = false;
				if (DEBUGGING) DB->Debug_Print("Command Line Playback : " + DisplayPathName);
            	}
            }
        }
    // Disconnect
    DisconnectButton->OnClick(DisconnectButton);
	}
	*/
#endif
//----------------------------------
//
//Create Program Control popup menu based on PCMAX
//
//PCPopupMenu
// remove any menu items
int PCCount = PCPopupMenu->Items->Count;
if (PCCount > 0) {
	for (int i = 0 ; i < PCCount; i++) {
    	//PCPopupMenu->Items->Delete(0);
        delete PCPopupMenu->Items->Items[0];
		}
    }
    
TMenuItem * M;
// PCMAX count of menu items
// read MRU from registry if they exist
TRegIniFile * const RegFile1 = new TRegIniFile(RegFileRootKey);
assert(RegFile1 != NULL);
__try {
    for (int PCNo = 1 ; PCNo <= PCMAX ; PCNo++) {
        M = new TMenuItem(PCPopupMenu);
        M->Name = "ProgramControl" + String(PCNo);
        int ItemIndex = RegFile1->ReadInteger("PC" + String(PCNo), "MRUItemIndex" + String(PCNo), -1);
        if (ItemIndex < 0) ItemIndex = 0;
        String S = RegFile1->ReadString("PC" + String(PCNo), "MRU" + String(PCNo) + String(ItemIndex + 1), "");
   		if (!S.IsEmpty()) {
            M->Caption = "PC#" + String(PCNo) + " - " + ExtractFileName(S);
            M->Hint = S;
			} else {
            M->Caption = "Program Control #" + String(PCNo);
            }
        M->Tag = PCNo;
        M->OnClick = ProgramControlClick;
        PCPopupMenu->Items->Add(M);
        }
    }
__finally {
	// Deallocate RegFile
	delete RegFile1;
	}    

// Separator
M = new TMenuItem(PCPopupMenu);
M->Name = "Separator1";
M->Caption = "-";
PCPopupMenu->Items->Add(M);

// Two menu items "Detected Tags Window" and "Layout Window"
M = new TMenuItem(PCPopupMenu);
M->Name = "ShowDetectedTags";
M->Caption = "Show Detected Tags Window";
M->Enabled = true;
M->Checked = false;
M->Hint = "Show all of the detected tags in a window";
M->OnClick = ShowDetectedTagsClick;
M->ShortCut = ShortCut(Word('D'), TShiftState() << ssCtrl);
PCPopupMenu->Items->Add(M);

M = new TMenuItem(PCPopupMenu);
M->Name = "ShowLayoutWindow";
M->Caption = "Show Layout Window";
M->Enabled = true;
M->Checked = false;
M->Hint = "Show the layout and occupancy flags in a window";
M->OnClick = ShowLayoutWindowClick;
M->ShortCut = ShortCut(Word('L'), TShiftState() << ssCtrl);
PCPopupMenu->Items->Add(M);
//
// Tell MainForm that we updated the Menu
TNotifyEvent();

//----------------------------------
//
//Create AIU Control popup menu based on AIUMAX and TIUMAX
//
#if AIUMAX > 5
AIU6Control1->Visible = true;
#endif
#if AIUMAX > 6
AIU7Control1->Visible = true;
#endif
#if AIUMAX > 7
AIU8Control1->Visible = true;
#endif
#if AIUMAX > 8
AIU9Control1->Visible = true;
#endif
#if AIUMAX > 9
AIU10Control1->Visible = true;
#endif
#if TIUMAX > 5
TIU61->Visible = true;
#endif
#if TIUMAX > 6
TIU71->Visible = true;
#endif
#if TIUMAX > 7
TIU81->Visible = true;
#endif

randomize();	// initialize random number generator for creation of sequence number
Sequence = random(256);	// Global sequence number for transmissions - remote seems to always start at 175
						// this unsigned char rolls 255->0
}
//---------------------------------------------------------------------------

__fastcall TMainForm::~TMainForm(void)
{
CloseHandle(TerminationCTEvent);	// not really needed - the system closes the handle automatically when the process terminates
CloseHandle(TerminationDLEvent);	// not really needed - the system closes the handle automatically when the process terminates
									// TerminationDLEvent clears and frees memory for dispatch list items
CloseHandle(TerminationPBEvent);	// not really needed - the system closes the handle automatically when the process terminates
CloseHandle(DispatchListReqEvent);	// not really needed - the system closes the handle automatically when the process terminates

DeleteCriticalSection(&Global_Flag_CS);
DeleteCriticalSection(&Signals_CS);

/*
// Display ThreadId in debug buffer - NOTE Debug_Controls form is already destroyed
if (DEBUGGING) {
	DB->Debug_Print("TMainForm:~MainThreadID " + IntToHex((int)MainThreadID, 8), clGreen);
   	// MainThreadID and GetCurrentThreadId() return the same value
	//DB->Debug_Print("TMainForm:~ThreadID " + IntToHex((int)GetCurrentThreadId(), 8), clGreen);
    }
*/
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormShow(TObject *Sender)
{
assert(Setup); 	// Setup must be constructed
// This sets the debugging option after DB is constructed
//
// Configuration Script
//
/*
	-- Remote Train Control Configuration File

	-- Enable debugging screens
	Debugging = true	-- Enable
	DBLevel = 0			-- Debug level 0 to 10

*/
// Lua state M is created in the MainForm constructor
if (MainForm->M != NULL) {
	if (lua_getglobal(MainForm->M, "Debugging") == LUA_TBOOLEAN) {
		if (lua_toboolean(MainForm->M, -1)) DebugButtonClick(NULL);	// Show Debug window - true or false
		}
	lua_pop(M,1);	// remove global

	if (lua_getglobal(MainForm->M, "DBLevel") == LUA_TNUMBER) {
		DB->DBLevel->Value = lua_tonumber(MainForm->M, -1);		// Debug Level
		}
	MainForm->GlobalDBLevel = DB->DBLevel->Value;
	lua_pop(M,1);	// remove global
	}

// Auto Connect to the TIU if requested
if (Setup->AutoConnect->Checked) ConnectButton->OnClick(ConnectButton);
// Auto Show the Hot Buttons if requested
if (Setup->AutoHotButtons->Checked) HotButtonClick(HotCommandButton);
//
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CloseAllWindows(void)
{
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

// stop the single timer thread
if (ptrTimer_Thread) {
	ptrTimer_Thread->StopRequest();
    TimerThreadLED->Checked = false;
    }
ptrTimer_Thread = NULL;

// Close manually created form for each TIU/AIU pair and Conventional Channel, DCS setup, and LashUp
try {
	for (int iTIU = 0; iTIU < TIUMAX ; iTIU++) { // TIU 1-5
    	try {
			if (ptrDCS[iTIU] != NULL) {
	    		delete ptrDCS[iTIU];
	    	    ptrDCS[iTIU] = NULL;
	    		}
            }
        catch (...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate DCS Setup Windows");
        	}

    	try {
			if (ptrEngDat[iTIU] != NULL) {
	    		delete ptrEngDat[iTIU];
	    	    ptrEngDat[iTIU] = NULL;
	    		}
			//if (ptrLashupDat[iTIU] != NULL) {
	    	//	delete ptrLashupDat[iTIU];
	    	//    ptrLashupDat[iTIU] = NULL;		// deleted in Lashup_Selection.cpp where it is created
	    	//	}
            }
        catch (...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate Engine or Lashup Data Pointers");
        	}

    	assert(LashUp == NULL);	// this pointer must not be used
        try {
	    	if (ptrLashUp[iTIU] != NULL) {
	    		delete ptrLashUp[iTIU];
	    	    ptrLashUp[iTIU] = NULL;
	    	    }
            }
        catch (...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate Create/Edit LashUp Windows");
        	}
    	assert(QuickCtrl == NULL);	// this pointer must not be used
        try {
	    	if (ptrQuick[iTIU] != NULL) {
	    		delete ptrQuick[iTIU];
	    	    ptrQuick[iTIU] = NULL;
	    	    }
            }
        catch (...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate Quick Control Windows");
        	}
    	assert(ConvMode == NULL);	// this pointer must not be used
        try {
			if (ptrConvMode[iTIU][0] != NULL) {	// Var Channel 1
	    		delete ptrConvMode[iTIU][0];
	    	    ptrConvMode[iTIU][0] = NULL;
	    		}
            }
        catch (...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate Variable 1 Conventional Mode Windows");
        	}
        try {
			if (ptrConvMode[iTIU][1] != NULL) {	// Var Channel 2
	    		delete ptrConvMode[iTIU][1];
	    	    ptrConvMode[iTIU][1] = NULL;
	    		}
            }
        catch(...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate Variable 2 Conventional Mode Windows");
        	}
    	assert(AIUControl == NULL);	// this pointer must not be used
        try {
		   	for (int iAIU=0 ; iAIU < AIUMAX ; iAIU++) { 	// AIU 1-5
		       	if (ptrAIUControl[iTIU][iAIU] != NULL) {
		           	delete ptrAIUControl[iTIU][iAIU];
		            ptrAIUControl[iTIU][iAIU] = NULL;
		            }
		        }
            }
        catch (...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate AIU Windows");
			}
	    }
    }
catch (...) {
    CaughtErrors->Value ++;
    DebugMessage("Failed to deallocate TIU arrays");
	}
// Close hot button windows
try {
	for (int iHot = 0; iHot < HOTPAGEMAX ; iHot++) { // Hot button pages 1-5
        try {
			if (ptrHot_Command[iHot] != NULL) {	// Hot Buttons - up to 5
	    		delete ptrHot_Command[iHot];
	    	    ptrHot_Command[iHot] = NULL;
	    		}
            }
        catch(...) {
	        CaughtErrors->Value ++;
        	DebugMessage("Failed to deallocate Hot Button Screens");
        	}
        }
    }
catch (...) {
    CaughtErrors->Value ++;
    DebugMessage("Failed to deallocate Hot Button Screens");
	}

// Close manually created forms for each LashUp
try {
	for (int iTIU = 0; iTIU < TIUMAX ; iTIU++) { // TIU 1-5
		for (int iEng = LASHUPMIN; iEng < ABSOLUTEMAX ; iEng++) {	// 101 to 120 -  to end of LashUps
			if (ptrVolume[iTIU][iEng] != NULL) {
		       	delete ptrVolume[iTIU][iEng];
		        ptrVolume[iTIU][iEng] = NULL;
		        }
		    if (ptrOp[iTIU][iEng] != NULL) {
		       	delete ptrOp[iTIU][iEng];
		        ptrOp[iTIU][iEng] = NULL;
		        }
		    if (ptrLUEngineList[iTIU][iEng] != NULL) {
		    	delete[] ptrLUEngineList[iTIU][iEng]; 	// pointer to a char array
		        ptrLUEngineList[iTIU][iEng] = NULL;
				}
		    if (ptrLUNameList[iTIU][iEng] != NULL) {
		    	delete[] ptrLUNameList[iTIU][iEng]; 	// pointer to a char array
		        ptrLUNameList[iTIU][iEng] = NULL;
				}
		    }
        }
    }
catch (...) {
    CaughtErrors->Value ++;
    DebugMessage("Failed to deallocate LashUp Operation & Sound Control Screens");
	}

// Close manually created Program Control Forms and Event FIFOs
try {
	for (int iPCNo = 1; iPCNo <= PCMAX ; iPCNo++) { // Program Control Windows 1-PCMAX
		if (ptrProgram_Control[iPCNo-1] != NULL) {
    		delete ptrProgram_Control[iPCNo-1];
        	ptrProgram_Control[iPCNo-1] = NULL;
        	}
		if (Event_FIFO[iPCNo-1] != NULL) {
    		delete Event_FIFO[iPCNo-1];
        	Event_FIFO[iPCNo-1] = NULL;
        	}
        }
    }
catch (...) {
    CaughtErrors->Value ++;
    DebugMessage("Failed to deallocate Program Control Screens");
	}

// Close manually created forms for each Engine
//assert(Volume == NULL);	// this pointer must not be used
//assert(Op == NULL);		// this pointer must not be used
try {
	for (int iTIU = 1; iTIU <= TIUMAX ; iTIU++) { // TIU 1-5
		for (int iEng = 0; iEng < LASHUPMIN ; iEng++) {	// Engine 0 to 100 - factory reset to All Engine Operation
			if (ptrVolume[iTIU-1][iEng] != NULL) {
		       	delete ptrVolume[iTIU-1][iEng];
		        ptrVolume[iTIU-1][iEng] = NULL;
		        }
		    if (ptrOp[iTIU-1][iEng] != NULL) {
		       	delete ptrOp[iTIU-1][iEng];
		        ptrOp[iTIU-1][iEng] = NULL;
		        }
		    if (ptrEngineInfo[iTIU-1][iEng] != NULL) {
		       	delete ptrEngineInfo[iTIU-1][iEng];
		        ptrEngineInfo[iTIU-1][iEng] = NULL;
		        }
		    if (ptrTrackSignal[iTIU-1][iEng] != NULL) {
		       	delete ptrTrackSignal[iTIU-1][iEng];
		        ptrTrackSignal[iTIU-1][iEng] = NULL;
		        }
		    if (ptrEngineStatus[iTIU-1][iEng] != NULL) {
		       	delete ptrEngineStatus[iTIU-1][iEng];
		        ptrEngineStatus[iTIU-1][iEng] = NULL;
		        }
		    if (ptrLUEngineList[iTIU-1][iEng] != NULL) {
	   		 	delete[] ptrLUEngineList[iTIU-1][iEng]; 	// pointer to a char array
	 	        ptrLUEngineList[iTIU-1][iEng] = NULL;
     	        }
		    if (ptrLUNameList[iTIU-1][iEng] != NULL) {
	   		 	delete[] ptrLUNameList[iTIU-1][iEng]; 	// pointer to a char array
	 	        ptrLUNameList[iTIU-1][iEng] = NULL;
     	        }
	    	}
        }
    }
catch (...) {
	CaughtErrors->Value ++;
    DebugMessage("Failed to deallocate Engine Operation and Sound Controls Screens");
	}
Screen->Cursor = OldCursor;
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
// Note: this function does not run when Terminate() is called. Does run when I press the [Exit] button
// which calls Close();
//
// Save the TIU specific data to a disk file
SaveTIUData(EngineData, TIUNo);

if (DisconnectButton->Enabled) DisconnectButton->OnClick(DisconnectButton);	// simulate the user pressing the Disconnect Button

if (bSaveRegistryInfo) {
	// write preferences to the Registry
	TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
	assert(RegFile != NULL);			// make sure that construction worked
	__try {
		RegFile->WriteInteger("Options", "TIU", TIUSpin->Value);
		RegFile->WriteInteger("Options", "Remote", RemoteSpin->Value);
    	RegFile->WriteInteger("Options", "LastHotComPage", LastHotComPage);
	    // AIUNo saved by SaveTIUData();
	    // EngineNo saved by SaveTIUData();

	    // Global over all TIU
		RegFile->WriteInteger("Options", "PlainTextFile", PlainTextFile->Checked);
		RegFile->WriteInteger("Options", "TIULog", TIULog->Checked);
		RegFile->WriteInteger("Options", "RFIDLog", RFIDLog->Checked);
		RegFile->WriteInteger("Options", "CRCSel", CRCSel->ItemIndex);
		RegFile->WriteString("Options", "WorkingDir" + String(IntToHex(__TCPLUSPLUS__, 4)), WorkingDir);
		RegFile->WriteInteger("Options", "SS", iSS);
		// Engine RAM Dump
	    RegFile->WriteInteger("Debug", "LastLine", LastLine->ValueAsInt);
    	RegFile->WriteInteger("Debug", "FirstLine", FirstLine->ValueAsInt);
    	RegFile->WriteInteger("Options", "LastHotComPage", LastHotComPage);
		//
	    }
	__finally {
		// Deallocate RegFile
		delete RegFile;
		}
	}
// Save strings from Combo Boxes to disk unless the files already exist
if (!FileExists(String(cStartDir) + "\\" + CarModelsFileName)) CarModelComboBox->Items->SaveToFile(String(cStartDir) + "\\" + CarModelsFileName);
if (!FileExists(String(cStartDir) + "\\" + EngineModelsFileName)) EngineModelComboBox->Items->SaveToFile(String(cStartDir) + "\\" + EngineModelsFileName);
if (!FileExists(String(cStartDir) + "\\" + RailroadNamesFileName)) RailroadNameComboBox->Items->SaveToFile(String(cStartDir) + "\\" + RailroadNamesFileName);
if (!FileExists(String(cStartDir) + "\\" + EngineTypesFileName)) EngineTypeComboBox->Items->SaveToFile(String(cStartDir) + "\\" + EngineTypesFileName);
if (!FileExists(String(cStartDir) + "\\" + TagLocationsFileName)) TagLocationComboBox->Items->SaveToFile(String(cStartDir) + "\\" + TagLocationsFileName);
//
// close configuration script
if (M != NULL) {
	lua_close(M);
    M = NULL;
    }
//
CloseAllWindows();
//
Action = caFree; //The form is closed and all allocated memory for the form is freed
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ConnectButtonClick(TObject *Sender)
{
//
// Connects to the TIU Serial Port, the RFID Serial Port (if enabled), the RFID WiFi Port (if enabled)
// and then to the TIU.
// If we can't connect to the TIU Serial or TIU WiFi Port then no attempt is made to connect to the TIU
// Starts the mDNS Server
//
// The Connect Button is disabled once the connection is successful
// The Connect Button is enabled at program start and after the Disconnect Button is pressed
//
// All requested connection must be successful for RTC to work
//
assert(Setup != NULL);		// setup must be created
static int retv = 0;
CTS_Count = DSR_Count = CD_Count = 0;
CTS_Counter->Text = String(CTS_Count);
DSR_Counter->Text = String(DSR_Count);
CD_Counter->Text = String(CD_Count);
//iRFIDbufCount = 0;	// reset RFID Serial Port Byte counters
TIUSpin->ColorDefault = clYellow;	// set TIU/AIU boxes color to yellow indicating unknown comm with the TIU
AIUSpin->ColorDefault = clYellow;

// Status bar
StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "";
StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "";
StatusBar->Panels->Items[TIU_STATUS]->Text = "";

if (Setup->Radio_COM_Mode->Checked && Setup->Radio_WiFi_Mode->Checked) {
    DebugMessage("Only one radio port can be selected, select either COM (USB Serial) or WiFi");
	return;
	}
//
if (Setup->Radio_COM_Mode->Checked) {
	if (IsRadioCOMConnected()) {
   		StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = ConnectedRadioLED->Caption + " Already connected";
		if (DEBUGGING) DB->Debug_Print("Connect Button: " + ConnectedRadioLED->Caption + "  Already connected");
    	}
    }
if (Setup->RFID_COM_Mode->Checked) {
	if (bRFIDCOMConnected) {
    	StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = ConnectedRFIDLED->Caption + "  Already connected";
		if (DEBUGGING) DB->Debug_Print("Connect Button: " + ConnectedRFIDLED->Caption + "  Already connected");
        }
    }
if (IsTIUConnected()) {
   	StatusBar->Panels->Items[TIU_STATUS]->Text = "TIU Already connected";
	if (DEBUGGING) DB->Debug_Print("Connect Button: TIU Already connected");
    }

DisconnectRequest = false;

// reset counters
NumDetectors->Value = 0;
NumReaders->Value = 0;

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

//---------------------- Dispatch List Thread -------------------------------
//
// start Dispatch_List_Thread suspended and then start it running
//
if (!ptrDL_Thread) {
    ResetEvent(TerminationDLEvent);	// reset any lingering events
    // Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
    // The Execute() routine actually runs before some lower level construction is finished.
    // Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
    // per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0
    ptrDL_Thread = new TDispatch_List_Thread(true, this, Handle, RemoteNo);
    assert(ptrDL_Thread);	// Check that constructor worked
    DLThreadLED->Checked = true;	// light Dispatch List LED
    #if __BORLANDC__ >= 0x720
    ptrDL_Thread->Start();			// start the thread for CB10.1
    #else
    ptrDL_Thread->Resume();			// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
    #endif
    }

//---------------------- Communications Thread -------------------------------

if (!ptrTIUCommThread) {
    ResetEvent(TerminationCTEvent);	// reset any lingering events

    // Start the Communications Thread (for both Serial and WiFi communicaions)

    // bTIUSerialPortConnected is set to true by the ISR from the connect() call in CommThread
    //
    // Create (but don't run) the Communications Thread
    // Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
    // The Execute() routine actually runs before some lower level construction is finished.
    // Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
    // per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0
    ptrTIUCommThread = new TCommThread(true, this, Handle, ConnectedRadioLED->Caption,
            Baud[Setup->BaudRateSel->ItemIndex],
            Parity[Setup->ParitySel->ItemIndex],
            Bits[Setup->BitsSel->ItemIndex],
            Stop[Setup->StopSel->ItemIndex],
            Setup->Radio_COM_Mode->Checked,
            Setup->Radio_WiFi_Mode->Checked,
            Setup->RFID_COM_Mode->Checked,
            Setup->RFID_WiFi_Mode->Checked,
            Setup->AIU_WiFi_Mode->Checked,
            &retv);
    assert(ptrTIUCommThread);	// Check that constructor worked
    //CommThreadLED->LED->ColorOn = clYellow;	// Debug only
    CommThreadLED->Checked = true;	// light Comm Thread LED Yellow and will go Green when Execute() runs
    if (ptrTIUCommThread) {
      // Successful Open
        #if __BORLANDC__ >= 0x720
        ptrTIUCommThread->Start();			// start the thread for CB10.1
        #else
        ptrTIUCommThread->Resume();			// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
        #endif
        }
    }
//---------------------- Serial Radio -------------------------------
//
if (Setup->Radio_COM_Mode->Checked && !IsRadioCOMConnected()) {
	// connect Radio Serial Communication Port
    //
    // note that port must be specified as       "\\\\.\\COM10"
    // see https://support.microsoft.com/en-us/kb/115831
    char ComName[10];
	snprintf(ComName, sizeof ComName, "COM%d", Setup->PortNum->ValueAsInt);
    const String MungedPortName = "\\\\.\\" + String(ComName);	// pass in communications port

    // Return 0 on success, > 0 on failure.
    retv = Serial_port->connect(MungedPortName.c_str(), Baud[Setup->BaudRateSel->ItemIndex],
    	Parity[Setup->ParitySel->ItemIndex], Bits[Setup->BitsSel->ItemIndex], Stop[Setup->StopSel->ItemIndex], true);
    if (retv) {
        // failure
	    StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Failed to open " + String(ComName) + " (" + String(retv) + ")";
	    DebugMessage("Failed to open " + String(ComName) + " (" + String(retv) + ")");
		Screen->Cursor = OldCursor;
        return;
        } else {
        // Successful Open - wait for it to complete
        while(!bRadioCOMConnected) {
        	Sleep(10);
        	Application->ProcessMessages();
            }
        StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Radio Serial Connected";
        // DTR and RTS start in off state
        Serial_port->DTR(false);
        Serial_port->RTS(false);
        // DTR must be set when using the Feather with 32U4 processor
		DTRSwitch->Checked = Setup->DTRAlwaysOn->Checked;		// On my Wired interface, DTR enables the Hex Buffer
        								// Setting DTR Switch to true resets the Arduino each time the COM port is opened.
                                        // This causes the first packet to be ignored and always timesout and gets an error
                                        // but then the retry is always successful.
        // Set DTR if requested
	    if (DTRSwitch->Checked && !WIREDMODE) {
        	ptrTIUCommThread->DTRSwitch(true);
            }
        if (Setup->DTRAlwaysOn->Checked) {
        	Sleep(2000);	// wait for DTR to take effect when using the Feather with 32U4 processor
            Application->ProcessMessages();
            }
        // set RTS if requested
        if (RTSSwitch->Checked) {
        	ptrTIUCommThread->RTSSwitch(true);
			// RTS can be used for interface control as needed
            }
		//
		//if (!TIUEMULATORMODE && !RADIOMODE) {
		if (WIREDMODE) {
			// turn off DTR if running in WIRED mode
			DTRSwitch->Checked = false;      // on my interface DTR enables the Hex Buffer
        	ptrTIUCommThread->DTRSwitch(false);
			}
        //
        }
	// bRadioCOMConnected is set to true by the ISR from the connect() call
	}
//---------------------- Serial RFID Detectors -------------------------------
//

if (Setup->RFID_COM_Mode->Checked && !IsRFIDCOMConnected()) {
	// connect RFID Serial  Port if so selected
    // note that port must be speicfied as       "\\\\.\\COM10"
    // see https://support.microsoft.com/en-us/kb/115831
    // bRFIDSerialPortConnected is set to true by the ISR
    char RFIDName2[10];
	snprintf(RFIDName2, sizeof RFIDName2, "COM%d", Setup->RFIDNum->ValueAsInt);

    const String MungedPortName = "\\\\.\\" + String(RFIDName2);

//    if (!RFID_serial_port) RFID_serial_port = new Tserial_event;
//    assert(RFID_serial_port);

    retv = RFID_serial_port->connect(MungedPortName.c_str(),
            TAG_READER_BAUD_RATE,								// always 115200 baud
            Parity[Setup->ParitySel->ItemIndex],
            Bits[Setup->BitsSel->ItemIndex],
            Stop[Setup->StopSel->ItemIndex],
            true);
    if (retv) {
        // failure
        StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "Failed to open " + String(RFIDName2) + " (" + String(retv) + ")";
        DebugMessage("Failed to open " + String(RFIDName2) + " (" + String(retv) + ")");
        Screen->Cursor = OldCursor;
        DisconnectButtonClick(Sender);		// make sure that ports are disconnected
        // enable Exit escape
        ExitButton->Enabled = true;
        return;
        } else {
        // else Successful Open
//        RFID_serial_port->setManager(SerialEventManager2);
        // the serial port driver turns off DTR and RTS to set the respective lines
        RFID_serial_port->DTR(false);
        RFID_serial_port->RTS(true);
        // Toggle DTR to reset the Arduino
        RFID_serial_port->DTR(true);
        Send_ASCII_TIU->Enabled = true;
        Send_Hex_RFID->Enabled = true;
        //if (ptrProgram_Control[0] ) ptrProgram_Control[0]->RFIDLED->Checked = true;
        StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "RFID Serial Connected";
        }
 	 // bRFIDCOMConnected is set to true by the ISR from the connect()
     }
//
if (Setup->RFID_COM_Mode->Checked || Setup->Radio_COM_Mode->Checked) {
    // give up to 250 ms for the serial ports to connect
    // We look at the radio port and the RFID port if they are enabled in Setup
    for (int step = 250/50; step > 0 ; step--) {
        Delay(50);	// delay for 50 ms
        if ((!Setup->Radio_COM_Mode->Checked || bRadioCOMConnected) && (!Setup->RFID_COM_Mode->Checked || bRFIDCOMConnected)) {   // check flags set by ISR
        	// was 	if (bTIUSerialPortConnected && (!Setup->RFID_Mode->Checked || bRFIDSerialPortConnected)) {   // check flag set by ISR
            //
            // start idle timer if requested to drop DTR or RTS after 10 seconds of idle
            // Don't drop DTR in TIU Emulator mode because DTR is used as Reset of Arduino
            //
            if ((Setup->DropDTRonIdle->Checked  && (!TIUEMULATORMODE && !RADIOMODE)) || Setup->DropRTSonIdle->Checked) {
                IdleTimer->Enabled = true;
                IdleTimer->Tag = 0;	// reset Timer Counter
                } else {
                IdleTimer->Enabled = false;
                }
            }
        }
    }

//---------------------- WiFi RFID -------------------------------
//
if (Setup->RFID_WiFi_Mode->Checked && !bRFIDUDPWiFiConnected) {
	#ifdef NOTNEEDED
	FServerAddr                = WSocketResolveHost("127.0.0.1");
    if (FServerAddr.s_addr == htonl(INADDR_LOOPBACK)) {
        // Replace loopback address by real localhost IP addr
		FServerAddr            = WSocketResolveHost(LocalHostName());
		}
    #endif
	if (DEBUGGINGX(6)) DB->Debug_Print("Local Host Name (WiFi RFID) = " + LocalHostName());
    #ifdef NOTNEEDED
	if (DEBUGGINGX(6)) DB->Debug_Print("Dotted Host Address = " + String(WSocket_inet_ntoa(FServerAddr)));
    #endif

    if (DEBUGGING) DB->Debug_Print("Selected WiFi RFID Network = " + Setup->IPComboBox->Text);
    //---------------------- WiFi RFID UDP ---------------------------
	RFIDudpRX->Proto    = "udp";
	RFIDudpRX->Addr     = "0.0.0.0";	// Listen on all available network connections simultaniously
	RFIDudpRX->Port     = 8899;
    RFIDudpRX->ReuseAddr = true;

	try {
		RFIDudpRX->Listen();
		while(bRFIDUDPWiFiRXConnected == false) {
			Sleep(10);
            Application->ProcessMessages();
			}		
        if (DEBUGGING) DB->Debug_Print("RFID RX WiFi listening on Port " + RFIDudpRX->Port);
		} catch (const ESocketException &E) {
		// The socket is probably already in use
			if (DEBUGGING) DB->Debug_Print("RFID RX WiFi error : " + E.Message);
			throw;
		}
    //---------------------- WiFi RFID TCP ---------------------------
    // Code from TCPICSTest.cpp
    RFIDtcpRX->ClientClass = __classid(TTcpSrvClient); // Use our component
    RFIDtcpRX->Banner      = "";
    RFIDtcpRX->Proto       = "tcp";     // Use TCP protocol
    RFIDtcpRX->Port        = 8897;
    RFIDtcpRX->Addr        = "0.0.0.0"; // Use any interface
    RFIDtcpRX->ListenBacklog   = 100;   // maximum backlog of clients waiting to connect
    RFIDtcpRX->MaxClients      = 100;   // maximum number of detectors
	RFIDtcpRX->ReuseAddr = true;

    if (DEBUGGING) DB->Debug_Print("TCP Server Port = " + RFIDtcpRX->Port);
    RFIDtcpRX->Listen();                    // Start listening
    bRFIDTCPWiFiConnected = true;
    if (DEBUGGING) DB->Debug_Print("Listening for TCP clients...");

    StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "RFID TCP WiFi Connected";
    }
//---------------------- WiFi Radio UDP -------------------------------
//
if (Setup->Radio_WiFi_Mode->Checked && !bRadioWiFiConnected) {
	#ifdef NOTNEEDED
	FServerAddr                = WSocketResolveHost("127.0.0.1");
    if (FServerAddr.s_addr == htonl(INADDR_LOOPBACK)) {
        // Replace loopback address by real localhost IP addr
		FServerAddr            = WSocketResolveHost(LocalHostName());
		}
    #endif
	if (DEBUGGINGX(6)) DB->Debug_Print("Local Host Name (WiFi Radio) = " + LocalHostName());
    #ifdef NOTNEEDED
	if (DEBUGGINGX(6)) DB->Debug_Print("Dotted Host Address = " + String(WSocket_inet_ntoa(FServerAddr)));
    #endif

    if (DEBUGGING) DB->Debug_Print("Selected WiFi Radio Network = " + Setup->IPComboBox->Text);

    //
    // Radio WiFI RX and TX are started automatically in CommThread.cpp
    //
    StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Radio WiFi Connected";
    }
// Wait for FIFO to be created
while (!ptrTIUCommThread->FIFO) {
	Sleep(10);
    Application->ProcessMessages();
    }
//
if (IsRadioConnected()) {
    // Disable most Serial port setup controls while we are connected
    Setup->RFID_COM_Mode->Enabled = false;
    Setup->Radio_COM_Mode->Enabled = false;
    Setup->RFID_WiFi_Mode->Enabled = false;
    Setup->Radio_WiFi_Mode->Enabled = false;
    Setup->Radio_WiFi_Multicast->Enabled = false;
    Setup->PortNum->Enabled = false;
    Setup->RFIDNum->Enabled = false;
    Setup->RadioIP->Enabled = false;
    Setup->RadioPort->Enabled = false;
    Setup->COMCheckListBox->Enabled = false;
    Setup->BaudRateSel->Enabled = false;
    Setup->ParitySel->Enabled = false;
    Setup->BitsSel->Enabled = false;
    Setup->StopSel->Enabled = false;
    //
    Send_ASCII_TIU->Enabled = true;
    Send_Hex_TIU->Enabled = true;
    // IIII  enable controls on successful Serial Port connection
    RTSSwitch->Enabled = true;
    DTRSwitch->Enabled = true;
    STOPem->Enabled = false;
    StartedConsists->Value = 0;
    StartupButton->Enabled = true;
    QuickButton->Enabled = true;
    Var1ModeButton->Enabled = true;
    Var2ModeButton->Enabled = true;
    ShutdownButton->Enabled = true;
    ReadEnginesButton->Enabled = true;
    // Program control is allowed even if the RFID tag system is not used
    Prog_Control_Button->Enabled = true;

    // enable Accessories/Switches Button & Stop All Button
    StopAllButton->Enabled = true;
    // if any LashUp window is up, enable its buttons
    for (int iTIU = 0 ; iTIU < TIUMAX ; iTIU++) {	// TIU 1-5
        if (ptrLashUp[iTIU] != NULL) {
            ptrLashUp[iTIU]->LashUpShutdownButton->Enabled = true;
            }
        }
    }
//-------------------- Start mDNS Server ---------------------------
if (Setup->Radio_WiFi_Mode->Checked || Setup->RFID_WiFi_Mode->Checked || Setup->AIU_WiFi_Mode->Checked) {
    // Start the mDNS responder - mDNS query if in WiFi mode
   	Setup->StartmDNSServices(Setup->RFID_WiFi_Mode->Checked, Setup->Radio_WiFi_Mode->Checked);
    }
//---------------------- TIU -------------------------------
//
if (!IsTIUConnected()) {
    //
    // This next piece of code caused an interesting problem. When I had just the AIUCount
    // readup enabled, the TIU would respond correctly the first time. If I just started RTC
    // connected, disconnected, exited, restarted, connected - the second connection would fail. This was
    // because the sequence number was the same for two packets in a row and the TIU ignored
    // the second packet (never responded).
    // I attempt to fix that by sending two commands at startup which will bump the sequence number
    //
    // if shift is pressed, skip these two calls and use debugging data
    //
	const SHORT ShiftKeyState = ConnectButton->Focused() ? GetKeyState(VK_SHIFT) : 0;	// Do a debug Connect to TIU
    // key pressed =  0xFF8X
    // key not pressed = 0x000X
    // "X" can be 0 or 1 - indicated key was toggled
    if (ShiftKeyState >= 0) {
      // if shift is not pressed, do a normal connect
      if (Setup->Hex->Checked && !Setup->TXOnlyComm->Checked) {	// Hex mode & two way comm required
        if (GetAIUCount(TIUNo)) { // read up the number of AIU, set bTIUConnected to true or false
            // if the GetAIUCount worked, read up the TIU version number - needed to form LashUp command correctly
            GetTIUVer(TIUNo);	// this also bumps the sequence number
            // success with above two calls will set TIUSpin to clLime
            // and bTIUConnected to true, failure will set TIUSpin to clRed
            }
        }
      } else {
      // if shift is pressed, use debug values
      bTIUConnected = true;
      ConnectedTIULED->Checked = true;
      TIUSpin->ColorDefault = clYellow;
      TIUVersion = 6.1;
      AIUSpin->Value = 1;	// no AIU
      AIUNo = 1;
      //AIUButton->Enabled = true;
      AIUSpin->MaxValue = AIUMAX;	// allow all AIU
      AIUCount[TIUNo - 1] = AIUSpin->MaxValue;
      AIUSpin->Enabled = true;
      StatusBar->Panels->Items[TIU_STATUS]->Text = "TIU Simulated";
      }
	// if the Get AIU command was successful, move forward
    if (IsTIUConnected()) {
        DisconnectButton->Enabled = true;
        ConnectButton->Enabled = false;	// on success, disable connect button
        ExitButton->Enabled = false;
        if (DEBUGGING) DB->Debug_Print("Connect Button: Connected");
        // Auto Show the Quick Controls if requested
        if (Setup->AutoQuickControls->Checked && IsTIUConnected()) QuickButtonClick(QuickButton);
        //
        StatusBar->Panels->Items[TIU_STATUS]->Text = "TIU Connected";
        } else {
        // Failed to get AIU count
        DisconnectButtonClick(Sender);		// make sure that ports are disconnected
        // DebugMessage("Can't connect to the TIU - Get AIU count failed");	// TODO is this an appropriate message?
        // enable Exit escape
        ExitButton->Enabled = true;
        }
    Screen->Cursor = OldCursor;
    return;    // At this point we are either connected or not - look at value of bTIUConnected or color of TIUSpin
	}
// TODO need to rewrite this
// check for failure
//if (!IsRadioCOMConnected() || (Setup->RFID_COM_Mode->Checked && !IsRFIDCOMConnected()) || !IsTIUConnected()) {
//	// no response after timeout
//	DisconnectButtonClick(Sender);		// make sure that ports are disconnected
//	if (DEBUGGING) DB->Debug_Print("Connect Button: Can't connect to the serial port (Timeout after 250ms)");
//	}
// enable Exit escape
ExitButton->Enabled = true;
Screen->Cursor = OldCursor;
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ExitButtonClick(TObject *Sender)
{
// TODO when compiled with BDB4.0, the computer sometimes beeps when the [Exit] button is pressed.
Close();  										// Close the Form. This calls Application->Terminate()
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DisconnectButtonClick(TObject *Sender)
{
//
assert(Setup != NULL);		// setup must be created
if (iPlayingBackCounter && !bEmergencyStop) {	// unless Emergency Stop was pressed
	DebugMessagePopup("Warning", "Cancel the Playback before trying to disconnect", "");
    return;
	}
// if recording, ask to save the recording
if (bRecording && !bEmergencyStop) {
    Setup->RecordStopButton->OnClick(Setup->RecordStopButton);	// unless Emergency Stop was pressed
	}

// if Program Control is running, ask that the user [STOP] any scripts before disconnecting
for (int iPCNo = 1 ; iPCNo <= PCMAX ; iPCNo++) {
	if (ptrProgram_Control[iPCNo - 1] != NULL) {
		if (ptrProgram_Control[iPCNo - 1]->ptrPC_Thread != NULL) {
			DebugMessagePopup("Warning", "[STOP] any Program Control Scripts before trying to disconnect", "");
    		return;
			}
        }
	}

// if any Engine Information or Engine Status threads are run, ask the the user [STOP] them before disconnecting
for (int iTIU = 1; iTIU <= TIUMAX ; iTIU++) { // TIU 1-5
	for (int iEng = 0; iEng < LASHUPMIN ; iEng++) {  	// Engine 0 to 120 - factory reset to end of engines
        if (ptrEngineInfo[iTIU-1][iEng]) {
            if (ptrEngineInfo[iTIU-1][iEng]->ptrEngineInfoThread != NULL) {
				DebugMessagePopup("Warning", "[Stop] any Engine Information windows before trying to disconnect", "");
    			return;
            	}
        	}
        if (ptrEngineStatus[iTIU-1][iEng]) {
            if (ptrEngineStatus[iTIU-1][iEng]->ptrEngineStatusThread != NULL) {
				DebugMessagePopup("Warning", "[Stop] any Engine Status windows before trying to disconnect", "");
    			return;
            	}
        	}
        if (ptrTrackSignal[iTIU-1][iEng]) {
            if (ptrTrackSignal[iTIU-1][iEng]->ptrTrackSignalThread != NULL) {
				DebugMessagePopup("Warning", "[Stop] any Track Signal windows before trying to disconnect", "");
    			return;
            	}
        	}
    	}
    }

DisconnectRequest = true;	// indicate that the [Disconnect] button was pressed
// Stop the mDNS Server
if (Setup->Radio_WiFi_Mode->Checked || Setup->RFID_WiFi_Mode->Checked || Setup->AIU_WiFi_Mode->Checked) {
    // Stop the mDNS responder/queryer if in WiFi mode
   	Setup->StopmDNSServices(Setup->RFID_WiFi_Mode->Checked, Setup->Radio_WiFi_Mode->Checked);
    }
// disconnect RFID UDP WiFi
if (bRFIDUDPWiFiConnected && RFIDudpRX) {
	RFIDudpRX->Close();
    }
// disconnect RFID TCP WiFi
if (bRFIDTCPWiFiConnected && RFIDtcpRX) {
    TTcpSrvClient *AClient;

    RFIDtcpRX->Close();    // first close the socket server
    // then close all clients (working backwards through the client list)
	int CC = RFIDtcpRX->ClientCount;
	if (CC > 0) {
		for (int I = CC - 1; I >= 0; I--) {
			AClient = (TTcpSrvClient *)(RFIDtcpRX->Client[I]);
			if (DEBUGGING) DB->Debug_Print("Closing " + AClient->PeerAddr);
			AClient->Shutdown(2);
			}
		AClient = NULL;
		}
    bRFIDTCPWiFiConnected = false;
    NumReaders->Value = 0;
    NumDetectors->Value = 0;
    }
// disconnect Radio WiFi
if (bRadioWiFiConnected) {
	// Radio WiFi disconnected in TCommThread.cpp
    // moved to onSessionClosed event
	// bRadioWiFiConnected = false;
    }

if (Setup->Radio_COM_Mode->Checked && !IsRadioCOMConnected()) {
   	StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = ConnectedRadioLED->Caption + " Not connected";
	if (DEBUGGING) DB->Debug_Print("Disconnect Button: " + ConnectedRadioLED->Caption + " Not connected");
    }
if (Setup->RFID_COM_Mode->Checked && !IsRFIDCOMConnected()) {
   	StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = ConnectedRFIDLED->Caption + " Not connected";
	if (DEBUGGING) DB->Debug_Print("Disconnect Button: " + ConnectedRFIDLED->Caption + " Not connected");
    }
if (!IsTIUConnected()) {
   	StatusBar->Panels->Items[TIU_STATUS]->Text = "TIU Not connected";
	if (DEBUGGING) DB->Debug_Print("Disconnect Button: TIU Not connected");
    }

TIUSpin->ColorDefault = clYellow;	// set TIU/AIU boxes color to yellow indicating unknown comm with the TIU
AIUSpin->ColorDefault = clYellow;

ExitButton->Enabled = false;		// don't allow [Exit] until disconnect is complete

bPlayingBack = bPlaybackCommError = false;		// clear record or playback flags to terminate those operations
iPlayingBackCounter = 0;						// indicate no playbacks are running

bRecording = false;
//
// if TIU is connected, disconnect it
//
if (IsTIUConnected()) {
	bTIUConnected = false;
    ConnectedTIULED->Checked = false;
    StatusBar->Panels->Items[TIU_STATUS]->Text = "TIU Disconnected";
    }
//
//
bool bretv;
// stop the Dispatch List and Playback Threads
assert(ptrTIUCommThread);					// must be created
bretv = SetEvent(TerminationDLEvent);  		// stop Dispatch List thread -  DL_Thread
assert(bretv);
bretv = SetEvent(TerminationPBEvent);  		// stop any (and all) Playback threads -  PBFileThread
assert(bretv);
DLThreadLED->Checked = false;	// extinquish Dispatch List LED
ptrDL_Thread = NULL;						// FreeOnTerminate is set
//---------------------------------------------------------------------------
// if Radio COM port is enabled and connected, disconnect it
//
// bRadioCOMConnected is set to false by the ISR
//
if (Setup->Radio_COM_Mode->Checked && IsRadioCOMConnected()) {
	DSR->Checked = false;
	CTS->Checked = false;
   	CD->Checked = false;
    // Disable controls when diconnected
	DTRSwitch->Enabled = false;
    RTSSwitch->Enabled = false;
	if (Setup->NumRetries->Value != 10) {
	    Serial_port->disconnect();					// Serial_port.disconnect() call in Comm_Thread destructor
        bretv = SetEvent(TerminationCTEvent);  		// destroy Comm thread - Comm_Thread & serial_event
        assert(bretv);
		ptrTIUCommThread = NULL;             		// FreeOnTerminate is set
        CommThreadLED->Checked = false;	// extinquish Comm Thread LED
		if (DEBUGGING) DB->Debug_Print("Disconnect Button: Radio Serial Port Disconnected");
        StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Radio Serial Disconnected";
        }
	}
//---------------------------------------------------------------------------
//
// if RFID COM port is enabled and connected, disconnect it
//
// bRFIDCOMConnected is set to false by the ISR
//
if (Setup->RFID_COM_Mode->Checked && IsRFIDCOMConnected()) {
    Send_ASCII_RFID->Enabled = false;
	Send_Hex_RFID->Enabled = false;
    /////////
  	if (Setup->NumRetries->Value != 10) {
    	RFID_serial_port->disconnect();			// RFID_serial_port->disconnect() call in Comm_Thread destructor
        bretv = SetEvent(TerminationCTEvent);  	// destroy Comm thread - Comm_Thread & serial_event
        assert(bretv);
		ptrTIUCommThread = NULL;             	// FreeOnTerminate is set
        CommThreadLED->Checked = false;	        // extinquish Comm Thread LED
		if (DEBUGGING) DB->Debug_Print("Disconnect Button: RFID Serial Port Disconnected");
        StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "RFID Serial Disconnected";
		}
    }
//---------------------------------------------------------------------------
// if RFID WiFi port is enabled, show a disconnect message
//
//
if (Setup->RFID_WiFi_Mode->Checked) {
    StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "RFID WiFi Disconnected";
	}

//---------------------------------------------------------------------------
// if Radio WiFi port is enabled, show a disconnect message
//
//
if (Setup->Radio_WiFi_Mode->Checked) {
    StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Radio WiFi Disconnected";
	}

//---------------------------------------------------------------------------

if (Setup->RFID_COM_Mode->Checked || Setup->Radio_COM_Mode->Checked) {
    //
    // give up to 250 ms for both serial ports to disconnect
    // We look at the radio COM port and the RFID COM port if they are enabled in Setup
    for (int step = 250/50; step > 0 ; step--) {
        Delay(50);	// delay for 50 ms
        if (!bRadioCOMConnected && !bRFIDCOMConnected) {    // check flags set by ISR to be both false
			break;	// out of for() loop as ports are disconnected
            }
        }
	// Serial Port Disconnect timeout if either connected flag is still set
    if (bRadioCOMConnected || bRFIDCOMConnected) {
		if (DEBUGGING) DB->Debug_Print("Disconnect Button: Can't disconnect from the serial ports (Timeout after 250ms)");
        }
    }
if (Setup->RFID_WiFi_Mode->Checked || Setup->Radio_WiFi_Mode->Checked) {
	// no special waiting is needed for the WiFi Ports but still must terminate the TCommThread
   	if (Setup->NumRetries->Value != 10) {
		if (DEBUGGING) DB->Debug_Print("Disconnect Button: Radio & RFID WiFi Ports Disconnected");
        }
    }

// Terminate the TCommThread
bretv = SetEvent(TerminationCTEvent);  		// destroy Comm thread - Comm_Thread & serial_event
assert(bretv);
ptrTIUCommThread = NULL;             		// FreeOnTerminate is set
CommThreadLED->Checked = false;	// extinquish Comm Thread LED

// We are disconnected; enable port setup controls so user can change selections
Setup->RFID_COM_Mode->Enabled = true;
Setup->Radio_COM_Mode->Enabled = true;
Setup->RFID_WiFi_Mode->Enabled = true;
Setup->Radio_WiFi_Mode->Enabled = true;
Setup->Radio_WiFi_Multicast->Enabled = true;
Setup->RadioIP->Enabled = Setup->Radio_WiFi_Mode->Checked;
Setup->RadioPort->Enabled = Setup->Radio_WiFi_Mode->Checked;
Setup->PortNum->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->RFIDNum->Enabled = Setup->RFID_COM_Mode->Checked;
Setup->COMCheckListBox->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->BaudRateSel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->ParitySel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->BitsSel->Enabled = Setup->Radio_COM_Mode->Checked;
Setup->StopSel->Enabled = Setup->Radio_COM_Mode->Checked;
// disable BUttons and Switches
Send_ASCII_TIU->Enabled = false;
Send_Hex_TIU->Enabled = false;
//RTSSwitch->Enabled = false;
//DTRSwitch->Enabled = false;
CTS->Checked = false;
DSR->Checked = false;
CD->Checked = false;
Ring->Checked = false;
STOPem->Enabled = false;
StartedConsists->Value = 0;
StartupButton->Enabled = false;
QuickButton->Enabled = false;
Var1ModeButton->Enabled = false;
Var2ModeButton->Enabled = false;
ShutdownButton->Enabled = false;
ReadEnginesButton->Enabled = false;
Prog_Control_Button->Enabled = true;

//////////////////////////
// disable Accessories/Switches Button & Stop All Button
// AIUButton->Enabled = false;
StopAllButton->Enabled = false;
// if any LashUp window is up, disable its buttons
for (int iTIU = 0 ; iTIU < TIUMAX; iTIU++) {	// TIU 1-5
    if (ptrLashUp[iTIU] != NULL) {
        ptrLashUp[iTIU]->LashUpStartUpButton->Enabled = false;
        ptrLashUp[iTIU]->LashUpShutdownButton->Enabled = false;
        }
    }
// DTR not dropped in TIU Emulator mode bacause we might reset the Arduino
if ((Setup->DropDTRonIdle->Checked  && !TIUEMULATORMODE) || Setup->DropRTSonIdle->Checked) {
    IdleTimer->Enabled = false;		// turn off DTR drop timer
    }
DisconnectButton->Enabled = false;
DisconnectRequest = false;			// clear [Disconnect] button flag
ConnectButton->Enabled = true;		// on success, enable connect button
ConnectButton->Caption = "&Connect";
// enable [Exit] Button escape
ExitButton->Enabled = true;
if (DEBUGGING) DB->Debug_Print("Disconnect Button: Disconnected");
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Send_ASCII_TIUClick(TObject *Sender)
{
// Send to TIU Port
// Debug only
LOCKEM()
//
// Take the String in Encoded_CommandB and send to the TIU port
// it will send any hex bytes, not just 5,6,9 & A & 4D

if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
    return;
    }


if (TIULog->Checked) {
    // display it in the TIU buffer
    TIUBuffer->SelAttributes->Color = clBlue;
    TIUBuffer->SelStart = TIUBuffer->GetTextLen();
    TIUBuffer->SelText = Encoded_CommandA->Text + TEXT("\r\n"); // CR LF
    TIULineCount->Caption = String(TIULineCount->Caption.ToInt() + 1);
    // Method 4 works under CB10.1 but not BCB 4.0
    TIUBuffer->HideSelection = false;
    TIUBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
    TIUBuffer->HideSelection = true;
    }
// Send it
// Does not record or handle errors
(void) ptrTIUCommThread->SendHexStrStr_FIFO(TIU_PORT, Encoded_CommandA->Text, Encoded_CommandA->Text.Length()/2);
}
//---------------------------------------------------------------------------

void  __fastcall TMainForm::Send_ASCII_RFIDClick(TObject *Sender)
{
LOCKEM()
// Send to RFID Port
// Debug only
//
// Take the String in Encoded_CommandB and send to the RFID port
// it will send any hex bytes, not just 5,6,9 & A & 4D

if (!IsRFIDConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
    return;
    }

if (RFIDLog->Checked) {
    // display it in the RFID buffer
    RFIDBuffer->SelAttributes->Color = clGreen;
    RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();
    RFIDBuffer->SelText = Encoded_CommandB->Text + TEXT("\r\n");    // CR LF
    // Method 4 works under CB10.1 but not BCB 4.0
    RFIDBuffer->HideSelection = false;
    RFIDBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    RFIDBuffer->HideSelection = true;
    }

// Send it and don't check for errors and don't wait for response
// Does not record or handle errors
(void) ptrTIUCommThread->SendHexStrStr_FIFO(RFID_PORT, Encoded_CommandB->Text, Encoded_CommandB->Text.Length()/2);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Send_Hex_TIUClick(TObject *Sender)
{
LOCKEM();
// Debug only
//
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
    return;
    }

String CommandA = CleanUpCommand(Encoded_CommandA->Text);	// unlike previous function, this one cleans the command string
if (DEBUGGING) DB->Debug_Print("Send_Hex_AClick: " + CommandA);

// don't check for errors and don't wait for response
// Does not record or handle errors
(void) ptrTIUCommThread->SendHexStrStr_FIFO(TIU_PORT, CommandA, CommandA.Length()/2);
// bump sequence number
H_Seq_Count = H_Seq_Count +1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Send_Hex_RFIDClick(TObject *Sender)
{
LOCKEM();
// Debug only
//
if (!IsRFIDConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
    return;
    }

String CommandB = CleanUpCommand(Encoded_CommandB->Text);	// unlike previous function, this one cleans the command string
if (DEBUGGING) DB->Debug_Print("Send_Hex_BClick: " + CommandB);

// don't check for errors and don't wait for response
// Does not record or handle errors
(void) ptrTIUCommThread->SendHexStrStr_FIFO(RFID_PORT, CommandB, CommandB.Length()/2);
// bump sequence number
H_Seq_Count = H_Seq_Count +1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::WritetoDiskClick(TObject *Sender)
{
// Write a RichEdit Control to the disk
// Note: started to edit to change to use the TStrings object Lines
// but the TStrings loses all of the RichEdit Colors and Fonts
//
String SourceBuffer;
TRichEdit * SelectedBuffer;
const TButton * ButtonUsed = (TButton *)Sender;

// setup pointer to buffer being saved
if (ButtonUsed->Name.Pos("RFID")) {
	SelectedBuffer = RFIDBuffer;
    SourceBuffer = "_RFID_";
    }
if (ButtonUsed->Name.Pos("TIU")) {
	SelectedBuffer = TIUBuffer;
    SourceBuffer = "_TIU_";
    }
if (ButtonUsed->Name.Pos("Debug")) {
	SelectedBuffer = DB->Debug_Messages;
    SourceBuffer = "_DEBUG_";
    }
if (ButtonUsed->Name.Pos("CRC")) {
	SelectedBuffer = CRCBuffer;
    SourceBuffer = "_CRC_";
    }

assert(SelectedBuffer);
if( SelectedBuffer->Lines->Count == 0) return;	// buffer is empty

SaveDialog1->InitialDir = WorkingDir;

String NameOfFile;
if (PlainTextFile->Checked) {  // Plain text - make .txt file
	SaveDialog1->DefaultExt = String("txt");
	SaveDialog1->Filter = "Text files (*.txt)|*.txt";
    SaveDialog1->FileName = FileNamePrefix->Text + SourceBuffer + "TIU" + String(TIUNo) + "_" + FileNameSuffix->Text + ".txt";
    } else {					// Rich text - make .rtf file
	SaveDialog1->DefaultExt = String("rtf");
	SaveDialog1->Filter = "Rich text files (*.rtf)|*.rtf";
    SaveDialog1->FileName = FileNamePrefix->Text + SourceBuffer + "TIU" + String(TIUNo) + "_"+ FileNameSuffix->Text + ".rtf";
    }
SaveDialog1->Title = "Save File";
SaveDialog1->Options = SaveDialog1->Options >> ofOverwritePrompt << ofEnableSizing;	// RTC will handle this case

if(SaveDialog1->Execute()) {
	NameOfFile = SaveDialog1->FileName;
    // Change working directory to whatever was selected
    WorkingDir = ExtractFilePath(NameOfFile);
    // if file exists, confirm first
    if (FileExists(NameOfFile)) {
        // ask for confirmation
        int result = Application->MessageBox(TEXT("File Exists - Do you want to overwrite?"), Caption.c_str(), MB_YESNO | MB_ICONQUESTION);
    	// if No, exit
    	if (result  == IDNO) {
            return;
        	}
    	}
    // At this point, it's either a new file or the user has indicated overwrite
    // text only file if PlainText is checked
    bool savPlainText = SelectedBuffer->PlainText;
    SelectedBuffer->PlainText = PlainTextFile->Checked;   // if Checked set PlainText to true
	SelectedBuffer->Lines->SaveToFile(NameOfFile);
	if (DEBUGGINGX(1)) DB->Debug_Print("WritetoDisk: " + NameOfFile + " - " + String(SelectedBuffer->Lines->Count) + " Lines");
    SelectedBuffer->PlainText = savPlainText;
	return;	// done
	} else {
    return;	// user canceled
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ReadfromDiskClick(TObject *Sender)
{
String SourceBuffer;
TRichEdit * SelectedBuffer;
const TButton * ButtonUsed = (TButton *)Sender;

// setup pointer to buffer being read into
if (ButtonUsed->Name.Pos("RFID")) {
	SelectedBuffer = RFIDBuffer;
    SourceBuffer = "_RFID_";
    }
if (ButtonUsed->Name.Pos("TIU")) {
	SelectedBuffer = TIUBuffer;
    SourceBuffer = "_TIU_";
    }
if (ButtonUsed->Name.Pos("Debug")) {
	SelectedBuffer = DB->Debug_Messages;
    SourceBuffer = "_DEBUG_";
    }
if (ButtonUsed->Name.Pos("CRC")) {
	SelectedBuffer = CRCBuffer;
    SourceBuffer = "_CRC_";
    }

OpenDialog1->InitialDir = WorkingDir;

String NameOfFile;
if (PlainTextFile->Checked) {  // Plain text - make .txt file
	OpenDialog1->DefaultExt = String("txt");
	OpenDialog1->Filter = "Text files (*.txt)|*.txt";
	OpenDialog1->FileName = FileNamePrefix->Text + SourceBuffer + "TIU" + String(TIUNo) + "_"+ FileNameSuffix->Text + ".txt";
    } else {
	OpenDialog1->DefaultExt = String("rtf");
	OpenDialog1->Filter = "Rich Text files (*.rtf)|*.rtf";
	OpenDialog1->FileName = FileNamePrefix->Text + SourceBuffer + "TIU" + String(TIUNo) + "_"+ FileNameSuffix->Text + ".rtf";
    }
OpenDialog1->Title = "Open File";
OpenDialog1->Options = OpenDialog1->Options << ofEnableSizing;	// Allow Resizing

if(OpenDialog1->Execute()) {
	SelectedBuffer->Clear();
	NameOfFile = OpenDialog1->FileName;
    // Change working directory to whatever was selected
    WorkingDir = ExtractFilePath(NameOfFile);
    // text only file
    bool savPlainText = SelectedBuffer->PlainText;
    SelectedBuffer->PlainText = PlainTextFile->Checked;   // if Checked set PlainText to true
    SelectedBuffer->Lines->LoadFromFile(NameOfFile);
    SelectedBuffer->PlainText = savPlainText;
    if (DEBUGGING) DB->Debug_Print("ReadFromDisk: " + NameOfFile);
	} else {
    return;	// user canceled
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::STOPemClick(TObject *Sender)
{
bStop = true;					// set flag, then
STOPem->Enabled = false;		// disable button
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ResetTIUCCButtonClick(TObject *Sender)
{
// Reset all TIU Port Counters
CTS_Count = DSR_Count = CD_Count = 0;
CTS_Counter->Text = String(CTS_Count);
DSR_Counter->Text = String(DSR_Count);
CD_Counter->Text = String(CD_Count);
// TIU related counts
TIU_Char_Count->Text = "0";
TIU_Total_Byte_Count->Value = 0;
TIULineCount->Caption = "0";
// Communications related counts
PacketsSent->ValueAsInt = 0;
sendDataBusy->ValueAsInt = 0;
Retries->ValueAsInt = 0;
NAKShort->ValueAsInt = 0;	// NAK counter #1
NAKLen->ValueAsInt = 0;		// NAK counter #2
NAK0T->ValueAsInt = 0;		// NAK counter #3
NAKCRC->ValueAsInt = 0;		// NAK counter #4
ReentrantErrors->ValueAsInt = 0;
CaughtErrors->ValueAsInt = 0;
EngNotFound->ValueAsInt = 0;
// FIFO's and Dispatch List
Dispatched_Cmd_Count->ValueAsInt = 0;
FIFO_Count->ValueAsInt = 0;
DL_FIFO_Total->ValueAsInt = 0;
DL_FIFO_Count->ValueAsInt = 0;
// Tag related Counters
Tags->ValueAsInt = 0;  // total number of packets received from the detectors (includes startup packets)
// reset Tag related counters
NumDetectors->Value = 0;
NumReaders->Value = 0;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ResetRFIDCCButtonClick(TObject *Sender)
{
// Clear RFID Byte Count
//iRFIDbufCount = 0;
RFID_Char_Count->Text = "0";
Tags->ValueAsInt = 0;  // total number of packets received from the detectors (includes startup packets)
}
//---------------------------------------------------------------------------
unsigned char __fastcall HexStringtoInt(unsigned char * Str)
{
// not used
assert(false);
// converts a 2 Hexidecimal character string to binary
static const char* const lut = "0123456789ABCDEF";
// Convert Hex character string to Hex
char a = Str[0];
const char* p = std::lower_bound(lut, lut + 16, a);
if (*p != a) {
   	if (DEBUGGING) DB->Debug_Print("HexStringtoInt: Error: A Not a Hex Digit : " + String((char *)Str), clRed);
    return false;
    }

char b = Str[1];
const char* q = std::lower_bound(lut, lut + 16, b);
if (*q != b) {
	if (DEBUGGING) DB->Debug_Print("HexStringtoInt: Error: A Not a Hex Digit : " + String((char *)Str), clRed);
	return false;
    }
return (((p - lut) << 4) | (q - lut));
}
//---------------------------------------------------------------------------

String Convert16B8BNumber(String W)
{
// Convert 4 digit hex encoded 16 bits string to 8 bit Morton Numbers in binary
// (Call bitswapRX() after calling this function)
/*
55	0000 0		56	0001 1		59	0010 2		5A  0011 3
65	0100 4		66	0101 5		69	0110 6		6A	0111 7
95	1000 8		96	1001 9		99	1010 10		9A	1011 11
A5	1100 12		A6	1101 13		A9	1110 14		AA	1111 15

This is an RLL(0,1) encoding
*/
	const static String Encoded[] = {"55",  "56",  "59",  "5A",  "65",  "66",  "69",  "6A",  "95",  "96",  "99",  "9A",  "A5",  "A6",  "A9",  "AA"};
	const static String Binary[] = {"0000","0001","0010","0011","0100","0101","0110","0111","1000","1001","1010","1011","1100","1101","1110","1111"};
    int Digit1, Digit2;
	String Dig1 = W.SubString(1,2);
    Dig1 = Dig1.UpperCase();
    for (Digit1 = 0 ; Digit1 < 16 ; Digit1++) {
        if (Dig1 == Encoded[Digit1]) {
        	break;
            }
    	}
	String Dig2 = W.SubString(3,2);
    Dig2 = Dig2.UpperCase();
    for (Digit2 = 0 ; Digit2 < 16 ; Digit2++) {
        if (Dig2 == Encoded[Digit2]) {
        	break;
            }
    	}
if (Digit1 == 16 || Digit2 == 16)
    return "00";
    else
    return Binary[Digit1] + Binary[Digit2];

}
//---------------------------------------------------------------------------

String Convert8B16BNumber(String W)
{
// convert 8 bit Morton Nnumbers in binary to 4 digit hex encoded 16 bit String
// (call bitswapTX() before calling this function)
// NO VCL calls in this function
//
	const static String Encoded[] = {"55",  "56",  "59",  "5A",  "65",  "66",  "69",  "6A",  "95",  "96",  "99",  "9A",  "A5",  "A6",  "A9",  "AA"};
	const static String Binary[] = {"0000","0001","0010","0011","0100","0101","0110","0111","1000","1001","1010","1011","1100","1101","1110","1111"};
    int Digit1, Digit2;
	String Dig1 = W.SubString(1,4);
    Dig1 = Dig1.UpperCase();
    for (Digit1 = 0 ; Digit1 < 16 ; Digit1++) {
        if (Dig1 == Binary[Digit1]) {
        	break;
            }
    	}
	String Dig2 = W.SubString(5,4);
    Dig2 = Dig2.UpperCase();
    for (Digit2 = 0 ; Digit2 < 16 ; Digit2++) {
        if (Dig2 == Binary[Digit2]) {
        	break;
            }
    	}
if (Digit1 == 16 || Digit2 == 16)
	return "0000";
    else
    return Encoded[Digit1] + Encoded[Digit2];
}
//---------------------------------------------------------------------------

String bitswapRX(String S)
{
//
// Do a bit swap on a received word
// Convert from Morton Numbers to Binary
//
String bitswap =
    S.SubString(1,1) +
    S.SubString(5,1) +
    S.SubString(2,1) +
    S.SubString(6,1) +
    S.SubString(3,1) +
    S.SubString(7,1) +
    S.SubString(4,1) +
    S.SubString(8,1);
return bitswap;
}

//---------------------------------------------------------------------------

String bitswapTX(String S)
{
//
// Do a bit swap on a word to be transmitted
// Convert from Binary into Morton Numbers
//
String bitswap =
    S.SubString(1,1) +
    S.SubString(3,1) +
    S.SubString(5,1) +
    S.SubString(7,1) +
    S.SubString(2,1) +
    S.SubString(4,1) +
    S.SubString(6,1) +
    S.SubString(8,1);
return bitswap;
}
//---------------------------------------------------------------------------

unsigned char bintodec8(String W)
{
// 8 bit binary String to 8 bit unsigned char
unsigned char result = 0;
result += (W.SubString(1,1) == "1") ? (unsigned char)128 : (unsigned char)0;
result += (W.SubString(2,1) == "1") ? (unsigned char)64 : (unsigned char)0;
result += (W.SubString(3,1) == "1") ? (unsigned char)32 : (unsigned char)0;
result += (W.SubString(4,1) == "1") ? (unsigned char)16 : (unsigned char)0;
result += (W.SubString(5,1) == "1") ? (unsigned char)8 : (unsigned char)0;
result += (W.SubString(6,1) == "1") ? (unsigned char)4 : (unsigned char)0;
result += (W.SubString(7,1) == "1") ? (unsigned char)2 : (unsigned char)0;
result += (W.SubString(8,1) == "1") ? (unsigned char)1 : (unsigned char)0;
return result;
}
//---------------------------------------------------------------------------

String dectobin8(int num)  // 8 bit to 8 bit binary String
{
// 8 bit to 8 bit binary String
const static int steps[] = {128, 64, 32, 16, 8, 4, 2, 1};
String binary;

num = num & 0x00FF;	// this routine only handles unsigned char (0-255)

if(num > 0)	{
	int stepLen = sizeof(steps) / sizeof(int);
	for(int x = 0; x < stepLen; x++) {
		if(num >= steps[x]) {
			num -= steps[x];
			binary += "1";
			} else
			binary += "0";
		}
	}
	else {
	binary = "00000000";
    }
return binary;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PasteIntoRFIDBufferClick(TObject *Sender)
{
// Remote Buffer popup menu item
// Text is Green
RFIDBuffer->SelAttributes->Color = clGreen;
RFIDBuffer->PasteFromClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CopyIntoRFIDBufferClick(TObject *Sender)
{
// Remote Buffer popup menu item
RFIDBuffer->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SelAllRFIDBufferClick(TObject *Sender)
{
// Remote Buffer popup menu item
RFIDBuffer->SelectAll();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SendHexFromRemoteBufferClick(TObject *Sender)
{
// Remote buffer popup menu
String CommandA = RFIDBuffer->SelText;
if (!CommandA.IsEmpty()) {
	CommandA = CleanUpCommand(CommandA);
    if (ValidCommand(CommandA)) {
    	// Does not record or handle errors
		(void) ptrTIUCommThread->SendHexStrStr_FIFO(TIU_PORT, CommandA, CommandA.Length()/2);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ExpandRemoteCommandClick(TObject *Sender)
{
// Remote Buffer popup menu item
// insert spaces into command line

// If there is a selected text active, then expand that selected text
// Expand the selection from the Main Buffer
TPoint CaretPos = RFIDBuffer->CaretPos;
const String OriginalCmd = RFIDBuffer->Lines->Strings[CaretPos.y];
//
String MyCmd = RFIDBuffer->SelText;
if (MyCmd.IsEmpty()) {
	// otherwise, if the cursor is on a line,
    // then expand the selection on that line
    // get MyCmd from the cursor line
    MyCmd = CleanUpCommand(OriginalCmd);
    } else {
    // Selection Active, so use the selection
    MyCmd = CleanUpCommand(MyCmd);
    }
//
CRCBuffer->SelAttributes->Color = RFIDBuffer->SelAttributes->Color;	// Copy with same color
CRCBuffer->Lines->Append(OriginalCmd);
//CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
CRCBuffer->HideSelection = false;
CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
CRCBuffer->HideSelection = true;

MyCmd = ExpandCommand(MyCmd); // expand and print the command in MyCmd
if (!MyCmd.IsEmpty()) {
	RFIDBuffer->SelText = MyCmd;
    CRCBuffer->SelAttributes->Color = clBlue;
	CRCBuffer->Lines->Append(MyCmd);
	// Method 4 works under CB10.1 but not BCB 4.0
    CRCBuffer->HideSelection = false;
    CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    CRCBuffer->HideSelection = true;
	}
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::ValidCommand(String Command)
{
if (!Command.IsEmpty()) {
	if (DEBUGGING) DB->Debug_Print("ValidCommand: " + Command);
   	Command = CleanUpCommand(Command);
    // must be a command of correct length
    const int CommandLen = Command.Length();
    if (CommandLen < 30) {		// minimum command lenght Sync + Engine + command + Seq + TIU/Rem + CRCH + CRCL + EOT
        if (DEBUGGING) DB->Debug_Print("ValidCommand: Error: Command less than 30 encoded characters", clRed);
        return false;
        }
    // Command must start with 5555
    if (Command.SubString(1,4) != "5555") {
    	if (DEBUGGING) DB->Debug_Print("ValidCommand: Error: Command does not start with 5555", clRed);
    	return false;
        }
    // Command must end with 4D, last character
    if (Command.SubString(CommandLen-1,2) != "4D") {
        if (DEBUGGING) DB->Debug_Print("ValidCommand: Error: Command does not end with 4D", clRed);
        return false;
    	}
    } else {
    // command is empty
    if (DEBUGGING) DB->Debug_Print("ValidCommand: Error: Command is empty", clRed);
    return false;
    }
return true;
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::ExpandCommand(String Command)
{
if (!Command.IsEmpty()) {
	//if (DEBUGGING) DB->Debug_Print("Expand Command: " + Command);
   	Command = CleanUpCommand(Command);
    //if (ValidCommand(Command)) {
		// insert spaces
        int CommandLen = Command.Length();
        int idxec2 = 5;
        for (int idxec = 5 ; idxec < CommandLen ; idxec += 4) {
        	Command.Insert(" ", idxec2);
            idxec2 += 5;
        	}
        // adjust spaces to line up the command bits
        if (CommandLen == 30) {
            Command.Insert("                    ", 16);
        	} else {
            if (CommandLen == 34) {
                Command.Insert("               ", 21);
            	} else {
                if (CommandLen == 38) {
                    Command.Insert("          ", 26);
                	} else {
                	if (CommandLen == 42) {
                        Command.Insert("     ", 31);
                		}
                    }
                }
            }
        return Command;
        // }
    }
return "";
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ExpandRemoteBufferClick(TObject *Sender)
{
// Remote Buffer popup menu item

// expand Commands in entire buffer
// insert spaces into command line
if (DEBUGGING) STOPem->Enabled = true;
bStop = false;

for (int idxeb = 0; idxeb < RFIDBuffer->Lines->Count; idxeb++) {
	if (bStop) {
       	break;
        }
    String FullLine = RFIDBuffer->Lines->Strings[idxeb];
    String Command = CleanUpCommand(RFIDBuffer->Lines->Strings[idxeb]);
    // find the command on the line if there is one
    int Endpos = FullLine.Pos("4D") + 1;
    if (Endpos == 1) {
    	continue; // no command on the line
        }
    Command = Command.SubString(1, Endpos);
    int CommandLen = Command.Length();

	if (!Command.IsEmpty()) {
		 String NewCommand = ExpandCommand(Command);
		 if (!NewCommand.IsEmpty()) {
         	// replace only the part of the Line that contains the command
            // to retain any comments
    	 	FullLine.Delete(1, CommandLen);
            FullLine.Insert(NewCommand, 1);
            RFIDBuffer->Lines->Strings[idxeb] = FullLine;
		    CRCBuffer->SelAttributes->Color = clBlue;
            CRCBuffer->Lines->Append(FullLine);
			// Method 4 works under CB10.1 but not BCB 4.0
    		CRCBuffer->HideSelection = false;
    		CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    		CRCBuffer->HideSelection = true;
        	}
	    }
	Application->ProcessMessages();
    }
STOPem->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Encode_TIUClick(TObject *Sender)
{
// encode a command
if (D_EngNo->Text == "-") return;
Encoded_CommandA->Clear();

// Sync Byte
Encoded_CommandA->Text = Encoded_CommandA->Text + "5555";

// Engine Number
Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_EngNo->Text.ToInt())));

// Command
Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Command->Text.ToInt())));

//  Data Bytes
if (D_Data1->Text  != "-" ) {
	Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data1->Text.ToInt())));
    }
if (D_Data2->Text  != "-" ) {
	Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data2->Text.ToInt())));
    }
if (D_Data3->Text  != "-" ){
	Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data3->Text.ToInt())));
    }
if (D_Data4->Text  != "-" ) {
	Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data4->Text.ToInt())));
    }
Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Seq_Count->Text.ToInt())));

// TIU_Remote Byte
Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_TIU_Remote->Text.ToInt())));

// CRC
Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_CRC_LSB->Text.ToInt())));
Encoded_CommandA->Text = Encoded_CommandA->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_CRC_MSB->Text.ToInt())));

// EOT
Encoded_CommandA->Text = Encoded_CommandA->Text + "4D";
//if (DEBUGGING) DB->Debug_Print("Encode_AClick: " + Encoded_CommandA->Text);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Encode_RemoteClick(TObject *Sender)
{
// encode a command
if (D_EngNo->Text == "-") return;
Encoded_CommandB->Clear();

// Sync Byte
Encoded_CommandB->Text = Encoded_CommandB->Text + "5555";

// Engine Number
Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_EngNo->Text.ToInt())));

// Command
Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Command->Text.ToInt())));

// Data Bytes
if (D_Data1->Text  != "-" ) {
	Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data1->Text.ToInt())));
    }
if (D_Data2->Text  != "-" ) {
	Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data2->Text.ToInt())));
    }
if (D_Data3->Text  != "-" ){
	Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data3->Text.ToInt())));
    }
if (D_Data4->Text  != "-" ) {
	Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Data4->Text.ToInt())));
    }
Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_Seq_Count->Text.ToInt())));

// TIU_Remote Byte
Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_TIU_Remote->Text.ToInt())));

// CRC
Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_CRC_LSB->Text.ToInt())));
Encoded_CommandB->Text = Encoded_CommandB->Text + Convert8B16BNumber(bitswapTX(dectobin8(D_CRC_MSB->Text.ToInt())));

// EOT
Encoded_CommandB->Text = Encoded_CommandB->Text + "4D";
//if (DEBUGGING) DB->Debug_Print("EncodeBClick: " + Encoded_CommandB->Text);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodeCommandBufferClick(TObject *Sender)
{
// Remote Buffer and TIU Buffer popup menu item

TRichEdit * SelectedBuffer;
TPopupMenu * MenuUsed;
MenuUsed = (TPopupMenu *)Sender;
if (MenuUsed->Name.Pos("RFID")) SelectedBuffer = RFIDBuffer;
if (MenuUsed->Name.Pos("TIU")) SelectedBuffer = TIUBuffer;

// Decode all commands in the Selected Buffer and write them to the CRC Buffer
if (DEBUGGING) STOPem->Enabled = true;
bStop = false;
//
const int count = SelectedBuffer->Lines->Count;

for (int idxdb = 0 ; idxdb < count ; idxdb++) {
   if (bStop) {
        break;
        }
	String Command = SelectedBuffer->Lines->Strings[idxdb];
    CRCBuffer->SelAttributes->Color = clBlue;
   	CRCBuffer->Lines->Append(Command);
	// Method 4 works under CB10.1 but not BCB 4.0
    CRCBuffer->HideSelection = false;
    CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    CRCBuffer->HideSelection = true;

    Command = CleanUpCommand(Command);
	if (ValidCommand(Command)) {
        DecodePrint(Command); // Decode and print the command found on the line
    	}
	Application->ProcessMessages();
	}
STOPem->Enabled = false;
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::CleanUpCommand(String Command)
{
	////if (DEBUGGING) DB->Debug_Print("CleanUpCommand: " + Command);
	// remove spaces if any are there
	while (true) {
    	const int position = Command.Pos(" ");
    	if (position == 0) break;
    	// remove space
    	Command.Delete(position, 1);
		}
    // remove zeroes
    while (true) {
    	const int position = Command.Pos("0");
    	if (position == 0) break;
    	// remove zero
    	Command.Delete(position, 1);
		}
    // make the string upper case
    Command = Command.UpperCase();
    // remove anything after the 4D if there is a 4D on the line
    const int Endpos = Command.Pos("4D") + 1;
    if (Endpos == 1) {
    	return Command; // no 4D on the line so this command will fail future checks
        }
    // remove everything after the 4D
    Command = Command.SubString(1, Endpos);
	return Command;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PrintCommand(void)
{
	// insert decoded command into the RFIDBuffer
    TCHAR os[100];
    CRCBuffer->SelAttributes->Color = clBlue;
	CRCBuffer->Lines->Append("   Sync  Eng  Cmd    1    2    3    4  Seq  T/R  LSB  MSB");
	CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);// remain at end of CRC Buffer
    CRCBuffer->SelAttributes->Color = clBlue;
	CRCBuffer->Lines->Append("   ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ----");
	CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
    _sntprintf(os, sizeof os, TEXT("      0%5d%5d%5.0d%5.0d%5.0d%5.0d%5d%5d%5d%5d"),	
            	D_EngNo->Text == "-" ? 0 : D_EngNo->Text.ToInt(),
                D_Command->Text == "-" ? 0 : D_Command->Text.ToInt(),
                D_Data1->Text == "-" ? 0 : D_Data1->Text.ToInt(),
                D_Data2->Text == "-" ? 0 : D_Data2->Text.ToInt(),
                D_Data3->Text == "-" ? 0 : D_Data3->Text.ToInt(),
                D_Data4->Text == "-" ? 0 : D_Data4->Text.ToInt(),
                D_Seq_Count->Text == "-" ? 0 : D_Seq_Count->Text.ToInt(),
                D_TIU_Remote->Text == "-" ? 0 : D_TIU_Remote->Text.ToInt(),
                D_CRC_LSB->Text == "-" ? 0 : D_CRC_LSB->Text.ToInt(),
                D_CRC_MSB->Text == "-" ? 0 : D_CRC_MSB->Text.ToInt()
                );
    CRCBuffer->SelAttributes->Color = clBlue;
    CRCBuffer->Lines->Append((String)os);
	CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
    _sntprintf(os, sizeof os, TEXT("   0x00 0x%2s 0x%2s %4s %4s %4s %4s 0x%2s 0x%2s 0x%2s 0x%2s"),
            	H_EngNo->Text.c_str(),
                H_Command->Text.c_str(),
				H_Data1->Text == "-" ? TEXT("    ") : (TEXT("0x") + H_Data1->Text).c_str(),
				H_Data2->Text == "-" ? TEXT("    ") : (TEXT("0x") + H_Data2->Text).c_str(),
				H_Data3->Text == "-" ? TEXT("    ") : (TEXT("0x") + H_Data3->Text).c_str(),
				H_Data4->Text == "-" ? TEXT("    ") : (TEXT("0x") + H_Data4->Text).c_str(),
                H_Seq_Count->Text.c_str(),
                H_TIU_Remote->Text.c_str(),
                H_CRC_LSB->Text.c_str(),
                H_CRC_MSB->Text.c_str()
                );
    CRCBuffer->SelAttributes->Color = clBlue;
    CRCBuffer->Lines->Append((String)os);
	CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer	
    // print command byte in ASCII
    String A_Command;
    if (H_Command->Text != "-") {
          A_Command = (char)D_Command->Text.ToInt();  // Convert to ASCII Character	
    	  A_Command = "'" + A_Command + "'";
          }
    // print data bytes in ASCII
    String A_Data1, A_Data2, A_Data3, A_Data4;
    if (H_Data1->Text != "-") {
          A_Data1 = (char)D_Data1->Text.ToInt();  // Convert to ASCII Character 	
    	  A_Data1 = "'" + A_Data1 + "'";
          }
    if (H_Data2->Text != "-") {
          A_Data2 = (char)D_Data2->Text.ToInt();  // Convert to ASCII Character	
    	  A_Data2 = "'" + A_Data2 + "'";
          }
    if (H_Data3->Text != "-") {
          A_Data3 = (char)D_Data3->Text.ToInt();  // Convert to ASCII Character	
    	  A_Data3 = "'" + A_Data3 + "'";
          }
    if (H_Data4->Text != "-") {
          A_Data4 = (char)D_Data4->Text.ToInt();  // Convert to ASCII Character	
    	  A_Data4 = "'" + A_Data4 + "'";
          }
    _sntprintf(os, sizeof os, TEXT("             %4s %4s %4s %4s %4s"),
    			H_Command->Text == "-" ? TEXT("    ") : A_Command.c_str(),
                H_Data1->Text == "-" ? TEXT("    ") : A_Data1.c_str(),
                H_Data2->Text == "-" ? TEXT("    ") : A_Data2.c_str(),
                H_Data3->Text == "-" ? TEXT("    ") : A_Data3.c_str(),
                H_Data4->Text == "-" ? TEXT("    ") : A_Data4.c_str()
                );
    CRCBuffer->SelAttributes->Color = clBlue;
    CRCBuffer->Lines->Append((String)os);
	// Method 4 works under CB10.1 but not BCB 4.0
    CRCBuffer->HideSelection = false;
    CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    CRCBuffer->HideSelection = true;
    }
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodePrint(String Command)
{
// Decode and print Command
if (DecodeMyCommand(Command)) {
	PrintCommand();
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodeCommandClick(TObject *Sender)
{
// Remote Buffer and TIU Buffer popup menu item

TRichEdit * SelectedBuffer;
TPopupMenu * MenuUsed;
MenuUsed = (TPopupMenu *)Sender;
if (MenuUsed->Name.Pos("RFID")) SelectedBuffer = RFIDBuffer;
if (MenuUsed->Name.Pos("TIU")) SelectedBuffer = TIUBuffer;

//If there is a selected text active, then decode that selected text
//Decode the selected command from the Main Buffer
TPoint CaretPos = SelectedBuffer->CaretPos;
String OriginalCmd = SelectedBuffer->Lines->Strings[CaretPos.y];
//
String MyCmd = SelectedBuffer->SelText;
if (MyCmd.IsEmpty()) {
	// otherwise, if the cursor is on a line with a valid command,
    // then decode the command on that line
    // get MyCmd from the cursor line
    MyCmd = CleanUpCommand(OriginalCmd);
	if (!ValidCommand(MyCmd)) {
    	return;
        }
    } else {
    // Selection Active, so use the selection
    MyCmd = CleanUpCommand(MyCmd);
    }
//
CRCBuffer->SelAttributes->Color = SelectedBuffer->SelAttributes->Color;	// Copy with same color
CRCBuffer->Lines->Append(OriginalCmd);
// Method 4 works under CB10.1 but not BCB 4.0
CRCBuffer->HideSelection = false;
CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
CRCBuffer->HideSelection = true;

DecodePrint(MyCmd); // Decode and print the command in MyCmd
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ClearDecode() {
// clear all of the decode fields - only the H_* field needs to be cleared
// Its event handler will clear the D_* and A_* fields
EngNo->Text = "-";
H_EngNo->Text = "-";

Command->Text = "-";
H_Command->Text = "-";

Data1->Text = "-";
H_Data1->Text = "-";

Data2->Text = "-";
H_Data2->Text = "-";

Data3->Text = "-";
H_Data3->Text = "-";

Data4->Text = "-";
H_Data4->Text = "-";

Seq_Count->Text = "-";
H_Seq_Count->Text = "-";

TIU_Remote->Text = "-";
H_TIU_Remote->Text = "-";

CRC_LSB->Text = "-";
H_CRC_LSB->Text = "-";

CRC_MSB->Text = "-";
H_CRC_MSB->Text = "-";

return;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::DecodeMyCommand(String MyCmd) {
// Decoded a command into separate fields.
//
// Return true if successful
// return false if error
//
//if (DEBUGGING) DB->Debug_Print("DecodeMyCommand: " + MyCmd);

ClearDecode();	// Clear all fields

if (!MyCmd.IsEmpty()) {
   	// remove spaces if any are there
    MyCmd = CleanUpCommand(MyCmd);
    if (ValidCommand(MyCmd)) {
        int CommandLen = MyCmd.Length();
    	//
    	EngNo->Text = MyCmd.SubString(5, 4);
		String C_EngNo = Convert16B8BNumber(EngNo->Text);
		String I_EngNo = bintodec8(bitswapRX(C_EngNo));
    	H_EngNo->Text = IntToHex(I_EngNo.ToInt(), 2);
    	EngNo->Text = MyCmd.SubString(5, 4);
    	//
    	Command->Text = MyCmd.SubString(9, 4);
        String C_Command = Convert16B8BNumber(Command->Text);
		String I_Command = bintodec8(bitswapRX(C_Command));
    	H_Command->Text = IntToHex(I_Command.ToInt(), 2);
    	Command->Text = MyCmd.SubString(9, 4);
    	// Decode first 4 data bytes
    	if (CommandLen > 30) {
	    	Data1->Text = MyCmd.SubString(13, 4);
	        String C_Data1 = Convert16B8BNumber(Data1->Text);
    		String D_Data1 = bintodec8(bitswapRX(C_Data1));
			H_Data1->Text = IntToHex(D_Data1.ToInt(), 2);
	    	Data1->Text = MyCmd.SubString(13, 4);
        	//
	    	if (CommandLen > 34) {
	   	    	Data2->Text = MyCmd.SubString(17, 4);
                String C_Data2 = Convert16B8BNumber(Data2->Text);
			    String D_Data2 = bintodec8(bitswapRX(C_Data2));
				H_Data2->Text = IntToHex(D_Data2.ToInt(), 2);
	   	    	Data2->Text = MyCmd.SubString(17, 4);
            	//
	        	if (CommandLen > 38) {
	              	Data3->Text = MyCmd.SubString(21, 4);
                    String C_Data3 = Convert16B8BNumber(Data3->Text);
				    String D_Data3 = bintodec8(bitswapRX(C_Data3));
					H_Data3->Text = IntToHex(D_Data3.ToInt(), 2);
	              	Data3->Text = MyCmd.SubString(21, 4);
   	              	//
	              	if (CommandLen > 42 ) {
	                    Data4->Text = MyCmd.SubString(25, 4);
                    	String C_Data4 = Convert16B8BNumber(Data4->Text);
					    String D_Data4 = bintodec8(bitswapRX(C_Data4));
						H_Data4->Text = IntToHex(D_Data4.ToInt(), 2);
	                    Data4->Text = MyCmd.SubString(25, 4);
                        //
                        }
                  	}
            	}
    		}
        // Sequence number, 4th byte from the end
    	Seq_Count->Text = MyCmd.SubString(CommandLen-17,4);
	   	String C_Seq_Count = Convert16B8BNumber(Seq_Count->Text);
		String D_Seq_Count = bintodec8(bitswapRX(C_Seq_Count));
	    H_Seq_Count->Text = IntToHex(D_Seq_Count.ToInt(), 2);
    	Seq_Count->Text = MyCmd.SubString(CommandLen-17,4);
    	//  TIU/Remote, 3rd byte from the end
    	TIU_Remote->Text = MyCmd.SubString(CommandLen-13,4);
		String C_TIU_Remote = Convert16B8BNumber(TIU_Remote->Text);
    	String D_TIU_Remote = bintodec8(bitswapRX(C_TIU_Remote));
		H_TIU_Remote->Text = IntToHex(D_TIU_Remote.ToInt(), 2);
    	TIU_Remote->Text = MyCmd.SubString(CommandLen-13,4);
    	//  LSB - Little Endian for the PC, 2nd byte from the end
    	CRC_LSB->Text = MyCmd.SubString(CommandLen-9,4);
    	String C_CRC_LSB = Convert16B8BNumber(CRC_LSB->Text);
    	String D_CRC_LSB = bintodec8(bitswapRX(C_CRC_LSB));
		H_CRC_LSB->Text = IntToHex(D_CRC_LSB.ToInt(), 2);
    	CRC_LSB->Text = MyCmd.SubString(CommandLen-9,4);
        // MSB, end byte
    	CRC_MSB->Text = MyCmd.SubString(CommandLen-5,4);
        String C_CRC_MSB = Convert16B8BNumber(CRC_MSB->Text);
    	String D_CRC_MSB = bintodec8(bitswapRX(C_CRC_MSB));
		H_CRC_MSB->Text = IntToHex(D_CRC_MSB.ToInt(), 2);
    	CRC_MSB->Text = MyCmd.SubString(CommandLen-5,4);
		//
    	CommandDecoded = true;
		return true;
    	}
   	}
CommandDecoded = false;
return false;
}
//---------------------------------------------------------------------------

/******************************************************************************/
/*                             Start of crcmodel.c                            */
/******************************************************************************/
/*                                        	                                  */
/* Author : Ross Williams (ross@guest.adelaide.edu.au.).                      */
/* Date   : 3 June 1993.                                                      */
/* Status : Public domain.                                                    */
/*                                                                            */
/* Description : This is the implementation (.c) file for the reference       */
/* implementation of the Rocksoft Model CRC Algorithm. For more              */
/* information on the Rocksoft Model CRC Algorithm, see the document         */
/* titled "A Painless Guide to CRC Error Detection Algorithms" by Ross        */
/* Williams (ross@guest.adelaide.edu.au.). This document is likely to be in   */
/* "ftp.adelaide.edu.au/pub/rocksoft".                                        */
/*                                                                            */
/* Note: Rocksoft is a trademark of Rocksoft Pty Ltd, Adelaide, Australia.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Implementation Notes                                                       */
/* --------------------                                                       */
/* To avoid inconsistencies, the specification of each function is not echoed */
/* here. See the header file for a description of these functions.            */
/* This package is light on checking because I want to keep it short and      */
/* simple and portable (i.e. it would be too messy to distribute my entire    */
/* C culture (e.g. assertions package) with this package.                     */
/*                                                                            */
/******************************************************************************/

/* The following definitions make the code more readable. */

#define BITMASK(X) (1L << (X))
#define MASK32 0xFFFFFFFFL
#define LOCAL static

/******************************************************************************/

//LOCAL ulong reflect P_((ulong v,int b));
LOCAL ulong reflect (ulong v,int b)
/* Returns the value v with the bottom b [0,32] bits reflected. */
/* Example: reflect(0x3e23L,3) == 0x3e26                        */
//ulong v;
//int   b;
{
 int   i;
 ulong t = v;
 for (i=0; i<b; i++)
   {
    if (t & 1L)
       v|=  BITMASK((b-1)-i);
    else
       v&= ~BITMASK((b-1)-i);
    t>>=1;
   }
 return v;
}

/******************************************************************************/

//LOCAL ulong widmask P_((p_cm_t));
LOCAL ulong widmask (p_cm_t p_cm)
/* Returns a longword whose value is (2^p_cm->cm_width)-1.     */
/* The trick is to do this portably (e.g. without doing <<32). */
//p_cm_t p_cm;
{
 return (((1L<<(p_cm->cm_width-1))-1L)<<1)|1L;
}

/******************************************************************************/

void cm_ini (p_cm_t p_cm)
//p_cm_t p_cm;
{
 p_cm->cm_reg = p_cm->cm_init;
}

/******************************************************************************/

void cm_nxt (p_cm_t p_cm, int ch)
//p_cm_t p_cm;
//int    ch;
{
 int   i;
 char os[CMDMAX];
 ulong uch  = (ulong) ch;
 ulong topbit = BITMASK(p_cm->cm_width-1);  // for 16 bit CRC topbit is 0x00008000

 if (p_cm->cm_refin) uch = reflect(uch,8);

 p_cm->cm_reg ^= (uch << (p_cm->cm_width-8));  // for 16 bit CRC, this is left shift 8
	if (Setup->CRC_Trace->Checked) {
		snprintf(os, sizeof os, "+ cm_nxt uch 0x%02X cm_reg 0x%04X cm_poly 0x%04X",
    		uch,
            p_cm->cm_reg,
            p_cm->cm_poly
            );
	    MainForm->CRCBuffer->SelAttributes->Color = clBlue;
    	MainForm->CRCBuffer->Lines->Append(os);
		MainForm->CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
    	}
 for (i=0; i<8; i++)
   {
    if (p_cm->cm_reg & topbit) {   // for 16 bit CRC topbit is 0x00008000
        p_cm->cm_reg = (p_cm->cm_reg << 1) ^ p_cm->cm_poly;
		if (Setup->CRC_Trace->Checked) {
		   MainForm->CRCBuffer->SelAttributes->Color = clBlue;
	       MainForm->CRCBuffer->Lines->Append("shift then XOR");
		   MainForm->CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
           }
    	} else {
		if (Setup->CRC_Trace->Checked) {
		   MainForm->CRCBuffer->SelAttributes->Color = clBlue;
    	   MainForm->CRCBuffer->Lines->Append("shift");
		   MainForm->CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
           }
        p_cm->cm_reg <<= 1;
        }
    p_cm->cm_reg &= widmask(p_cm);

    // trace printout
	if (Setup->CRC_Trace->Checked) {
		snprintf(os, sizeof os, "- cm_nxt uch 0x%02X cm_reg 0x%04X cm_poly 0x%04X",
    		uch & 0xFF,
        	p_cm->cm_reg,
        	p_cm->cm_poly
        	);
	    MainForm->CRCBuffer->SelAttributes->Color = clBlue;
	    MainForm->CRCBuffer->Lines->Append(os);
		MainForm->CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
        }
    }
}

/******************************************************************************/
// this routine not used
void cm_blk (p_cm_t p_cm, p_ubyte_ blk_adr, ulong blk_len)
//p_cm_t   p_cm;
//p_ubyte_ blk_adr;
//ulong    blk_len;
{
 while (blk_len--) cm_nxt(p_cm, *blk_adr++);
}

/******************************************************************************/

ulong cm_crc (p_cm_t p_cm)
//p_cm_t p_cm;
{
 if (p_cm->cm_refot)
    return p_cm->cm_xorot ^ reflect(p_cm->cm_reg, p_cm->cm_width);
 else
    return p_cm->cm_xorot ^ p_cm->cm_reg;
}

/******************************************************************************/
// this routine not used
ulong cm_tab (p_cm_t p_cm, int index)
//p_cm_t p_cm;
//int    index;
{
 int   i;
 ulong r;
 ulong topbit = BITMASK(p_cm->cm_width-1);
 ulong inbyte = (ulong) index;

 if (p_cm->cm_refin) inbyte = reflect(inbyte,8);
 r = inbyte << (p_cm->cm_width-8);
 for (i=0; i<8; i++)
    if (r & topbit)
       r = (r << 1) ^ p_cm->cm_poly;
    else
       r<<=1;
 if (p_cm->cm_refin) r = reflect(r,p_cm->cm_width);
 return r & widmask(p_cm);
}

/******************************************************************************/
/*                             End of crcmodel.c                              */
/******************************************************************************/
//---------------------------------------------------------------------------
/*
Example of call from RetroXFMRE when creating a packet to send to the TIU
	// build command
	unsigned char Cmd_String[64];
	// Sync
	Cmd_String[0] = 0;
	// Engine Number
	Cmd_String[1] = EngineNo + 1;
	// Command
	int len = strlen((char *)Command_String);	// ASCII Command in Command_String, eg: "m4"
	for (int index = 0 ; index < len; index++) {
  		Cmd_String[2 + index] = Command_String[index];
  		}
	// Sequence Number
	Cmd_String[2 + len] = SeqNo++;
	// TIU / Remote
	Cmd_String[3 + len] = (((TIU - 1) & 0x000F) << 4) | (Remote & 0x000F);
	// calcualte CRC, skipping over the Sync
	unsigned int CRC = crc752(Cmd_String+1, (3 + len), false);
	byte CRCH = (CRC >> 8) & 0xFF;
	Cmd_String[4 + len] = CRCH;
	byte CRCL = CRC & 0xFF;
	Cmd_String[5 + len] = CRCL;

Example of call from RTCEngineE when creating a packet to send to the Remote

  const char bEM[] PROGMEM =  {0x00, 0x04, "0", "5", "0", "0"};  // TIU version Number (hardcoded)
  // calculate CRC
  unsigned char R[sizeof(bEM) + 1];
  int len = sizeof(bEM);  // not counting NULL
  BuildResponse(sizeof(bEM), bEM, R, true); =
  unsigned int CRC = crc752(R, len + 1, true);
  // Send CRC first
  byte CRCH = (CRC >> 8) & 0xFF;
  byte CRCL = CRC & 0xFF;
  SendBinaryResponse(CRCH);
  SendBinaryResponse(CRCL);
  // Send rest of Response packet
  for (int i = 0 ; i < len + 1 ; i++) {
    SendBinaryResponse(R[i]);	// convert to Morton, endcode and send the byte
    }
*/
//---------------------------------------------------------------------------
#define poly      0x8408    // Polynomial used to compute CRC
//
// CRC Calculator
// Based on routine written by Paul Reynolds, used with permission
//
// modified by Mark DiVecchio to handle larger packets and to handle both forward
// and reverse order CRC generation
//
unsigned int __fastcall crc752(byte message[], byte nBytes, bool Direction) {
// Compute crc using 16 bit MCRF45X with poly 0x8408, with left to right processing of input.
// Direction: true : packets from the TIU
//            false : packets to the TIU
//
unsigned int    remainder = 0xFFFF;  // Set remainder to initial remainder
short           bytePtr;
byte            bitCount, crcTemp, nextBit;
//byte maskreg[8] = {1, 2, 4, 8, 16, 32, 64, 128};
byte maskreg[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

const short numBits = nBytes * 8;

// This FOR loop does all of the CRC computation work.
for (short i = 0; i < numBits; i++) {
    // process input byte from leftmost byte to rightmost byte
    bitCount = i & 0x07 ; // (was i % 8) repeatedly counts from zero to seven as FOR loop iterates
    bytePtr = i >> 3; // (was i / 8) points at bytes, from left to right, shifting left one every eight iterations.
    nextBit = (message[Direction ? bytePtr : nBytes - bytePtr - 1] & maskreg[bitCount]); //This will find the next data bit to apply, not yet shifted.
    nextBit = ((nextBit >> bitCount) & 0x01); //This will set nextBit to zero or one, depending on bit masked out, above.
    crcTemp = remainder ^ nextBit; //xor the last bit of remainder with next_bit
    if (crcTemp & 0x01) {
      remainder = remainder >> 1; //Shift the remainder by 1 to right
      remainder = remainder ^ poly ; //xor current remainder with chosen polynomial (poly)
      } else remainder = remainder >> 1;
  }
return Direction ? remainder : ((remainder & 0xFF00) >> 8) | ((remainder & 0x00FF) << 8);
}
//---------------------------------------------------------------------------

int __fastcall TMainForm::CreateCommandString(Cmd * Cmd_Array) // Create Command String from the HEX display boxes
{
	Cmd_Array->idx = 0;
    Cmd_Array->Str[Cmd_Array->idx++] = 0; // Sync
	Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_EngNo->Text.c_str(), NULL , 16); // D_EngNo->Text.ToInt();
	Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_Command->Text.c_str(), NULL , 16); // D_Command->Text.ToInt();
	if ((H_Data1->Text != "-") && (!H_Data1->Text.IsEmpty())){
		Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_Data1->Text.c_str(), NULL , 16); // D_Data1->Text.ToInt();
        }
	if ((H_Data2->Text != "-") && (!H_Data2->Text.IsEmpty())){
		Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_Data2->Text.c_str(), NULL , 16); // D_Data2->Text.ToInt();
        }
	if ((H_Data3->Text != "-") && (!H_Data3->Text.IsEmpty())){
		Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_Data3->Text.c_str(), NULL , 16); // D_Data3->Text.ToInt();
		}
	if ((H_Data4->Text != "-") && (!H_Data4->Text.IsEmpty())){
		Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_Data4->Text.c_str(), NULL , 16); // D_Data4->Text.ToInt();
        }
	Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_Seq_Count->Text.c_str(), NULL , 16); // D_Seq_Count->Text.ToInt();
	Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_TIU_Remote->Text.c_str(), NULL , 16); // D_TIU_Remote->Text.ToInt();
	Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_CRC_LSB->Text.c_str(), NULL , 16); // D_CRC_LSB->Text.ToInt();
	Cmd_Array->Str[Cmd_Array->idx++] = (char) _tcstoul(H_CRC_MSB->Text.c_str(), NULL , 16); // D_CRC_MSB->Text.ToInt();
    Cmd_Array->Str[Cmd_Array->idx] = NULL;    // NULL Terminated
    Cmd_Array->ready = true;
return Cmd_Array->idx; // return item count in array
}
//---------------------------------------------------------------------------

unsigned int __fastcall TMainForm::CRCGen(Cmd * Command, int StartChar, int EndChar, bool Quiet)
{
//
// This function generates CRC for packets to sent to the TIU for CRC-16/MCRF4XX
//
// NOTE: Cmd_String may contain binary zero
//
cm_t var;
p_cm_t p_cm = &var;
p_cm->cm_width = 16;
p_cm->cm_poly  = 0x1021L;
p_cm->cm_init  = 0x0000FFFFL;
p_cm->cm_refin = TRUE;
p_cm->cm_refot = TRUE;
p_cm->cm_xorot = 0L;

cm_ini(p_cm);

// feed in the data bytes in reverse order
for (int ii = EndChar ; ii >= StartChar ; ii--) {
	cm_nxt(p_cm, Command->Str[ii]);
    }
//
ulong const crc_result = cm_crc(p_cm) & 0xFFFF;
#ifdef MOREDEBUGGING
char os[MAX_PATH];
unsigned char const msb_byte_c1 = (crc_result & 0xFF00) >> 8;
unsigned char const lsb_byte_c1 = (crc_result & 0x00FF);
#endif

// Insert the CRC byte in the command string - LSB then MSB
Command->Str[EndChar+1] = (crc_result & 0x00FF);
Command->Str[EndChar+2] = (crc_result & 0xFF00) >> 8;
// Insert the NULL
Command->Str[EndChar+3] = NULL;	    // NULL Terminated
// then put it in Command Decode area in preparation for sending
if (DEBUGGING) {
	H_CRC_LSB->Text = IntToHex(Command->Str[EndChar+1], 2);
	H_CRC_MSB->Text = IntToHex(Command->Str[EndChar+2], 2);
    }

#ifdef MOREDEBUGGING
// This routine called from Comm Thread so these VCL calls should not be used
if (DEBUGGING) {
	// don't display CRC details
	snprintf(os, sizeof os, "CRC Gen : 0x%02X 0x%02X", lsb_byte_c1,	msb_byte_c1);
	if (!Quiet) DB->Debug_Print(os);
	if (!Quiet) {
		CRCBuffer->SelAttributes->Color = clBlue;
		CRCBuffer->Lines->Append(os);
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
		CRCBuffer->SelAttributes->Color = clBlue;
		CRCBuffer->Lines->Append ("Done.");
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	    }
  	}
#endif
return crc_result;
}
//---------------------------------------------------------------------------

unsigned int __fastcall TMainForm::CRCCheck(Cmd * Cmd_String, int StartChar, int EndChar, bool Quiet)
{
//
// This function checks packets received from the Remote for CRC for CRC-16/MCRF4XX
//
// NOTE: Cmd_String may contain binary zero
//
cm_t var;
p_cm_t p_cm = &var;
p_cm->cm_width = 16;
p_cm->cm_poly  = 0x1021L;
p_cm->cm_init  = 0x0000FFFFL;
p_cm->cm_refin = TRUE;
p_cm->cm_refot = TRUE;
p_cm->cm_xorot = 0L;

cm_ini(p_cm);

// feed in the data bytes in reverse order
for (int ii = EndChar ; ii >= StartChar ; ii--) {
	cm_nxt(p_cm, Cmd_String->Str[ii]);
    }
// feed in the CRC bytes - LSB then MSB
cm_nxt(p_cm, Cmd_String->Str[EndChar+1]);
cm_nxt(p_cm, Cmd_String->Str[EndChar+2]);
//
ulong const crc_result = cm_crc(p_cm) & 0xFFFF;
#ifdef MOREDEBUGGING
// Can be used for Debugging
char os[MAX_PATH];
if (DEBUGGING) {
	unsigned char msb_byte_c1 = (crc_result & 0xFF00) >> 8;
	unsigned char lsb_byte_c1 = (crc_result & 0x00FF);
	// don't display CRC details
	//snprintf(os, sizeof os, "CRC Check: 0x%02X 0x%02X                   0x%04X 0x%04X %5s %5s 0x%04X",
	snprintf(os, sizeof os, "CRC Check : 0x%02X 0x%02X", lsb_byte_c1, msb_byte_c1);
	DB->Debug_Print(os);
	if (!Quiet) {
		CRCBuffer->SelAttributes->Color = clBlue;
		CRCBuffer->Lines->Append(os);
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
		CRCBuffer->SelAttributes->Color = clBlue;
		CRCBuffer->Lines->Append ("Done.");
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	    }
    }
#endif
return crc_result;
}
//---------------------------------------------------------------------------

unsigned int __fastcall TMainForm::CRCEvaluation(Cmd * Cmd_String, int StartChar, int EndChar, bool Quiet)
{
//
// This function checks packets received from the TIU for CRC for CRC-16/MCRF4XX
//
// NOTE: Cmd_String may contain binary zero
//
//   A9A995955666555555595555555556555955
//     FC  C0  13  00  04  00  00  02  08
// reveng -c -m CRC-16/MCRF4XX 13000400000208c0fc
//
cm_t var;
p_cm_t p_cm = &var;
p_cm->cm_width = 16;
p_cm->cm_poly  = 0x1021L;
p_cm->cm_init  = 0x0000FFFFL;
p_cm->cm_refin = TRUE;
p_cm->cm_refot = TRUE;
p_cm->cm_xorot = 0L;

cm_ini(p_cm);

// feed in the data bytes in forward order    *******
for (int ii = StartChar + 2 ; ii <= EndChar ; ii++) {
	cm_nxt(p_cm, Cmd_String->Str[ii]);
    }
// feed in the CRC bytes - LSB then MSB     ******
cm_nxt(p_cm, Cmd_String->Str[StartChar+1]);
cm_nxt(p_cm, Cmd_String->Str[StartChar]);
//
ulong const crc_result = cm_crc(p_cm) & 0xFFFF;

//  ******   crc_result must be zero

#ifdef MOREDEBUGGING
// Can be used for Debugging
char os[MAX_PATH];
if (DEBUGGING) {
	unsigned char msb_byte_c1 = (crc_result & 0xFF00) >> 8;
	unsigned char lsb_byte_c1 = (crc_result & 0x00FF);
	snprintf(os, sizeof os, "CRC Eval : 0x%02X 0x%02X", lsb_byte_c1, msb_byte_c1);
    DB->Debug_Print(os);

	if (crc_result != 0x0000) {
        DB->Debug_Print("CRC Error", clRed);
        Quiet = false;	// to display the result
    	}
	if (!Quiet) {
		MainForm->CRCBuffer->SelAttributes->Color = clBlue;
        if (crc_result != 0x0000) {
        	CRCBuffer->Lines->Append("CRC Error");
			CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
            }
		CRCBuffer->Lines->Append(os);
		CRCBuffer->SelAttributes->Color = clBlue;
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	    }
    }
#endif
return crc_result;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

unsigned int __fastcall TMainForm::CRCEvaluation(byte * Cmd_String, int StartChar, int EndChar, bool Quiet){
//
// This function checks packets received from the TIU for CRC for CRC-16/MCRF4XX
//
// NOTE: Cmd_String may contain binary zero
//
//   A9A995955666555555595555555556555955
//     FC  C0  13  00  04  00  00  02  08
// reveng -c -m CRC-16/MCRF4XX 13000400000208c0fc
//
cm_t var;
p_cm_t p_cm = &var;
p_cm->cm_width = 16;
p_cm->cm_poly  = 0x1021L;
p_cm->cm_init  = 0x0000FFFFL;
p_cm->cm_refin = TRUE;
p_cm->cm_refot = TRUE;
p_cm->cm_xorot = 0L;

cm_ini(p_cm);

// feed in the data bytes in forward order    *******
for (int ii = StartChar + 2 ; ii <= EndChar ; ii++) {
	cm_nxt(p_cm, Cmd_String[ii]);
    }
// feed in the CRC bytes - LSB then MSB     ******
cm_nxt(p_cm, Cmd_String[StartChar+1]);
cm_nxt(p_cm, Cmd_String[StartChar]);
//
ulong const crc_result = cm_crc(p_cm) & 0xFFFF;

//  ******   crc_result must be zero

#ifdef MOREDEBUGGING
// This routine called from Comm Thread so these VCL calls should not be used
char os[MAX_PATH];
if (DEBUGGING) {
	unsigned char msb_byte_c1 = (crc_result & 0xFF00) >> 8;
	unsigned char lsb_byte_c1 = (crc_result & 0x00FF);
	snprintf(os, sizeof os, "CRC Eval : 0x%02X 0x%02X", lsb_byte_c1, msb_byte_c1);
    DB->Debug_Print(os);

	if (crc_result != 0x0000) {
        DB->Debug_Print("CRC Error", clRed);
        Quiet = false;	// to display the result
    	}
	if (!Quiet) {
		MainForm->CRCBuffer->SelAttributes->Color = clBlue;
        if (crc_result != 0x0000) {
        	CRCBuffer->Lines->Append("CRC Error");
			CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
            }
		CRCBuffer->Lines->Append(os);
		CRCBuffer->SelAttributes->Color = clBlue;
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	    }
    }
#endif
return crc_result;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ProcessCRCButtonClick(TObject *Sender)
{
Cmd Command;
PrintCommand();	// print the command
switch (CRCSel->ItemIndex) {
	case 0:
		Command.idx = CreateCommandString(&Command);	// Fill Cmd_String from the Display Boxes
		(void) CRCGen(&Command, 1, Command.idx - 3, false);	// from Cmd_String - Generate CRC's - verbosely
		break;
    case 1:
		Command.idx = CreateCommandString(&Command);	// Fill Cmd_String from the Display Boxes
		(void) CRCCheck(&Command, 1, Command.idx - 3, false);// from Cmd_String - Check CRC - verbosely
		break;
	}
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::PrintCommandString(String Command) {
    Command = CleanUpCommand(Command);
	if (ValidCommand(Command)) {
	// then Decode it
		if (DecodeMyCommand(Command)) {
            // insert decoded command into the CRCBuffer
            TCHAR os[100];
			_sntprintf(os, sizeof os, TEXT("   0%5d%5d  %3s  %3s  %3s  %3s%5d%5d   %02X   %02X"),	
            	D_EngNo->Text.ToInt(),
                D_Command->Text.ToInt(),
				D_Data1->Text == "-" ? TEXT("   ") : D_Data1->Text.c_str(),
				D_Data2->Text == "-" ? TEXT("   ") : D_Data2->Text.c_str(),
				D_Data3->Text == "-" ? TEXT("   ") : D_Data3->Text.c_str(),
                D_Data4->Text == "-" ? TEXT("   ") : D_Data4->Text.c_str(),
                D_Seq_Count->Text.ToInt(),
                D_TIU_Remote->Text.ToInt(),
                D_CRC_LSB->Text.ToInt(),
                D_CRC_MSB->Text.ToInt()
                );
		    CRCBuffer->SelAttributes->Color = clGreen;
            CRCBuffer->Lines->Append((String)os);
			CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
        	}
    	}
return true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ClearCRCButtonClick(TObject *Sender)
{
CRCBuffer->Clear();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ProcessCRCIntoCRCBufferClick(TObject *Sender)
{
// Remote buffer popup menu item
// for the entire Remote buffer, decode each line then run the CRC generator on it
// All results into CRCBuffer	
// Decode all commands in the Main Buffer and write them to the CRC Buffer
if (DEBUGGING) STOPem->Enabled = true;
bStop = false;
//
const int count = RFIDBuffer->Lines->Count;

for (int idxdb = 0 ; idxdb < count ; idxdb++) {
    if (bStop) {
       	break;
        }
	String Command = RFIDBuffer->Lines->Strings[idxdb];
    CRCBuffer->SelAttributes->Color = clBlue;
   	CRCBuffer->Lines->Append(Command);
    Command = CleanUpCommand(Command);
	CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	if (ValidCommand(Command)) {
        // Decode command found on line
        DecodeMyCommand(Command);
		ProcessCRCButton->OnClick(ProcessCRCButton);
    	}
	Application->ProcessMessages();
	}
STOPem->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_EngNoChange(TObject *Sender)
{
if ((H_EngNo->Text != "-") && (!H_EngNo->Text.IsEmpty())) {
    D_EngNo->Text = _tcstoul(H_EngNo->Text.c_str(), NULL , 16);
    // then clear the Encoded field because it's not valid anymore
    EngNo->Text = "-";
    } else {
    D_EngNo->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_CommandChange(TObject *Sender)
{
if ((H_Command->Text != "-") && (!H_Command->Text.IsEmpty())) {
    D_Command->Text = _tcstoul(H_Command->Text.c_str(), NULL , 16);
    A_Command->Text = (char)D_Command->Text.ToInt();  // Convert to ASCII Character
    A_Command->Text = "'" + A_Command->Text + "'";
    // then clear the Encoded field because it's not valid anymore
    Command->Text = "-";
    } else {
    D_Command->Text = "-";
    A_Command->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_Seq_CountChange(TObject *Sender)
{
if ((H_Seq_Count->Text != "-") && (!H_Seq_Count->Text.IsEmpty())) {
    D_Seq_Count->Text = _tcstoul(H_Seq_Count->Text.c_str(), NULL , 16);
    // then clear the Encoded field because it's not valid anymore
    Seq_Count->Text = "-";
    } else {
    D_Seq_Count->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_TIU_RemoteChange(TObject *Sender)
{
if ((H_TIU_Remote->Text != "-") && (!H_TIU_Remote->Text.IsEmpty())) {
    D_TIU_Remote->Text = _tcstoul(H_TIU_Remote->Text.c_str(), NULL , 16);
    // then clear the Encoded field because it's not valid anymore
    TIU_Remote->Text = "-";
    } else {
    D_TIU_Remote->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_CRC_LSBChange(TObject *Sender)
{
if ((H_CRC_LSB->Text != "-") && (!H_CRC_LSB->Text.IsEmpty())) {
    D_CRC_LSB->Text = _tcstoul(H_CRC_LSB->Text.c_str(), NULL , 16);
    // then clear the Encoded field because it's not valid anymore
    CRC_LSB->Text = "-";
    } else {
    D_CRC_LSB->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_CRC_MSBChange(TObject *Sender)
{
if ((H_CRC_MSB->Text != "-") && (!H_CRC_MSB->Text.IsEmpty())) {
    D_CRC_MSB->Text = _tcstoul(H_CRC_MSB->Text.c_str(), NULL , 16);
    // then clear the Encoded field because it's not valid anymore
    CRC_MSB->Text = "-";
    } else {
    D_CRC_MSB->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_Data1Change(TObject *Sender)
{
if ((H_Data1->Text != "-") && (!H_Data1->Text.IsEmpty())) {
    D_Data1->Text = _tcstoul(H_Data1->Text.c_str(), NULL , 16);
    A_Data1->Text = (char)D_Data1->Text.ToInt();  // Convert to ASCII Character
    A_Data1->Text = "'" + A_Data1->Text + "'";
    // then clear the Encoded field because it's not valid anymore
    Data1->Text = "-";
    } else {
    D_Data1->Text = "-";
    A_Data1->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_Data2Change(TObject *Sender)
{
if ((H_Data2->Text != "-") && (!H_Data2->Text.IsEmpty())) {
    D_Data2->Text = _tcstoul(H_Data2->Text.c_str(), NULL , 16);
    A_Data2->Text = (char)D_Data2->Text.ToInt();  // Convert to ASCII Character
    A_Data2->Text = "'" + A_Data2->Text + "'";
    // then clear the Encoded field because it's not valid anymore
    Data2->Text = "-";
    } else {
    D_Data2->Text = "-";
    A_Data2->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_Data3Change(TObject *Sender)
{
if ((H_Data3->Text != "-") && (!H_Data3->Text.IsEmpty())) {
    D_Data3->Text = _tcstoul(H_Data3->Text.c_str(), NULL , 16);
    A_Data3->Text = (char)D_Data3->Text.ToInt();  // Convert to ASCII Character
    A_Data3->Text = "'" + A_Data3->Text + "'";
    // then clear the Encoded field because it's not valid anymore
    Data3->Text = "-";
    } else {
    D_Data3->Text = "-";
    A_Data3->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::H_Data4Change(TObject *Sender)
{
if ((H_Data4->Text != "-") && (!H_Data4->Text.IsEmpty())) {
    D_Data4->Text = _tcstoul(H_Data4->Text.c_str(), NULL , 16);
    A_Data4->Text = (char)D_Data4->Text.ToInt();  // Convert to ASCII Character
    A_Data4->Text = "'" + A_Data4->Text + "'";
    // then clear the Encoded field because it's not valid anymore
    Data4->Text = "-";
    } else {
    D_Data4->Text = "-";
    A_Data4->Text = "-";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ClearTIUButtonClick(TObject *Sender)
{
// clear TIU Buffer
TIUBuffer->Clear();
// Clear TIU Buffer Byte Count
TIU_Char_Count->Text = "0";
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ClearRemoteButtonClick(TObject *Sender)
{
// clear RFID Buffer
RFIDBuffer->Clear();
// Clear Received Byte Count
//iRFIDbufCount = 0;
// Clear RFID Port Buffer Byte Count
RFID_Char_Count->Text = "0";
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Clear_Decode_ButtonClick(TObject *Sender)
{
ClearDecode();	// Clear all fields	
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CutFromCRCBufferClick(TObject *Sender)
{
// CRC Buffer popup menu item
CRCBuffer->CutToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CopyFromCRCBufferClick(TObject *Sender)
{
// CRC Buffer popup menu item
CRCBuffer->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PasteIntoCRCBufferClick(TObject *Sender)
{
// CRC Buffer popup menu item
CRCBuffer->PasteFromClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SelectAllCRCBufferClick(TObject *Sender)
{
// CRC Buffer popup menu item
CRCBuffer->SelectAll();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CutFromTIUBufferClick(TObject *Sender)
{
// TIU Buffer popup menu item
TIUBuffer->CutToClipboard();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CopyFromTIUBufferClick(TObject *Sender)
{
// TIU Buffer popup menu item
TIUBuffer->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PasteIntoTIUBufferClick(TObject *Sender)
{
// TIU Buffer popup menu item
// Text is Red
TIUBuffer->SelAttributes->Color = clRed;
TIUBuffer->PasteFromClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SelectAllTIUBufferClick(TObject *Sender)
{
// TIU Buffer popup menu item
TIUBuffer->SelectAll();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CutFromRFIDBufferClick(TObject *Sender)
{
// Remote Buffer popup menu item
RFIDBuffer->CutToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AppendCRCBuffertoRemoteBuffer1Click(TObject *Sender)
{
// CRC Buffer popup menu item
RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();	// Cursor to end of text
// copy CRC Buffer to Main Buffer
CRCBuffer->CopyToClipboard();
RFIDBuffer->PasteFromClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodeStringClick(TObject *Sender)
{
// Remote Buffer and TIU Buffer popup menu item
// Decode an encoded string
// This will decode and print out a string regardless of the contents.
// It will process the entire string to the end
//
//If there is a selected text active, then decode that selected text
//Decode the selected string from the Main Buffer
//

TRichEdit * SelectedBuffer;
TPopupMenu * MenuUsed;
MenuUsed = (TPopupMenu *)Sender;
if (MenuUsed->Name.Pos("RFID")) SelectedBuffer = RFIDBuffer;
if (MenuUsed->Name.Pos("TIU")) SelectedBuffer = TIUBuffer;

TPoint CaretPos = SelectedBuffer->CaretPos;
String OriginalString = SelectedBuffer->Lines->Strings[CaretPos.y];
//
String MyString = SelectedBuffer->SelText;
if (MyString.IsEmpty()) {
	// otherwise, if the cursor is on a line with a valid string,
    // then decode the command on that line
    // get MyString from the cursor line
    MyString = CleanUpCommand(OriginalString);
    } else {
    // Selection Active, so use the selection
    MyString = CleanUpCommand(MyString);
    }
//
CRCBuffer->SelAttributes->Color = SelectedBuffer->SelAttributes->Color;	// Copy with same color
CRCBuffer->Lines->Append(OriginalString);
CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
(void) sDecodeString(MyString, true); // Decode and print (into CRCBuffer) the command in MyString
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::sDecodeString(String AnyString, bool PrinttoCRCBuffer)
{
//
// Decode any string, if PrinttoCRCBuffer, print to CRC Buffer
//
//if (DEBUGGING) DB->Debug_Print("DecodeString: " + AnyString);
String os = "", oc = "";
if (!AnyString.IsEmpty()) {
    // break the string into 4 character groups and print the values in
    // hex and (if possible) as a character
	int SLen = AnyString.Length();
	if (SLen == 0) return "";
    // only process strings which start with a valid Encoded Character 5, 6, 9, or A
    if (!AnyString.IsDelimiter("569A", 1)) return "";
    //
    for (int idx = 1 ; idx < SLen; idx += 4) {
       	if (SLen - idx < 3) continue; // don't decode unless we have 4 characters
		String sVal = bintodec8(bitswapRX(Convert16B8BNumber(AnyString.SubString(idx, 4))));
        int iVal = sVal.ToInt();	
	   	String hVal = IntToHex(iVal & 0xFF, 2);
        os += "  ";
    	os += hVal;
    	// print string byte in ASCII
    	String A_Char;
        if (iVal > 32 && iVal < 128) {
	        A_Char = (char)iVal;  // Convert to ASCII Character
			oc = oc + " '" + A_Char + "'";
            } else {
            oc += "    ";
            }
        }
    if (PrinttoCRCBuffer) {
    	CRCBuffer->SelAttributes->Color = clBlue;	// Blue for Decodes
	    CRCBuffer->Lines->Append(os);
	    CRCBuffer->SelAttributes->Color = clBlue;	// Blue for Decodes
	   	CRCBuffer->Lines->Append(oc);
		// Method 4 works under CB10.1 but not BCB 4.0
        CRCBuffer->HideSelection = false;
        CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
        CRCBuffer->HideSelection = true;
        }
	}
return os;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

String __fastcall byteDecodeBinaryString(byte * AnyString, int Len, bool PrinttoCRCBuffer)
{
//
// Decode a binary string, if PrinttoCRCBuffer, print to CRC Buffer
// string can contain a binary 0x00 since a length is passed in the call
//
String os = "", oc = "";
if (AnyString != NULL) {
    // print the values in hex and (if possible) as a character
	if (Len == 0) return "";
    //
    for (int idx = 0 ; idx < Len; idx += 2) {
       	if (Len - idx < 1) continue; // don't decode unless we have 2 bytes
        int iVal = RXLookUp(AnyString[idx] << 8 | AnyString[idx + 1]);
	   	String hVal = IntToHex(iVal & 0xFF, 2);
        os += "  ";
    	os += hVal;
    	// print string byte in ASCII
    	String A_Char;
        if (iVal > 32 && iVal < 128) {
	        A_Char = (char)iVal;  // Convert to ASCII Character
			oc = oc + " '" + A_Char + "'";
            } else {
            oc += "    ";
            }
        }
    /*
    if (PrinttoCRCBuffer) {
    	CRCBuffer->SelAttributes->Color = clBlue;	// Blue for Decodes
	    CRCBuffer->Lines->Append(os);
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	    CRCBuffer->SelAttributes->Color = clBlue;	// Blue for Decodes
	   	CRCBuffer->Lines->Append(oc);
		CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
        }
    */
	}
return os;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodeStringBufferClick(TObject *Sender)
{
// Remote Buffer and TIU Buffer popup menu item
TRichEdit * SelectedBuffer;
TPopupMenu * MenuUsed;
MenuUsed = (TPopupMenu *)Sender;
if (MenuUsed->Name.Pos("RFID")) SelectedBuffer = RFIDBuffer;
if (MenuUsed->Name.Pos("TIU")) SelectedBuffer = TIUBuffer;
// Decode all strings in the Buffer and write them to the CRC Buffer
//
if (DEBUGGING) STOPem->Enabled = true;
bStop = false;
//
const int count = SelectedBuffer->Lines->Count;

for (int idxdb = 0 ; idxdb < count ; idxdb++) {
   if (bStop) {
        break;
        }
	String MyString = SelectedBuffer->Lines->Strings[idxdb];
    CRCBuffer->SelAttributes->Color = SelectedBuffer->SelAttributes->Color;	// Copy with same color
   	CRCBuffer->Lines->Append(MyString);
    MyString = CleanUpCommand(MyString);
	// Method 4 works under CB10.1 but not BCB 4.0
    CRCBuffer->HideSelection = false;
    CRCBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
    CRCBuffer->HideSelection = true;
    //
	(void) sDecodeString(MyString, true); // Decode and print (into CRC Buffer) the string in MyString
    Application->ProcessMessages();
	}
STOPem->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RTSSwitchClick(TObject *Sender)
{
if (!IsRadioCOMConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
//ptrTIUCommThread->RTSSwitch(RTSSwitch->Checked);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DTRSwitchClick(TObject *Sender)
{
if (!IsRadioCOMConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
//ptrTIUCommThread->DTRSwitch(DTRSwitch->Checked);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DTRSwitchStatusChanged(TObject *Sender)
{
ptrTIUCommThread->DTRSwitch(DTRSwitch->Checked);
if (DEBUGGINGX(6)) DB->Debug_Print("DTRSwitchStatus: " + ConnectedRadioLED->Caption + " DTR " + (int)DTRSwitch->Checked);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RTSSwitchStatusChanged(TObject *Sender)
{
ptrTIUCommThread->RTSSwitch(RTSSwitch->Checked);
if (DEBUGGINGX(6)) DB->Debug_Print("RTSSwitchStatus: " + ConnectedRadioLED->Caption + " RTS " + (int)RTSSwitch->Checked);
}
//---------------------------------------------------------------------------

String sEncodeAString(Cmd CS)
{
// Encode a command into a string that can be transmitted
// NO VCL calls in this function
String Encoded_Command;
// encode a command from CS - then the command can be sent - NOTE : CS may contain binary zero
// Sync Byte at the beginning
Encoded_Command = "5555";
// Encode the command
for (int i = 0 ; i < CS.idx ; i++) {
	Encoded_Command = Encoded_Command + Convert8B16BNumber(bitswapTX(dectobin8(CS.Str[i])));
	}
// EOT
Encoded_Command = Encoded_Command + "4D";
//if (DEBUGGING) DB->Debug_Print("EncodeAString: " + Encoded_Command);
return Encoded_Command;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::UtilitiesButtonClick(TObject *Sender)
{
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Utiliies Form should be manually created as it is rarely needed
if (!Utilities) {
	Utilities = new TUtilities(this);     // Pass "this" as the owner
	assert(Utilities != NULL);			// make sure that construction worked
   	}
Utilities->Show();	// since I used the C++ pointer, it will get automatically deleted at close
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetupButtonClick(TObject *Sender)
{
assert(Setup != NULL);		// setup must be created
// Setup Form is auto-create
Setup->Show();
if (Setup->DisableEmergencyStopButton->Checked) {
	ESTOP->Enabled = false;
    ESTOP->Hint = "Disabled in Setup";
    ESTOP->Cursor = crNo;
    }
Ignore_FF_Char = Setup->Ignore_FF->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DebugButtonClick(TObject *Sender)
{
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Debug Form should be auto create because the DB pointer is used everywhere
if (!DB) {
	DB = new TDB(this);     // Pass "this" as the owner
	assert(DB != NULL);		// make sure that construction worked
    }
// since I used the C++ pointer, it will get automatically deleted at close
DB->ShowIt(RemoteNo, EngineNo, TIUNo);
//
SetZOrder(true); 	// put window behind main form
// open up mainform
Width = SavedDebugWidth;
Height = SavedDebugHeight;
DebugPageControl->ActivePage = WorkingTabSheet;

// Display ThreadId in debug buffer
if (DEBUGGING) {
	DB->Debug_Print("TMainForm:MainThreadID " + IntToHex((int)MainThreadID, 8), clGreen);
   	// MainThreadID and GetCurrentThreadId() return the same value
	//DB->Debug_Print("TMainForm:ThreadID " + IntToHex((int)GetCurrentThreadId(), 8), clGreen);
       // MainInstance returns 0x00400000
	//DB->Debug_Print("TMainForm:MainInstance 0x" + IntToHex((int)MainInstance, 8), clGreen);
       // IsMultiThread returns true
	//DB->Debug_Print("TMainForm:IsMultiThread " + String(IsMultiThread ? "true" : "false"), clGreen);
	/*
    http://docwiki.embarcadero.com/RADStudio/Rio/en/Predefined_Macros
    http://docwiki.embarcadero.com/RADStudio/Rio/en/Predefined_Macros#C.2B.2B_Compiler_Versions_in_Predefined_Macros

	Compiler 			__BORLANDC__ value
	Borland C++ Builder 4 	0x0540
	Borland C++ Builder 5 	0x0550
	Borland C++ 5.5 		0x0550
	Borland C++ 5.5.1 		0x0551
	Borland C++ Builder 6 	0x0560
	Developer Studio 2006	0x0570
	C++Builder Seattle 10	0x0710
	C++Builder Berlin 10.1	0x0720
	C++Builder Tokyo 10.2	0x0730
	C++Builder Rio 10.3		0x0740
	*/
	DB->Debug_Print("__BORLANDC__ 0x" + IntToHex(__BORLANDC__, 4));
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartupButtonClick(TObject *Sender)
{
// https://stackoverflow.com/questions/31649315/c-builder-button-click-function-reentry
LOCKEM()

// This can be called even if the Engine is already running. It just restores windows
// Allow Engines Only
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= 100);	// 0 factory reset, 1-99, 100 "all engine operation"

// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
	}

AltKeyState = StartupButton->Focused() ? GetKeyState(VK_MENU) : 0;		// Run Startup All
ShiftKeyState = StartupButton->Focused() ? GetKeyState(VK_SHIFT) : 0;	// Use Extended Startup
CtrlKeyState = StartupButton->Focused() ? GetKeyState(VK_CONTROL) : 0;	// Not used yet
// key pressed =  0xFF8X
// key not pressed = 0x000X
// "X" can be 0 or 1 - indicated key was toggled
// if either ALT key is pressed whilst clicking "Startup" button, run "Startup All"
if (AltKeyState < 0) {
	StartupAllActiveEngines1->OnClick(StartupButton);
	AltKeyState = ShiftKeyState = CtrlKeyState = 0;
    return;
    }
StartupButton->Enabled = false;
//
if (ptrQuick[TIUNo - 1]) {
	// Quick Controls is running
	if (ptrQuick[TIUNo - 1]->RunningState[EngineNo] == clLime) {
		EngineStartUp(EngineNo, TIUNo, true, 0); 	// engine is already started so don't send actual startup commands
		} else {
		EngineStartUp(EngineNo, TIUNo, false, 0);   // Startup Engine: Send startup commands and switch setup
		}
	} else {
	// Quick Controls is not running
	EngineStartUp(EngineNo, TIUNo, false, 0);   // Startup Engine: Send startup commands and switch setup
	}
//
AltKeyState = ShiftKeyState = CtrlKeyState = 0;
StartupButton->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineStartUp(int Engine, int TIU, bool AlrdyStrted, int iDelayTimeSeconds)
{
// startup engine given by Engine on TIU
// if AlrdyStrted is true, RTC will not send "m4", "u4"/"u6", "d0" or "s0" commands.
//
// RTC will still send commands for all engine settings as stored in the Engine List
// handles Engines from 0 to 100 (Factory Reset, 1-99, 100=All Engine Operation)
// Delay Time in Seconds
//
// TODO: Bug - if AlrdyStrted is true, we need to read up the goal speed in the engine RAM 0x62-0x63
// and set the speed to that value. This will correctly set the speed slider on the
// operations window.
//
assert(Engine >= (Setup->AllowZero->Checked? 0 : 1) && Engine <= 100);	// 0/1-99
if (!IsTIUConnected()) return;	// if TIU Port is not connected fail

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
//
// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);
//
// show Operation Form
// Start the Operations form for engines 0-100
//
// make sure that the engine is on the track
//
// If engine does not respond, show an error and don't start the operations window
//
int respCode = 0x00;
if (!AlrdyStrted) {
	bool retv = ptrTIUCommThread->Send_String_FIFO(Engine, "d0", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, NULL, &respCode);
	/*
     Last Response  : 5996996659595565 (0x10)
                    :   49  99  0C  10
                                    ^^  Byte 3 of the response should be 0x10 to 0x1X (TIU number), 0x00 is engine not on track
	*/
	if (!retv || (respCode == 0x00)) {
    	// d0 command failed
	    Screen->Cursor = OldCursor;		// restore cursor
		new TMessageDisplayThread(false, this, Handle, "Warning", "Engine " + String(Engine) + " not on track (Engine Start Up)", "", 2000);
		return;	// startup failed - don't show Op window
		}
    }

// Enable controls until Engine Number is changed or shutdown occurs
SoundCtrlsButton->Enabled = true;
OpButton->Enabled = true;

// Manually create form and show it
// Many Operation Control Forms - if the form is not already created, then create it
if (ptrOp[TIU-1][Engine] == NULL) {	// Engine 0-100 (0 is factory reset, 1-99, 100 for "all engine operation")
	// Create the Operation Screen
	if (ptrLUEngineList[TIU-1][Engine] != NULL) {
    	delete[] ptrLUEngineList[TIU-1][Engine];	// if it's left over, delete the string - pointer to a char array
		ptrLUEngineList[TIU-1][Engine] = NULL;		// never a LashUpEngineList for a regular engine
        }
	if (ptrLUNameList[TIU-1][Engine] != NULL) {
    	delete[] ptrLUNameList[TIU-1][Engine];		// if it's left over, delete the string - pointer to a char array
		ptrLUNameList[TIU-1][Engine] = NULL;		// never an Engine Name for a regular engine
        }
	ptrOp[TIU-1][Engine] = new TOp(this, RemoteNo, Engine, TIU, EngineData->Cells[ENG_NAME][Engine], NULL);   // create the Form and show it
								// Pass "this" as the owner
    							// 1-99 Engine Number, 100 "all engine operaton"
	assert(ptrOp[TIU-1][Engine] != NULL);	// make sure that construction worked
    ptrOp[TIU-1][Engine]->ParentHandle = Handle; // for return messages
    //
    // Begin Startup Command sequence
    //
    // This is a guess at determining if this is a PS2 or PS3 engine. It appears that RAM 0x50 at the
    // start is ==0x00 for all PS3 engines and !=0x00 for all PS2 engines. This might not be always true.
    //
    if (DEBUGGING || (EngineData->Cells[ENG_PS][Engine] != PS2 && EngineData->Cells[ENG_PS][Engine] != PS3)) {
	    byte ival;
    	FetchByte(0x50, &ival, Engine, TIU);
	    if (ival == 0x00) {
    	    EngineData->Cells[ENG_PS][Engine] = PS3;
            if (DEBUGGING) DB->Debug_Print("Engine reports PS3", clBlack);
        	} else {
	        EngineData->Cells[ENG_PS][Engine] = PS2;
            if (DEBUGGING) DB->Debug_Print("Engine reports PS2", clBlack);
    	    }
        } else {
        if (EngineData->Cells[ENG_PS][Engine] == PS2) {
	        if (DEBUGGING) DB->Debug_Print("Engine previously reported PS2", clBlack);
        	} else
            if (EngineData->Cells[ENG_PS][Engine] == PS3) {
				if (DEBUGGING) DB->Debug_Print("Engine previously reported PS3", clBlack);
            	} else
                if (DEBUGGING) DB->Debug_Print("Engine unknown type", clBlack);
        }
    //
    // Setup volume control defaults before startup occurs
    //
	if ((Engine != 0) && (Engine < LASHUPMIN) && (EngineData->Cells[ENG_NAME][Engine] != "*")) {   // skip if EngineData not initialized
      try {
          // Set Engine volumes.
          // TODO Needed because Horn & Bell Sounds are muted in a LashUp and I don't seem to restore them correctly
          //
          char str[15];
		  snprintf(str, sizeof str, "v0%d", EngineData->Cells[VOL_MASTER][Engine].ToInt()); 	// Master Volume
		  (void) ptrTIUCommThread->Send_String_FIFO(Engine, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
		  //ptrOp[TIU-1][Engine]->LastSentMasterVolumeValue = EngineData->Cells[VOL_MASTER][Engine].ToInt();

		  snprintf(str, sizeof str, "v1%d", EngineData->Cells[VOL_ENGINE][Engine].ToInt()); 	// Engine Sounds
		  (void) ptrTIUCommThread->Send_String_FIFO(Engine, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);

		  snprintf(str, sizeof str, "v2%d", EngineData->Cells[VOL_ACCENT][Engine].ToInt()); 	// Accent Sounds
		  (void) ptrTIUCommThread->Send_String_FIFO(Engine, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);

		  snprintf(str, sizeof str, "v3%d", EngineData->Cells[VOL_HORN][Engine].ToInt()); 	// Horn Sounds
		  (void) ptrTIUCommThread->Send_String_FIFO(Engine, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);

   		  snprintf(str, sizeof str, "v4%d", EngineData->Cells[VOL_BELL][Engine].ToInt());	// Bell Sounds
		  (void) ptrTIUCommThread->Send_String_FIFO(Engine, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
	      }
        catch ( EConvertError & oops ){
          // first time this Engine is used, the EngineData StringGrid is empty and the ToInt() function
          // returns an EConvertError exception
          DebugMessagePopup("Warning", "Empty Engine Data Field - Repaired", "");
          CaughtErrors->Value ++;
	      }
        catch (...) {
          CaughtErrors->Value ++;
          DebugMessagePopup("Error", "Unknown Startup Problem in Volume Settings", "");
          }
		}
    //
	// Startup    // Command "u4" for normal, "u6" for Extended
	// if either SHIFT key is pressed whilst clicking "Startup" button, use the extended startup sound
    // If the high-order bit is 1, the key is down; otherwise, it is up. If the low-order bit is 1, the key is toggled.
    // (HO bit means a mask value of 0x0080
	if (!AlrdyStrted) {
			// key pressed =  0xFF8X
			// key not pressed = 0x000X
			// "X" can be 0 or 1 - indicated key was toggled
		if (Setup->ExtendedSUSD->Checked || (ShiftKeyState < 0)) {
			(void)ptrTIUCommThread->Send_String_FIFO(Engine, "u6", TIU, RemoteNo, 11, LONG_WAIT, RECORD_OK);
		    } else {
			(void)ptrTIUCommThread->Send_String_FIFO(Engine, "u4", TIU, RemoteNo, 11, LONG_WAIT, RECORD_OK);
		    }
		// set DIR forward  "d0"
		(void)ptrTIUCommThread->Send_String_FIFO(Engine, "d0", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
		// set speed to zero unless it already is
		if (ptrOp[TIU-1][Engine]->SpeedSlider->Value >= 1.0) {
		    // BUG workaround - slider starts out at 10 	// causes the 's0' command to be sent because the slider is not at zero
			ptrOp[TIU-1][Engine]->SetSpeedSliderValue(0.0);
		    }
	    // Startup Complete
        StartedConsists->Value++;
        } else {
        // if AlrdyStrted is set, just set the slider to the RAM speed value
        // TODO: this code does not work and was removed.
		#ifdef DOESNOTWORK
		/*
        // read up the speed and set the slider
        unsigned short Spd = 0;
        bool bretv = FetchWord(0x62, &Spd, Engine, TIU);
        Spd = Spd >> 3;	// divide by 8
        if (bretv && (Spd < Setup->MaxSpeed->Value)) {	// protect against large (most likely out of range) values
	        // disconnect event handler while processing the SpeedSliderValueChange();
	        ptrOp[TIU-1][Engine]->SetSpeedSliderValue(Spd);
            // Restore Event Handler
            }
		// direction switch is not changed because that will make the engine stop
		*/
        #endif
        }
    //
	// setup control defaults on Operations form after startup occurs
    //
    if ((Engine != 0) && (Engine < LASHUPMIN) && (EngineData->Cells[ENG_NAME][Engine] != "*")) {   // skip if EngineData not initialized
      try {
          // this code causes these three command to be sent:
          //	Dq Da1 Dd1
          //
          if (Setup->StandardAccDec->Checked) {
          	// set Acc and Dec values to 1 in all modes of operation
        	ptrOp[TIU-1][Engine]->SetStandardAccDecValues();
          	} else {
	      	if (!AllEngineMode) {
		  	  ptrOp[TIU-1][Engine]->AccSpin->Enabled = true;
	      	  ptrOp[TIU-1][Engine]->DecSpin->Enabled = true;
			  // read Acc/Dec rates from engine and display them
	    	  int iAcc, iDec;	// routine returns default values if we cannot communicate.
    		  (void) GetAccDecValues(Engine, TIU, &iAcc, &iDec);
        	  SetNumSpin(ptrOp[TIU-1][Engine]->AccSpin, iAcc);		// Sends Acc Rate to the Engine
        	  SetNumSpin(ptrOp[TIU-1][Engine]->DecSpin, iDec);		// Sends Dec Rate to the Engine
			  } else {
			  // disable Acc and Dec controls
	    	  ptrOp[TIU-1][Engine]->AccSpin->Enabled = false;
	    	  ptrOp[TIU-1][Engine]->DecSpin->Enabled = false;
			  }
            }
          // Engine Type - this MUST be done first before the other switches are set
          SetComboBox(ptrOp[TIU-1][Engine]->EngineTypeComboBox, EngineData->Cells[ENG_TYPE][Engine].ToInt());
       	  // a lot of this can be removed if I can figure out how to ask the engine
          // what settings are stored in it.
          // According to Barry's book, the following are stored in the Engine:
          //   DCS Engine #
          //   Lighting including headlight, marker lights, interior lights, number board lights, ditch lights, and beacon
          //   Sounds appropriate to the engine
          //   Sound levels
          //   Chuff Rate and Proto Chuff setting
          //   Smoke level
          //   Accelerations and Deceleration rates
          //   original soft key settings
          //   custom name
          //   Chronometer, odometer, measured route, and battery condition
          // ENG_HDLT contains five encoded values - Headlight State, Rule 17, Bell on with horn, Auto Bell on Startup and
          // Auto Whistle/Horn on Startup
          // RAM 0x25 bit 0
          SetSwitch(ptrOp[TIU-1][Engine]->HDLTSwitch, EngineData->Cells[ENG_HDLT][Engine].ToIntDef(false) & 0x01);
          SetLED(ptrOp[TIU-1][Engine]->SimulatedRule17LED, (EngineData->Cells[ENG_HDLT][Engine].ToIntDef(false) >> 1) & 0x01);
          SetLED(ptrOp[TIU-1][Engine]->BellOnWithHornLED, (EngineData->Cells[ENG_HDLT][Engine].ToIntDef(false) >> 2) & 0x01);
          SetLED(ptrOp[TIU-1][Engine]->AutoBellLED, (EngineData->Cells[ENG_HDLT][Engine].ToIntDef(false) >> 3) & 0x01);
          SetLED(ptrOp[TIU-1][Engine]->AutoWhistleLED, (EngineData->Cells[ENG_HDLT][Engine].ToIntDef(false) >> 4) & 0x01);
          //  Smoke - off first then setting
          // RAM 0x25 bit 7
          SetSwitch(ptrOp[TIU-1][Engine]->SmokeSwitch, true);
          SetSwitch(ptrOp[TIU-1][Engine]->SmokeSwitch, false);	// start with smoke/exhaust off
          // set set to saved value unless Smoke/Exhaust is forced OFF at startup
          if (!Setup->SmokeOFFatStartup->Checked) SetSwitch(ptrOp[TIU-1][Engine]->SmokeSwitch, EngineData->Cells[ENG_SMOKE][Engine].ToIntDef(false));
          // RAM 0x25 bit 3
          SetSwitch(ptrOp[TIU-1][Engine]->ClickClackSwitch, EngineData->Cells[ENG_FCC][Engine].ToIntDef(false));
          // ENG_SMOKELEVEL contains two encoded values - SmokeGroup (2 bits) and AutoLaboredSmokeSwitch (1 bit)
          // RAM 0x24 bits 1,0
          SetRadioGroup(ptrOp[TIU-1][Engine]->SmokeGroup, EngineData->Cells[ENG_SMOKELEVEL][Engine].ToIntDef(false) & 0x03);
          SetLED(ptrOp[TIU-1][Engine]->AutoLaboredSmokeLED, (EngineData->Cells[ENG_SMOKELEVEL][Engine].ToIntDef(false) >> 2) & 0x01);
          //
          // RAM 0x24 bit 2
          SetSwitch(ptrOp[TIU-1][Engine]->CabChatSwitch, EngineData->Cells[ENG_CHAT][Engine].ToIntDef(false));
          // RAM 0x24 bit 6
          SetSwitch(ptrOp[TIU-1][Engine]->ProtoChuffSwitch, EngineData->Cells[ENG_PROTOCHUFF][Engine].ToIntDef(false));
		  // the 'g' command does different things for diesels than for steam engines - handled inside of ChuffRateSwitchChanged()
          // ie - ChuffRateSwitch() does nothing if this is a diesel engine.
          SetSwitch(ptrOp[TIU-1][Engine]->ChuffRateSwitch, EngineData->Cells[ENG_CHUFFRATE][Engine].ToIntDef(false));
          // RAM 0x24 bit 4
          SetSwitch(ptrOp[TIU-1][Engine]->BrakesSwitch, EngineData->Cells[ENG_BRAKES][Engine].ToIntDef(false));
          // for ENG_TYPE, see above
          int mx = EngineData->Cells[ENG_MAXSPEED][Engine].ToIntDef(60);
          if (mx < 10) mx = Setup->MaxSpeed->Value; 	// maximum value
          ptrOp[TIU-1][Engine]->MaxSpeedSpin->Value = mx;
          ptrOp[TIU-1][Engine]->MaxSpeedSpin->MaxValue = mx;
		  ptrOp[TIU-1][Engine]->SpeedSlider->SignalSettings->DigitalTo = ptrOp[TIU-1][Engine]->MaxSpeedSpin->Value;
		  ptrOp[TIU-1][Engine]->SpeedSlider->SignalSettings->ValueTo = ptrOp[TIU-1][Engine]->MaxSpeedSpin->Value;
		  ptrOp[TIU-1][Engine]->SpeedSlider->ScaleSettings->Steps = ptrOp[TIU-1][Engine]->MaxSpeedSpin->Value / 10;
          //
          if (EngineData->Cells[ENG_PS][Engine].ToIntDef(PS2) == PS2) ptrOp[TIU-1][Engine]->SetEnginePS2->Checked = true;
          if (EngineData->Cells[ENG_PS][Engine].ToIntDef(PS2) == PS3) ptrOp[TIU-1][Engine]->SetEnginePS3->Checked = true;
	      }
        catch ( EConvertError & oops ){
          // first time this Engine is used, the EngineData StringGrid is empty and the ToInt() function
          // returns an EConvertError exception
          DebugMessagePopup("Warning", "Empty Engine Data Field - Repaired", "");
          CaughtErrors->Value ++;
	      }
        catch (...) {
          CaughtErrors->Value ++;
          DebugMessagePopup("Error", "Unknown Startup Problem in Engine Data Settings", "");
          }
		}

	} else {    // it's already created, just show it
	ptrOp[TIU-1][Engine]->Show();
	}
//
// if either CTRL key is pressed whilst clicking "Startup" button, show the Sound form also
if (Setup->AutoSoundCtrls->Checked || (CtrlKeyState < 0)) {
	SoundCtrlsButton->OnClick(SoundCtrlsButton);
    }
if (iDelayTimeSeconds > 0) Delay(iDelayTimeSeconds * 1000);	// Delay Time in Seconds convert to Delay time in microseconds
Screen->Cursor = OldCursor;		// restore cursor
// Note: Engine startup usually continues for several seconds after this routine
// finished. If you click on another button before the engine is ready for a
// a command, the engine seems to just ignore the command.
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartupAllActiveEngines1Click(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

StartupAllActiveEngines(TIUNo, 5);	// Wait X seconds between each engine startup
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartupAllActiveEngines(int TIU, int iDelayTimeSeconds)
{
TMessagePopup *MessagePopup = NULL;
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
if (!MessagePopup){
	MessagePopup = new TMessagePopup(this, Handle, NULL, NULL);     // Pass "this" as the owner
	assert(MessagePopup != NULL);		// make sure that construction worked
	}
MessagePopup->Show();
MessagePopup->MessageLine->Caption = "   Starting all Active Engines   ";
MessagePopup->StatusLine->Caption = "   Please Wait   ";
MessagePopup->Repaint();

// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);

DWORD OldEngineValue = EngineSpin->Value;

// Start All Active Engines - 1-99 (DCS 2-100)
// (total of 99 engines)
// NOTE: Only startup active engines that have not yet been started
for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
   	if ((EngineData->Cells[ENG_NAME][iEng] != "*") && (EngineData->Cells[ENG_STATE][iEng] == ACTIVE) && (ptrOp[TIUNo-1][iEng] == NULL)) {
    	EngineSpin->Value = iEng; 	// show the engine in the Engine List and Engine Number Spinner
        MessagePopup->StatusLine->Caption = "   Engine # " + String(iEng) + "   ";
		EngineStartUp(iEng, TIUNo, false, 0);	// Startup Engine: send startup commands and switch setup and open Operations window
		MessagePopup->Repaint();
        if (iDelayTimeSeconds > 0) Delay(iDelayTimeSeconds * 1000);	// wait X seconds after each startup for startup sequence to run
		}
    }
EngineSpin->Value = OldEngineValue;
Screen->Cursor = OldCursor;		// restore cursor

delete MessagePopup;

//DebugMessage("Startup All Active Engines Complete");
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartupAllActiveEngines2Click(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Run Startup All
StartupAllActiveEngines1->OnClick(StartupAllActiveEngines2);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ShutdownAllEngines1Click(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Run Shutdown All with three second delay between the shutdown commands
ShutdownAllEngines(TIUNo, 3);
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::ShutdownAllEngines(int TIU, int iDelayTimeSeconds)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

TMessagePopup *MessagePopup = NULL;
TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
if (!MessagePopup){
	MessagePopup = new TMessagePopup(this, Handle, NULL, NULL);     // Pass "this" as the owner
	assert(MessagePopup != NULL);		// make sure that construction worked
	}
MessagePopup->Show();
MessagePopup->MessageLine->Caption = "   Shutting Down all Engines   ";
MessagePopup->StatusLine->Caption = "   Please Wait   ";
MessagePopup->Repaint();

// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);

DWORD OldEngineValue = EngineSpin->Value;
// Send Shutdown command to all active engines in the engine list
for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
    if (ptrOp[TIUNo-1][iEng] != NULL) {
	    // if the engine is running (that is, the operations window is up)
	    // Shutdown
    	EngineSpin->Value = iEng;
        MessagePopup->StatusLine->Caption = "   Engine # " + String(iEng) + "   ";
		MessagePopup->Repaint();
		ConsistShutDown(iEng, TIUNo, iDelayTimeSeconds);	// Shutdown Engine and close Operations Window
        									// wait iDelayTimeSeconds seconds after each shutdown
        } else {
        // Send Shutdown command directly to all engines in the engine list (active and non-running)
       	if (EngineData->Cells[ENG_NAME][iEng] != "*" && (EngineData->Cells[ENG_STATE][iEng] == ACTIVE)) {
            MessagePopup->StatusLine->Caption = "   Engine # " + String(iEng) + "   ";
			MessagePopup->Repaint();
            // NOTE: the TIU will NAK each shutdown command if the engine is not actually on the layout
			(void) ptrTIUCommThread->Send_String_FIFO(iEng, "u5", TIUNo, RemoteNo,  11, SHORT_WAIT, RECORD_OK);	// Shutdown
           	}
        }
    }
EngineSpin->Value = OldEngineValue;
Screen->Cursor = OldCursor;		// restore cursor

delete MessagePopup;

//DebugMessage("Shutdown All Engines Complete");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ShutDownActiveEngines(int TIU, int iDelayTimeSeconds)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
TMessagePopup *MessagePopup = NULL;
TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
if (!MessagePopup){
	MessagePopup = new TMessagePopup(this, Handle, NULL, NULL);     // Pass "this" as the owner
	assert(MessagePopup != NULL);		// make sure that construction worked
	}
MessagePopup->Show();
MessagePopup->MessageLine->Caption = "   Shutting Down Active Engines   ";
MessagePopup->StatusLine->Caption = "   Please Wait   ";
MessagePopup->Repaint();

// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);

DWORD OldEngineValue = EngineSpin->Value;
// Send Shutdown command to all engines in the engine list (running or active or not)
for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
    if (ptrOp[TIUNo-1][iEng] != NULL) {
	    // if the engine is running, use the speed scroller
	    // Shutdown
    	EngineSpin->Value = iEng;
        MessagePopup->StatusLine->Caption = "   Engine # " + String(iEng) + "   ";
		MessagePopup->Repaint();
		ConsistShutDown(iEng, TIUNo, iDelayTimeSeconds);	// Shutdown Engine and close Operations Window
        }
    }
EngineSpin->Value = OldEngineValue;
Screen->Cursor = OldCursor;		// restore cursor

delete MessagePopup;

//DebugMessage("Shutdown Active Engines Complete");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ShutdownButtonClick(TObject *Sender)
{
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

AltKeyState = StartupButton->Focused() ? GetKeyState(VK_MENU) : 0;		// Run Shutdown All
ShiftKeyState = StartupButton->Focused() ? GetKeyState(VK_SHIFT) : 0;	// Use Extended Shutdown
CtrlKeyState = StartupButton->Focused() ? GetKeyState(VK_CONTROL) : 0;	// Not used yet

// key pressed =  0xFF8X
// key not pressed = 0x000X
// "X" can be 0 or 1 - indicated key was toggled
// if either ALT key is pressed whilst clicking "Shutdown" button, run "Shutdown All"
if (AltKeyState < 0) {
	ShutdownAllEngines1->OnClick(ShutdownButton);
	AltKeyState = ShiftKeyState = CtrlKeyState = 0;
    return;
    }
// You can shutdown a consist even it was not started up
ConsistShutDown(EngineNo, TIUNo);
AltKeyState = ShiftKeyState = CtrlKeyState = 0;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ConsistShutDown(int Consist, int TIU, int iDelayTimeSeconds)
{
// shutdown engine/lashup given by Consist on TIU
// handles Engines & LashUps from 0 to 120
assert(Consist >= (Setup->AllowZero->Checked? FACTORYRESET : ENGINEMIN) && Consist <= LASHUPMAX);	// 0 factory reset, 1-99, 100 "all engine operation, 101-120 LashUps
assert (TIU >= 1 && TIU <= TIUMAX);
// if TIU Port is not connected fail
if (!IsTIUConnected()) return;
//
// You can shutdown a consist even it was not started up
//
// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);
//
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

// make sure that we are in command mode if we are shutting down an engine
if (Consist < LASHUPMIN) {
	if (!ptrTIUCommThread->Send_String_FIFO(Consist, "m4", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, ptrLUEngineList[TIU-1][Consist])) {
	    Screen->Cursor = OldCursor;
		//new TMessageDisplayThread(false, this, Handle, "Error", "Engine " + String(Consist) + " not on track (Consist Shutdown)", "", 2000);
        MainForm->DebugMessagePopup("Error", "Engine " + String(Consist) + " not on track (Consist Shutdown)");
		return; // shutdown failed TODO: This leaves the Op window on the screen, maybe we should take it down
    	}
    }
// set speed to zero
if (!ptrTIUCommThread->Send_String_FIFO(Consist, "s0", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, ptrLUEngineList[TIU-1][Consist])) {
    Screen->Cursor = OldCursor;
	//new TMessageDisplayThread(false, this, Handle, "Error", "Engine/LashUp " + String(Consist) + " not on track (Consist Shutdown)", "", 2000);
    MainForm->DebugMessagePopup("Error", "Engine/LashUp " + String(Consist) + " not on track (Consist Shutdown)");
	return; // shutdown failed TODO: This leaves the Op window on the screen, maybe we should take it down
	}
// Shutdown   // Command "u5" for normal, "u7" for extended
// if either SHIFT key is pressed whilst clicking "Shutdown" button, use the extended shutdown sound
if (Setup->ExtendedSUSD->Checked || (ShiftKeyState < 0)) {
	(void)ptrTIUCommThread->Send_String_FIFO(Consist, "u7", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, ptrLUEngineList[TIU-1][Consist]);
    } else {
	(void)ptrTIUCommThread->Send_String_FIFO(Consist, "u5", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, ptrLUEngineList[TIU-1][Consist]);
    }
// if SPW is enabled, make sure that it is turned off
if (ptrOp[TIU-1][Consist] != NULL) {
	if (ptrOp[TIU-1][Consist]->ProtoWhistleSwitch->Checked) {
        ptrOp[TIU-1][Consist]->ProtoWhistleSwitch->Checked = false;
        }
	}
//
// Close Volume Control window if it is open
// Save information in EngineData and clear the pointer to NULL
if (ptrVolume[TIU-1][Consist] != NULL) {
	if ((Consist != 0) && (Consist < LASHUPMIN) && (EngineData->Cells[ENG_NAME][Consist] != "*")) {	// there must be an Engine Name in [ENG_NAME]
        assert(Consist >= 1 && Consist <= ENGINEMAX);	// 1-99
		// Engine Sound on/off is never saved, volume is restored before engine shutdown
        if (ptrVolume[TIU-1][Consist]->EngineSoundOnSwitch->Checked == false)
	        ptrVolume[TIU-1][Consist]->EngineSoundOnSwitch->Checked = true; // restore the volume setting if necessary

		// Save information in EngineData, Engines 1-99 and "all engine operation"  (0-100)
		EngineData->Cells[VOL_MASTER][Consist] = ptrVolume[TIU-1][Consist]->MasterVolume->Value;
		EngineData->Cells[VOL_ENGINE][Consist] = ptrVolume[TIU-1][Consist]->EngineVolume->Value;
		EngineData->Cells[VOL_ACCENT][Consist] = ptrVolume[TIU-1][Consist]->AccentVolume->Value;
		EngineData->Cells[VOL_HORN][Consist] = ptrVolume[TIU-1][Consist]->HornWhistleVolume->Value;
		EngineData->Cells[VOL_BELL][Consist] = ptrVolume[TIU-1][Consist]->BellVolume->Value;

		EngineData->Cells[VOL_SPARE1][Consist] = 0;
		EngineData->Cells[VOL_SPARE2][Consist] = 0;
		EngineData->Cells[VOL_SPARE3][Consist] = 0;
        }
	// Close the form
    assert(Consist >= (Setup->AllowZero->Checked? 0 : 1) && Consist <= LASHUPMAX);	// 0 factory reset, 1-99, 100 "all engine operation, 101-120 LashUps
    delete ptrVolume[TIU-1][Consist];
    ptrVolume[TIU-1][Consist] = NULL;
	}
// Close Operations Window if it is open
// Save information in EngineData and clear the pointer to NULL
// Save only if we are shutting down an engine
if (ptrOp[TIU-1][Consist] != NULL) {
	if ((Consist != 0) && (Consist < LASHUPMIN) && (EngineData->Cells[ENG_NAME][Consist] != "*")) {	// there must be an Engine Name in [ENG_NAME]
		// Save information in EngineData, Engines 1-99 and "all engine operation"  (0-100)
        assert(Consist >= 1 && Consist <= 100);	// 1-100
        // ENG_HDLT contains five encoded values - Headlight State, Simulated Rule 17,  Bell on with horn, Auto Bell and AUto Whistle
		EngineData->Cells[ENG_HDLT][Consist] = (short)ptrOp[TIU-1][Consist]->HDLTSwitch->Checked |
        		(ptrOp[TIU-1][Consist]->SimulatedRule17LED->Checked << 1) | (ptrOp[TIU-1][Consist]->BellOnWithHornLED->Checked << 2) |
                (ptrOp[TIU-1][Consist]->AutoBellLED->Checked << 3) | (ptrOp[TIU-1][Consist]->AutoWhistleLED->Checked << 4);
		EngineData->Cells[ENG_SMOKE][Consist] = (short)ptrOp[TIU-1][Consist]->SmokeSwitch->Checked;
		EngineData->Cells[ENG_FCC][Consist] = (short)ptrOp[TIU-1][Consist]->ClickClackSwitch->Checked;
        // ENG_SMOKELEVEL contains two encoded values - SmokeGroup (2 bits) and Auto Labored Smoke (1 bit)
		EngineData->Cells[ENG_SMOKELEVEL][Consist] = ptrOp[TIU-1][Consist]->SmokeGroup->ItemIndex |
        		(ptrOp[TIU-1][Consist]->AutoLaboredSmokeLED->Checked << 2);
   		EngineData->Cells[ENG_CHAT][Consist] = (short)ptrOp[TIU-1][Consist]->CabChatSwitch->Checked;
   		EngineData->Cells[ENG_PROTOCHUFF][Consist] = (short)ptrOp[TIU-1][Consist]->ProtoChuffSwitch->Checked;
   		EngineData->Cells[ENG_CHUFFRATE][Consist] = (short)ptrOp[TIU-1][Consist]->ChuffRateSwitch->Checked;
   		EngineData->Cells[ENG_BRAKES][Consist] = (short)ptrOp[TIU-1][Consist]->BrakesSwitch->Checked;
   		EngineData->Cells[ENG_DITCH][Consist] = ptrOp[TIU-1][Consist]->DitchComboBox->ItemIndex;
   		EngineData->Cells[ENG_BEACON][Consist] = (short)ptrOp[TIU-1][Consist]->BeaconLightSwitch->Checked;
   		EngineData->Cells[ENG_AUX1][Consist] = (short)ptrOp[TIU-1][Consist]->ab3ab2Switch->Checked;
   		EngineData->Cells[ENG_AUX2][Consist] = (short)ptrOp[TIU-1][Consist]->ab5ab4Switch->Checked;
    	EngineData->Cells[ENG_AUX3][Consist] = (short)ptrOp[TIU-1][Consist]->ab7ab6Switch->Checked;
    	EngineData->Cells[ENG_MARKER][Consist] = (short)ptrOp[TIU-1][Consist]->MarkerLightSwitch->Checked;
   		EngineData->Cells[ENG_TYPE][Consist] = (short)ptrOp[TIU-1][Consist]->EngineTypeComboBox->ItemIndex;	// Steam or Diesel or Electric
   		EngineData->Cells[ENG_MAXSPEED][Consist] = ptrOp[TIU-1][Consist]->MaxSpeedSpin->Value;	// Max speed for this engine
   		// EngineData->Cells[ENG_STATE][Consist] = ACTIVE;	// don't change ENG_STATE - only changed by READ pushbutton
        													// and LashUp create and breakup and manually from the Engine popup menu
		// EngineData->Cells[ENG_PS][Consist] = 0;		// don't change ENG_PS - only changed by Engine Startup and Operations popup Menu
        }
	// Close the form
    delete ptrOp[TIU-1][Consist];
    ptrOp[TIU-1][Consist] = NULL;
    StartedConsists->Value--;
    if (Consist >= LASHUPMIN) { 		// if a LashUp, delete the LUEngineList and LUNameList entry if they exist
        if (ptrLUNameList[TIU-1][Consist]) {
            delete ptrLUNameList[TIU-1][Consist];
            ptrLUNameList[TIU-1][Consist] = NULL;
            }
        if (ptrLUEngineList[TIU-1][Consist]) {
            delete ptrLUEngineList[TIU-1][Consist];
            ptrLUEngineList[TIU-1][Consist] = NULL;
            }
        }
    }

if (Consist == EngineNo) {
	SoundCtrlsButton->Enabled = false;	// disable Volume Button
	OpButton->Enabled = false;	// disable Operations Button
	}
//
// Redraw the Engine List or the LashUp List
if (Consist <= 100) {
	EngineData->Repaint();
    } else {
    if (ptrLashUp[TIU - 1]) ptrLashUp[TIU - 1]->LashUpData->Repaint();
    }
//
if (iDelayTimeSeconds > 0) Delay(iDelayTimeSeconds * 1000);	// Delay Time in Seconds convert to Delay time in microseconds
Screen->Cursor = OldCursor;
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMShutDown(TMessage &Message)
{
// process Engine/LashUp Shutdown message - Engines/LashUp 0-120
const int TIU = Message.WParam;		// set TIU Number from Word Parameter
const int Consist = Message.LParam;	// set Consist Number from Long Parameter
assert (TIU >= 1 && TIU <= TIUMAX);
assert (Consist >= (Setup->AllowZero->Checked? FACTORYRESET : ENGINEMIN) && Consist <= LASHUPMAX);	// 0 factory reset, 1-99, 100 "all engine operation, 101-120 LashUps
if (ptrOp[TIU-1][Consist] == NULL) {  // Consist not running - caller must validate
    return;
	}
// TIU may not equal TIUNo, check and reload data if necessary
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);
// CM_SHUTDOWN message can only shutdown a started engine
ConsistShutDown(Consist, TIU);
AltKeyState = ShiftKeyState = CtrlKeyState = 0;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMOpControl(TMessage &Message)
{
// process Engine Open Operations Control message - Engines 0-100  (LashUps are handled by another message)
const int TIU = Message.WParam;		// set TIU Number from Word Parameter

// TIU may not equal TIUNo, check and reload data if necessary
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);

const int Consist = Message.LParam;	// set Consist Number from Long Parameter
assert (TIU >= 1 && TIU <= TIUMAX);
assert (Consist >= (Setup->AllowZero->Checked? 0 : 1) && Consist <= 100);	// 0 factory reset, 1-99, 100 "all engine operation
assert(ptrOp[TIU-1][Consist] != NULL);	// not running - caller must validate
									   	// that the consist is running
EngineStartUp(Consist, TIU, true, 0);	// engine is already started so don't send actual startup commands
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMRadioCOMStatus(TMessage &Message)
{
// Thread to VCL interface - Handles all VCL operations for the Radio COM
const uint32 event = Message.LParam;	// set event from Long Parameter
const uint32 DoneReason = Message.WParam;	// Reason for Disconnect
String Msg;

switch(event) {
    case  SERIAL_CONNECTED  :
        Msg = "Connected to " + ConnectedRadioLED->Caption;
        if (DEBUGGING) DB->Debug_Print(Msg);
        //StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Radio Serial Connected";
        ConnectedRadioLED->Checked = true;
        ConnectedRadioLED->Flashing = false;
        break;

    case  SERIAL_DISCONNECTED  :
        Msg = "Disconnected from " + ConnectedRadioLED->Caption + " (" + String(DoneReason) + ")";
        if (DEBUGGING) DB->Debug_Print(Msg);
        //StatusBar->Panels->Items[RADIO_PORT_STATUS]->Text = "Radio Serial Disconnected";
        TIUVer1->Caption = "Display Selected TIU Ver # ";		// clear TIU number in RTCPopupMenu
        ConnectedRadioLED->Checked = false;
        ConnectedRadioLED->Flashing = false;
        if (!DisconnectRequest) {
            // this is the case where tserial_event detected an error/failure of some sort and sent a SERIAL_DISCONNECTED event
			TIUSpin->ColorDefault = clYellow;	// set TIU box color to default indicating unknown comm with the TIU
			AIUSpin->ColorDefault = clYellow;
			if (DEBUGGING) MainForm->DebugMessagePopup("Error", "TIU port unexpectedly disconnected (" + String(DoneReason) + ")");
            }
        break;

    case  SERIAL_DATA_SENT  :
        // Nothing needs to be done here
        break;

    case  SERIAL_RING_ON    :
        MainForm->Ring->Checked = true;
        break;

    case  SERIAL_RING_OFF   :
        MainForm->Ring->Checked = false;
        break;

    case  SERIAL_CD_ON      :
        MainForm->CD->Checked = true;
        CD_Count += 1;
        MainForm->CD_Counter->Text = String(CD_Count);
        break;

    case  SERIAL_CD_OFF     :
        MainForm->CD->Checked = false;
        break;

    case  SERIAL_CTS_ON		:
        MainForm->CTS->Checked = true;
        CTS_Count += 1;
        MainForm->CTS_Counter->Text = String(CTS_Count);
        break;

    case  SERIAL_CTS_OFF		:
        MainForm->CTS->Checked = false;
        break;

    case  SERIAL_DSR_ON		:
        MainForm->DSR->Checked = true;
        DSR_Count += 1;
        MainForm->DSR_Counter->Text = String(DSR_Count);
        break;

    case  SERIAL_DSR_OFF		:
        MainForm->DSR->Checked = false;
        break;

    case  SERIAL_ERR			:
        TIUPortCE = Message.WParam;
        MainForm->Err->Checked = true;
        MainForm->ExitButton->Enabled = true;  // allow exit on Serial Error
        break;
        /*  From winbase.h Values returned from ClearCommError()
        #define CE_RXOVER           0x0001  // Receive Queue overflow
        #define CE_OVERRUN          0x0002  // Receive Overrun Error
        #define CE_RXPARITY         0x0004  // Receive Parity Error
        #define CE_FRAME            0x0008  // Receive Framing error
        #define CE_BREAK            0x0010  // Break Detected
        #define CE_TXFULL           0x0100  // TX Queue is full
        #define CE_PTO              0x0200  // LPTx Timeout
        #define CE_IOE              0x0400  // LPTx I/O Error
        #define CE_DNS              0x0800  // LPTx Device not selected
        #define CE_OOP              0x1000  // LPTx Out-Of-Paper
        #define CE_MODE             0x8000  // Requested mode unsupported
        */
    case  SERIAL_BREAK			:
        MainForm->Break->Checked = true;
        break;
    }

}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMRFIDCOMStatus(TMessage &Message)
{
// Thread to VCL interface - Handles all VCL operations for the RFID COM
const uint32 event = Message.LParam;		// set event from Long Parameter
const uint32 DoneReason = Message.WParam;	// disconnect reason from Word Parameter
String Msg;

switch(event) {
    case  SERIAL_CONNECTED  :
        Msg = "Connected to " + ConnectedRFIDLED->Caption;
        if (DEBUGGING) DB->Debug_Print(Msg);
        //StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "RFID Serial Connected";
        ConnectedRFIDLED->Checked = true;
        ConnectedRFIDLED->Flashing = false;
        bRFIDCOMConnected = true;	// this is the only place this is set true
        bRFIDSendComplete = true;
        //DisconnectButton->Enabled = true;
        //cRFIDbuf[0] = '\0';
        //iRFIDbufCount = 0;
        break;

    case  SERIAL_DISCONNECTED  :
        Msg = "Disconnected from " + ConnectedRFIDLED->Caption + " (" + String(DoneReason) + ")";
        if (DEBUGGING) DB->Debug_Print(Msg);
        //StatusBar->Panels->Items[RFID_PORT_STATUS]->Text = "RFID Serial Disconnected";
        if (!DisconnectRequest) {
            // this is the case where tserial_event detected an error/failure of some sort and sent a SERIAL_DISCONNECTED event
            if (DEBUGGING) MainForm->DebugMessagePopup("Error", "RFID port unexpectedly disconnected (" + String(DoneReason) + ")");
            }
        ConnectedRFIDLED->Checked = false;
        ConnectedRFIDLED->Flashing = false;
        bRFIDCOMConnected = false;	// this is the only place this is set false
        bRFIDSendComplete = true;                
        //DisconnectButton->Enabled = false;
        //cRFIDbuf[0] = '\0';
        //iRFIDbufCount = 0;
        break;

    case  SERIAL_DATA_SENT  :
        bRFIDSendComplete = true;
        break;

    case  SERIAL_RING_ON    :
        assert(false);
        break;

    case  SERIAL_RING_OFF   :
        assert(false);
        break;

    case  SERIAL_CD_ON      :
        assert(false);
        break;

    case  SERIAL_CD_OFF     :
        assert(false);
        break;

    case  SERIAL_CTS_ON		:
        assert(false);
        break;

    case  SERIAL_CTS_OFF	:
        assert(false);
        break;

    case  SERIAL_DSR_ON		:
        assert(false);
        break;

    case  SERIAL_DSR_OFF	:
        assert(false);
        break;

    case  SERIAL_ERR		:
        RFIDPortCE = Message.WParam;
        assert(false);
        break;
        /*  From winbase.h Values returned from ClearCommError()
        #define CE_RXOVER           0x0001  // Receive Queue overflow
        #define CE_OVERRUN          0x0002  // Receive Overrun Error
        #define CE_RXPARITY         0x0004  // Receive Parity Error
        #define CE_FRAME            0x0008  // Receive Framing error
        #define CE_BREAK            0x0010  // Break Detected
        #define CE_TXFULL           0x0100  // TX Queue is full
        #define CE_PTO              0x0200  // LPTx Timeout
        #define CE_IOE              0x0400  // LPTx I/O Error
        #define CE_DNS              0x0800  // LPTx Device not selected
        #define CE_OOP              0x1000  // LPTx Out-Of-Paper
        #define CE_MODE             0x8000  // Requested mode unsupported
        */
    case  SERIAL_BREAK		:
        assert(false);
        break;
    }
}
//---------------------------------------------------------------------------
// these next routines setup the switches and other controls

void __fastcall TMainForm::SetSwitch(TAbRockerSwitch * Switch, int Checked)
{
assert(Checked == 1 || Checked == 0); 	// Check for actual true and false numeric values
if (Switch->Checked == Checked) {
	assert(Switch->OnStatusChanged != NULL);
	Switch->OnStatusChanged(Switch);
	} else {
	Switch->Checked = Checked;
	}
}

//--------------------------------

void __fastcall TMainForm::SetRadioGroup(TRadioGroup * RadioGroup, int Position)
{
assert(Position >= 0 && Position <= 12); 	// limit radiogroups to 13 positions
RadioGroup->ItemIndex = Position;
assert(RadioGroup->OnClick != NULL);
RadioGroup->OnClick(RadioGroup);
}
//--------------------------------

void __fastcall TMainForm::SetComboBox(TComboBox * ComboBox, int Position)
{
assert(Position >= -1 && Position <= 5); 	// limit combo boxes to 6 positions
if (Position == -1) Position = 0;
ComboBox->ItemIndex = Position;
assert(ComboBox->OnClick != NULL);
ComboBox->OnClick(ComboBox);
}
//--------------------------------

void __fastcall TMainForm::SetSlider(TAbVSlider * Slider, float Value)
{
assert(Value >= 0 && Value <= SPEEDMAX); 	// limit sliders to  max value of 120
if (Slider->Value == Value) {
	assert(Slider->OnValueChange != NULL);
	Slider->OnValueChange(Slider);
	} else {
	Slider->Value = Value;
	}
}
//--------------------------------

void __fastcall TMainForm::SetLED(TAbLED * LED, int Checked)
{
assert(Checked == 1 || Checked == 0);  	// Check for actual true and false numeric values
if (LED->Checked == Checked) {
	assert(LED->OnStatusChanged != NULL);
	LED->OnStatusChanged(LED);
	} else {
	LED->Checked = Checked;
	}
}
//--------------------------------

void __fastcall TMainForm::SetToggle(TAbToggleSwitch * Switch, int Checked)
{
assert(Checked == 1 || Checked == 0);  	// Check for actual true and false numeric values
if (Switch->Checked == Checked) {
	assert(Switch->OnClick != NULL);
	Switch->OnClick(Switch);
	} else {
	Switch->Checked = Checked;
	}
}
//--------------------------------
void __fastcall TMainForm::SetNumSpin(TAbNumSpin * Spinner, int Value)
{
assert(Value >= 0 && Value <= SPEEDMAX);		// limit spinners to max value of 120
if (Spinner->Value == Value) {
	assert(Spinner->OnValueChanged != NULL);
	Spinner->OnValueChanged(Spinner);
	} else {
	Spinner->Value = Value;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::OpButtonClick(TObject *Sender)
{
// Start the Operations form for engines 0-100
//
// https://stackoverflow.com/questions/31649315/c-builder-button-click-function-reentry
LOCKEM()
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= 100);	// 0 factory reset, 1-99, 100 "all engine operation"
if (!IsConsistRunning(EngineNo, TIUNo)) return;	// not running

// Manually create form and show it
// Many Operation Control Forms - if the form is not already created, then create it
//
if (ptrOp[TIUNo-1][EngineNo] == NULL) {	// EngineNo 0-100 (0 is factory reset, 1-99, 100 for "all engine operation")
	if (ptrLUEngineList[TIUNo-1][EngineNo] != NULL) {
    	delete[] ptrLUEngineList[TIUNo-1][EngineNo];	// if it's left over, delete the string - pointer to a char array
		ptrLUEngineList[TIUNo-1][EngineNo] = NULL;		// never a LashUpEngineList for a regular engine
        }
	ptrOp[TIUNo-1][EngineNo] = new TOp(this, RemoteNo, EngineNo, TIUNo, EngineData->Cells[ENG_NAME][EngineNo], NULL);   // create the Form and show it
								// Pass "this" as the owner
    							// 1-99 Engine Number, 100 "all engine operaton"
	assert(ptrOp[TIUNo-1][EngineNo] != NULL);	// make sure that construction worked
    ptrOp[TIUNo-1][EngineNo]->ParentHandle = Handle; // for return messages
    //
	} else {    // it's already created
	ptrOp[TIUNo-1][EngineNo]->Show();
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMSoundControl(TMessage &Message)
{
// process Engine/LashUp Open Sound Control message - Engines 0-100
// LashUps have their own message : CM_LASHUPSOUNDCTRLS
const int TIU = Message.WParam;		// set TIU Number from Word Parameter

// TIU may not equal TIUNo, check and reload data if necessary
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);

const int Consist = Message.LParam;	// set Consist Number from Long Parameter
assert (TIU >= 1 && TIU <= TIUMAX);
assert (Consist >= (Setup->AllowZero->Checked? 0 : 1) && Consist <= 100);	// 0 factory reset, 1-99, and 100 "all engine operation
assert(ptrOp[TIU-1][Consist] != NULL);	// not running - caller must validate
									   	// that the consist is running
OpenSoundControl(Consist, TIU);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SoundCtrlsButtonClick(TObject *Sender)
{
OpenSoundControl(EngineNo, TIUNo);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::OpenSoundControl(int Consist, int TIU)
{
assert(Consist >= (Setup->AllowZero->Checked? 0 : 1) && Consist <= 100);	// 0 factory reset, 1-99, 100 "all engine operation"
if (!IsConsistRunning(Consist, TIU)) return;	// not running
LOCKEM()
// Manually create form and show it
assert(Consist >= (Setup->AllowZero->Checked? 0 : 1) && Consist <= 100);	// 0 factory reset, 1-99, 100 "all engine operation"

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

// Many Volume Control Forms
if (ptrVolume[TIU-1][Consist] == NULL) {
	ptrVolume[TIU-1][Consist] = new TVolume(this, RemoteNo, Consist, TIU, EngineData->Cells[ENG_NAME][Consist], NULL);     // Pass "this" as the owner
	assert(ptrVolume[TIU-1][Consist] != NULL);	// make sure that construction worked
    ptrVolume[TIU-1][Consist]->ParentHandle = Handle; // for return messages
    // setup defaults on Operations form after startup  (String)
    if ((Consist != 0) && (EngineData->Cells[ENG_NAME][Consist] != "*")) {  // don't load uninitialized data
        if (ptrOp[TIU-1][Consist]->EngineTypeComboBox->ItemIndex == ELECTRIC_ENGINE) {
            ptrVolume[TIU-1][Consist]->BoilerStartupButton->Visible = true;
            ptrVolume[TIU-1][Consist]->BoilerReleaseButton->Visible = true;
        	}
    	try {
            SetSlider(ptrVolume[TIU-1][Consist]->MasterVolume, EngineData->Cells[VOL_MASTER][Consist].ToInt());
            SetSlider(ptrVolume[TIU-1][Consist]->EngineVolume, EngineData->Cells[VOL_ENGINE][Consist].ToInt());
            SetSlider(ptrVolume[TIU-1][Consist]->AccentVolume, EngineData->Cells[VOL_ACCENT][Consist].ToInt());
            SetSlider(ptrVolume[TIU-1][Consist]->HornWhistleVolume, EngineData->Cells[VOL_HORN][Consist].ToInt());
            // Horn or Whistle label
            if (EngineData->Cells[ENG_TYPE][Consist].ToIntDef(STEAM_ENGINE) == STEAM_ENGINE) {
            	ptrVolume[TIU-1][Consist]->HornWhistleVolume->SignalSettings->Name1 = "Whistle"; 	// Steam Engine
                } else {
            	ptrVolume[TIU-1][Consist]->HornWhistleVolume->SignalSettings->Name1 = "Horn";	// Diesel or Electric Engine
				}
            SetSlider(ptrVolume[TIU-1][Consist]->BellVolume, EngineData->Cells[VOL_BELL][Consist].ToInt());
  			// Mute is never saved and on Shutdown, the volume is returned to normal
            ptrVolume[TIU-1][Consist]->Repaint();
        	}
        catch ( EConvertError & oops  ){
            // first time this Engine is used, the EngineData StringGrid is empty and the ToInt() function
            // returns an EConvertError exception which we ignore and use the control's default values.
            DebugMessagePopup("Warning", "Empty Engine Sound Field - Repaired");
	        CaughtErrors->Value ++;
    	    }
        catch (...) {
            DebugMessagePopup("Error", "Unknown Startup Problem");
	        CaughtErrors->Value ++;
        	}
        }
	} else {   // it's already created
	ptrVolume[TIU-1][Consist]->Show();
	}
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineSpinValueChanged(TObject *Sender)
{
// this routine is called whenever something or someone changes EngineSpin->Value
assert((EngineSpin->Value >= 0) && (EngineSpin->Value <= 100));	// 0 factory reset, 1-99, 100 "all engine operation"

// This routine gets called during construction of the MainForm
// window. It is called only for the selected engine (shown as blue)

if (EngineSpin->Value == ALLENGINE) {
	// if TIU Port is not connected fail
	if (!IsTIUConnected()) {
		ShowMessage("You must be connected before you can select All Engine Operation");
        EngineSpin->Value = EngineNo;	// reset Engine Number to previous value
		// re-process engine number
        return;
    	}
    }
if (DEBUGGING) DB->Debug_Print("Engine Number changed from " + String(EngineNo) + " to " + String(EngineSpin->Value));
// the next line should be the ONLY place that EngineNo is changed (besides it's initialization to -1)
EngineNo = EngineSpin->Value;
//
if (EngineNo == 0) {	// Engine 0 (DCS #1) is a Factory Reset Engine
    NameofEngine->Text = "Factory Reset Engine";
    FileNamePrefix->Text = NameofEngine->Text;
    } else {
	NameofEngine->Text = EngineData->Cells[ENG_NAME][EngineNo];
	FileNamePrefix->Text = NameofEngine->Text;
	// Highlight the engine in the StringGrid
	EngineData->Col = ENG_NAME;
	EngineData->Row = EngineNo;
    EngineUpDown->Position = ENGINEUPDOWNLIMIT - EngineNo;
    }

// Enable buttons if an operations form is already created
if (ptrOp[TIUNo-1][EngineNo] != NULL) {
	SoundCtrlsButton->Enabled = true;
	OpButton->Enabled = true;
	} else {
	SoundCtrlsButton->Enabled = false;
	OpButton->Enabled = false;
	}

// Check for All Engine Mode
// TODO - There might be a way to support All Engine Mode while recording - think about it
if (EngineNo == ALLENGINE) {
	if (bRecording) {
    	ShowMessage("All Engine Operation is not supported when Recording");
       	AllEngineMode = false;
        return;
        }
	if (Setup != NULL && Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
		MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
	    return;
    	}
    // setup All Engine Operation for the attached TIU
	// I0 command to get engine list, this command takes a long time to get a response
    byte AEbyteResp[CMDMAX]; 	// Array for decoded binary response
    int iTIUCount;
   	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "I0", TIUNo, RemoteNo, 39, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, AEbyteResp);
    if (iTIUCount != 39) {	// the I0 command returns 39 characters
		// didn't get enough data
		DebugMessage("Failed to get Engine list to setup All Engine Operation");
		AllEngineMode = false;
    	} else {
	    // Got the data - enable All Engine Operation for Engines Attached to TIU
        // pull out the 100 Bit long Engine Encoding starting at character 11 for 26 characters
        // 99A6A9655A5A5555595A5555555555555555555555555555555555555555555555565565
		//   D9  B8  0F  00  0D  00  00  00  00  00  00  00  00  00  00  00  01  10
        // the string is in Hex so we really need two characters to get one 8 bit hex value
		String ThirtyNineHexString = byteBinaryStringToHexString(AEbyteResp, 18);
		// J13HexString must be AnsiString because the c_str() taken needs to be a char type.
		AnsiString J13HexString = "J" + ThirtyNineHexString.SubString(11, 26);   // 13 hex digits (26 characters)
        //
        // Due to an error in the TIU response to the I0 command, if an AIU is attached to the
        // TIU, the bit in the response string representing engine #99 (DCS 100) is on. That
        // bit needs to be turned off. In this case, the response may look like this:
		//	       14780C000D48000000000000000000000010
        // The command into All Engine Operation might look like this:
        //        (178) 100 J48000000000000000000000010 1 12
        // That  8  here      ^ might cause a problem so it will have to be removed
        // this means that Engines 99, 98, 97 & 96 cannot be set into All Engine Mode
        // right now, we we clear the byte of the string that contains these 4 engines so they can't be used
		if (DEBUGGING) DB->Debug_Print("All Engine Operation: " + J13HexString);
        // TODO I may be clearing the wrong byte here
		J13HexString[3] = '0';	// The AnsiString or String [] operator assumes a base index of 1.
		if (DEBUGGING) DB->Debug_Print("All Engine Operation: " + J13HexString);
        // NOTE: additional odd things might happen if more than one AIU is attached. I've not tested that case.
        //
        //
		if (ptrTIUCommThread->Send_String_FIFO(EngineNo, J13HexString.c_str(), TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD)) {    // AnsiString required
			DebugMessage("   All Engine Operation enabled - press Startup   ");
			//Application->MessageBox("All Engine Operation enabled", Caption.c_str(), MB_OK);
        	AllEngineMode = true;	// Use Engine Number of 100 (DCS 101 - hex 65) to command all engines
            } else {
            DebugMessage("  TIU failed to enter All Engine Operation  ");
            AllEngineMode = false;
            }
        return;
        }
	} else {
	AllEngineMode = false;
    }
return;
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::sEncodedStringToHexString(String Cmd)
{
// decode the encoded string into Hex
const int SLen = Cmd.Length();
String DecodedString = "";
//  Decode
    for (int idx = 1 ; idx < SLen; idx += 4) {
       	if (SLen - idx < 3) continue; // don't decode unless we have 4 characters
		String sVal = bintodec8(bitswapRX(Convert16B8BNumber(Cmd.SubString(idx, 4))));
        int iVal = sVal.ToInt();	
        String hVal = IntToHex(iVal & 0xFF, 2);
    	DecodedString += hVal;
        }
return DecodedString;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineDataClick(TObject *Sender)
{
if (EngineData->Row == -1 || EngineData->Col == -1) {
	EngineData->Row = ENG_TOPROW;  	// ignore out of bounds (if possible)
    EngineUpDown->Position = ENGINEUPDOWNLIMIT - ENG_TOPROW;
    }
// Set the Engine Number into the Control and the working variable
EngineSpin->Value = EngineData->Row;	// causes EngineNo to change
// set the Engine Name into the Control
NameofEngine->Text = EngineData->Cells[ENG_NAME][EngineNo];
FileNamePrefix->Text = NameofEngine->Text;
// Enable buttons if an operations form is created
if (ptrOp[TIUNo-1][EngineNo] != NULL) {
	SoundCtrlsButton->Enabled = true;
	OpButton->Enabled = true;
	} else {
	SoundCtrlsButton->Enabled = false;
	OpButton->Enabled = false;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineDataMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
int ARow, ACol;
if (Button == mbRight) {
	EngineData->MouseToCell(X, Y, ACol, ARow);
    EngineData->Row = ARow;
	EngineData->Col = ACol;
    }

EngineDataClick(Sender);
#ifdef NOTNEEDED
if (EngineData->Row == -1 || EngineData->Col == -1) {
	EngineData->Row = ENG_TOPROW;  	// ignore out of bounds (if possible)
    EngineUpDown->Position = ENGINEUPDOWNLIMIT - ENG_TOPROW;
    }
// Set the Engine Number into the Control and the working variable
EngineSpin->Value = EngineData->Row;	// causes EngineNo to change
// set the Engine Name into the Control
NameofEngine->Text = EngineData->Cells[ENG_NAME][EngineNo];
FileNamePrefix->Text = NameofEngine->Text;
// Enable buttons if an operations form is created
if (ptrOp[TIUNo-1][EngineNo] != NULL) {
	SoundCtrlsButton->Enabled = true;
	OpButton->Enabled = true;
	} else {
	SoundCtrlsButton->Enabled = false;
	OpButton->Enabled = false;
	}
#endif
// Popup the PopupMenu
if (Button == mbRight) {
	EngineData->PopupMenu->Popup(Left + X, Top + Y);
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineDataKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
// if ENTER is pressed, startup the Engine
if (Key == VK_RETURN) {
	if (EngineData->Cells[ENG_NAME][EngineNo] != "*") {
		StartupButton->OnClick(StartupButton);
    	} else {
    	// Edit the Engine Name Field
    	EditEngineName->OnClick(EditEngineName);
    	}
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineDataDrawCell(TObject *Sender, int ACol,
      int ARow, TRect &Rect, TGridDrawState State)
{
// Draw the contents of each cell
EngineData->Canvas->Font->Size = 12;
// http://bcbjournal.org/articles/vol3/9904/Dealing_with_sets.htm?PHPSESSID=2f96cbddf92baa7dc77f7a33591f59b5
EngineData->Canvas->Font->Style = EngineData->Canvas->Font->Style >> fsItalic;
// if the cell is selected, highlight it
if( State.Contains(gdSelected) ) {
    EngineData->Canvas->Font->Color = clWhite;
	EngineData->Canvas->Brush->Color = clBlue;		// White on Blue
	} else {
    if (ptrOp[TIUNo-1][ARow] != NULL) {
		EngineData->Canvas->Font->Color = clBlack;	// Started up Engines are Black on Lime
		EngineData->Canvas->Brush->Color = clLime;
    	} else {
		EngineData->Canvas->Font->Color = clBlack;	// Normal Black on White
		EngineData->Canvas->Brush->Color = clWhite;
        }
	}
if ((ARow != 0) && (ARow != ALLENGINE) && (ACol == ENG_NAME) && (EngineData->Cells[ENG_STATE][ARow] == INACTIVE)) {
	// if Engine is inactive make the cell Italic.
    // Skip the column title (Row 0) and All Engine Operation (Row 100)
    EngineData->Canvas->Font->Style = EngineData->Canvas->Font->Style << fsItalic;
	}
// fill the rectangle
EngineData->Canvas->FillRect(Rect);
// draw the text
String text = EngineData->Cells[ACol][ARow];
EngineData->Canvas->TextRect(Rect, Rect.Left, Rect.Top, text);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ReadEnginesButtonClick(TObject *Sender)
{
// https://stackoverflow.com/questions/31649315/c-builder-button-click-function-reentry
LOCKEM()
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
ReadEnginesButton->Enabled = false;

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

static TMessagePopup *MessagePopup = NULL;

if (DEBUGGING) STOPem->Enabled = true;
bStop = false;
// please wait message
if (!MessagePopup) {
	MessagePopup = new TMessagePopup(this, Handle, NULL, NULL);     // Pass "this" as the owner
	assert(MessagePopup != NULL);		// make sure that construction worked
    }
MessagePopup->Show();
MessagePopup->MessageLine->Caption = "  Reading Engines  ";
MessagePopup->Repaint();
// Try to load data from all engines on Layout and powered on
int Found = 0;

// Send I0 command to get 100 bit string
// I0 command to get engine list, this command takes a long time to get a response
byte REbyteResp[CMDMAX]; 	// Array for decoded binary response
int iTIUCount;
bool bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "I0", TIUNo, RemoteNo, 39, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, REbyteResp);
if (!bretv) {   // retry once
    if (DEBUGGING) DB->Debug_Print("ReadEnginesButtonClick() Retrying I0");
	bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "I0", TIUNo, RemoteNo, 39, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, REbyteResp);
	}

if (!bretv || iTIUCount != 39) {	// the I0 command returns 39 characters
	// command failed or didn't get enough data
	DebugMessage("         Failed to get list of Engines from TIU         ");
	} else {
	// Got a good response
    // pull out the 100 Bit long Engine Encoding starting at character 11 for 26 characters
    // the string is in Hex so we really need two characters to get one 8 bit hex value
	String EncodedBinary = byteBinaryStringToString(REbyteResp, 18);
	// 46 means skip the first 5 hex digits + 5 bits
    // then pick up the next 99 bits (it's 98 Engines
    // due to the bug in the TIU code when an AIU is attached.)
    String OneHundredBitString = EncodedBinary.SubString(46, 99);  // 99 bits
    // it appears that connecting an AIU causes engine bit 100 to be on and that confuses
    // this routine. See note just above and in for loop below. Routine has been modified
    // to only look up to DCS engine 99
	// Count the number of 1 bits on in the Binary string
    int EngCount = OneBitsinBinaryString(OneHundredBitString);
	// start with next to LSB of string, which is DCS engine #2 (LSB of string is DCS engine #1 or a Factory Reset Engine
	// go through the string from DSC engine #2 to DCS engine #99  (Engines 1-98)
	for (int DCSEngNo = 2 ; DCSEngNo <= 99 ; DCSEngNo++) {  // DCS Engines 2 through 99 (max 98 of them) Skip DSC #100
						 // must be one higher than actual Engine Number
                         // NOTE NOTE Due to error in TIU, if any AIU is connected, then
                         // the bit for engine 99 (DCS 100) is on. Need to skip that bit for now
                         // that is, check only up to Engine 98 (DCS 99)
        SetConsistState(DCSEngNo - 1, TIUNo, INACTIVE);		// set inactive at start of loop, Engines 1-98
        													// Uses Engine Number
		// For each bit that is on, send an Ixxx command to get the engine name
		if (bStop) {
	       	break;
	        }
        // if the appropriate bit is not on, then continue the loop
        // The [] operator of a String or AnsiSTring assumes a base index of 1
        if (OneHundredBitString[(99 + 1) - DCSEngNo] != '1') {
        	// if we are keeping Engines even if they are not in the TIU, then don't
            // clear the engine name. If we are not keeping engines, then reset the
            // engine name
		    if (!Setup->KeepEngines->Checked) {
            	EngineData->Cells [ENG_NAME] [DCSEngNo - 1] = "*";
	            // if the Engine is selected, we must clear out a couple of other fields
				if (EngineNo == (DCSEngNo - 1)) {
	    	    	NameofEngine->Text = "*";
	        	    FileNamePrefix->Text = NameofEngine->Text;
					}
                }
	    	continue;  // Engine not in TIU, go on to next engine
            }
        // status message
	    MessagePopup->StatusLine->Caption = "Engine # " + String(DCSEngNo - 1) +
	    	" of " + String(EngCount) + " Engines Total";
        MessagePopup->Repaint();
		// Convert Engine Number into chars for the command
	    // NOTE: the Engine Number as digits in the command do matter
        char str[10];
        (void)snprintf(str, sizeof str, "I%d", DCSEngNo);	// Uses the DCS Engine Number
		bool bretv = ptrTIUCommThread->Send_String_FIFO(DCSEngNo - 1, str, TIUNo, RemoteNo, 173, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, REbyteResp);
		if (!bretv) {	// retry once
           	if (DEBUGGING) DB->Debug_Print("ReadEnginesButtonClick() Retrying " + String(str));
			bretv = ptrTIUCommThread->Send_String_FIFO(DCSEngNo - 1, str, TIUNo, RemoteNo, 173, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, REbyteResp);
			}

		if (!bretv || iTIUCount != 173) {	// the I command returns 173 characters if an engine is found with that number
			// command failed or didn't get enough data so clear the ENG_NAME cell in the Engine List
		    EngineData->Cells [ENG_NAME] [DCSEngNo - 1] = "*";
            // if the Engine is selected, we must clear out a couple of other fields
			if (EngineNo == (DCSEngNo - 1)) {
	        	NameofEngine->Text = "*";
	            FileNamePrefix->Text = NameofEngine->Text;
				}
            DebugMessage("Read Engine Failed - Engine " + String(DCSEngNo - 1));
	    	continue;  // no engine name response, go on to next engine
	        }
	    // Got a good response
	    EngineData->Cells [ENG_NAME] [DCSEngNo - 1] = byteExtractEngineName(REbyteResp);  // put it into the EngineData TStringGrid
		// Get the info about whether this is a diesel or steam engine, use
        // that info to enable or disable certain switches
        //
        // Beginning of response to Ixxx command:
        // 99A55AA95A5A555556A555555596665956A665556565695A......
        //   D8  5E  0F  00  52  00  41  26  53  20  30  2D
        //                     ^^^^ is 0x00 for steam, 0x05/0x85 for Diesel and Gas/Electric
        unsigned char iEngType = REbyteResp[5] & 0x0F;
        // Eng Type = 0x00 for Steam
        // Eng Type = 0x05 for Diesel
        // Eng Type = 0x85 for newer PS3 Diesels
        // Eng Type = 0x25 for electric (PCC Car)
        // there may be others but I haven't seen them
        if (DEBUGGING) DB->Debug_Print("Found Engine " + String(DCSEngNo - 1) + " : 0x" + IntToHex(iEngType, 2));
        switch(iEngType) {
            case 0x00:
                EngineData->Cells [ENG_TYPE] [DCSEngNo - 1] = STEAM_ENGINE;
                break;
            case 0x05:
            case 0x85:
                EngineData->Cells [ENG_TYPE] [DCSEngNo - 1] = DIESEL_ENGINE;
                break;
            case 0x25:
                EngineData->Cells [ENG_TYPE] [DCSEngNo - 1] = ELECTRIC_ENGINE;
                break;
            default:
	        	DebugMessage("Found an unknown engine type of 0x" + IntToHex(iEngType, 2) + ". Setting it to Diesel");
                EngineData->Cells [ENG_TYPE] [DCSEngNo - 1] = DIESEL_ENGINE;
                break;
        	}
        //
        // This is a guess at determining if this is a PS2 or PS3 engine
        //
		byte ival;
		FetchByte(0x50, &ival, DCSEngNo -1, TIUNo);
		if (ival == 0x00) {
			EngineData->Cells[ENG_PS][DCSEngNo - 1] = PS3;
			} else {
			EngineData->Cells[ENG_PS][DCSEngNo - 1] = PS2;
			}

        SetConsistState(DCSEngNo - 1, TIUNo, ACTIVE);		// set active
		if (EngineNo == (DCSEngNo - 1)) {
	        // if the Engine is selected, we must set a couple of other fields
	    	NameofEngine->Text = EngineData->Cells [ENG_NAME] [DCSEngNo - 1];
	        FileNamePrefix->Text = NameofEngine->Text;
			}
	    Found++;
        }
	}

// redraw the control
EngineData->Repaint();
//
STOPem->Enabled = false;
delete MessagePopup;
MessagePopup = NULL;
Screen->Cursor = OldCursor;		// restore Cursor

// update users of the Engine List
PutEngData();

if (Found == 0) {
	DebugMessage("Load Engine Data Complete - No Engines Found");
	} else {
    DebugMessage("Load Engine Data Complete - Found " + String(Found) + " Engine(s)");
    }
//
ReadEnginesButton->Enabled = true;
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PutEngData(void)
{
// Update the users of the Engine List
// if Quick Ctrls for this TIU is running, update its Engine List
if (ptrQuick[TIUNo - 1] != NULL) {
    ptrQuick[TIUNo - 1]->GetEngData();
	}
// if Lashup Selection for this TIU is running, update its Engine List
if (ptrLashUp[TIUNo - 1] != NULL) {
    ptrLashUp[TIUNo - 1]->ReloadButtonClick(NULL);
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AddEnginesClick(TObject *Sender)
{
// Add Engine - Add a Factory Reset engine to the engine list
//
// Look for Engine with DCS number 1 and change the number.
// Ask for the new engine number
//
// https://stackoverflow.com/questions/31649315/c-builder-button-click-function-reentry
LOCKEM()
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// disable the AddEngines menu item while processing AddEnginesClick();
AddEngines->Enabled = false;

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

static TMessagePopup *MessagePopup = NULL;

if (DEBUGGING) STOPem->Enabled = true;
bStop = false;
// please wait message
if (!MessagePopup) {
	MessagePopup = new TMessagePopup(this, Handle, NULL, NULL);     // Pass "this" as the owner
	assert(MessagePopup != NULL);		// make sure that construction worked
    }
MessagePopup->Show();
MessagePopup->MessageLine->Caption = "  Adding Engines  ";
MessagePopup->Repaint();
// Look for Factory Reset engines and change their engine number to a valid unused number
int Found = 0;

// Send I0 command to get 100 bit string
// I0 command to get engine list, this command takes a long time to get a response
byte REbyteResp[CMDMAX]; 	// Array for decoded binary response
int iTIUCount;
// Uses while() to find all Factory Reset engines that might be on the track
while(1) {
    (void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "I0", TIUNo, RemoteNo, 39, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, REbyteResp);

    if (iTIUCount != 39) {	// the I0 command returns 39 characters
        // didn't get enough data
        DebugMessage("         Failed to get list of Engines from TIU         ");
        break;
        } else {
        // Got a good response
        // pull out the 100 Bit long Engine Encoding starting at character 11 for 26 characters
        // the string is in Hex so we really need two characters to get one 8 bit hex value
        String EncodedBinary = byteBinaryStringToString(REbyteResp, 18);
        // 46 means skip the first 5 hex digits + 5 bits
        // then pick up the next 99 bits (it's 98 Engines
        // due to the bug in the TIU code when an AIU is attached.)
        String OneHundredBitString = EncodedBinary.SubString(46, 99);  // 99 bits
        // Look at LSB of string which is DCS engine #1 or a Factory Reset engine
        // if the bit is off, there are no more engines to add
        // if the bit is on, then add the found Factory Reset engine
        // The [] operator of a String or AnsiSTring assumes a base index of 1
        if (OneHundredBitString[(99 + 1) - 1] == '1') {
            // Factory Reset engine is found
            // readup its data including engine name
            // if the bit is on, send an I1 command to get the engine name
            if (bStop) {
                break;
                }
            // NOTE: EngineNo in the command does not seem to matter
            // NOTE: the Engine Number as digits in the command do matter
            (void)ptrTIUCommThread->Send_String_FIFO(0, "I1", TIUNo, RemoteNo, 173, I0WAIT, NO_RECORD, NULL, NULL, &iTIUCount, REbyteResp);
            if (iTIUCount != 173) {	// the I command returns 173 characters if an engine is found with that number
                DebugMessage("    Add Engine Failed    ");
                break;  // no engine name response
                }
            // Got a good response, get the engine name and then ask for new engine number
            String EngineName = byteExtractEngineName(REbyteResp);
            // uses while() to ask again if given engine number is in use
            while(1) {
                NumberQuery->Caption = "Change Factory Reset Engine Number";
                char * G = Amp((const char *)EngineName.c_str());
                NumberQuery->Question->Caption = G;
                free(G);
                NumberQuery->SpinLabel->Caption = "New Engine Number:";
                NumberQuery->NumberSpin->MaxValue = ENGINEMAX;
                NumberQuery->NumberSpin->MinValue = (Setup->AllowZero->Checked? 0 : 1);
                NumberQuery->NumberSpin->Value = 1;	// current engine number
                NumberQuery->NumberSpin->Visible = true;
                int retv = NumberQuery->ShowModal();
                if (retv == mrCancel) break;
                // make sure the new engine number slot is empty
                int NewEngineNumber = NumberQuery->NumberSpin->Value;
                if (NewEngineNumber == 0) break;		// no change
                if (EngineData->Cells[ENG_NAME][NewEngineNumber] != "*") {
                    ShowMessage("        This Engine Number is in Use        ");
                    continue;
                    }
                //
                EngineData->Cells [ENG_NAME] [NewEngineNumber] = EngineName;  // put it into the EngineData TStringGrid
                SetConsistState(NewEngineNumber, TIUNo, ACTIVE);		// set active
                if (EngineNo == (NewEngineNumber)) {
                    // if the Engine is selected, we must set a couple of other fields
                    NameofEngine->Text = EngineData->Cells [ENG_NAME] [NewEngineNumber];
                    FileNamePrefix->Text = NameofEngine->Text;
                    }

	            // change engine number
                // send Exxx command where xx is the new Engine Number
                unsigned char Command_String[MAX_PATH];
                (void)snprintf((char *)Command_String, sizeof Command_String, "E%d", NewEngineNumber + 1);		// build command string
                // NOTE: 0 in the command is the Engine to change to NewEngineNumber
                // NOTE: the Engine Number as digits xxx do matter   (xxx = 2-100 for Engine Numbers 1-99, xxx = 1 for Factory Reset Engine #0)
                // Wait for first 11 characters of response
                if (!ptrTIUCommThread->Send_String_FIFO(0, Command_String, TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD)) {
                    DebugMessage("Failed to change Engine number of Factory Reset Engine");
                    break;	// don't make any changes
                    }

                // Get the info about whether this is a diesel or steam engine, use
                // that info to enable or disable certain switches
                //
                // Beginning of response to Ixxx command:
                // 99A55AA95A5A555556A555555596665956A665556565695A......
                //   D8  5E  0F  00  52  00  41  26  53  20  30  2D
                //                     ^^^^ is 0x00 for steam, 0x05/0x85 for Diesel and Gas/Electric
                unsigned char iEngType = REbyteResp[5] & 0x0F;
                // Eng Type = 0x00 for Steam
                // Eng Type = 0x05 for Diesel
                // Eng Type = 0x85 for newer PS3 Diesels
                // Eng Type = 0x25 for electric (PCC Car)
                // there may be others but I haven't seen them
                if (DEBUGGING) DB->Debug_Print("Found Factory Reset Engine Type : 0x" + IntToHex(iEngType, 2));
                switch(iEngType) {
                    case 0x00:
                        EngineData->Cells [ENG_TYPE] [NewEngineNumber] = STEAM_ENGINE;
                        break;
                    case 0x05:
                    case 0x85:
                        EngineData->Cells [ENG_TYPE] [NewEngineNumber] = DIESEL_ENGINE;
                        break;
                    case 0x25:
                        EngineData->Cells [ENG_TYPE] [NewEngineNumber] = ELECTRIC_ENGINE;
                        break;
                    default:
                        DebugMessage("Found an unknown engine type of 0x" + IntToHex(iEngType, 2) + ". Setting it to Diesel");
                        EngineData->Cells [ENG_TYPE] [NewEngineNumber] = DIESEL_ENGINE;
                        break;
                    }
                //
                // This is a guess at determining if this is a PS2 or PS3 engine
                //
                byte ival;
                FetchByte(0x50, &ival, NewEngineNumber, TIUNo);
                if (ival == 0x00) {
                    EngineData->Cells[ENG_PS][NewEngineNumber] = PS3;
                    } else {
                    EngineData->Cells[ENG_PS][NewEngineNumber] = PS2;
                    }
	            Found++;
                break;
	            }
            // Loop to see if another Factory Reset Engine is present
            //
            } else {
            break;
            }
        }
    }

// redraw the control
EngineData->Repaint();
//
STOPem->Enabled = false;
delete MessagePopup;
MessagePopup = NULL;
Screen->Cursor = OldCursor;		// restore Cursor

if (Found == 0) {
	DebugMessage("Add Engine Complete - No Engines Found");
	} else {
    DebugMessage("Add Engine Complete - Found " + String(Found) + " Engine(s)");
    }
// Reenable the menu item
AddEngines->Enabled = true;
return;
}
//---------------------------------------------------------------------------

int __fastcall TMainForm::QueryConsistState(int Consist, int TIU)
{
// Return the active/inactive state of an Engine on a TIU
// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);
//
if (Consist == 0) return (ACTIVE);	// factory reset engine not allowed
assert((Consist >= 1) && (Consist <= ENGINEMAX));	// 1-99
assert (TIU >= 1 && TIU <= TIUMAX);
return(EngineData->Cells[ENG_STATE][Consist].ToInt());	
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetConsistState(int Consist, int TIU, int State)
{
// Correct TIU data is loaded by FormActive() in LashUpForm
// if a different TIU's Engine Data is loaded, save it and load the requested TIU's Engine Data
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);
//
assert((State == ACTIVE) || (State == INACTIVE));
if (Consist == 0) return;	// factory reset engine not allowed
assert((Consist >= 1) && (Consist <= ENGINEMAX));	// 1-99
assert (TIU >= 1 && TIU <= TIUMAX);
if (EngineData->Cells[ENG_STATE][Consist] == ACTIVE) {		// Adjust Counts
	if (State == INACTIVE) {
        ActiveEdit->Value--;
		InactiveEdit->Value++;
    	}
	}
if (EngineData->Cells[ENG_STATE][Consist] == INACTIVE) {		// Adjust Counts
	if (State == ACTIVE) {
        ActiveEdit->Value++;
		InactiveEdit->Value--;
    	}
	}

EngineData->Cells[ENG_STATE][Consist] = State;	// set active or inactive
// redraw the control
EngineData->Repaint();
return;
}
//---------------------------------------------------------------------------

int __fastcall TMainForm::OneBitsinBinaryString(String BitString)
{
// Count the number of '1' bits in a Binary string
int Answer = 0;
const int Len = BitString.Length();
// char& __fastcall operator [](const int idx) returns the character in the string at index value idx.
// The [] operator assumes a base index of 1
for (int i = 1 ; i <= Len ; i++) {
	if (BitString[i] == '1' ) Answer ++;	// [] returns a char
	}
return Answer;
}
//---------------------------------------------------------------------------

int __fastcall TMainForm::OneBitsinHexString(String BitString)
{
assert(false);// this routine not used anymore
// Count the number of '1' bits in a Hex string
int Answer = 0;
const int Len = BitString.Length();
// The [] operator of an AnsiString or String assumes a base index of 1
for (int i = 1 ; i <= Len ; i++) {
	switch(toupper(BitString.c_str()[i])) {
			case '0':
				Answer += 0;
				break;
			case '1':
			case '2':
			case '4':
   			case '8':
				Answer += 1;
				break;
			case '3':
			case '5':
			case '6':
   			case '9':
   			case 'A':
   			case 'C':
				Answer += 2;
                break;
   			case '7':
   			case 'B':
   			case 'D':
   			case 'E':
				Answer += 3;
                break;
   			case 'F':
   				Answer += 4;
                break;
		}
	}
return Answer;
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::sExtractEngineName(String Cmd)
{
// pull out the Engine Name
String EngineNamefromTIU = sEncodedStringToASCIIString(Cmd);
return EngineNamefromTIU.SubString(7,16);	// starting at character 7 for 16 characters
// Note: character 6 in the string contains the engine number
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::sEncodedStringToASCIIString(String Cmd)
{
// decode the encoded string into an String
const int SLen = Cmd.Length();
String DecodedString = "";
//  Decode
    for (int idx = 1 ; idx < SLen; idx += 4) {
		if (SLen - idx < 3) continue; // don't decode unless we have 4 characters
		String sVal = bintodec8(bitswapRX(Convert16B8BNumber(Cmd.SubString(idx, 4))));
        int iVal = sVal.ToInt();
    	// print string byte in ASCII
    	String A_Char;
        if (iVal >= 32 && iVal < 128) {   // include SPACE and all characters up to 128
	        A_Char = (char)iVal;  // Convert to ASCII Character
			DecodedString = DecodedString + A_Char;
            } else {
            DecodedString += " ";
            }
        }
return DecodedString;
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::sEncodedStringToBinaryString(String Cmd)
{
// decode the encoded string into binary
const int SLen = Cmd.Length();
String DecodedString = "";
//  Decode
	for (int idx = 1 ; idx < SLen; idx += 4) {
		if (SLen - idx < 3) continue; // don't decode unless we have 4 characters
		String sVal = bitswapRX(Convert16B8BNumber(Cmd.SubString(idx, 4)));
    	DecodedString += sVal;
        }
return DecodedString;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ESTOPClick(TObject *Sender)
{
EmergencyStopAllEngines();	// Stop Engines and take down windows
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMEmergencyStop(TMessage &Message)
{
// process Emergency Stop message unless disabled
if (Setup->DisableEmergencyStopButton->Checked) return;
//
EmergencyStopAllEngines();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EmergencyStopAllEngines(void)
{
// playback might be running so stop it
//
bPlayingBack = bPlaybackCommError = false;	// in one second, the timer will be stopped and
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// clear the dispatchlist
if (ptrDL_Thread) {
	ptrDL_Thread->ClearDispatchList();
    }
    
// emergency Stop command to TIU, no wait for OK response or timeout
if (!RADIOMODE) {
	(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "o0", TIUNo, RemoteNo, 0, 0, RECORD_OK);
    }

if (RADIOMODE) {
	// do all TIU only when radio is working, otherwise, we are just connected by wire to one TIU
	for (int iTIU = 1 ; iTIU <= TIUMAX ; iTIU++) {
		// TIU numbers 1-TIUMAX
		(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "o0", iTIU, RemoteNo, 0, 0, RECORD_OK);
        }
    }

SetEvent(TerminationPBEvent);  	// stop any (and all) Playback threads
								// PBFileThread from Setup_Controls.cpp and from Hot_Com.cpp
// stop the Program Control Threads
for (int iPCNo = 1 ; iPCNo <= PCMAX ; iPCNo++) {
	if (ptrProgram_Control[iPCNo-1] != NULL && ptrProgram_Control[iPCNo-1]->ptrPC_Thread != NULL) {
		(void) ptrProgram_Control[iPCNo-1]->ptrPC_Thread->StopRequest();		// Thread will free memory on terminate
        ptrProgram_Control[iPCNo-1]->StopButtonOp->Visible = false;				// remove STOP button on Program Control window
    	}
    }

// clear the dispatchlist
if (ptrDL_Thread) {
	ptrDL_Thread->ClearDispatchList();
    }

// set global flag
bEmergencyStop = true;
if (DEBUGGING) DB->Debug_Print("Emergency Stop");
(void) Application->MessageBox(TEXT("Emergency Stop : Cycle Power on the TIU and then restart the program"), Caption.c_str(), MB_OK | MB_ICONSTOP);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::HiddenDblClick(TObject *Sender)
{
// Doo, doo, doo, lookin' out my back door -- John Fogerty
DebugButton->OnClick(DebugButton);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUCountButtonClick(TObject *Sender)
{
// Also called by the ConnectButtonClick() routine
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    AIUSpin->Value = 0;
    AIUNo = 0;
    // AIUButton->Enabled = false;
    AIUSpin->MaxValue = AIUMAX;	// allow all AIU
    AIUCount[TIUNo - 1] = AIUSpin->MaxValue;
    AIUSpin->Enabled = true;
    return;
    }
(void) GetAIUCount(TIUNo);	// Get the AIU Count
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIUVerButtonClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
   	TIUVerButton->Caption = "00.00";
    // set version number into RTCPopupMenu
    TIUVer1->Caption = "Display Selected TIU Ver # ";
    return;
    }
(void) GetTIUVer(TIUNo); 	// Get the TIU Version Number
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::GetAIUCount(int TIU)
{
assert (TIU >= 1 && TIU <= TIUMAX);
String AIUVal;

int iTIUCount = -1;
int iAIU;
bool retv;
byte AIUbyteResp[CMDMAX]; 	// Array for decoded binary response
AIUbyteResp[0] = NULL;
int respCode = 0x00;
// panStamp requires a long wait because after power on, the first command resets it and we must wait for the reset
// TODO Debug why panStamp goes into a reset on the first packet
(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "x", TIU, RemoteNo, 15, SHORT_WAIT * 5, NO_RECORD, NULL, &respCode, &iTIUCount, AIUbyteResp); 	// send 'x' command to the TIU
//
if (iTIUCount == 15) {	// this command seems to return a response code of 0x00 so that is not an error indication
    // Got the right number of bytes
	// pull out the value
	// A9A9 956A 5666 5555 5556 5556
    //                     ^^^^      Number of bytes in response : 1
	//                          ^^^^ AIU  Count -- returned by the TIU
    AIUVal = AIUbyteResp[5];
    // watchout for EConvertError
    try {
    	AIUCountButton->Caption = "AIU(s): " + AIUVal;
    	if (DEBUGGING) DB->Debug_Print(AIUCountButton->Caption);
        iAIU = AIUVal.ToInt();	// Throws an error if we can't read the AIU Count	
        }
    catch(...) {
        bTIUConnected = false;	// indicate problem in communications with TIU
        ConnectedTIULED->Checked = false;
        if (DEBUGGING) ptrTIUCommThread->PrintItem("GetAIUCount: Read AIU Count from TIU " + String(TIU) + " Invalid Value");
        retv = false;
        iAIU = 0;
        }
    if (iAIU == 0) {
    	AIUSpin->Value = 0;	// no AIU present
        AIUNo = 0;
        //AIUButton->Enabled = DEBUGGING;	// no AIU present
        AIUSpin->Enabled = false;
        AIUSpin->MaxValue = 0;		// max number of AIU permitted
        AIUSpin->MinValue = 0;		// min number of AIU permitted
    	} else {
        //AIUButton->Enabled = true;	// some AIU present
        AIUSpin->Enabled = true;
        AIUSpin->MaxValue = iAIU;	// max number of AIU permitted
        AIUSpin->MinValue = 1;		// min number of AIU permitted
        if (AIUNo == 0) {
        	AIUNo = 1;	// if current AIU is 0, then set it to one
            AIUSpin->Value = 1;
            }
        if (AIUNo > iAIU) {
        	AIUSpin->Value = iAIU;	// reset AIU number if too high
            AIUNo = iAIU;
            }
        }

    bTIUConnected = true;	// indicate successful communications with the TIU
    ConnectedTIULED->Checked = true;

    if (AIUSpin->Value == 0) {
    	AIUSpin->Enabled = false;	// no AIU present
        AIUSpin->ColorDefault = clInactiveBorder;
        } else {
    	AIUSpin->Enabled = true;	// some AIU present
	    AIUSpin->ColorDefault = clLime;
        }
	retv = true;
    } else {
    // Read AIU count failed
    if (DEBUGGING) ptrTIUCommThread->PrintItem("GetAIUCount: Read AIU Count from TIU " + String(TIU) + " Failed");
    // no AIU
    AIUSpin->Value = 0;	// no AIU
    AIUNo = 0;
    // AIUButton->Enabled = false;
    AIUSpin->MaxValue = AIUMAX;	// allow all AIU
    AIUSpin->Enabled = false;
    bTIUConnected = false;	// indicate problem in communications with TIU
    ConnectedTIULED->Checked = false;
    TIUSpin->ColorDefault = clRed;	// set TIU/AIU boxes color to red
    AIUSpin->ColorDefault = clRed;
    retv = false;
    }
// adjust AIU Control popup menu to enable only real AIU
//AIUControl1->Enabled = iAIU >= 1;
//AIUControl2->Enabled = iAIU >= 2;
//AIUControl3->Enabled = iAIU >= 3;
//AIUControl4->Enabled = iAIU >= 4;
//AIUControl5->Enabled = iAIU >= 5;
//
AIUCount[TIUNo - 1] = AIUSpin->MaxValue;
return retv;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::GetTIUVer(int TIU)
{
assert (TIU >= 1 && TIU <= TIUMAX);
bool retv;
int iTIUCount = -1;
int respCode = 0x00;
byte TIUbyteResp[CMDMAX] = "\0'"; 	// Array for decoded binary response
(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "!", TIU, RemoteNo, 21, SHORT_WAIT, NO_RECORD, NULL, &respCode, &iTIUCount, TIUbyteResp); 	// send '!' command to the TIU
//
if (iTIUCount == 21) {	// this command seems to return a response code of 0x00 so that is not an error indication
	// watch out for EConvertError
    try {
    	// pull out the digits
        //5AA5 969A 5A5A 5555 5559 6565 6569 6665 6565
        //                    ^^^^                      Number of bytes in response : 4
        //                          '0'  '4'  '2'  '0' 	<- characters returned by v4.20
        //66A5 9969 5A5A 5555 5559 6565 6569 6666 5555
        //                          '0'  '4'  '3'  \0 	<- characters returned by v4.30
        String D1 = (char) TIUbyteResp[5];
        String D2 = (char) TIUbyteResp[6];
        String D3 = (char) TIUbyteResp[7];
        String D4 = (char) TIUbyteResp[8];
        // NOTE: a bug in the TIU 4.30 software causes this response to come
        // out "04.3 ". Earlier version shows correctly as "04.20" - Verified
        // with Remote
        // The next line fixes the error so it is displayed correctly
		if (D4.IsEmpty() || D4 == " ") D4 = "0";
        // delete leading zero
        if (D1 == "0") D1 =   " ";
        TIUVerButton->Caption = D1 + D2 + "." + D3 + D4;
        // set version number into RTCPopupMenu
        TIUVer1->Caption = "TIU" + String(TIU) + " " + "v" + TIUVerButton->Caption;
        if (DEBUGGING) DB->Debug_Print(TIUVer1->Caption);
        // set version number into Connect button
        ConnectButton->Caption="v" + TIUVerButton->Caption;
        // Save Major and Minor version number as a floating point number
        TIUVersion = TIUVerButton->Caption.ToDouble();
        bTIUConnected = true;	// indicate successful communications with the TIU
        ConnectedTIULED->Checked = true;
		TIUSpin->ColorDefault = clLime;
        retv = true;
        }
    catch(...) {
        DebugMessage("GetTIUVer: Read TIU Version Number Invalid from TIU " + String(TIU));
        TIUVersion = 0;
        bTIUConnected = false;	// indicate problem in communications with TIU
        ConnectedTIULED->Checked = false;
        retv = false;
    	}
    } else {
    // Read TIU Version number Failed
    DebugMessage("GetTIUVer: Read TIU Version Number Failed from TIU " + String(TIU));
    bTIUConnected = false;	// indicate problem in communications with TIU
    ConnectedTIULED->Checked = false;
    TIUSpin->ColorDefault = clRed;	// set TIU box color to red
    retv = false;
    }
return retv;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ADButtonClick(TObject *Sender)
{
// Called by the button in Debug mode only
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= ENGINEMAX);	// 0 factory reset, 1-99
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }
AccLabel->Caption = "Acc Rate: ?";
DecLabel->Caption = "Dec Rate: ?";
int iAcc, iDec;	// use return values only if communications is successful
if (GetAccDecValues(EngineNo, TIUNo, &iAcc, &iDec)) {
	AccLabel->Caption = "Acc Rate: " + String(iAcc) + " Smph/s";	// change with DaXX, XX = 1,10
	DecLabel->Caption = "Dec Rate: " + String(iDec) + " Smph/s";	// change with DdXX
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EditEngineNameClick(TObject *Sender)
{
// Engine Data popup menu item
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= ENGINEMAX);	// 0/1-99
// if TIU Port is not connected do nothing
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Prepare the Engine Name field for editing
NameofEngine->Cursor = crHandPoint;
NameofEngine->ReadOnly = false;
NameofEngine->SetFocus();
NameofEngine->Hint = "After selecting from the menu, use this field to edit the Engine Name - Press [Enter] to send to engine";

return;
#ifdef OLDEDITCODEMOVED
// Edit Engine Name (requires AnsiString because c_str() is taken and must be char type
AnsiString NewEngineName = InputBox("Edit Engine #" + String(EngineNo) + " Name", "Only the first 16 characters will be sent to the Engine:", NameofEngine->Text);
if (NewEngineName == NameofEngine->Text) return;	// no change
// must be more than 1 character
if (NewEngineName.Length() == 1) {
	ShowMessage("New Engine Name must be more than one character");
    return;	// no change
    }
//
if (EngineNo > 0) EngineData->Cells[ENG_NAME][EngineNo] = NewEngineName;
// send new engine name to engine
NameofEngine->Text = NewEngineName;
FileNamePrefix->Text = NameofEngine->Text;

// send Command with exactly 16 characters for new Engine name
unsigned char Command_String[MAX_PATH];
// Limit engine name sent to TIU to exactly 16 characters
snprintf((char *)Command_String, sizeof Command_String, "Ns%-16.16s", NewEngineName.c_str());	// exactly 16 characters
if (!ptrTIUCommThread->Send_String_FIFO(EngineNo, Command_String, TIUNo, RemoteNo, 11, LONG_WAIT, NO_RECORD)) {
	ShowMessage("Failed to set new Engine name (Engine " + String(EngineNo) + ")");
    return;
    }
SetConsistState(EngineNo, TIUNo, ACTIVE);		// set active
#endif
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineDataPopupMenuPopup(TObject *Sender)
{
// Perform right click function then bring up the popup menu
if ((EngineNo == 0) && (Setup->AllowZero->Checked == false)) return;
EditEngineName->Caption = "Edit Engine #" + String(EngineNo) + " Name";
EditEngineNumber->Caption = "Edit Engine #" + String(EngineNo) + " Number";
DeleteEngine->Caption = "Delete Engine #" + String(EngineNo);
SetEngineintoCommandMode->Caption = "Set Engine #" + String(EngineNo) + " into Command Mode";
SetEngineintoLegacyMode->Caption = "Set Engine #" + String(EngineNo) + " into Legacy Mode";
SetEngineintoConventionalMode->Caption = "Set Engine #" + String(EngineNo) + " into Conventional Mode";
SetEngineActive1->Caption = "Set Engine #" + String(EngineNo) + " Active";
SetEngineInactive1->Caption = "Set Engine #" + String(EngineNo) + " Inactive";
EngineInformation1->Caption = "Engine #" + String(EngineNo) + " Information";
EngineTrackSignal1->Caption = "Engine #" + String(EngineNo) + " Track Signal (ping)";
EngineStatus1->Caption = "Engine #" + String(EngineNo) + " Status";
SetEnginePS2->Caption = "Set Engine #" + String(EngineNo) + " as PS2";
SetEnginePS2->Checked = (MainForm->EngineData->Cells[ENG_PS][EngineNo] == PS2);
SetEnginePS3->Caption = "Set Engine #" + String(EngineNo) + " as PS3";
SetEnginePS3->Checked = (MainForm->EngineData->Cells[ENG_PS][EngineNo] == PS3);
FactoryReset1->Caption = "Factory Reset Engine #" + String(EngineNo);
FeatureReset1->Caption = "Feature Reset Engine #" + String(EngineNo);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EditEngineNumberClick(TObject *Sender)
{
// Engine Data popup menu item
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= ENGINEMAX);	// 0/1-99
// if TIU Port is not connected do nothing
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// change the engine number - allowing changing the engine number for engine #0
if ((EngineNo != 0) && (EngineData->Cells[ENG_NAME][EngineNo] == "*")) {
	ShowMessage("         No Engine with this number         ");
	return;
    }

NumberQuery->Caption = "Edit Engine #" + String(EngineNo) + " Number";
NumberQuery->Question->Caption = "Select the New Engine Number";
NumberQuery->SpinLabel->Caption = "New Engine Number:";
NumberQuery->NumberSpin->MaxValue = ENGINEMAX;
NumberQuery->NumberSpin->MinValue = (Setup->AllowZero->Checked? 0 : 1);
NumberQuery->NumberSpin->Value = EngineNo;	// current engine number
NumberQuery->NumberSpin->Visible = true;
int retv = NumberQuery->ShowModal();
if (retv == mrCancel) return;
// make sure the new engine number slot is empty
int NewEngineNumber = NumberQuery->NumberSpin->Value;
if (EngineNo == NewEngineNumber)return;		// no change
if ((NewEngineNumber != 0) && (EngineData->Cells[ENG_NAME][NewEngineNumber] != "*")) {
	ShowMessage("        This Engine Number is in Use        ");
	return;
    }

// Shutdown the old Engine
ShutdownButton->OnClick(ShutdownButton); 	// closes Op and Volume Forms if they exist
Delay(1 * 1000);
// send Exxx command where xx is the new Engine Number
// Old or New engine number can be 0 which would indicate that it was a factory reset engine
unsigned char Command_String[MAX_PATH];
(void)snprintf((char *)Command_String, sizeof Command_String, "E%d", NewEngineNumber + 1);		// build command string
// NOTE: EngineNo in the command is the Engine to change to NewEngineNumber
// NOTE: the Engine Number as digits xxx do matter   (xxx = 2-100 for Engine Numbers 1-99)
// Wait for first 11 characters of response
if (!ptrTIUCommThread->Send_String_FIFO(EngineNo, Command_String, TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD)) {
	DebugMessage("Failed to change Engine number (Engine " + String(EngineNo) + ")");
    return;	// don't make any changes
    }

// change the global EngineNo, the Engine Name box and the selection on the Engine Data List

// transfer the rows data from EngineNo to NewEngineNumber
if (EngineNo != 0 && (NewEngineNumber != 0)) {
	assert(EngineData->Rows[EngineNo]->Count == ENGDAT_COUNT);
	EngineData->Rows[NewEngineNumber]->CommaText = EngineData->Rows[EngineNo]->CommaText;
	EngineData->Cells[ENG_NUM][NewEngineNumber] = NewEngineNumber; 	// fix the engine number in the StringGrid
    }

// clear out the old Engine Info
if (EngineNo != 0) {
        EngineData->Rows[EngineNo]->CommaText = String(EngineNo) + ENGDAT_INIT;
        assert(EngineData->Rows[EngineNo]->Count == ENGDAT_COUNT);
        }

// transfer the new engine number
EngineSpin->Value = NewEngineNumber;	// causes EngineNo to change
//
if (EngineNo != 0) {
	NameofEngine->Text = EngineData->Cells[ENG_NAME][EngineNo];
    FileNamePrefix->Text = NameofEngine->Text;
	// highlight the selection
	EngineData->Col = ENG_NAME;
	EngineData->Row = EngineNo;
    EngineUpDown->Position = ENGINEUPDOWNLIMIT - EngineNo;
    // Enable buttons if an operations form is created
	if (ptrOp[TIUNo-1][EngineNo] != NULL) {
		SoundCtrlsButton->Enabled = true;
		OpButton->Enabled = true;
		} else {
		SoundCtrlsButton->Enabled = false;
		OpButton->Enabled = false;
		}
    } else {
    // if EngineNo == 0
    NameofEngine->Text = "Factory Reset Engine";
    FileNamePrefix->Text = NameofEngine->Text;
    }
// update users of the Engine List
PutEngData();}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DeleteEngineClick(TObject *Sender)
{
// Engine Data popup menu item
// Delete an engine from the Engine List
if (EngineData->Cells[ENG_NAME][EngineNo] == "*") {
	ShowMessage("         No Engine with this number         ");
	return;
    }
String AskString = "Do you really want to delete " + EngineData->Cells[ENG_NAME][EngineNo] + "?";
int retv = Application->MessageBox(AskString.c_str(), TEXT("Confirmation"), MB_YESNO | MB_ICONQUESTION);
if (retv == IDNO) return;
// if consist is running, pop up error message
if (IsConsistRunning(EngineNo, TIUNo)) {
	ShowMessage("Shutdown the engine before you try to delete it");
    return;
    }
// otherwise delete it
// then clear out the Engine List
EngineData->Cells[ENG_NAME][EngineNo] = "*";
NameofEngine->Text = "*";
FileNamePrefix->Text = NameofEngine->Text;
SetConsistState(EngineNo, TIUNo, INACTIVE);		// set inactive

// update users of the Engine List
PutEngData();
}
//---------------------------------------------------------------------------

unsigned int HextoInt(char c)
{
// Convert a Hex digit to a binary number
  if (c >= '0' && c <= '9') return      (c - '0');
  if (c >= 'A' && c <= 'F') return 10 + (c - 'A');
  if (c >= 'a' && c <= 'f') return 10 + (c - 'a');
  return -1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FeatureResetButtonClick(TObject *Sender)
{
if (EngineNo <= 0 || EngineNo >= 100) return;	// Engines only

if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Engine Feature Reset
String FeatureResetMsg = "Are you sure you want to Feature Reset engine " + String(EngineNo) + " - " +
	EngineData->Cells[ENG_NAME][EngineNo] + "?";
int retv = Application->MessageBox(FeatureResetMsg.c_str(),	TEXT("Warning"), MB_YESNO | MB_ICONQUESTION);
if (retv == IDYES) {
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "F0", TIUNo, RemoteNo, 11, VERY_LONG_WAIT, RECORD_OK);
    }
// clear out the old Engine Info (not including the Name) - must be added again with READ button
EngineData->Rows[EngineNo]->CommaText = String(EngineNo) + ENGDAT_INIT;
assert(EngineData->Rows[EngineNo]->Count == ENGDAT_COUNT);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FactoryResetButtonClick(TObject *Sender)
{
if (EngineNo <= 0 || EngineNo >= 100) return;	// Engines only

if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// Engine Factory Reset
String FactoryResetMsg = "Are you sure you want to Factory Reset engine " + String(EngineNo) + " - " +
	NameofEngine->Text + " (Engine Number will reset to 0)?";
int retv = Application->MessageBox(FactoryResetMsg.c_str(), TEXT("Warning"), MB_YESNO | MB_ICONQUESTION);
if (retv == IDYES)
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "F1", TIUNo, RemoteNo, 11, VERY_LONG_WAIT * 20, NO_RECORD); 	// don't record this reset
// clear out the old Engine Info including the Name
if (EngineNo != 0) {
        EngineData->Rows[EngineNo]->CommaText = String(EngineNo) + ENGDAT_INIT;
        assert(EngineData->Rows[EngineNo]->Count == ENGDAT_COUNT);
        }
EngineData->Cells[ENG_NAME][EngineNo] = "*";
NameofEngine->Text = "*";
FileNamePrefix->Text = NameofEngine->Text;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::UndoRFIDBufferClick(TObject *Sender)
{
// implement Undo with a message to the RFIDBuffer RichEdit
//SendMessage(RFIDBuffer->Handle, EM_UNDO, 0, 0);
RFIDBuffer->Undo();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::UndoTIUBufferClick(TObject *Sender)
{
// implement Undo with a message to the TIUBuffer RichEdit
//SendMessage(TIUBuffer->Handle, EM_UNDO, 0, 0);
TIUBuffer->Undo();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::UndoCRCBufferClick(TObject *Sender)
{
// implement Undo with a message to the CRCBuffer RichEdit
//SendMessage(CRCBuffer->Handle, EM_UNDO, 0, 0);
CRCBuffer->Undo();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::I0ButtonClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

// Add Engine
byte I0byteResp[CMDMAX]; 	// Array for decoded binary response
int TIUbufCount;
(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "I0", TIUNo, RemoteNo, 39, I0WAIT, NO_RECORD, NULL, NULL, &TIUbufCount, I0byteResp);
// Returns a 104 bit string (13 bytes) with right most bit being
// engine 1 (a factory reset engine) of the engine number to be added - only need the rightmost 99 bits
if (TIUbufCount < 39) {	// the I0 command returns 39 characters
	// didn't get enough data
    DebugMessage("Failed to get Engine list");
    } else {
    // Got a good response;
    // pull out the 100 Bit long Engine Encoding starting at character 11 for 26 characters
    // the string is in Hex so we really need two characters to get one 8 bit hex value
	String EncodedBinary = byteBinaryStringToString(I0byteResp, 18);
    // skip the first five hex characters + 5 bits
    String OneHundredBitString = EncodedBinary.SubString(46, 99);	// 99 bits
    // NOTE NOTE there is a problem with the 100 bits. When an AIU is plugged
    // into the TIU, engine bit DCS #100 (Engine 99) comes on
    // EG:                    .
    //   F4  D0  0F  00  0D  08  00  00  00  00  00  00  00  00  00  00  01  10
    // it's the 08 in the first of the 13 values that make up the 100 bits
             // NOTE NOTE Due to error in TIU, if any AIU is connected, then
             // the bit for engine 99 (DCS 100) is on. Need to skip that bit for now
             // that is, check only up to Engine 98 (DCS 99)
	// Count the number of 1 bits on in the Binary string
    int EngCount = OneBitsinBinaryString(OneHundredBitString);
    if (DEBUGGING) DB->Debug_Print("Found " + String(EngCount) + " Engine(s)");
    I0Button->Caption = String(EngCount) + " Engine(s)";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SaveCRCClick(TObject *Sender)
{
WritetoDiskClick(WriteCRC);
}
//---------------------------------------------------------------------------

unsigned int __fastcall TMainForm::EncodedStringToDecimalInt(String Cmd, int Index, int Len)
{
// decode the encoded string into Binary
// Index starts at 1
unsigned int iVal = 0;
//  Decode
for (int idx = Index ; idx < (Index + Len); idx += 4) {
   	if ((Index + Len ) - idx < 3) continue; // don't decode unless we have 4 characters
	String sVal = bintodec8(bitswapRX(Convert16B8BNumber(Cmd.SubString(idx, 4))));
    iVal = (iVal << 8 ) | (sVal.ToInt() & 0xFF);
    }
return iVal;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodeLastResponseButtonClick(TObject *Sender)
{
Last_Response_Decoded->Text = sDecodeString(Last_Response->Text, false);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::VarModeButtonClick(TObject *Sender)
{
const TButton * ButtonUsed = (TButton *)Sender;
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// open Var control based on button pushed
if (ButtonUsed->Name.Pos("1")) {
	StartUpVarControl(1);
    }
if (ButtonUsed->Name.Pos("2")) {
	StartUpVarControl(2);
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartUpVarControl(int Chan)
{
// https://stackoverflow.com/questions/31649315/c-builder-button-click-function-reentry
LOCKEM()
ChannelNo = Chan;	// global location of Channel Number
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
//
// Manually create form and show it
assert(TIUNo >= 1 && TIUNo <= TIUMAX);
assert(ChannelNo >= 1 && ChannelNo <= CONVMAX);
// Many Operation Control Forms
if (ptrConvMode[TIUNo-1][ChannelNo-1] == NULL) {    // create the Form and show it - TIU 1-5 - 	Var ChannelNo 1-2
	ptrConvMode[TIUNo-1][ChannelNo-1] = new TConvMode(this, RemoteNo, EngineNo, TIUNo, ChannelNo);
																  // Pass "this" as the owner
	assert(ptrConvMode[TIUNo-1][ChannelNo-1] != NULL);		// make sure that construction worked
    ptrConvMode[TIUNo-1][ChannelNo-1]->ParentHandle = Handle; // for return messages
    // setup defaults on Operations form after startup
	} else {    // it's already created
	ptrConvMode[TIUNo-1][ChannelNo-1]->Show();
	}
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DCSSetupClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// DCS Setup Form should be manually created as it is rarely needed
assert(TIUNo >= 1 && TIUNo <= TIUMAX);
// One DCS Setup Form per TIU (5 total)
if (ptrDCS[TIUNo-1] == NULL) {	// TIU 1-5
	ptrDCS[TIUNo-1] = new TDCS_Setup(this, RemoteNo, EngineNo, TIUNo);     // Pass "this" as the owner
	assert(ptrDCS[TIUNo-1] != NULL);	// make sure that construction worked
    ptrDCS[TIUNo-1]->ParentHandle = Handle; // for return messages
	} else {   // it's already created
	ptrDCS[TIUNo-1]->Show();
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::LashUpButtonClick(TObject *Sender)
{
assert(Setup != NULL);		// setup must be created
// LashUp Form should be manually created as it is rarely needed
// if the form is not created, then create it, otherwise show the previously created form
//
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
//
assert(TIUNo >= 1 && TIUNo <= TIUMAX);
if (ptrLashUp[TIUNo-1] == NULL) { 								// TIU 1-5
	ptrLashUp[TIUNo-1] = new TLashUp(this, RemoteNo, TIUNo);	// Pass "this" as the owner
	assert(ptrLashUp[TIUNo-1] != NULL);			// make sure that construction worked
    ptrLashUp[TIUNo-1]->ParentHandle = Handle;	// for return messages
   	ptrLashUp[TIUNo-1]->Show();					// show the form when the button is pushed
	} else {   // it's already created
	ptrLashUp[TIUNo-1]->Show();					// show the form when the button is pushed
	}
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMStartLashUp(TMessage &Message)
{
// process Start Lashup message
//
unsigned char LashUpEngineList[LASHUP_ENGINE_LIST_LENGTH];	// Bytes containing the LashUp in the form 0x2C 0x01 0x82 0x03 0xFF
(void) strncpy((char *)LashUpEngineList, (char *) Message.WParam, sizeof LashUpEngineList); 	// binary engine list from Word parameter

const int LUNo = Message.LParam & 0x00FF;	// set Lashup Number from Long Parameter -  101-120
if (LUNo == 0) return;						// no LashUp Number
assert(LUNo >= LASHUPMIN && LUNo <= LASHUPMAX);	// value 101 to 120 - DCS Engine Number always 102 (0x66)
const int TIU = (Message.LParam & 0xFF00) >> 8;	// set TIU Number from Long Parameter - TIU Number 1-5
assert(TIU >= 1 && TIU <= TIUMAX);

// TIU may not equal TIUNo, check and reload data if necessary
ChangeTIUData(EngineData, TIU);
assert(TIU == TIUNo);
//
String Name;
if (ptrLashUp[TIU-1]) { 			// LashUp name is put into LashUpData->Cells by LashUp_Selection.cpp
	Name = ptrLashUp[TIU-1]->LashUpData->Cells[LASHUP_NAME][LUNo - 100];            // Name of the LashUp
    }
if (ptrLUNameList[TIU-1][LUNo]) { 	// LashUp name is put here by Program Control routine LashUpStartUp
	Name = ptrLUNameList[TIU-1][LUNo];
    }

if (Name == "*") return;				// no LashUp name
if (LashUpEngineList[0] == '\0') return;	// no LashUpEngineList

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;

// Manually create the Operation Form and show it
//
// Many Operation Control Forms for LashUps
if (ptrOp[TIU-1][LUNo] == NULL) {
    //  ---------------------------------------------------------------------
    //  ---------------------------------------------------------------------
    // New Location of "U" command
	// Build Hex lash up string 0x2C 0x00 0x00 0x00 0xFF
	// create and setup the operations window
	// Create lashup "U" command followed by lash up string  0A 03 84 FF
	// 5555669955AA5A5556569559AAAA996556569696A9694D
	//   00  66  55  0A  03  84  FF  98  03  C3  BC		DCS v5.00, DCS v6.00 is different, see below
	//          'U'
    int respCode = 0x00;
	bool retv = ptrTIUCommThread->Send_String_FIFO(LUNo, "U", TIU, RemoteNo, 11, VERY_LONG_WAIT, RECORD_OK, LashUpEngineList + 1, &respCode);
		// uses LashUpEngineList but skiping over the 0x2C, starts right at the head engine number
		// LashUp is always sent to the TIU as engine number 101 (DCS EngineNo #102 0x66).
        // It takes quite a while for the TIU to setup the LashUp (almost 2 seconds)
        // NOTE NOTE : this command is handled as a special case inside of ptrTIUCommThread->Send_String_FIFO()
        // and Playback - "U"
	//  ---------------------------------------------------------------------
	if (!retv || (respCode == 0x00)) {
    	// U command failed
	    Screen->Cursor = OldCursor;		// restore cursor
		//new TMessageDisplayThread(false, this, Handle, "Error", "LashUp " + String(LUNo) + " not on track (LashUp Start Up)", "", 2000);
		//MainForm->DebugMessagePopup( "Error", "LashUp " + String(LUNo) + " not on track (LashUp Start Up)");
		new TMessageDisplayThread(false, this, Handle, "Warning", "LashUp " + String(LUNo) + " not on track (LashUp Start Up)", "", 2000);
		return;	// startup failed - don't show Op window
        }
	//  ---------------------------------------------------------------------
	/*  DCS V5.00 Commands look like this: -------------------------------------------
	startup - "u4," command followed by lashup string  0A 03 84 FF
	5555669965AA656969595A5556569559AAAA99665656A9969A9A4D
	  00  66  75  34  2C  0A  03  84  FF  99  03  E9  CF
	         'u' '4' ','
	whistle is command "w2," followed by lash up string
	5555669966AA666569595A5556569559AAAA9A655656965AA9A94D
	  00  66  77  32  2C  0A  03  84  FF  9A  03  87  FC
	         'w' '2' ','
	555566996695569956995699559969595A5556569559AAAA9A66565656A699AA4D
	  00  66  62  46  46  46  44  2C  0A  03  84  FF  9B  03  53  DD
	         'b' 'F' 'F' 'F' 'D' ','
	--------------------------------------------------------------------- */
    /*  DCS V6.00 Commands look like this: ------------------------------------------
    The way the Lash Up engine list is sent to the TIU was changed dramatically
    in DCS v6.00.

    In version 5.00 and earlier, the lash up engine list was sent as
    binary numbers, preceeded by a comma "," 0x2C and terminated
    by 0xFF. Here is an example of DCS engines 15 and 16 (Remote engines 14 and 15)
    with the second engine running in reverse.

	--V5.00
	5555669965AA656969595A5A9565AAAAA555555955959AA54D
	  00  66  75  34  2C  0F  90  FF  A0  04  40  DA
	         'u' '4' ','								V5.00 uses DCS Engine #

    Here are two examples using DCS engines 10, 3 and 4 (with engine 4 reverse running):

	5555669966AA666569595A5556569559AAAA9A655656965AA9A94D
	  00  66  77  32  2C  0A  03  84  FF  9A  03  87  FC	whistle is command "w2," followed by lash up string
	         'w' '2' ','
	555566996695569956995699559969595A5556569559AAAA9A66565656A699AA4D
	  00  66  62  46  46  46  44  2C  0A  03  84  FF  9B  03  53  DD
	         'b' 'F' 'F' 'F' 'D' ','

    In version 6.00, the lash up engine list is sent as ASCII characters but
    still preceeded by a comma "," 0x2C and terminated by 0xFF. The ASCII is
    the hex representation of the engine number as appeared in v5.00 commands.
    Here is an example of DCS engines 15 and 16 (Remote engines 14 and 15):

    --V6.00
	5555669965AA656969596565569965666565AAAA956955596AA9999A4D
	  00  66  75  34  2C  30  46  31  30  FF  94  04  7E  CD
	         'u' '4' ',' '0' 'F' '1' '0'				V6.00 uses ASCII DCS Engine number

    In both cases, reverse running is indicated by adding 0x80 (decimal 128) to the
    engine number before the command is created (H.O. bit is turned on).
    --------------------------------------------------------------------- */

    //  ---------------------------------------------------------------------
	ptrLUEngineList[TIU-1][LUNo] = new char [LASHUP_ENGINE_LIST_LENGTH];
	assert(ptrLUEngineList[TIU-1][LUNo] != NULL);		// make sure that memory allocation worked
    strncpy(ptrLUEngineList[TIU-1][LUNo], LashUpEngineList, LASHUP_ENGINE_LIST_LENGTH);
    //  ---------------------------------------------------------------------
	ptrOp[TIU-1][LUNo] = new TOp(this, RemoteNo, LUNo, TIU, Name, ptrLUEngineList[TIU-1][LUNo]);   // create the Form and show it
											  // Pass "this" as the owner
                                              // LashUpNo 101-120 (always DCS Number #102 0x66)
	assert(ptrOp[TIU-1][LUNo] != NULL);		// make sure that construction worked
    ptrOp[TIU-1][LUNo]->ParentHandle = Handle; // for return messages - back to MainForm
	// Startup
    int HeadEngine = (LashUpEngineList[1] & 0x7F) - 1;
	int iAcc, iDec; // routine returns default values if we cannot communicate.
	if (!Setup->StandardAccDec->Checked) {  // if "Standard Acc/Dec" is not checked
		// Use the Acc/Dec rates from the lead engine for all of the other engines
		// Read Acc/Dec rates from head engine and display them, and send them to the LashUp
		// GetAccDecValues() returns default values if we cannot communicate.
		// Note: LashUpEngineList[] already contains DCS engine numbers so we must return it
		// to an engine number because GetAccDecValues calls ptrTIUCommThread->Send_String_FIFO which increments it
		(void) GetAccDecValues(HeadEngine, TIU, &iAcc, &iDec);
		// TODO - Consider saving the Acc/Dec rates for the Middle and Tail Engines in the LashUp and restoring them on shutdown
	    }
	// Volume setting matches the head engine
    // Setup volume control defaults before startup occurs
    //
    try {
      // Set Engine sound volumes.
      // TODO Needed because Horn & Bell Sounds are muted in a LashUp and I don't seem to restore them correctly
      //
      char str[15];
      snprintf(str, sizeof str, "v0%d", EngineData->Cells[VOL_MASTER][HeadEngine].ToInt()); 	// Master Volume
      (void) ptrTIUCommThread->Send_String_FIFO(LUNo, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, LashUpEngineList);
      //ptrOp[TIU-1][LUNo]->LastSentMasterVolumeValue = EngineData->Cells[VOL_MASTER][HeadEngine].ToInt();

      snprintf(str, sizeof str, "v1%d", EngineData->Cells[VOL_ENGINE][HeadEngine].ToInt()); 	// Engine Sounds
      (void) ptrTIUCommThread->Send_String_FIFO(LUNo, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, LashUpEngineList);

      snprintf(str, sizeof str, "v2%d", EngineData->Cells[VOL_ACCENT][HeadEngine].ToInt()); 	// Accent Sounds
      (void) ptrTIUCommThread->Send_String_FIFO(LUNo, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, LashUpEngineList);

      snprintf(str, sizeof str, "v3%d", EngineData->Cells[VOL_HORN][HeadEngine].ToInt()); 	// Horn Sounds
      (void) ptrTIUCommThread->Send_String_FIFO(LUNo, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, LashUpEngineList);

      snprintf(str, sizeof str, "v4%d", EngineData->Cells[VOL_BELL][HeadEngine].ToInt());	// Bell Sounds
      (void) ptrTIUCommThread->Send_String_FIFO(LUNo, str, TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, LashUpEngineList);
      }
    catch ( EConvertError & oops ){
      // first time this Engine is used, the EngineData StringGrid is empty and the ToInt() function
      // returns an EConvertError exception
      DebugMessagePopup("Warning", "Empty Engine Data Field in Lashup Startup- Repaired", "");
      CaughtErrors->Value ++;
      }
    catch (...) {
      CaughtErrors->Value ++;
      DebugMessagePopup("Error", "Unknown Startup Problem in Lashup Volume Settings", "");
      }
	// Command "u4" for normal, "u6" for Extended
	// if either SHIFT key is pressed whilst clicking "Startup" button, use the extended startup sound
	//
	if (Setup->ExtendedSUSD->Checked || (ShiftKeyState < 0)) {
		(void)ptrTIUCommThread->Send_String_FIFO(LUNo, "u6", TIU, RemoteNo, 11, LONG_WAIT, RECORD_OK, LashUpEngineList);
	    } else {
		(void)ptrTIUCommThread->Send_String_FIFO(LUNo, "u4", TIU, RemoteNo, 11, LONG_WAIT, RECORD_OK, LashUpEngineList);
		}
	// set DIR forward  "d0"
	(void)ptrTIUCommThread->Send_String_FIFO(LUNo, "d0", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK, LashUpEngineList);
	// set speed to zero unless it alread is
	if (ptrOp[TIU-1][LUNo]->SpeedSlider->Value >= 0.01) {
	    // BUG workaround - slider starts out at 10 	// causes the 's0' command to be sent because the slider is at 10
		ptrOp[TIU-1][LUNo]->SetSpeedSliderValue(0.0);
	    }
    // setup the Op Window
    // Engine type matches the head engine
    // (LashUpEngineList[1] & 0x7F)  is the DCS number of the head engine
    // Note: LashUpEngineList[] contains DCS engine numbers so we have to subtract one
    // Note: in lashup mode, many of the settings in the engines are controlled automatically by the TIU
    // based on the sequence of engines inthe engine list (ie, head - middle - tail).
    // These settings include : bell, headlight, coupler operation and whistle (and maybe more).
    int HeadEngineType = EngineData->Cells[ENG_TYPE][HeadEngine].ToInt();
    SetComboBox(ptrOp[TIU-1][LUNo]->EngineTypeComboBox, HeadEngineType);
	// 5 LED switches
    SetLED(ptrOp[TIU-1][LUNo]->SimulatedRule17LED, (EngineData->Cells[ENG_HDLT][HeadEngine].ToIntDef(false) >> 1) & 0x01);
    SetLED(ptrOp[TIU-1][LUNo]->BellOnWithHornLED, (EngineData->Cells[ENG_HDLT][HeadEngine].ToIntDef(false) >> 2) & 0x01);
    SetLED(ptrOp[TIU-1][LUNo]->AutoBellLED, (EngineData->Cells[ENG_HDLT][HeadEngine].ToIntDef(false) >> 3) & 0x01);
    SetLED(ptrOp[TIU-1][LUNo]->AutoWhistleLED, (EngineData->Cells[ENG_HDLT][HeadEngine].ToIntDef(false) >> 4) & 0x01);
    SetLED(ptrOp[TIU-1][LUNo]->AutoLaboredSmokeLED, (EngineData->Cells[ENG_SMOKELEVEL][HeadEngine].ToIntDef(false) >> 2) & 0x01);
	// Smoke setting matches the head engine - off first then setting
    SetSwitch(ptrOp[TIU-1][LUNo]->SmokeSwitch, true);
    SetSwitch(ptrOp[TIU-1][LUNo]->SmokeSwitch, false);
    if (!Setup->SmokeOFFatStartup->Checked) SetSwitch(ptrOp[TIU-1][LUNo]->SmokeSwitch, EngineData->Cells[ENG_SMOKE][HeadEngine].ToIntDef(false));
    // For max speed, the the lowest max of the engines in the LashUp
    // (LashUpEngineList, 2nd byte is DCS engine number so need to subtract one)
    int mx = Setup->MaxSpeed->Value; 	// start at overall max speed
    for (int i = 1; i <= MAXLASHUP ; i++) {
    	if ((LashUpEngineList[i] & 0x7F) == 0x7F) break;
	    // (LashUpEngineList, 2nd byte is DCS engine number so need to subtract one)
        int nextmx = EngineData->Cells[ENG_MAXSPEED][(LashUpEngineList[i] & 0x7F) - 1].ToInt();
        if (nextmx < mx) mx = nextmx;
    	}
    if (mx < 10) mx = Setup->MaxSpeed->Value; 	// maximum value
    ptrOp[TIU-1][LUNo]->MaxSpeedSpin->Value = mx;
    ptrOp[TIU-1][LUNo]->MaxSpeedSpin->MaxValue = mx;
	ptrOp[TIU-1][LUNo]->SpeedSlider->SignalSettings->DigitalTo = ptrOp[TIU-1][LUNo]->MaxSpeedSpin->Value;
	ptrOp[TIU-1][LUNo]->SpeedSlider->SignalSettings->ValueTo = ptrOp[TIU-1][LUNo]->MaxSpeedSpin->Value;
	ptrOp[TIU-1][LUNo]->SpeedSlider->ScaleSettings->Steps = ptrOp[TIU-1][LUNo]->MaxSpeedSpin->Value / 10;
	if (Setup->StandardAccDec->Checked) {
		// set Acc and Dec values to 1
        ptrOp[TIU-1][LUNo]->SetStandardAccDecValues();
	    } else {
	    // Load Acc and Dec values as read from the first engine into the LashUp
	    SetNumSpin(ptrOp[TIU-1][LUNo]->AccSpin, iAcc);		// Sends Acc Rate to the LashUp
	    SetNumSpin(ptrOp[TIU-1][LUNo]->DecSpin, iDec);		// Sends Dec Rate to the LashUp
        }
    StartedConsists->Value++;
	} else {    // it's already created
	ptrOp[TIU-1][LUNo]->Show();
	}

// if either CTRL key is pressed whilst clicking "Startup" button, show the Sound form also
if (CtrlKeyState < 0) {
	// Send CM_LASHUPSOUNDCTRLS message to MainForm which will open the Volume Control Form
	// for the message, use the same message that was sent to this routine
	(void) PostMessage(Handle, CM_LASHUPSOUNDCTRLS, Message.WParam, Message.LParam);
    }
AltKeyState = ShiftKeyState = CtrlKeyState = 0;
Screen->Cursor = OldCursor;
return;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::GetAccDecValues(int Engine, int TIU, int * iAcc, int * iDec)
{
assert (TIU >= 1 && TIU <= TIUMAX);
bool retv;
if (Setup->TXOnlyComm->Checked) {	// Two way communications required
	if (Setup->StandardAccDec->Checked) {
	    *iAcc = 1; *iDec = 1;
    	} else {
	    *iAcc = 4; *iDec = 2;
        }
    return true;
    }
byte ADbyteResp[CMDMAX]; 	// Array for decoded binary response
if (ptrTIUCommThread->Send_String_FIFO(Engine, "Dq", TIU, RemoteNo, 17, MED_WAIT, NO_RECORD, NULL, NULL, NULL, ADbyteResp)) {
	// pull out the digits from the Response
	// A969 5A6A 6656 5555 5655 5559 5655
	//                          Acc  Dec
    *iAcc = ADbyteResp[5];
    *iDec = ADbyteResp[6];
    if (*iAcc == 0) *iAcc = (Setup->StandardAccDec->Checked) ? 1 : 4;	// default
    if (*iDec == 0) *iDec = (Setup->StandardAccDec->Checked) ? 1 : 2;	// default
    if (DEBUGGING) DB->Debug_Print("GetAccDecValues: Acc/Dec = " + String(*iAcc) + "/" + String(*iDec));

    retv = true;
	} else {
	if (Setup->StandardAccDec->Checked) {
	    *iAcc = 1; *iDec = 1;
    	} else {
	    *iAcc = 4; *iDec = 2;
        }
    DebugMessage("  Get Acc/Dec Rates Failed  ");
	retv = false;
    }
return retv;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMLashUpSoundCtrls(TMessage &Message)
{
//
// process Lashup Volume Control message for LashUps 101-120
//
char LashUpEngineList[LASHUP_ENGINE_LIST_LENGTH];			// Bytes containing the LashUp in the form 0x2C 0x01 0x82 0x03 0xFF
(void) strncpy((char *)LashUpEngineList, (char *) Message.WParam, sizeof LashUpEngineList); 	// binary engine list from Word parameter

const int LUNo = Message.LParam & 0x00FF;	// set Lashup Number from Long Parameter
if (LUNo == 0) return;				// no LashUp

assert(LUNo >= LASHUPMIN && LUNo <= LASHUPMAX);	// value 101 to 120 - DCS Engine Number always 102 (0x66)
int TIU = (Message.LParam & 0xFF00) >> 8;		// set TIU Number from Long Parameter
assert(TIU >= 1 && TIU <= TIUMAX); 	// TIU Number is 1 to 5
assert (ptrOp[TIU-1][LUNo] != NULL);	// Lashup must be running

String Name;
if (ptrLashUp[TIU-1]) { 			// LashUp name is put into LashUpData->Cells by LashUp_Selection.cpp
	Name = ptrLashUp[TIU-1]->LashUpData->Cells[LASHUP_NAME][LUNo - 100];            // Name of the LashUp
    }
if (ptrLUNameList[TIU-1][LUNo]) { 	// LashUp name is put here by Program Control routine LashUpStartUp
	Name = ptrLUNameList[TIU-1][LUNo];
    }

if (Name == "*") return;				// no LashUp name
if (LashUpEngineList[0] == '\0') return;	// no LashUpEngineList
assert(ptrOp[TIU-1][LUNo] != NULL);	// engine must be running

// Manually create form and show it
// Many Volume Control Forms
if (ptrVolume[TIU-1][LUNo] == NULL) {
	ptrVolume[TIU-1][LUNo] = new TVolume(this, RemoteNo, LUNo, TIU, Name, ptrLUEngineList[TIU-1][LUNo]);     // Pass "this" as the owner
	assert(ptrVolume[TIU-1][LUNo] != NULL);		// make sure that construction worked
	ptrVolume[TIU-1][LUNo]->ParentHandle = Handle; // for return messages - back to MainForm
	// TODO - Consider using the volume setting for the head engine as the volume for the consist
	} else {   // it's already created
	ptrVolume[TIU-1][LUNo]->Show();
	}
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsConsistRunning(int Consist, int TIU)
{
assert((Consist >= (Setup->AllowZero->Checked ? 0 : 1)) && (Consist <= LASHUPMAX));	// value 0 to 120 - Any engine or LashUp
assert((TIU >= 1) && (TIU <= TIUMAX)); 	// TIU Number is 1 to 5
if (ptrOp[TIU-1][Consist] != NULL) return true;	// if a pointer to the Op Window is not NULL, then the consist is running
return false;	// not running
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsLashUpRunning(int LUNo, int TIU)
{
assert(LUNo >= LASHUPMIN && LUNo <= LASHUPMAX);	// value 101 to 120 - Any LashUp
assert(TIU >= 1 && TIU <= TIUMAX); 	// TIU Number is 1 to 5
if (ptrLUEngineList[TIU-1][LUNo] != NULL) return true;	// if a pointer to the LashUpEngineList is not NULL, then the LashUp is running
return false;	// not running
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsTIUConnected()
{
// returns true if the TIU has responded to commands
return bTIUConnected;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsRadioCOMConnected()
{
// returns true if the Radio serial port has opened successfully
return bRadioCOMConnected;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsRFIDCOMConnected()
{
// returns true if the RFID serial port has opened successfully
return bRFIDCOMConnected;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsRadioWiFiConnected()
{
// returns true if the Radio serial port has opened successfully
return bRadioWiFiConnected;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::IsRFIDWiFiConnected()
{
// returns true if the RFID serial port has opened successfully
return bRFIDUDPWiFiConnected;
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::IsRadioConnected()
{
// returns true if the Radio (Serial or WiFi) has opened successfully
return bRadioCOMConnected  || bRadioWiFiConnected;
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::IsRFIDConnected()
{
// returns true if the RFID (Serial or WiFi) has opened successfully
return bRFIDCOMConnected || bRFIDUDPWiFiConnected;
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::IsEngineActive(int iEngineNo)
{
// returns true if the engine is active
if ((EngineData->Cells[ENG_NAME][iEngineNo] != "*") && (EngineData->Cells[ENG_STATE][iEngineNo] == ACTIVE)) return true;
return false;
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::DoesEngineExist(int iEngineNo)
{
// returns true if the engine exists
if (EngineData->Cells[ENG_NAME][iEngineNo] != "*") return true;
return false;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::InfoButtonClick(TObject *Sender)
{
// Splash Window Information Test
Splash = new TSplash(Application);
Splash->Memo1->Lines->Strings[0] = "Remote Train Control - " + String(Version);
Splash->Show();			// Show Splash Window
Splash->Update();		// Paint Splash Window
iSS = 0;					// Turn off future Splash Window
AboutRTC->Checked = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::LashUpShow(int LashUp, int TIU)
{
assert (LashUp >= LASHUPMIN && LashUp <= LASHUPMAX);
assert(TIU >= 1 && TIU <= TIUMAX); 	// TIU Number is 1 to 5
if (ptrOp[TIU-1][LashUp] == NULL) return;	//engine not running
ptrOp[TIU-1][LashUp]->Show();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEngineintoCommandModeClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "m4", TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEngineintoLegacyModeClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "m8", TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEngineintoConventionalModeClick(
      TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "m2", TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TMCCButtonClick(TObject *Sender)
{
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// This is test code only - final implementation should be using a selection form just like for LashUps
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
//
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
//
if (!TMCCOp) {
	TMCCOp = new TTMCCOp(this, RemoteNo, EngineNo, TIUNo, "TMCC Engine " + String(EngineNo), NULL);     // Pass "this" as the owner
	assert(TMCCOp != NULL);   		// make sure that construction worked
    TMCCOp->ParentHandle = Handle;  // for return messages

    (void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "T0D", TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
	// TODO - Send saved settings to TMCC Engine
   	}
TMCCOp->Show();	// since I used the C++ Builder's generated pointer, it will get automatically
				// deleted at close if I don't do it first (in CMTMCCShutdown())
// TODO TMCC operation is not tested and not complete
// Need to add engine selection screen
// Need to add ptrTMCCOp[] array and its support
//
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMTMCCShutdown(TMessage &Message)
{
// TODO TMCC operation is not fully coded
// process TMCC Engine Shutdown message - Engines 1-99
const int TIU = Message.WParam;		// set TIU Number from Word Parameter
const int Consist = Message.LParam;	// set Consist Number from Long Parameter
assert (TIU >= 1 && TIU <= TIUMAX);
assert (Consist >= 1 && Consist <= ENGINEMAX);	// 1-99 TMCC Engines
assert (TMCCOp != NULL);	// not running - caller must validate that the consist is running
(void)ptrTIUCommThread->Send_String_FIFO(Consist, "T50", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
(void)ptrTIUCommThread->Send_String_FIFO(Consist, "T0F", TIU, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
// Close the form
delete TMCCOp;
TMCCOp = NULL;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::IdleTimerTimer(TObject *Sender)
{
// Disconnect after 10 seconds of idle if we are connected
if ((Setup->DropDTRonIdle->Checked || Setup->DropRTSonIdle->Checked) && IsTIUConnected()) {
	IdleTimer->Tag++;
    if (IdleTimer->Tag >= 10) {
        if (Setup->DropDTRonIdle->Checked && (!TIUEMULATORMODE && !RADIOMODE)) DTRSwitch->Checked = false;
        if (Setup->DropRTSonIdle->Checked) RTSSwitch->Checked = false;
        ConnectedRadioLED->Flashing = true;
        IdleTimer->Tag = 0;		// restart Timer Counter
        }
    }
// Idle timer keeps on running
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMTIUChar(TMessage &Message)
{
//
// Display the encoded TIU character received in the TIUBuffer - CM_TIUCHAR
// NOTE: Used only for USB serial communications
//
const unsigned char Char1 = Message.WParam & 0x00FF;		// set received Char from Word Parameter
const int TIUbufCount = Message.LParam;			// set byte count from Long Parameter
// put the character in the buffer - no filtering is done
// convert Int to Hex  -----------------------------------------------
const String strHex = IntToHex(Char1 & 0xFF, 2);
// if the previous text was not red, output a CRLF and set text to Red
if (TIULog->Checked) {
    if (DEBUGGING) {
        if (TIUBuffer->SelAttributes->Color != clRed) {
            TIUBuffer->SelAttributes->Color = clRed;
            TIUBuffer->SelStart = TIUBuffer->GetTextLen();
            //TIUBuffer->SelText = TEXT("\r\n");
            //TIULineCount->Caption = String(TIULineCount->Caption.ToInt() + 1);
            // Method 4 works under CB10.1 but not BCB 4.0
            TIUBuffer->HideSelection = false;
            TIUBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
            TIUBuffer->HideSelection = true;
            }
        // put the character into the buffer;
        TIUBuffer->SelText = strHex;
        //
        if ((Char1 == 0x4D)) {  // CR LF on 0x4D EOT character
            TIUBuffer->SelStart = TIUBuffer->GetTextLen();
            TIUBuffer->SelText = TEXT("\r\n");
            TIULineCount->Caption = String(TIULineCount->Caption.ToInt() + 1);
            // Method 4 works under CB10.1 but not BCB 4.0
            TIUBuffer->HideSelection = false;
            TIUBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
            TIUBuffer->HideSelection = true;
            }
        }
    }
//
if (DEBUGGING) {
    MainForm->TIU_Char_Count->Text = TIUbufCount;  // show the new count
    MainForm->TIU_Total_Byte_Count->Value += 1;
    }
//
if (RFIDLog->Checked && Setup->Duplex->Checked && (Setup->RFID_COM_Mode->Checked || Setup->RFID_WiFi_Mode->Checked) ) {	// TODO is this needed?
	// also output to RFID Buffer
    // if the previous text was not red, output a CRLF and set text to Red
    if (DEBUGGING && RFIDBuffer->SelAttributes->Color != clRed) {
    	RFIDBuffer->SelAttributes->Color = clRed;
        RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();
        RFIDBuffer->SelText = TEXT("\r\n");
        // Method 4 works under CB10.1 but not BCB 4.0
        RFIDBuffer->HideSelection = false;
        RFIDBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
        RFIDBuffer->HideSelection = true;
	    }
    // Hex display character in RFID Buffer
	RFIDBuffer->SelText = strHex;
    if (DEBUGGING && (Char1 == 0x4D)) {  // CR LF on 0x4D EOT character
        RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();
    	RFIDBuffer->SelText = TEXT("\r\n");
        // Method 4 works under CB10.1 but not BCB 4.0
        RFIDBuffer->HideSelection = false;
        RFIDBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
        RFIDBuffer->HideSelection = true;
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMTIUASCIIChar(TMessage &Message)
{
//
// Display TIU ASCII character received in the TIUBuffer - CM_TIUASCIICHAR
// Handles one character at a time.
//
assert(false);  // this function is not used anymore
const char Char1 = Message.WParam & 0x00FF;		// set received Char from Word Parameter
const int TIUbufCount = Message.LParam;			// set byte count from Long Parameter
if (TIULog->Checked) {
    if (DEBUGGING) {
        // if the previous text was not red, output a CRLF
        // and set text to Red
        if (TIUBuffer->SelAttributes->Color != clRed) {
            TIUBuffer->SelStart = TIUBuffer->GetTextLen();
            TIUBuffer->SelText = TEXT("\r\n");
            TIULineCount->Caption = String(TIULineCount->Caption.ToInt() + 1);
            // Method 4 works under CB10.1 but not BCB 4.0
            TIUBuffer->HideSelection = false;
            TIUBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
            TIUBuffer->HideSelection = true;
            TIUBuffer->SelAttributes->Color = clRed;
            }
        // put the character into the buffer;
        TIUBuffer->SelStart = TIUBuffer->GetTextLen();
        TIUBuffer->SelText = Char1;
        if (Char1 == '\n') {
            // Method 4 works under CB10.1 but not BCB 4.0
            TIUBuffer->HideSelection = false;
            TIUBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
            TIUBuffer->HideSelection = true;
            }
        }
    }
if (DEBUGGING) {
    MainForm->TIU_Char_Count->Text = TIUbufCount;   // show the new count
    MainForm->TIU_Total_Byte_Count->Value += 1;
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMRFIDASCIIChar(TMessage &Message)
{
// process RFID ASCII Character Received message - CM_RFIDASCIICHAR
//
if (RFIDLog->Checked) {
    const char Char1 = Message.WParam & 0x00FF;		// set received Char from Word Parameter
    const int iRFIDbufCount = Message.LParam;		// set count from Long Parameter
    assert((Char1 >= '0' && Char1 <= 'F') || Char1 == '\n');
    // set text to Green
    RFIDBuffer->SelAttributes->Color = clGreen;
    // put the character into the buffer;
    RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();
    RFIDBuffer->SelText = Char1;
    //
    if (Char1 == '\n') {
        RFID_Char_Count->Text = iRFIDbufCount;  		// show the count
        // Method 4 works under CB10.1 but not BCB 4.0
        RFIDBuffer->HideSelection = false;
        RFIDBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
        RFIDBuffer->HideSelection = true;
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::LoadTIUData (TStringGrid * ED, int TIU)
{
// Load all TIU specific data
// Before this routine is called, previous TIU data must be saved
// Fill the Engine Data Grid from a disk file
// read Engine Number from Registry if it exists
assert(TIU >= 1 && TIU <= TIUMAX);
assert(ED != NULL);
//
if(bTIUDataLoaded[TIU - 1]) return;	// if already loaded then do nothing
//
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);		// make sure that construction worked
__try {
	// AIUNo and EngineNo loaded on a per TIU basis
	AIUSpin->Value = RegFile->ReadInteger("Options", "AIU" + String(TIU), AIUSpin->Value);
    AIUNo = AIUSpin->Value;
	int EngineTmp = RegFile->ReadInteger("Options", "Engine" + String(TIU), EngineSpin->Value);
    if (EngineTmp == ALLENGINE) EngineTmp = 1;	// don't allow startup in All Engine Operation
	EngineSpin->Value = EngineTmp;   		// Note: this calls EngineValueChanged(Sender)	// causes EngineNo to change
    if (EngineNo != EngineTmp) EngineSpinValueChanged(EngineSpin);	// On first use of EngineNo,
    															// we need to manually cause EngineNo to change
	RFIDudpRX->Port = RegFile->ReadString("Options", "WiFi" + String(TIU), RFIDudpRX->Port);
	RFIDtcpRX->Port = RegFile->ReadString("Options", "TCP" + String(TIU), RFIDtcpRX->Port);
    }
__finally {
	// Deallocate RegFile
	delete RegFile;
    }

ActiveEdit->Value = 0;
InactiveEdit->Value = 0;

assert (EngineNo != -1);		// Engine number must be set by now

const String EngineDataFile = String(cStartDir)  + "\\" + "EngineData" + String(TIU) + ".ini";

__try {
    if (ptrEngDat[TIU-1] == NULL) {
        ptrEngDat[TIU-1] = new TStringList();	// deleted in MainForm->CloseAllWindows()
        assert(ptrEngDat[TIU-1] != NULL);		// make sure that construction worked
        ptrEngDat[TIU-1]->Sorted = false;
        }
   	try {
    	if (FileExists(EngineDataFile)) {
   			// read it
   	    	ptrEngDat[TIU-1]->LoadFromFile(EngineDataFile);
   	    	// transfer data into TStringGrid from the TStringList
            ED->Rows[ENG_LABEL_ROW]->CommaText = ptrEngDat[TIU-1]->Strings[ENG_LABEL_ROW];
            assert(ED->Rows[ENG_LABEL_ROW]->Count == ENGDAT_COUNT);
   	    	for (int EngNum = ENG_TOPROW ; EngNum < ED->RowCount ; EngNum++) {
	    		ED->Rows[EngNum]->CommaText = ptrEngDat[TIU-1]->Strings[EngNum];
                assert(ED->Rows[EngNum]->Count == ENGDAT_COUNT);
                // Validate Data
                const int field = ValidateEngineData(ED, EngNum);
				if (field != -1) throw((EngNum * 100) + field);
				//
                if (EngNum != ALLENGINE) {
	                if ((ED->Cells[ENG_NAME][EngNum] != "*") && (ED->Cells[ENG_STATE][EngNum] == ACTIVE)) ActiveEdit->Value++;		// Adjust Counts
    	            if ((ED->Cells[ENG_NAME][EngNum] != "*") && (ED->Cells[ENG_STATE][EngNum] == INACTIVE)) InactiveEdit->Value++;
                    }
   	    		}
            if (DEBUGGING) DB->Debug_Print("Engine Data Loaded from " + EngineDataFile);
   	    	} else {
           	// Engine Data Files does not exist so create an empty TStringGrid
   			// Titles and Engine Numbers- needed when the grid is first created
   			ED->Rows[ENG_LABEL_ROW]->CommaText = ENGDAT_TITLE;
            assert(ED->Rows[ENG_LABEL_ROW]->Count == ENGDAT_COUNT);
   			for (int EngNum = ENG_TOPROW; EngNum < ED->RowCount ; EngNum++) {
       			// Insert Engine Number and switch setting, Engine Name is "*"
       			ED->Rows[EngNum]->CommaText = String(EngNum) + ENGDAT_INIT;
                assert(ED->Rows[EngNum]->Count == ENGDAT_COUNT);
   				}
   			ED->Cells [ENG_NAME][ENG_ALLROW] = "All Engines";
            if (DEBUGGING) DB->Debug_Print("Engine Data Created (Empty)");
           	}

       	}
        catch (int field) {
       		DebugMessage("Saved Engine Data File Field (" + String(field) + ") is Corrupt : " + EngineDataFile);
       		bEngineDataCorrupt[TIU-1] = true;	// loaded but corrupt
       		CaughtErrors->Value ++;
        	}
        catch ( ... ){
       		DebugMessage("Saved Engine Data File is Corrupt : " + EngineDataFile);
       		bEngineDataCorrupt[TIU-1] = true;	// loaded but corrupt
       		CaughtErrors->Value ++;
       		}
    }
__finally {
    //delete EngDat;
	bTIUDataLoaded[TIU - 1] = true;
    }

// set the Engine Name into the Control
NameofEngine->Text = ED->Cells[ENG_NAME][EngineNo];
FileNamePrefix->Text = NameofEngine->Text;
// highlight the value in the StringGrid
ED->Col = ENG_NAME;
ED->Row = EngineNo;

// Make sure the selected engine is visible in the middle row if possible
if (EngineNo < 3)
	ED->TopRow = 1;
    else
	ED->TopRow = EngineNo - 2;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SaveTIUData (TStringGrid * ED, int TIU)
{
// Save all TIU specific data
assert(TIU >= 1 && TIU <= TIUMAX);
assert(ED != NULL);
assert(bTIUDataLoaded[TIU - 1]);

// if the Engine Data is not loaded then do nothing.
if (!bTIUDataLoaded[TIU - 1]) return;

const String EngineDataFile = String(cStartDir)  + "\\" + "EngineData" + String(TIU) + ".ini";

if (bTIUDataLoaded[TIU - 1] && !bEngineDataCorrupt[TIU - 1] && ptrEngDat[TIU - 1]) { // don't save if it's not loaded or if we could not read it
	//TStringList * const EngDat = new TStringList();
	//assert(EngDat != NULL);		// make sure that construction worked
    ptrEngDat[TIU - 1]->Clear();
	__try {
    	try {
			// Save Engine Data to disk with overwrite
   			ptrEngDat[TIU - 1]->Sorted = false;
   			// transfer data from TSringGrid to TStringList
            assert(ED->Rows[ENG_LABEL_ROW]->Count == ENGDAT_COUNT);
        	ptrEngDat[TIU - 1]->Add(ED->Rows[ENG_LABEL_ROW]->CommaText);
   			for (int EngNum = ENG_TOPROW ; EngNum < ED->RowCount ; EngNum++) {
   	            // Validate Data
                const int field = ValidateEngineData(ED, EngNum);
				if (field != -1) throw((EngNum * 100) + field);
                assert(ED->Rows[EngNum]->Count == ENGDAT_COUNT);
				// Engine Data is valid so move it into the EngDat Object
	   			// transfer data from TSringGrid to TStringList
       				ptrEngDat[TIU - 1]->Add(ED->Rows[EngNum]->CommaText);
                }
   			// write it
   			ptrEngDat[TIU - 1]->SaveToFile(EngineDataFile);
			if (DEBUGGING) DB->Debug_Print("Engine Data Saved to " + EngineDataFile);
    		}
            catch (int field) {
            	DebugMessage("Edited Engine Data Field (" + String(field) + ") is Corrupt - Not Saved : " + EngineDataFile);
            	bEngineDataCorrupt[TIU - 1] = true;
            	CaughtErrors->Value ++;
            	}
            catch (...) {
            	DebugMessage("Edited Engine Data is Corrupt - Not Saved : " + EngineDataFile);
       			bEngineDataCorrupt[TIU - 1] = true;
       			CaughtErrors->Value ++;
            	}
        }
    __finally {
	   	//delete EngDat;
		bTIUDataLoaded[TIU - 1] = false;
    	}
// Debugging Only
#ifdef STRINGSDEBUG
// from:  http://www.bcbjournal.org/articles/vol4/0009/Working_with_string_grids.htm
  std::auto_ptr<TStrings> SaveStrings(new TStringList());
  const int col_count = ED->ColCount;
  for (int index = 0; index < col_count; ++index) {
    SaveStrings->AddStrings(ED->Cols[index]);
    }
  SaveStrings->SaveToFile("SaveEngineData");
#endif
    }
if (bSaveRegistryInfo) {
	// write selected Engine Number and selected AIU Number to the Registry
	TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
	assert(RegFile != NULL);		// make sure that construction worked
	__try {
		// AIU Count, Engine numbers and WiFi Port saved on a per TIU basis
		RegFile->WriteInteger("Options", "AIU" + String(TIU), AIUSpin->Value);
		RegFile->WriteInteger("Options", "Engine" + String(TIU), EngineSpin->Value);
		RegFile->WriteString("Options", "WiFi" + String(TIU), RFIDudpRX->Port);
		RegFile->WriteString("Options", "TCP" + String(TIU), RFIDtcpRX->Port);
	    }
	__finally {
		// Deallocate RegFile
		delete RegFile;
		}
    }
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::ChangeTIUData(TStringGrid * ED, int TIUTo)
{
//
// Change the EngineData specified by ED to the TIU given by TIUTo
//
assert(TIUNo >= 1 && TIUNo <= TIUMAX);
assert(TIUTo >= 1 && TIUTo <= TIUMAX);
if (TIUNo == TIUTo) return false;	// already set to this TIU
// otherwise, switch TIUs
SaveTIUData(ED, TIUNo);
TIUSpin->Value = TIUTo;
TIUNo = TIUTo;
LoadTIUData(ED, TIUTo);
// Reconnect to TIU if we were connected
if (IsTIUConnected() && Setup->Hex->Checked && !Setup->TXOnlyComm->Checked) {	// Hex mode & two way comm required
	if (GetAIUCount(TIUNo)) { // read up the number of AIU
		// if the GetAIUCount worked, read up the TIU version number - needed to form LashUp command correctly
		GetTIUVer(TIUNo);	// this also bumps the sequence number
		}
	// success with above two calls will set TIUSpin to clLime
	// and bTIUConnected to true, failure will set TIUSpin to clRed
	}
//
return true;	// return true if TIUNo was changed
}

//---------------------------------------------------------------------------

int __fastcall TMainForm::ValidateEngineData(TStringGrid * ED, int EngNum)
{
assert(ED != NULL);
assert(EngNum >= 1 && EngNum <= 100);	// 1-99 and 100 "All Engine Operation"
int val;

// Index
val = ED->Cells[ENG_NUM][EngNum].ToInt();
if (val != EngNum) return(ENG_NUM);
// Engine Name
val = ED->Cells[ENG_NAME][EngNum].Length();
if (val == 0) return(ENG_NAME);
if ((val == 1) && (ED->Cells[ENG_NAME][EngNum] != "*")) return(ENG_NAME);
//HDLT
//"Head Light" | ("Rule 17" << 1) | ("Bell on with Horn" << 2) | ("Auto Bell/Whistle" << 3)
val = ED->Cells[ENG_HDLT][EngNum].ToInt();
if (val < 0 || val > 31) return(ENG_HDLT);	// Headlight State (1 bit - mask: 0x01), Simulated Rule 17(1 bit - mask: 0x02) &
	// Bell on with Horn(1 bit - mask 0x04) & Auto Bell (1 bit - mask 0x08) & Auto Whistle (1 bit - mask 0x10)
//LIN
val = ED->Cells[ENG_PS][EngNum].ToInt();
if (val < UDEF0 || val > PS4) return(ENG_PS);
//Smoke
val = ED->Cells[ENG_SMOKE][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_SMOKE);
//FCC
val = ED->Cells[ENG_FCC][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_FCC);
//"Smoke Level (2 bits)" | ("Auto Labored Smoke" << 2)
val = ED->Cells[ENG_SMOKELEVEL][EngNum].ToInt();	// Smoke Level (2 bits - mask: 0x03) and Auto Labored Smoke (1 bit - mask: 0x04)
if (val < 0 || val > 7) return(ENG_SMOKELEVEL);
//Chat
val = ED->Cells[ENG_CHAT][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_CHAT);
//PC
val = ED->Cells[ENG_PROTOCHUFF][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_PROTOCHUFF);
//Rate
val = ED->Cells[ENG_CHUFFRATE][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_CHUFFRATE);
//Brake
val = ED->Cells[ENG_BRAKES][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_BRAKES);
//Ditch
val = ED->Cells[ENG_DITCH][EngNum].ToInt();
if (val < -1 || val > 3) return(ENG_DITCH);	// Ditch Light Combo Box (mask: 0x03)
//Auxiliary Lights 1
val = ED->Cells[ENG_AUX1][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_AUX1);
//Auxiliary Lights 2
val = ED->Cells[ENG_AUX2][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_AUX2);
//Beacon
val = ED->Cells[ENG_BEACON][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_BEACON);
//Auxiliary Lights 3
val = ED->Cells[ENG_AUX3][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_AUX3);
//Marker Light
val = ED->Cells[ENG_MARKER][EngNum].ToInt();
if (val != 0 && val != 1) return(ENG_MARKER);
//Type
val = ED->Cells[ENG_TYPE][EngNum].ToInt();
if (val != STEAM_ENGINE && val != DIESEL_ENGINE && val != ELECTRIC_ENGINE) return(ENG_TYPE);
//"Max Speed"
val = ED->Cells[ENG_MAXSPEED][EngNum].ToInt();
if (val < 0 || val > SPEEDMAX) return(ENG_MAXSPEED);
//Inactive
val = ED->Cells[ENG_STATE][EngNum].ToInt();
if (val != ACTIVE && val != INACTIVE) return(ENG_STATE);
//Master
val = ED->Cells[VOL_MASTER][EngNum].ToInt();
if (val < 0 || val > VOLMAX) return(VOL_MASTER);
//Engine
val = ED->Cells[VOL_ENGINE][EngNum].ToInt();
if (val < 0 || val > VOLMAX) return(VOL_ENGINE);
//Accent
val = ED->Cells[VOL_ACCENT][EngNum].ToInt();
if (val < 0 || val > VOLMAX) return(VOL_ACCENT);
//Horn
val = ED->Cells[VOL_HORN][EngNum].ToInt();
if (val < 0 || val > VOLMAX) return(VOL_HORN);
//Bell
val = ED->Cells[VOL_BELL][EngNum].ToInt();
if (val < 0 || val > VOLMAX) return(VOL_BELL);
//SP1 not checked
//SP2 not checked
//SP3 not checked
return -1;	// valid
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEngineActive1Click(TObject *Sender)
{
SetConsistState(EngineNo, TIUNo, ACTIVE);		// set active
// update users of the Engine List
PutEngData();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEngineInactive1Click(TObject *Sender)
{
SetConsistState(EngineNo, TIUNo, INACTIVE);		// set inactive
// update users of the Engine List
PutEngData();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Z0ButtonClick(TObject *Sender)
{
// using this button only for testing
Setup->StartmDNSServices(Setup->RFID_WiFi_Mode->Checked, Setup->Radio_WiFi_Mode->Checked);
/*
if (!TIUEMULATORMODE) return;	// TIU Emulator must be connected
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// ZO should cause a slow response with a received CRC error (NAK 4)
(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "Z0", TIUNo, RemoteNo, 11, MED_WAIT, NO_RECORD); 	// send 'Z0' command to the TIU
*/
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RelaySoundButtonClick(TObject *Sender)
{
PlaySound(MAKEINTRESOURCE(MOUSE_CLICK), HInstance, SND_RESOURCE | SND_ASYNC);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIUFeatureResetButtonClick(TObject *Sender)
{
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// TIU Feature Reset
String FeatureResetMsg = "Are you sure you want to Feature Reset TIU " + String(TIUNo) + "?";
int retv = Application->MessageBox(FeatureResetMsg.c_str(),	TEXT("Warning"), MB_YESNO | MB_ICONQUESTION);
if (retv == IDYES) {
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "F2", TIUNo, RemoteNo, 11, VERY_LONG_WAIT * 20, RECORD_OK);
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIUFactoryResetButtonClick(TObject *Sender)
{
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// TIU Factory Reset
String FactoryResetMsg = "Are you sure you want to Factory Reset TIU " + String(TIUNo) + "?";
int retv = Application->MessageBox(FactoryResetMsg.c_str(),	TEXT("Warning"), MB_YESNO | MB_ICONQUESTION);
if (retv == IDYES)
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "F3", TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
// https://docs.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
// Make sure KeyPreview is true for this form
assert(KeyPreview);
switch (Key) {
	case 'D' :
    	if (Shift.Contains(ssCtrl) && Shift.Contains(ssShift)) {
    		DebugButton->OnClick(DebugButton);
            Key = 0;	// handled
			}
				// Added as menu shortcut
                else if (Shift.Contains(ssCtrl)) {
				ShowDetectedTagsClick(Sender);
            	Key = 0;	// handled
            	}
		break;
	case 'U' :
    	if (Shift.Contains(ssCtrl) && Shift.Contains(ssShift)) {
        	UtilitiesButton->OnClick(UtilitiesButton);
            Key = 0;	// handled
            }
    	break;
    // Added as menu shortcut
    case 'L' :
    	if (Shift.Contains(ssCtrl)) {
        	ShowLayoutWindowClick(Sender);
            Key = 0;	// handled
            }
    	break;
    case VK_F1:
        if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][1] != "*") {
			// Set Engine 1 and call Startup
    	    EngineSpin->Value = 1;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][1] == ACTIVE) { StartupButton->OnClick(StartupButton); }
    	   	break;
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][1] != "*")) {
			// Set LashUp 1 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 1;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
		if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F2:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][2] != "*") {
			// Set Engine 2 and call Startup
    	    EngineSpin->Value = 2;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][2] == ACTIVE) { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][2] != "*")) {
			// Set LashUp 2 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 2;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F3:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][3] != "*") {
			// Set Engine 3 and call Startup
    	    EngineSpin->Value = 3;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][3] == ACTIVE) { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][3] != "*")) {
			// Set LashUp 3 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 3;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F4:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][4] != "*") {
			// Set Engine 4 and call Startup
    	    EngineSpin->Value = 4;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][4] == ACTIVE) { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][4] != "*")) {
			// Set LashUp 4 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 4;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F5:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][5] != "*") {
			// Set Engine 5 and call Startup
    	    EngineSpin->Value = 5;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][5] == ACTIVE) { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][5] != "*")) {
			// Set LashUp 5 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 5;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F6:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][6] != "*") {
			// Set Engine 6 and call Startup
    	    EngineSpin->Value = 6;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][6] == ACTIVE)  { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][6] != "*")) {
			// Set LashUp 6 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 6;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F7:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][7] != "*") {
			// Set Engine 7 and call Startup
    	    EngineSpin->Value = 7;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][7] == ACTIVE)  { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][7] != "*")) {
			// Set LashUp 7 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 7;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F8:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][8] != "*") {
			// Set Engine 8 and call Startup
    	    EngineSpin->Value = 8;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][8] == ACTIVE)  { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][8] != "*")) {
			// Set LashUp 8 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 8;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
  		if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F9:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][9] != "*") {
			// Set Engine 9 and call Startup
    	    EngineSpin->Value = 9;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][9] == ACTIVE)  { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][9] != "*")) {
			// Set LashUp 9 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 9;
        	ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
        if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F10:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) && EngineData->Cells[ENG_NAME][10] != "*") {
			// Set Engine 10 and call Startup
    	    EngineSpin->Value = 10;
            // if engine is active, call startup
            if (EngineData->Cells[ENG_STATE][10] == ACTIVE)  { StartupButton->OnClick(StartupButton); }
            }
		// if the Lashup Form is created and this is Ctrl-Fx and the LashUp is named then...
		if ((ptrLashUp[TIUNo-1] != NULL) && Shift.Contains(ssCtrl) && (ptrLashUp[TIUNo-1]->LashUpData->Cells[LASHUP_NAME][10] != "*")) {
			// Set LashUp 10 and call Startup
    	    ptrLashUp[TIUNo-1]->LUp->Value = 10;
			ptrLashUp[TIUNo-1]->LashUpStartUpButton->OnClick(ptrLashUp[TIUNo-1]->LashUpStartUpButton);
           	break;
            }
  		if (!Shift.Contains(ssAlt)) Key = 0;	// handled unless it's in conjuction with ALT key
    	break;
    case VK_F11:
        if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt) ) {
        	SetupButton->OnClick(SetupButton);
		    Key = 0;	// handled unless it's in conjuction with ALT or SHIFT key
            }
        if (Shift.Contains(ssCtrl)) {
        	StartUpVarControl(1);
		    Key = 0;	// handled
            }
    	break;
    case VK_F12:
    	if (!Shift.Contains(ssCtrl) && !Shift.Contains(ssAlt)) {
	    	QuickButton->OnClick(QuickButton);
			Key = 0;	// handled unless it's in conjuction with ALT or SHIFT key
        	}
        if (Shift.Contains(ssCtrl)) {
        	StartUpVarControl(2);
		    Key = 0;	// handled
            }
	   	break;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StopAllButtonClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
// Send s0 command to all active engines in the engine list
for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
    if (ptrOp[TIUNo-1][iEng] != NULL) {
	    // if the engine is running, use the speed scroller
	    // Speed 0
		ptrOp[TIUNo-1][iEng]->SpeedSlider->Value = 0;
        ptrOp[TIUNo-1][iEng]->SetSpeedSliderValue(0.0);				// causes the 's' command to be sent if the speed is not already zero
		ptrOp[TIUNo-1][iEng]->MotionSoundRequest->Checked = true;	// request sound on next train movement
       	} else {
        // Send s0 command directly to all active engines in the engine list (active and not running)
       	if (EngineData->Cells[ENG_NAME][iEng] != "*" && (EngineData->Cells[ENG_STATE][iEng] == ACTIVE)) {
            // NOTE: the TIU will NAK each speed command if the engine is not actually on the layout
			(void) ptrTIUCommThread->Send_String_FIFO(iEng, "s0", TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);	// Speed 0
           	}
        }
    }
Screen->Cursor = OldCursor;		// restore cursor
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::MuteAllClick(TObject *Sender)
{
// Toggle Mute Setting
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
MuteAll->Checked = !MuteAll->Checked;
if (MuteAll->Checked == true) {
	// mute all started engines not LashUps
    for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
    	// if the Volume Control window is up, use that
        if (ptrVolume[TIUNo-1][iEng] != NULL) {
            ptrVolume[TIUNo-1][iEng]->EngineSoundOnSwitch->Checked = false;	// Mute
        	continue;	// on to next engine
            }
        // if the engine is started, Mute it
        if (ptrOp[TIUNo-1][iEng] != NULL) {
			// v0 = Master Volume, v1 = Engine Volume, v2 = Accent Volume, v3 = Horn Volume, v4 = Bell Volume
			char str[15];
			snprintf(str, sizeof str, "v1%d", 0);
			(void) ptrTIUCommThread->Send_String_FIFO(iEng, str, TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
			//
			snprintf(str, sizeof str, "v2%d", 0);
			(void) ptrTIUCommThread->Send_String_FIFO(iEng, str, TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
			//
        	}
    	}
    } else {
    // unmute all muted engines not LashUps
    for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
    	// if the Volume Control window is up, use that
        if (ptrVolume[TIUNo-1][iEng] != NULL) {
            ptrVolume[TIUNo-1][iEng]->EngineSoundOnSwitch->Checked = true;	// unMute
        	continue;	// on to next engine
            }
        // if the engine is started, unMute it to volume value saved in the Engine Data file
        if (ptrOp[TIUNo-1][iEng] != NULL) {
			// v0 = Master Volume, v1 = Engine Volume, v2 = Accent Volume, v3 = Horn Volume, v4 = Bell Volume
			char str[15];
			snprintf(str, sizeof str, "v1%d", EngineData->Cells[VOL_ENGINE][iEng].ToInt());
			(void) ptrTIUCommThread->Send_String_FIFO(iEng, str, TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
			//
			snprintf(str, sizeof str, "v2%d", EngineData->Cells[VOL_ACCENT][iEng].ToInt());
			(void) ptrTIUCommThread->Send_String_FIFO(iEng, str, TIUNo, RemoteNo, 11, SHORT_WAIT, RECORD_OK);
			//
        	}
    	}
    }
}
//---------------------------------------------------------------------------
/**
 * RX Decoding : RLL(1,0) to Binary
 *
 * RLL(1,0) to 8 bit Morton encoded to 8 bit binary
 * returns a -1 on fail
 */
int RXLookUp(unsigned int EncodedWord) {
  // RLL(1,0) to 8 bit Morton
  unsigned char b1 = cvt8B4B(EncodedWord >> 12) << 6;
  unsigned char b2 = cvt8B4B(EncodedWord >> 8) << 4;
  unsigned char b3 = cvt8B4B(EncodedWord >> 4) << 2;
  unsigned char b4 = cvt8B4B(EncodedWord);
  int MortonWord = b1 | b2 | b3 | b4;
  if (MortonWord == -1) return -1; // fail
  // 8 bit Morton to 8 bit binary
  return RXMorton(MortonWord);
}
//---------------------------------------------------------------------------
// part of 8B to 4B RLL(1,0) conversion
// returns a -1 on fail
int cvt8B4B(unsigned char b) {
  b = b & 0x0F;
  if (b == 0x0A) return 0x03;
  if (b == 0x09) return 0x02;
  if (b == 0x06) return 0x01;
  if (b == 0x05) return 0x00;
  return(-1);
}
//---------------------------------------------------------------------------
/**
 * TX encoding binary to RLL(1,0)
 *
 * 8 bit binary to 8 bit Morton Encoded to RLL(1,0)
 * returns a -1 on fail
 */
int TXLookUp(unsigned char BinaryWord) {
  // 8 bit binary to 8 bit Morton
  int MortonWord = TXMorton(BinaryWord);
  // 8 bit Morton to RLL(1,0)
  unsigned int b1 = cvt4B8B(MortonWord >> 6) << 12;
  unsigned int b2 = cvt4B8B(MortonWord >> 4) << 8;
  unsigned int b3 = cvt4B8B(MortonWord >> 2) << 4;
  unsigned int b4 = cvt4B8B(MortonWord);
  int EncodedWord = b1 | b2 | b3 | b4;
  if (EncodedWord == -1) return -1; // fail
  return EncodedWord;
}
//---------------------------------------------------------------------------
// part of RLL(1,0) 4B to 8B conversion
// returns a -1 on fail
int cvt4B8B(unsigned char b) {
  b = b & 0x03;
  if (b == 0x03) return 0x0A;
  if (b == 0x02) return 0x09;
  if (b == 0x01) return 0x06;
  if (b == 0x00) return 0x05;
  return(-1);
}
//---------------------------------------------------------------------------
// FROM:  https://fgiesen.wordpress.com/2009/12/13/decoding-morton-codes/
//---------------------------------------------------------------------------
// "Insert" a 0 bit after each of the 4 low bits of x
unsigned char Part1By1(unsigned char x) {
  x &= 0x0f;                  // x = ---- 3210
  x = (x ^ (x <<  2)) & 0x33; // x = --32 --10
  x = (x ^ (x <<  1)) & 0x55; // x = -3-2 -1-0
  return x;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
/*
 * Convert a Morton byte to binary
 */
unsigned char RXMorton(unsigned char MortonWord) {
  return (Part1By1((MortonWord & 0xF0) >> 4) << 1) | Part1By1(MortonWord & 0x0F);
}
//---------------------------------------------------------------------------
// Inverse of Part1By1 - "delete" all odd-indexed bits
unsigned char Compact1By1(unsigned char x) {
  x &= 0x55;                  // x = -3-2 -1-0
  x = (x ^ (x >>  1)) & 0x33; // x = --32 --10
  x = (x ^ (x >>  2)) & 0x0f; // x = ---- 3210
  return x;
}
//---------------------------------------------------------------------------
/*
 * Convert an 8 bit binary word to Morton
 */
unsigned char TXMorton(unsigned char BinaryWord) {
  return (Compact1By1(BinaryWord >> 1 ) << 4) | Compact1By1(BinaryWord);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMTIUComm(TMessage &Message)
{
// Test Communications with TIU and set color of TIUSpin
const int TIU = Message.WParam;		// set TIU Number from Word Parameter
assert (TIU >= 1 && TIU <= TIUMAX);
//
if (IsTIUConnected()) {
	const TCursor OldCursor = Screen->Cursor; 	// save cursor
	Screen->Cursor = crHourGlass;
    if (!Setup->RFID_COM_Mode->Checked && !Setup->TXOnlyComm->Checked) {	// one port mode & two way comm required
	    if (GetAIUCount(TIU)) { // read up the number of AIU
   			// if the GetAIUCount worked, read up the TIU version number
			GetTIUVer(TIU);
	    	}
    	}
    Screen->Cursor = OldCursor;
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::BatteryStatusButtonClick(TObject *Sender)
{
// TODO - Process and display the response to the Battery Status command. It does not yet display the correct indication Hi/OK/Low
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

BatteryStatusButton->Caption = "Battery";

// Send Battery Status sequence (memory locations 0x82-0x83)
unsigned short BATHOWord;
if (DEBUGGING) DB->Debug_Print("Battery Status Query", clGreen);
bool bretv = FetchWord(0x82, &BATHOWord, EngineNo, TIUNo);
if (!bretv) return;
String BATStr = "OK";						// Decoding is not correct
if ((BATHOWord & 0x00FF) > 0xD0) BATStr = "High";
if ((BATHOWord & 0x00FF) < 0x90) BATStr = "Low";
BatteryStatusButton->Caption = BATStr;
if (DEBUGGING) DB->Debug_Print("Battery Status : " + BATStr, clGreen);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DTVButtonClick(TObject *Sender)
{
// It does not yet display the correct voltage
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

DTVButton->Caption = "DTV";

// Send DTV sequence (memory locations 0x4C-0x4D)
unsigned short lVal;
if (DEBUGGING) DB->Debug_Print("DTV Query", clGreen);
bool bretv = FetchWord(0x4C, &lVal, EngineNo, TIUNo);
if (!bretv) return;
const float RawValue = lVal;
float volts;
// Quadradic equation from cubic regression (volts = ax^2 + bx + c) of data points. See "CRC Analysis.wb3" page "DTV"
// RawValue = 8x^2 - 8x + 32		R^2 = 1.0 (Coefficient of Determination)
// Put equation into the form:  0 = 8x^2 - 8x + (32 - RawValue)
if (RawValue >= 32) {
	// Using the quadradic equation to solve:
	float a = 8;
	float b = -8;
	float c = 32 - RawValue;		// for RawValue >=32 (which is volts = 1.0)
	// then  b^2 - 4ac Only the positive root is valid
	float radicand = b*b - 4*a*c;
	volts = (-b + sqrt(radicand))/(2*a);
    } else {
    volts = 0.0;
    }

char DTVStr[64];
snprintf(DTVStr, sizeof DTVStr, "%4.1f V", volts);
DTVButton->Caption = DTVStr;
if (DEBUGGING) DB->Debug_Print("Track Voltage : " + String(DTVStr), clGreen);
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::DODButtonClick(TObject *Sender)
{
// read up the Odometer
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

DODButton->Caption = "DOD";

// Get scaling factor Memory locations 0x08-0x09
unsigned short sVal;
if (DEBUGGING) DB->Debug_Print("Scaling Factor Query", clGreen);
bool bretv = FetchWord(0x08, &sVal, EngineNo, TIUNo);
if (!bretv) return;
const float fSF = sVal;
// Get DOD Memory locations 0x34-0x37
if (DEBUGGING) DB->Debug_Print("DOD Query", clGreen);
unsigned long lVal;
bretv = FetchInt(0x34, &lVal, EngineNo, TIUNo);
if (!bretv) return;
if (fSF == 0.0) return;		// prevent divide by zero
//const double DOD = (DODFACTOR * lVal)/fSF;	// multiplier found by experimentation;
const double fDOD = DODEQUATION(lVal, fSF);	// multiplier found by experimentation;
char DODStr[64];
snprintf(DODStr, sizeof DODStr, "%8.2f M", fDOD);
DODButton->Caption = DODStr;
if (DEBUGGING) {
	DB->Debug_Print("Odometer : " + String(DODStr), clGreen);
	DB->Debug_Print(Format((String)"DODButton=%.6f raw=%d DODFACTOR=%.6f ScaleFactor=%f ", ARRAYOFCONST((fDOD, (int)lVal, /*DODFACTOR*/ 512.0/DTOFACTOR, fSF))), clBlue);
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DTOButtonClick(TObject *Sender)
{
// Read up the Trip Odometer
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

DTOButton->Caption = "DTO";

// Fetch scaling factor Memory locations 0x08-0x09
unsigned short sVal;
if (DEBUGGING) DB->Debug_Print("Scaling Factor Query", clGreen);
bool bretv = FetchWord(0x08, &sVal, EngineNo, TIUNo);
if (!bretv) return;
const float fSF = sVal;
// Get DTO Memory locations 0x3C-0x3F
if (DEBUGGING) DB->Debug_Print("DTO Query", clGreen);
unsigned long lVal;
bretv = FetchInt(0x3C, &lVal, EngineNo, TIUNo);
if (!bretv) return;
if (fSF == 0.0) return;		// prevent divide by zero
//const double DTO = lVal/(fSF * DTOFACTOR);
const double fDTO = DTOEQUATION(lVal, fSF);
char DTOStr[64];
snprintf(DTOStr, sizeof DTOStr, "%10.4f M", fDTO);
DTOButton->Caption = DTOStr;
const double Sfeet = fDTO * 5280.0;
snprintf(DTOStr, sizeof DTOStr, "%15.2f Sfeet", Sfeet);
DTOButton->Hint = DTOStr;
if (DEBUGGING) {
	DB->Debug_Print("Trip Odometer : " + DTOButton->Caption + " (" + DTOButton->Hint + ")", clGreen);
	DB->Debug_Print(Format((String)"DTOButton=%.6f raw=%d DTOFACTOR=%.6f ScaleFactor=%f ", ARRAYOFCONST((fDTO, (int)lVal, DTOFACTOR, fSF))), clBlue);
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FRMButtonClick(TObject *Sender)
{
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// Read up the Maintenance Clock
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

if (DEBUGGING) DB->Debug_Print("FRM Query", clGreen);
FRMButton->Caption = "FRM";

// Fetch FRM (memory locations 0x324-0x325)
unsigned short x;
bool bretv = FetchWord(0x324, &x, EngineNo, TIUNo);
if (!bretv) return;
x = x >> 4; 	// divide by 16
char FRMStr[64];
if (x == 0) {
    snprintf(FRMStr, sizeof FRMStr, "Now");
    } else {
    if (x <= 16) {
        x = ((x - 1) * 160) + 60;
        } else {
        snprintf(FRMStr, sizeof FRMStr, "Invalid");
        }
    snprintf(FRMStr, sizeof FRMStr, "%d M", x);
    }
FRMButton->Caption = FRMStr;
if (DEBUGGING) DB->Debug_Print("Maintenance Required : " + String(FRMStr), clGreen);
// Send DOD sequences is done by the Remote. I don't know why
//DODButtonClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DCHButtonClick(TObject *Sender)
{
    // ---------------------------------------------------------------------------------------
    // I believe that this calculation is correct. RTC result shows hours to be the same value
    // as the Remote but minutes and seconds are different. I believe there is an error
    // in the Remote's code.
    // ---------------------------------------------------------------------------------------
    // if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// Read up the Chronometer
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

if (DEBUGGING) DB->Debug_Print("DCH Query", clGreen);

DCHButton->Caption = "DCH";
// Fetch DCH (memory locations 0x38-0x3B)
unsigned long lVal;
bool bretv = FetchInt(0x38, &lVal, EngineNo, TIUNo);
if (!bretv) return;
double t = lVal;
//t /= DCHFACTOR;	// estimated ticks per second
double fDCH = DCHEQUATION(t);	// estimated ticks per second
const int tt = static_cast<int> (fDCH);
const unsigned int hours = (tt/(60 * 60));
const unsigned int minutes = (tt/60) % 60;
const unsigned int seconds = tt % 60;
// ARRAYOFCONST needs to have the unsigned int cast to an int
DCHButton->Caption = Format((String)"%5dh:%2.2dm:%2.2ds", ARRAYOFCONST(((int)hours, (int)minutes, (int)seconds)));
if (DEBUGGING) {
	DB->Debug_Print("Chronometer : " + DCHButton->Caption, clGreen);
	DB->Debug_Print(Format((String)"DCHButton=%.6f raw=%d DCH FACTOR=%.6f", ARRAYOFCONST((fDCH, (int)lVal, DCHFACTOR))), clBlue);
    }
}
//---------------------------------------------------------------------------

bool __fastcall TMainForm::Query(char * CommandString, byte * QBuf, int Engine, int TIU, int Remote)
{
// Send a Query "q" or "qXXXX" command to the TIU and wait for a response, then display it on the Debug window
// return true on success or false on failure
bool retv = ptrTIUCommThread->Send_String_FIFO(Engine, CommandString, TIU, Remote, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, QBuf);
if (retv && (QBuf != NULL)) {
	// dump from QBuf
	if ((DEBUGGINGX(1) && (strlen(CommandString) == 1)) || DEBUGGINGX(5)) byteBinaryDump(&QBuf[5], "Response to " + String(CommandString) + ":");
    }
return retv;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::byteBinaryDump(byte * Response, String Title)
{
if (DEBUGGINGX(1)) {
	DB->Debug_Print("Packet " + Title +
    	byteBinaryStringToHexString(Response, 4), clBlue);
    if (DEBUGGINGX(9)) {
		DB->Debug_Print("32 bit " + Title +
    		String(Response[0] << 24 | Response[1] << 16 | Response[2] << 8 | Response[3]), clBlue);
		DB->Debug_Print("16 bit " + Title +
			String(Response[0] << 8 | Response[1]) + " " +
			String(Response[2] << 8 | Response[3]), clBlue);
		DB->Debug_Print(" 8 bit " + Title +
			String(Response[0]) + " " +
			String(Response[1]) + " " +
			String(Response[2]) + " " +
			String(Response[3]), clBlue);
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::sBinaryDump(String Response, String Title)
{
if (DEBUGGINGX(1)) {
	DB->Debug_Print("Packet " + Title +
    	sEncodedStringToHexString(Response), clRed);
	DB->Debug_Print("32 bit " + Title +
    	String(EncodedStringToDecimalInt(Response, 21, 16)), clRed);
	DB->Debug_Print("16 bit " + Title +
		String(EncodedStringToDecimalInt(Response, 21, 8)) + " " +
		String(EncodedStringToDecimalInt(Response, 29, 8)), clRed);
	DB->Debug_Print(" 8 bit " + Title +
		String(EncodedStringToDecimalInt(Response, 21, 4)) + " " +
		String(EncodedStringToDecimalInt(Response, 25, 4)) + " " +
		String(EncodedStringToDecimalInt(Response, 29, 4)) + " " +
		String(EncodedStringToDecimalInt(Response, 33, 4)), clRed);
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EditTIUNumber1Click(TObject *Sender)
{
// TIU popup menu item
assert(TIUNo >= 1 && TIUNo <= TIUMAX);	// TIU 1-5
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

NumberQuery->Caption = "Edit TIU #" + String(TIUNo) + " Number";
NumberQuery->Question->Caption = "Select the New TIU Number";
NumberQuery->SpinLabel->Caption = "New TIU Number:";
NumberQuery->NumberSpin->MaxValue = TIUMAX;
NumberQuery->NumberSpin->MinValue = 1;
NumberQuery->NumberSpin->Value = TIUNo;	// current TIU number
NumberQuery->NumberSpin->Visible = true;
int retv = NumberQuery->ShowModal();
if (retv == mrCancel) return;
// make sure the new TIU number slot is empty
int NewTIUNumber = NumberQuery->NumberSpin->Value;
if (TIUNo == NewTIUNumber) return;		// no change
//
if (DEBUGGING) DB->Debug_Print("Changing TIU #" + String(TIUNo) + " to #" + String(NewTIUNumber));

// send Bx command where x is the new TIU Number
unsigned char Command_String[12];
(void)snprintf((char *)Command_String, sizeof Command_String, "B%d", NewTIUNumber - 1);		// build command string with TIU #0-4
// NOTE: TIUNo in the command is the TIU to change to NewTIUNumber

// Wait for first 11 characters of response (takes about 250 ms)
if (!ptrTIUCommThread->Send_String_FIFO(EngineNo, Command_String, TIUNo, RemoteNo, 11, LONG_WAIT, NO_RECORD)) {
	DebugMessage("Failed to change TIU number (TIU " + String(TIUNo) + " not found)");
    return;	// don't make any changes
    }

// change the TIU number we are using - this calls TIUSpinValueChanged
TIUSpin->Value = NewTIUNumber;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ResetSetupButtonClick(TObject *Sender)
{
// Registry Reset
String RegistryResetMsg = "Are you sure you want to Reset all of your Preferences?";
int retv = Application->MessageBox(RegistryResetMsg.c_str(), TEXT("Warning"), MB_YESNO | MB_ICONQUESTION);
if (retv != IDYES) return;
//
// clear preferences and Exit
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);		// make sure that construction worked
__try {
   	RegFile->EraseSection("Copyright");
   	RegFile->EraseSection("Debug");
   	RegFile->EraseSection("EngineInfo");
   	RegFile->EraseSection("EngineStatus");
   	RegFile->EraseSection("Layout");
	RegFile->EraseSection("Options");
   	RegFile->EraseSection("PC1");
   	RegFile->EraseSection("PC10");
   	RegFile->EraseSection("PC2");
   	RegFile->EraseSection("PC3");
   	RegFile->EraseSection("PC4");
   	RegFile->EraseSection("PC5");
   	RegFile->EraseSection("PC6");
   	RegFile->EraseSection("PC7");
   	RegFile->EraseSection("PC8");
   	RegFile->EraseSection("PC9");
	RegFile->EraseSection("Quick");
	RegFile->EraseSection("SIgnal");
	RegFile->EraseSection("Speed");
	RegFile->EraseSection("Tag");
	RegFile->EraseSection("TIU");
   	RegFile->EraseSection("TrackSignal");
    }
__finally {
	delete RegFile;
	}
bSaveRegistryInfo = false;	// remind all forms not to save setting in Registry
Application->Terminate();	// FormClose does not run but Hide() in Setup does run
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RTCHelpMenuClick(TObject *Sender)
{
// Help window - I use the IDE defined pointer so this Form will be
// automatically de-allocated when the program is terminated.
if (RTCHelp == NULL) {
	RTCHelp = new TRTCHelp(this);
	assert(RTCHelp != NULL);		// make sure that construction worked
    }
RTCHelp->Show();		// show Help window
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetAllEnginesActiveClick(TObject *Sender)
{
// Set All Engines Active 1-99 (DCS 2-100)
// (total of 99 engines)
//
for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
   	if ((EngineData->Cells[ENG_NAME][iEng] != "*") && (EngineData->Cells[ENG_STATE][iEng] == INACTIVE)) {
        EngineData->Cells[ENG_STATE][iEng] = ACTIVE;
		}
    }
ActiveEdit->Value = ActiveEdit->Value + InactiveEdit->Value;
InactiveEdit->Value = 0;
EngineData->Repaint();	// redraw
// update users of the Engine List
PutEngData();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetAllEnginesInactiveClick(TObject *Sender)
{
// Set All Engines Active 1-99 (DCS 2-100)
// (total of 99 engines)
//
for (int iEng = 1; iEng <= ENGINEMAX ; iEng++) {
   	if ((EngineData->Cells[ENG_NAME][iEng] != "*") && (EngineData->Cells[ENG_STATE][iEng] == ACTIVE)) {
        EngineData->Cells[ENG_STATE][iEng] = INACTIVE;
		}
    }
InactiveEdit->Value = InactiveEdit->Value + ActiveEdit->Value;
ActiveEdit->Value = 0;
EngineData->Repaint();	// redraw
// update users of the Engine List
PutEngData();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RSButtonClick(TObject *Sender)
{
assert(Setup != NULL);		// setup must be created
// Route/Scene Form should be manually created as it is rarely needed
// if the form is not created, then create it, otherwise show the previously created form
//
const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
//
if (RouteScene == NULL) {	// since I used the C++ Builder pointer, it will get automatically deleted at close
	RouteScene = new TRouteScene(this, RemoteNo, TIUNo, AIUNo, EngineNo);     // Pass "this" as the owner
	assert(RouteScene != NULL);			// make sure that construction worked
    RouteScene->ParentHandle = Handle;	// for return messages
	} else {   							// it's already created
	RouteScene->Show();
	}
Screen->Cursor = OldCursor;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetupActivateRoutesandScenes1Click(TObject *Sender)
{
RSButton->OnClick(RSButton);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineDataDblClick(TObject *Sender)
{
// Startup the Engine or if field is empty, Ejdit Engine Name
if (!IsTIUConnected()){
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// double clicking in column header or row header is ignored
if (EngineData->Col == ENG_NUM || EngineData->Row == ENG_LABEL_ROW) return;

// Engine Name column double clicked
if (EngineData->Col == ENG_NAME) {
	if (EngineData->Cells[ENG_NAME][EngineData->Row] == "*") {
	    EditEngineName->OnClick(EditEngineName);  // edit engine name
        return;
    	}
    }
// otherwise, startup the Engine
StartupButton->OnClick(StartupButton);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DebugMessage(String MessageLine)
{
// show the message with OK button and print it to the debug window
ShowMessage(MessageLine);
if (DEBUGGING) DB->Debug_Print(MessageLine);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DebugMessagePopup(String Caption, String MessageLine, String StatusLine, int dly)
{
// Show a message popup of message "Caption" with additional lines "MessageLine" and "StatusLine"
// dly defaults to 2 seconds (2000 milliseconds) if not specified.
//
if (dly < 1) dly = 1;
if (dly > 10000) dly = 10000; 	// max of 10 seconds
// don't show popup if running a Playback
if (bPlayingBack) return;
Application->ProcessMessages();	// empty the message queue
// Keep it on the screen for dly milliseconds (10 seconds max, 2 second default)
// no buttons to push
// Start Thread
// FreeOnTerminate is set in constructor so we don't need to worry about freeing memory
new TMessageDisplayThread(false, this, Handle, Caption, MessageLine, StatusLine, dly);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AlreadyStartedClick(TObject *Sender)
{
// https://stackoverflow.com/questions/31649315/c-builder-button-click-function-reentry
LOCKEM()
// This can be called even if the Engine is already running. It just restores windows
// Allow Engines Only
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= 100);	// 0 factory reset, 1-99, 100 "all engine operation"

// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
EngineStartUp(EngineNo, TIUNo, true, 0);	// Engine is already started so don't send actual startup commands
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RTCWebPage1Click(TObject *Sender)
{
//ShellExecute(NULL, "open", "http://www.silogic.com/trains/RTC_Running.html", NULL, NULL, SW_SHOWNORMAL);
// OR
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/RTC_Running.html"), NULL, NULL, SW_SHOWNORMAL);
// OR
//ShellExecuteA(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/RTC_Running.html"), NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

void __fastcall byteEncodeAString(Cmd * CSource, Cmd * CDest)
{
// Encode a string. Uses a CMD structure which includes the string ((char)Str) and a byte count (idx)
//byte Encoded_Command[CMDMAX];
// encode a command from CSource into CDest - then the command can be sent - NOTE : CSource may contain 0x00
assert(CSource->idx < ((CMDMAX - 1 ) / 2));
// Sync Byte
//Encoded_Command[0] = 0x55;	// sync
//Encoded_Command[1] = 0x55;
CDest->Str[0] = 0x55;	// sync
CDest->Str[1] = 0x55;
// Encode the command
int i;
for (i = 0 ; i < CSource->idx ; i++) {
	int Encoded = TXLookUp(CSource->Str[i]);
	//Encoded_Command[2 + (2 * i)] = (Encoded & 0xFF00) >> 8;
    //Encoded_Command[3 + (2 * i)] = (Encoded & 0x00FF);
	CDest->Str[2 + (2 * i)] = (Encoded & 0xFF00) >> 8;
    CDest->Str[3 + (2 * i)] = (Encoded & 0x00FF);
	}
// EOT
//Encoded_Command[2 + (2 * i)] = 0x4D;	// EOT
//Encoded_Command[3 + (2 * i)] = 0x00;	// NULL but not needed as CSource contains a length
CDest->Str[2 + (2 * i)] = 0x4D;	// EOT
CDest->Str[3 + (2 * i)] = 0x00;	// NULL but not needed as CSource contains a length
CDest->idx = (CSource->idx * 2) + 3;
//memcpy(CDest->Str, Encoded_Command, CDest->idx);	// copy the encoded string into the CMD structure
return;
}
//---------------------------------------------------------------------------

String __fastcall byteHextoString(byte * str)
{
String CM = "";
// str will not contain 0x00 character
int iLen = strlen(str);
//int iLen = strcspn(str, "M") + 1;	// locate the 0x4D
// convert byte string to a character String type for display in Buffer
for (int i = 0 ; i < iLen ; i++) {
    CM += IntToHex(str[i] & 0xFF, 2);
	}
return CM;
}
//---------------------------------------------------------------------------

String __fastcall TMainForm::byteExtractEngineName(byte * Resp)
{
// pull out the Engine Name from Response - always 16 characters
String EngineNamefromTIU = byteBinaryStringToASCIIString(&Resp[6], 16);
if (DEBUGGING) DB->Debug_Print("Engine Name: " + EngineNamefromTIU);
return EngineNamefromTIU;	// starting at character 6 for 16 characters
// Note: character 5 in the string contains the engine number
}
//---------------------------------------------------------------------------

String __fastcall byteBinaryStringToASCIIString(byte * Cmd, int Count)
{
// Convert the binary string into a String
String DecodedString = "";
//  Convert
    for (int idx = 0 ; idx < Count; idx++) {
		int iVal = Cmd[idx];
    	// print string byte in ASCII
    	String A_Char;
        if (iVal >= 32 && iVal < 128) {   // include SPACE and all characters up to 128
	        A_Char = (char)iVal;  // Convert to ASCII Character
			DecodedString += A_Char;
            } else {
            DecodedString += " ";
            }
        }
return DecodedString;
}
//---------------------------------------------------------------------------

String show_binary(unsigned int u, int num_of_bits)
{
  String a = "";

  unsigned int t = pow(2.00, num_of_bits - 1);   // t is the max number that can be represented

  for(t; t>0; t = t/2)           // t iterates through powers of 2
      if(u >= t){                // check if u can be represented by current value of t
          u -= t;
          a += "1";               // if so, add a 1
      }
      else {
          a += "0";               // if not, add a 0
      }

  return a ;                     // returns string
}
//---------------------------------------------------------------------------

String __fastcall byteBinaryStringToString(byte * Cmd, int Count)
{
// Convert the binary string into a String
String ConvertedString = "";
//  Convert
	for (int idx = 0 ; idx < Count; idx++) {
        // convert byte to binary string
		String sVal = show_binary(Cmd[idx], 8);	// convert byte to binary string
    	ConvertedString += sVal;
        }
if (DEBUGGINGX(8)) DB->Debug_Print("Binary String: " + ConvertedString);
return ConvertedString;
}
//---------------------------------------------------------------------------

String __fastcall byteBinaryStringToHexString(byte * Cmd, int Len)
{
// decode the encoded string into Hex
// string can contain 0x00 as a length is passed to this routine
// This routine must not call any VCL objects
String ConvertedString = "";
//  Decode
    for (int idx = 0 ; idx < Len; idx++) {
        String hVal = IntToHex(Cmd[idx] & 0xFF, 2);
    	ConvertedString += hVal;
        }
return ConvertedString;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::HotButtonClick(TObject *Sender)
{
DoHotButton(HotButtonPopup->Items->Items[LastHotComPage - 1], LastHotComPage);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DoHotButton(TObject * Sender, int Page)
{
// Create Hot Button page 1 to HOTPAGEMAX
//
assert(Page >= 1 && Page <= HOTPAGEMAX);
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Hot Command Form should be manually created as it is rarely needed
if (!ptrHot_Command[Page-1]) {
    ptrHot_Command[Page-1] = new THot_Command(this, RemoteNo, EngineNo, TIUNo, AIUNo, Page);   // create the Form & Pass "this" as the owner
    assert(ptrHot_Command[Page-1] != NULL);			// make sure that construction worked
    ptrHot_Command[Page-1]->ParentHandle = Handle; 	// for return messages
   	}
ptrHot_Command[Page-1]->Show();

if (Sender) {
	reinterpret_cast<TMenuItem *>(Sender)->Checked = true;	// add check mark to Hot Buttons Menu
	}

LastHotComPage = Page;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::HotButtons1Click(TObject *Sender)
{
DoHotButton(HotButtons1, 1);		// Page 1 (Default)
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::HotButtons2Click(TObject *Sender)
{
DoHotButton(HotButtons2, 2);		// Page 2
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::HotButtons3Click(TObject *Sender)
{
DoHotButton(HotButtons3, 3);		// Page 3
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::HotButtons4Click(TObject *Sender)
{
DoHotButton(HotButtons4, 4);		// Page 4
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::HotButtons5Click(TObject *Sender)
{
DoHotButton(HotButtons5, 5);		// Page 5
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMHotCom(TMessage &Message)
{
// Open a page of Hot Buttons
int Page = Message.WParam;
assert(Page >= 1 && Page <= HOTPAGEMAX);
DoHotButton(HotButtonPopup->Items->Items[Page - 1], Page);
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::CMAIU(TMessage &Message)
{
// Open a page of TIU/AIU Controls
int NextAIU = Message.WParam;
int NextTIU = Message.LParam;
assert(NextTIU >= 1 && NextTIU <= TIUMAX);
StartAIUControl(NULL, NextTIU, NextAIU);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ReadEEPROMButtonClick(TObject *Sender)
{
// note: this command works from the DCS Loader running on a PC. Does not work here.
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// try to read file name from EEPROM at 0x1E98  for 32 bytes
(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "R001E98000020", TIUNo, RemoteNo, 11 + 32, MED_WAIT, NO_RECORD); 	// send 'R' command to the TIU
//
// try to read Engine name from EEPROM at 0x1E22 for 16 bytes
(void) ptrTIUCommThread->Send_String_FIFO(EngineNo, "R001E22000010", TIUNo, RemoteNo, 11 + 16, MED_WAIT, NO_RECORD); 	// send 'R' command to the TIU
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CopyrightClick(TObject *Sender)
{
// Copy the Form to the clipboard.
// In BCB4.0, there is no SetSize method associated with TBitmap
// so we have to use Width and Height separately.
//
// see:
// https://stackoverflow.com/questions/3190777/how-do-i-copy-a-form-as-an-image-to-the-clipboard
//
Graphics::TBitmap * bm;
bm = new Graphics::TBitmap;
try
	{
    #if __BORLANDC__ >= 0x720
	// Defined for CB10.1 C++ programs that use the VCL.
	bm->SetSize(ClientWidth, ClientHeight);
    #else
    bm->Width = ClientWidth;
    bm->Height = ClientHeight;
    #endif
	BitBlt(bm->Canvas->Handle, 0, 0, ClientWidth, ClientHeight, Canvas->Handle, 0, 0, SRCCOPY);
    Clipboard()->Assign(bm);
    }
__finally
	{
	delete bm;
    }
}
//---------------------------------------------------------------------------
// Delay
//
// Waits the specified amount of time while processing messages
// Uses MsgWaitForMultipleObjects()
//
// Parameters:
// dwDelay - number of milliseconds to wait  0 < dwDelay <= 100000
//
// Returns:
// no return value.
//
//
void __fastcall TMainForm::Delay(DWORD dwDelay)
{
assert (dwDelay >= 0 && dwDelay <= 100000);
//
if (DEBUGGING && (dwDelay > 99)) {
	DB->Debug_Print("Delay " + String(dwDelay));
    }
// Cannot use Sleep in the main Thread because the UI relies on messages
DWORD dwStart, dwNow, dwWait;
//
dwStart = GetTickCount();
for (;;) {
	// Wait for a any message to show up
	dwWait = MsgWaitForMultipleObjects(0, NULL, FALSE, dwDelay,
		QS_KEY | QS_PAINT | QS_POSTMESSAGE | QS_SENDMESSAGE | QS_TIMER | QS_MOUSEMOVE | QS_HOTKEY);
    if (bEmergencyStop) break;	// get out in a hurry
	//
	if (dwWait == WAIT_TIMEOUT) {	// Timed out?
    	break;
        }
	MyProcessMessages();		// process all messages except mouse button messages
	dwNow = GetTickCount(); 	// adjust the delay time remaining
	dwStart = dwNow - dwStart;
	if (dwStart >= dwDelay) break;
	dwDelay -= dwStart;
	dwStart = dwNow;
    if (bEmergencyStop) break;	// get out in a hurry
    }
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::MyProcessMessages(void)
{
// replacement for the ProcessMessages() routine
// this routine processs all messages except mouse button messages
//
// https://stackoverflow.com/questions/41539678/are-event-handlers-re-entrant-in-embarcadero-cbuilder?rq=1
/*
typedef struct MSG {     // msg
    HWND   hwnd;
    UINT   message;
    WPARAM wParam;
    LPARAM lParam;
    DWORD  time;
    POINT  pt;
};
//
// Pump the message queue for all messages except mouse button messages
// from 513 to 521 (0x0201 to 0x0209)

0000 		000 		WM_NULL
...
0200 		512 		WM_MOUSEFIRST
0200 		512 		WM_MOUSEMOVE
0201 		513 		WM_LBUTTONDOWN
0202 		514 		WM_LBUTTONUP
0203 		515 		WM_LBUTTONDBLCLK
0204 		516 		WM_RBUTTONDOWN
0205 		517 		WM_RBUTTONUP
0206 		518 		WM_RBUTTONDBLCLK
0207 		519 		WM_MBUTTONDOWN
0208 		520 		WM_MBUTTONUP
0209 		521 		WM_MBUTTONDBLCLK
0209 		521 		WM_MOUSELAST
020a 		522 		WM_MOUSEWHEEL
*/
bool MsgAvailable;
MSG msg;
while (true) {
	MsgAvailable = PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE);	// message available ? but leave in queue
   	if (!MsgAvailable) break;	// no messages available
   	if (msg.message <= WM_MOUSEMOVE) {	// Is it a non mouse button message?
       	// Message from WM_NULL to and including WM_MOUSEMOVE
       	GetMessage(&msg, NULL, WM_NULL, WM_MOUSEMOVE);	// and remove from the queue
		TranslateMessage(&msg);
		DispatchMessage(&msg);
        continue;	// check for another message
        }

  	if (msg.message >= (WM_MOUSELAST+1)) {	// Is it a non mouse button message?
    	// Display the message Base and Offset if Debug is on
       	// Message from WM_MOUSELAST+1 to the last message possible, skipping a few that occur but we don't care to see
        if (DEBUGGINGX(5)) {
	        int MsgBase, MsgOffset;
	        // first decode the message base and offset
		    if (msg.message >= 0x8000) {	// WM_APP = 0x8000
            	// Application Message
    		    MsgBase = WM_APP;
        		MsgOffset = msg.message - WM_APP;
		    	} else {
    		    // Windows Message
				MsgBase = 0x0000;
	        	MsgOffset = msg.message;
	    	    }
    	    //
	    	if ((msg.message != 15) &&
            	(msg.message != (CM_TIUCHAR)) &&	    // CM_TIUCHAR
            	(msg.message != (CM_RFIDASCIICHAR)) &&  // CM_RFIDASCIICHAR
            	(msg.message != (CM_UI)) &&	            // CM_UI
            	(msg.message != (CM_TIUASCIICHAR))		// CM_TIUASCIICHAR
                	) {
                DB->Debug_Print("MyProcessMessages: 0x" + IntToHex(MsgBase, 4) + ":0x" + IntToHex(MsgOffset, 4));
                }
            }
       	GetMessage(&msg, NULL, WM_MOUSELAST+1, 0xFFFFFFFF);	// and remove from the queue
		TranslateMessage(&msg);
		DispatchMessage(&msg);
        continue;	// check for another message
        }
    // if all that's left is mouse button messages, get out
    if (msg.message > WM_MOUSEMOVE && msg.message <= WM_MOUSELAST) break;
    }
return;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::charstringtobinary(char * Command, char * Packet, int Len)
{
//
// Input is Command, a (char *) to an array of hexidecimal digits in ASCII. Null Terminated. Must
//	always be an even number of digits and be equal to (Len * 2).
//	eg: "55555A566A99656966669996696A5A55695A4D"
// Input is Len, an integer count of the number of bytes in the Packet array.
//	eg:	Len = strlen(Command)/2;
// Output is Packet, a (char *) to an array of the binary representation of Command, converting/
//	two hexidecimal digits into one 8 bit byte. It will contain Len bytes. It may contain
//	a 0x00 byte.
//	eg: "UUZVjeiffijZuiZM" where the 'M' is the 0x4D End of Packet
//
assert(Command);
assert(Packet);
assert(Len < 254);

static const char* const lut = "0123456789ABCDEF";
// Convert Hex character string to Hex
for (int i = 0; i < (Len * 2); i += 2) {
    char a = Command[i];
   	const char* p = std::lower_bound(lut, lut + 16, a);
    if (*p != a) {
   		if (DEBUGGING) DB->Debug_Print("Test: Error: " + ConnectedRadioLED->Caption + " Not a Hex Digit : " + String((char *)Command), clRed);
        return;
        }
    char b = Command[i + 1];
    const char* q = std::lower_bound(lut, lut + 16, b);
    if (*q != b) {
    	if (DEBUGGING) DB->Debug_Print("Test: Error: " + ConnectedRadioLED->Caption + " Not a Hex Digit : " + String((char *)Command), clRed);
        return;
        }
	Packet[i/2] = (((p - lut) << 4) | (q - lut));
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineUpDownClick(TObject *Sender, TUDBtnType Button)
{
// A bug in BCB4.0 - when this routine is called, the value of Position has
// already changed. Correct operation in CB10.1 "use the OnClick event handler
//to respond to changes just before the value of Position is changed."
if (Button == Comctrls::btPrev) {
    if (EngineData->Row < (ENGINEUPDOWNLIMIT - 1))
		EngineData->Row = ENGINEUPDOWNLIMIT - EngineUpDown->Position + UPDOWNBUGFIX;
	}
if (Button == Comctrls::btNext) {
    if (EngineData->Row > 1 )
    	EngineData->Row = ENGINEUPDOWNLIMIT - EngineUpDown->Position - UPDOWNBUGFIX;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RestorethescrollbarontheEngineList1Click(
      TObject *Sender)
{
EngineData->ScrollBars = ssVertical;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DecodeStringsButtonClick(TObject *Sender)
{
// Decode all strings in the Buffer and write them to the CRC Buffer
//
// this does the same thing as the popup menu item but only from the TIU Buffer
//
if (DEBUGGING) STOPem->Enabled = true;
bStop = false;
//
const int count = TIUBuffer->Lines->Count;

for (int idxdb = 0 ; idxdb < count ; idxdb++) {
   if (bStop) {
        break;
        }
	String MyString = TIUBuffer->Lines->Strings[idxdb];
    CRCBuffer->SelAttributes->Color = TIUBuffer->SelAttributes->Color;	// Copy with same color
   	CRCBuffer->Lines->Append(MyString);
    MyString = CleanUpCommand(MyString);
	CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
	(void) sDecodeString(MyString, true); // Decode and print (into CRC Buffer) the string in MyString
    Application->ProcessMessages();
	}
STOPem->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormActivate(TObject *Sender)
{
#if __BORLANDC__ >= 0x720      // CB10.1 only
AlphaBlend = false; // Remove Transparency
#else
// Remove WS_EX_LAYERED from this window styles
SetWindowLong(Handle, GWL_EXSTYLE, GetWindowLong(Handle, GWL_EXSTYLE) & ~WS_EX_LAYERED);
#endif
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormDeactivate(TObject *Sender)
{
if (Setup->TransparentSpin->Value == Setup->TransparentSpin->MaxValue) return;
if (LocalTransparencyOverride) return;
#if __BORLANDC__ >= 0x720      // CB10.1 only
AlphaBlendValue = (255 * Setup->TransparentSpin->Value) / 100;
AlphaBlend = true;  // Add Transparency
#else
// Set WS_EX_LAYERED on this window
SetWindowLong(Handle, GWL_EXSTYLE, GetWindowLong(Handle, GWL_EXSTYLE) | WS_EX_LAYERED);
//SetLayeredWindowAttributes
m_pSetLayeredWindowAttributes(Handle, NULL, (255 * Setup->TransparentSpin->Value) / 100, LWA_ALPHA);
#endif
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Engine_List_LabelClick(TObject *Sender)
{
LocalTransparencyOverride = !LocalTransparencyOverride;	// Toggle transparency locally
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUSpinValueChanged(TObject *Sender)
{
//
if (DEBUGGING) DB->Debug_Print("AIU Number changed from " + String(AIUNo) + " to " + String(AIUSpin->Value));
AIUNo = AIUSpin->Value;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIUSpinValueChanged(TObject *Sender)
{
// Called when the user changes the TIU Number
//
if (TIUNo == -1) return;				// main does the intializing
if (TIUNo == TIUSpin->Value) return;	// nop if the Value really didn't change
int OldTIUNo = TIUNo;

const TCursor OldCursor = Screen->Cursor; 	// save cursor
Screen->Cursor = crHourGlass;
//
// Save current Engine File
if (bTIUDataLoaded[OldTIUNo - 1]) {	// if engine data is loaded, save it to disk
	SaveTIUData(EngineData, OldTIUNo);
    }
// Close and delete the Setup Window (if it's created)
if (Setup != NULL) {
	Setup->Close();
	delete Setup;
	Setup = NULL;
    }
// Load up new Engine File for this new TIU
TIUNo = TIUSpin->Value;	// new TIUNo - this should be the only place where the user can change the TIU number
LoadTIUData(EngineData, TIUNo);
if(!IsTIUConnected()) {
	TIUSpin->ColorDefault = clYellow;	// set TIU box color to yellow indicating unknown comm with the TIU
	AIUSpin->ColorDefault = clYellow;
    }
//
// create the Setup window for the new TIU number
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Setup Form should be manually created
if (!Setup) {	// since I used the C++ pointer, it will get automatically deleted at close
	Setup = new TSetup(this, TIUNo, RemoteNo);     // Pass "this" as the owner
	assert(Setup != NULL);
   	}
if (DEBUGGING) DB->Debug_Print("TIU Number " + ((RADIOMODE) ? String("(Radio/WiFi Mode)") : String("(Wired Mode)")) + " changed from " + String(OldTIUNo) + " to " + String(TIUNo));
// ------------------------------------------------------------
// Reconnect to TIU if we were connected
if (IsTIUConnected() && Setup->Hex->Checked && !Setup->TXOnlyComm->Checked) {	// Hex mode & two way comm required
	if (GetAIUCount(TIUNo)) { // read up the number of AIU
		// if the GetAIUCount worked, read up the TIU version number - needed to form LashUp command correctly
		GetTIUVer(TIUNo);	// this also bumps the sequence number
		}
	// success with above two calls will set TIUSpin to clLime
	// and bTIUConnected to true, failure will set TIUSpin to clRed
	}
//
if (!IsTIUConnected()) {
	TIUVerButton->Caption = "TIU Ver #";
	TIUVer1->Caption = "Display Selected TIU Ver # ";
	AIUCountButton->Caption = "AIU Count";
    } //else {
	// DisconnectButton->OnClick(DisconnectButton);	// TODO I don't remember why I called Disconnect here
    //}
Screen->Cursor = OldCursor;
// Auto Connect to the TIU if requested
if (Setup->AutoConnect->Checked) ConnectButton->OnClick(ConnectButton);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RemoteSpinValueChanged(TObject *Sender)
{
// The remote number should be different than any actual remote that
// you have. I suggest using 15.
if (DEBUGGING) DB->Debug_Print("Remote Number changed from " + String(RemoteNo) + " to " + String(RemoteSpin->Value));
// besides initalization and command line, the next line is the only place that RemoteNo is chanegd.
RemoteNo = RemoteSpin->Value;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMUI(TMessage &Message)
{
//
// Messages from threads to do some processing in the VCL
//
int msgtype = Message.WParam;
UI_Struct * MessageX = (UI_Struct *) Message.LParam;

switch (msgtype) {
	case M_PrintItem:
		// Message to RichEdit Control
		DB->Debug_Messages->SelAttributes->Color = MessageX->PrintColor;
        DB->Debug_Messages->SelStart = DB->Debug_Messages->GetTextLen();
        DB->Debug_Messages->SelText = String(MessageX->Line) + TEXT("\n");
		DB->Debug_Messages->SelAttributes->Color = clBlack;	// reset color to Black
		// remain at end of buffer
		//
		// Method 1 does not work under CB10.1
		//DB->Debug_Messages->SelStart = DB->Debug_Messages->GetTextLen();
		//DB->Debug_Messages->Perform(EM_SCROLLCARET, 0, 0);
		// Method 2 does not work under CB10.1
		//DB->Debug_Messages->SelStart = DB->Debug_Messages->Text.Length();
		//SendMessage(DB->Debug_Messages->Handle, EM_SCROLLCARET,0,0);
		// Method 3 works under both BCB 4.0 and CB10.1 But sometimes does not get to end of buffer
		//DB->Debug_Messages->Perform(EM_SCROLL, SB_LINEDOWN, 0);	// remain at end of buffer
		// Method 4 works under CB10.1 but not BCB 4.0
		//http://codeverge.com/embarcadero.delphi.general/d2009-richedit/1040993
		//
        if (DB->AutoScroll->Checked) {
        	DB->Debug_Messages->HideSelection = false;
        	DB->Debug_Messages->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
        	DB->Debug_Messages->HideSelection = true;
        	}
		// alternative
		//SendMessage(DB->Debug_Messages->Handle, WM_VSCROLL, SB_BOTTOM, 0);
		// from  https://stackoverflow.com/questions/6372193/how-to-autoscroll-to-the-end-of-richedit
        delete MessageX;	// delete the Message after processing it
    	break;

    case M_Bump:
		if (MessageX->Countr != NULL) MessageX->Countr->Value++;
		delete MessageX;	// delete the Message after processing it
		break;

    case M_Dec:
		if (MessageX->Countr != NULL) MessageX->Countr->Value--;
        delete MessageX;	// delete the Message after processing it
 	    break;

    case M_SetColor:
    	if (MessageX->Spinner != NULL) MessageX->Spinner->ColorDefault = MessageX->ColorDefault;
        delete MessageX;	// delete the Message after processing it
		break;

    case M_SetTextValue:
    	if (MessageX->EditBox != NULL) MessageX->EditBox->Text = MessageX->Line;
        delete MessageX;	// delete the Message after processing it
        break;

    case M_GetTextValue:
    	Application->ProcessMessages();		// make sure that VCL component EditBox is up to date (this worked!)
        if (MessageX->EditBox != NULL)
        	MessageX->Line = MessageX->EditBox->Text;	// return text to Comm_Thread
            else
            MessageX->Line = "Data not Available";
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_SetRecordingTextValue:
		//RecordingBuffer->Lines->Append(MessageX->Line);
        RecordingBuffer->SelStart = RecordingBuffer->GetTextLen();
        RecordingBuffer->SelText = MessageX->Line;
        // Method 4 works under CB10.1 but not BCB 4.0
		RecordingBuffer->HideSelection = false;
		RecordingBuffer->Perform(EM_SCROLLCARET,0 ,0);	// remain at end of Recording Buffer
		RecordingBuffer->HideSelection = true;
		delete MessageX;	// delete the Message after processing it
		break;

    case M_ShowinTIUBuffer:
        if (TIULog->Checked) {
            TIUBuffer->SelAttributes->Color = MessageX->ColorDefault;
            // Display it in buffer
            TIUBuffer->SelStart = TIUBuffer->GetTextLen();
            if (MessageX->ColorDefault == clLime) {
                TIUBuffer->SelText = TEXT("\r\n") + String(MessageX->Line) + TEXT("\r\n");
                TIULineCount->Caption = String(TIULineCount->Caption.ToInt() + 1);
                } else {
                TIUBuffer->SelText = String(MessageX->Line) + TEXT("\r\n");
                }
            TIULineCount->Caption = String(TIULineCount->Caption.ToInt() + 1);
            TIUBuffer->SelAttributes->Color = clBlack;		// restore color to Black
            // Method 4 works under CB10.1 but not BCB 4.0
            TIUBuffer->HideSelection = false;
            TIUBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
            TIUBuffer->HideSelection = true;

            if (RFIDLog->Checked && Setup->Echo->Checked) {
                RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();
                RFIDBuffer->SelText = String(MessageX->Line) + TEXT("\r\n");
                // Method 4 works under CB10.1 but not BCB 4.0
                RFIDBuffer->HideSelection = false;
                RFIDBuffer->Perform(EM_SCROLLCARET,0 ,0);		// remain at end of RFID Buffer
                RFIDBuffer->HideSelection = true;
                }
            }
        delete MessageX;	// delete the Message after processing it
		break;

    case M_PopupItem:
    	DebugMessagePopup("Information", MessageX->Line, "", MessageX->iValue);	// Value is time to display in milliseconds
        delete MessageX;	// delete the MessageX after processing it
        break;

    case M_ShowinBuffer:
    	if (MessageX->Buffer != NULL) {
	    	MessageX->Buffer->SelAttributes->Color = MessageX->PrintColor;
			// Display it in buffer
            MessageX->Buffer->SelStart = MessageX->Buffer->GetTextLen();
            MessageX->Buffer->SelText = String(MessageX->Line) + TEXT("\n");
			MessageX->Buffer->SelAttributes->Color = clBlack;				// return color to Black
            if (MessageX->bValue) { // keep cursor at bottom of buffer

                //
            	// remain at end of buffer
            	//
        	    // Method 1 does not work under CB10.1
            	//MessageX->Buffer->SelStart = MessageX->Buffer->GetTextLen();
            	//MessageX->Buffer->Perform(EM_SCROLLCARET, 0, 0);
            	// Method 2 does not work under CB10.1
        	    //MessageX->Buffer->SelStart = MessageX->Buffer->Text.Length();
            	//SendMessage(MessageX->Buffer->Handle, EM_SCROLLCARET,0,0);
            	// Method 3 works under both BCB 4.0 and CB10.1 But sometimes line down doesn't get to end of buffer
            	//MessageX->Buffer->Perform(EM_SCROLL, SB_LINEDOWN, 0);	// remain at end of buffer
        	    // Method 4 works under CB10.1 but not BCB 4.0
            	//http://codeverge.com/embarcadero.delphi.general/d2009-richedit/1040993
	    		MessageX->Buffer->HideSelection = false;
		    	MessageX->Buffer->Perform(EM_SCROLLCARET,0,0);    	// remain at end of Buffer
			    MessageX->Buffer->HideSelection = true;
                }
			}
        delete MessageX;	// delete the Message after processing it
		break;

    case M_SetNumSpinValue:		// needs to be tested
    	if (MessageX->Spinner != NULL) MessageX->Spinner->Value = MessageX->iValue;
        delete MessageX;	// delete the Message after processing it
        break;

    case M_SetNumEditValue:
    	if (MessageX->Countr != NULL) MessageX->Countr->ValueAsInt = MessageX->iValue;
        delete MessageX;	// delete the Message after processing it
        break;

    case M_GetCntrValue:
    	Application->ProcessMessages();		// make sure that VCL component TAbNumEdit is up to date (this worked!)
        if (MessageX->Countr != NULL)
        	MessageX->Line = MessageX->Countr->Text;	// return text to Comm_Thread
            else
            MessageX->Line = "0";
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_SetPCLED:	// Set a LED to Visible & Checked, then set the color based on iState
    					// Line if not empty, sets the LED's Caption
                        // Used on the LED's on the Program Control window
    	if (MessageX->LED != NULL) {
            MessageX->LED->Visible = true;
            MessageX->LED->Checked = true;
            switch(MessageX->iState) {
            	case luaGreen:	// clGreen
                    MessageX->LED->LED->ColorOn = clLime;
                	break;
                case luaYellow:	// clYellow
                    MessageX->LED->LED->ColorOn = clYellow;
                	break;
                case luaClear:	// clClear makes the LED disappear
                	// clear
		            MessageX->LED->Visible = false;
        		    MessageX->LED->Checked = false;
                    break;
                case luaRed:	// clRed
                default:	// all other values
                    MessageX->LED->LED->ColorOn = clRed;
                	break;
                // clBlack, clBlue and clViolet are not valid and display as clRed
                }
            if (MessageX->Line != NULL && !MessageX->Line.IsEmpty()) {
            	MessageX->LED->Caption = MessageX->Line;
                }
        	}
        delete MessageX;	// delete the Message after processing it
        break;

    case M_SetNumEditColor:
    	if (MessageX->Countr != NULL) MessageX->Countr->ColorDefault = MessageX->ColorDefault;
        delete MessageX;	// delete the Message after processing it
        break;

    case M_GetComboBoxText:
    	Application->ProcessMessages();		// make sure that VCL component EditBox is up to date (this worked!)
        if (MessageX->ComboBox != NULL) {
            MessageX->Line = "";	// default return value
        	if (MessageX->iValue >= 0 && MessageX->iValue < MessageX->ComboBox->Items->Count) {
	        	MessageX->Line = MessageX->ComboBox->Items->Strings[MessageX->iValue];	// return text to Comm_Thread
                }
            }
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_ClearCntr:
		if (MessageX->Countr != NULL) MessageX->Countr->Value = 0;
        delete MessageX;	// delete the Message after processing it
 	    break;

    case M_SetCntr:
		if (MessageX->Countr != NULL) MessageX->Countr->Value = MessageX->fValue;
        delete MessageX;	// delete the Message after processing it
 	    break;

    case M_OpenAIUWindow:
    	// Open an AIU window
        {
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // make the window visible since this was done by direct request
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->Show();
            // Bring the PC window back to the front
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			} else {
            // make the window visible since this was done by direct request
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->Show();
			}
        delete MessageX;	// delete the Message after processing it
        }
    	break;

    case M_GetAIUSwitchState:
        // fetch the state of the given TIU/AIU/Channel Switch
        {
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
        // Make sure the AIU window is open but not visible
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // Bring the PC window back to the front by sending the new window to the back
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			}
        //
        int iChan = MessageX->iChan;
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL) {
	        MessageX->bValue = (ptrAIUControl[iTIUNo-1][iAIUNo-1]->AccData->Cells[SW_STATE][iChan] == "1") ? true : false;
            } else MessageX->bValue = false;
        }
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_GetAIUAccState:
        // fetch the state of the given TIU/AIU/Channel Accessory
        {
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
        // Make sure the AIU window is open but not visible
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // Bring the PC window back to the front by sending the new window to the back
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			}
        //
        int iChan = MessageX->iChan;
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL) {
	        MessageX->bValue = (ptrAIUControl[iTIUNo-1][iAIUNo-1]->AccData->Cells[ACC_STATE][iChan] == "1") ? true : false;
            } else MessageX->bValue = false;
        }
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_GetAIUSwitchName:
        // fetch the name of the given TIU/AIU/Channel Switch
        {
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
        // Make sure the AIU window is open but not visible
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // Bring the PC window back to the front by sending the new window to the back
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			}
        //
        int iChan = MessageX->iChan;
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL) {
	        MessageX->Line = ptrAIUControl[iTIUNo-1][iAIUNo-1]->AccData->Cells[SW_NAME][iChan];
            } else MessageX->Line = "*";
        }
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_GetAIUAccName:
        // fetch the name of the given TIU/AIU/Channel Accessory
        {
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
        // Make sure the AIU window is open but not visible
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // Bring the PC window back to the front by sending the new window to the back
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			}
        //
        int iChan = MessageX->iChan;
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL) {
	        MessageX->Line = ptrAIUControl[iTIUNo-1][iAIUNo-1]->AccData->Cells[ACC_NAME][iChan];
            } else MessageX->Line = "*";
        }
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_MakeItSo:
    	// set all accessories and switches per their AIU states
        {
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
        // Make sure the AIU window is open but not visible
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // Bring the PC window back to the front by sending the new window to the back
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			}
        // press the MakeItSo button on the selected AIU
        ptrAIUControl[iTIUNo-1][iAIUNo-1]->MakeItSoButtonClick(NULL);
        }
        delete MessageX;	// delete the Message after processing it
		break;

    case M_SetVirtualAIUAccState:
        // set the state of the given TIU/AIU/Channel - Accessory channels only
        {
        // TODO Needs to be tested
        int iTIUNo = MessageX->iTIU;
        int iAIUNo = MessageX->iAIU;
		int iChan = MessageX->iChan;
		bool bState = MessageX->bValue;
        // Make sure the AIU window is open but not visible
        if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
			ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, 1, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
			assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    		ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = MainForm->Handle; // for return messages
            // Bring the PC window back to the front by sending the new window to the back
            ptrAIUControl[iTIUNo-1][iAIUNo-1]->SendToBack();
            ptrProgram_Control[MessageX->PCNo-1]->SetFocus();
			}
        // set the accessory status of any AIU from 1 to AIUMAX - can be virtual
        ptrAIUControl[iTIUNo-1][iAIUNo-1]->AccData->Cells[ACC_STATE][iChan] = (bState ? "1" : "0");
        }
        delete MessageX;	// delete the Message after processing it
		break;

    case M_GetStringGridText:
		// return Engine name and PS LEvel
    	Application->ProcessMessages();		// make sure that VCL component StringList is up to date (this worked!)
        if (MessageX->StringGrid != NULL) {
        	MessageX->Line = MessageX->StringGrid->Cells[ENG_NAME][MessageX->iValue];
            MessageX->iValue =  MessageX->StringGrid->Cells[ENG_PS][MessageX->iValue].ToInt();;
            }
        // reurns engine name and engine PSx value
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_SetLabelText:
    	if (MessageX->Label != NULL) MessageX->Label->Caption = MessageX->Line;
        delete MessageX;	// delete the Message after processing it
    	break;

    case M_PopupMessage:
		// display a popup message window
        // Keep it on the screen for iValue milliseconds
		// no buttons to push
		// Start Thread
		// FreeOnTerminate is set in constructor so we don't need to worry about freeing memory
		new TMessageDisplayThread(false, this, Handle, MessageX->Caption, MessageX->Line, MessageX->Status, MessageX->iValue);
        delete MessageX;	// delete the Message after processing it
    	break;

   	// Flags -------------------------------------------------------------------
	case M_GetFlag:
    	// Get a boolean flag value from flag's "Tag" property
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->iState = LayoutForm->GetFlag(MessageX->iValue);
            } else {
            MessageX->iState = 0;   // simulates that the block is empty
            }
        // returns flag value
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_SetFlag:
    	// Set a integer flag value into flag's "Tag" property and
        // if that value == 0 then set the "Checked" property indicating an unoccupied block
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->iState = LayoutForm->SetFlag(MessageX->iValue, MessageX->iState);
            }
        // returns flag value
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
		break;

    case M_GetFlagName:
    	// Get a flag name from flag's "Caption" property
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->Line = LayoutForm->GetFlagName(MessageX->iValue);
            } else {
            MessageX->Line = "Flag #" + String(MessageX->iValue);
            }
        // returns flag name
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_SetFlagName:
    	// Set a flag name into flag's "Caption" property
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->iState = LayoutForm->SetFlagName(MessageX->iValue, MessageX->Line);
            } else {
            MessageX->iState =  0;   // simulates that the block is empty
            }
        // returns flag value
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

	// Fixed Block Signals -----------------------------------------------------
	case M_GetAspect:
    	// Get the Aspect value, the number of lights on the signal head and the signal type
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->iAspect = LayoutForm->GetAspect(MessageX->iSignalNo, MessageX->iNumLights, MessageX->iType);
            } else {
            MessageX->iAspect =  0;   // simulates that the block is empty
            }
        // returns signal aspect
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_SetAspect:
    	// Set a Aspect value into Block signal's "Tag" and "Aspect" properties
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->iAspect = LayoutForm->SetAspect(MessageX->iSignalNo, MessageX->iAspect);
            }
        // returns signal aspect
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
		break;

    case M_GetSignalName:
    	// Get a Block's signal name from Signal's "Caption" property, the value on
        // the numberplate and the Signal's Type
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->Line = LayoutForm->GetSignalName(MessageX->iSignalNo, MessageX->iNumberPlate, MessageX->iType);
            } else {
            MessageX->Line = "Signal #"+ String(MessageX->iSignalNo);
            }
		// returns signal name, number plate and signal type
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_CreateSignal:
    	// Create a Block signal
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            MessageX->iAspect = LayoutForm->CreateSignal(MessageX->iSignalNo, MessageX->iDir, MessageX->iType, MessageX->iCol,
        	    MessageX->iRow, /* SignalName */ MessageX->Line, MessageX->iNumLights, MessageX->iAspect, MessageX->iNumberPlate);
            }
        // returns signal aspect
        // MessageX struct is deleted in Comm_Thread.cpp after the return value is read
    	break;

    case M_ShowLayout:
        // create and show the Layout window if not already done
		if (!LayoutForm) {
            ShowLayoutWindowClick(NULL);
            }
        delete MessageX;	// delete the Message after processing it
    	break;

	case M_DeleteAllSignals:
		//if (!LayoutForm) ShowLayoutWindowClick(NULL);	// create and show the Layout window if not already done
		if (LayoutForm) {
            LayoutForm->DeleteAllSignals();
            }
		delete MessageX;	// delete the Message after processing it
		break;

	// Cab Signals -------------------------------------------------------------
    case M_SetCabAspect:
        ptrProgram_Control[MessageX->PCNo-1]->SetCabSignal(MessageX->iSignalNo, MessageX->iAspect, "", -1, -1);
    	delete MessageX;	// delete the Message after processing it
        break;

    case M_SetCabLabels:
        ptrProgram_Control[MessageX->PCNo-1]->SetCabSignal(MessageX->iSignalNo, MessageX->iAspect, MessageX->Line, MessageX->iNumberPlate,
        	MessageX->iType);
    	delete MessageX;	// delete the Message after processing it
        break;

    case M_CreateCab:
        ptrProgram_Control[MessageX->PCNo-1]->CreateCabSignals(MessageX->iSignalNo, MessageX->iDistant);
    	delete MessageX;	// delete the Message after processing it
        break;

    case M_GetCabAspect:	// and Type
        ptrProgram_Control[MessageX->PCNo-1]->GetCabAspect(MessageX->iSignalNo, MessageX->iAspect, MessageX->iNumLights, MessageX->iType);
		// returns signal aspect, number of lights on signal head and signal type
    	// MessageX struct is deleted in Comm_Thread.cpp after the return value is read
        break;

    case M_SetLED:		// Set LED's On Color from on iState
                        // Set LED's Checked from iValue
    					// Set LED's Flashing from iTYpe
    					// Line, if not empty, sets the LED's Caption
    	// Used on the LED's on the Main window (ConnectedRadioLED, ConnectedRFIDLED, and ConnectedTIULED)
    	if (MessageX->LED != NULL) {
            MessageX->LED->LED->ColorOn = (TColor) MessageX->iState;
            MessageX->LED->Checked = MessageX->iValue;
            MessageX->LED->Flashing = MessageX->iType;
            if (MessageX->Line != NULL && !MessageX->Line.IsEmpty()) {
            	MessageX->LED->Caption = MessageX->Line;
                }
        	}
        delete MessageX;	// delete the Message after processing it
        break;

    case M_AddCntr:
        if (MessageX->Countr != NULL) MessageX->Countr->Value += MessageX->iValue;
        delete MessageX;	// delete the Message after processing it
        break;
        
    default:
	    if (DEBUGGING) {
        	DB->Debug_Print("MainForm::CMUI Unknown User Interface Message", clRed);
            } else {
            assert(false);
            }
        delete MessageX;	// delete the message
		break;
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMProcessRFIDData(TMessage &Message)
{
// process ASCII Character message received from Tag Reader port - CM_PROCESSRFIDDATA
// By putting the message into the RFIDFIFO. Then send RFIDReqEvent to get
// tag processed by PC_Thread.cpp. Tags are sent to all instances of PC_Thread.
// Handles both Serial Tag Detectors (called from Comm_Thread) and WiFi Tag Detectors (called from MainForm)
//
// Note: Message.LParam points to a memory block of 57 bytes to hold the longest
// possible tag packet
//
char * sTagBuf = (char *) Message.LParam;	// set address of buffer to process from Long Parameter
assert (sTagBuf != NULL);
int length = strlen(sTagBuf);

if (DEBUGGING) {
	String TagLine = String((char *)sTagBuf);
	if (DEBUGGINGX(4)) DB->Debug_Print("Tag Reader Message " + TagLine + " (" + String(length) + " bytes)", clBlue);
	if (RFIDLog->Checked) {
		// display it in the RFID buffer
		RFIDBuffer->SelAttributes->Color = clGreen;
		RFIDBuffer->SelStart = RFIDBuffer->GetTextLen();
		RFIDBuffer->SelText = TagLine + TEXT("\r\n");    // CR LF
		// Method 4 works under CB10.1 but not BCB 4.0
		RFIDBuffer->HideSelection = false;
		RFIDBuffer->Perform(EM_SCROLLCARET, 0, 0);	// remain at end of buffer
		RFIDBuffer->HideSelection = true;
		}
	}
Tags->Value++;				// increment totals number of packets received (includes tags and startup packets)
//
/*
Note: on TCP connection to detectors, 3 packets are received from each detector and 1 packet is received for each reader.
//
FF05333401010270F0 (18 bytes)  HELLOPACKET
FF05000007104465746563746F722D3520202020202078F0 (48 bytes) DETECTORNAME
FF05010008103139322E3136382E312E31323120202008F0 (48 bytes) DETECTORIP
FF050000040009F0 (16 bytes)  PN53xFOUND for each attached tag reader
FF05010004000AF0 (16 bytes)  PN53xFOUND
//
  full tag packet tags are 56 bytes long
FF0500002114864C134A88080101030108011B0220203135363132F0
  UID only tag packet (0x03) is 24 bytes long
FF0500030304864C134A3EF0
*/
// TODO Check the Checksum at some point, but error rate is very low
//
// add to detected data tags list unless it is already there
if (length == 56) {
	DetectedRFIDTags->Add(String(sTagBuf).SubString(13, 40));
    }
// add to UID detected data tags list unless it is already there
if (length == 24) {
	DetectedRFIDTags->AddUID(String(sTagBuf).SubString(13, 8));
    }
unsigned char PacketType = Hex2Digit_to_Decimal(&sTagBuf[PACKETTYPEIDX]);
switch (PacketType) {
    // if it is a UID only tag, look it up in the TagsDetectedList
    case TAGDETECTED:
        {
        assert(length == 24);
        // keep count of UID only tags
        DetectedRFIDTags->UIDTagCount->Value++;
        // Look up the UID - NewPacket must be AnsiString because .c_str() is taken below
        AnsiString NewPacket = DetectedRFIDTags->LookUpTagPacket(String(sTagBuf).SubString(13, 8));
        if (!NewPacket.IsEmpty()) {
            // if we get a good tag lookup, use that looked up tag
            // NewPacket is 40 characters, 8 bytes of UID and 32 bytes of Enginedata
            // TODO Should the Checksum be calculated rather than just set to zero?
            NewPacket = String(sTagBuf).SubString(1, 8) + "2114" + NewPacket + "00F0"; // command 0x21 and byte count 0x14 = 20 (4 bytes UID & 8 bytes Engine Data)
            length = NewPacket.Length();
            assert(length == 56);
            strncpy(sTagBuf, NewPacket.c_str(), NewPacket.Length());
            //
            if (DEBUGGINGX(6)) {
                String PrintLine;
                PrintLine = "Detector " + String(Hex2Digit_to_Decimal(&sTagBuf[DETECTORIDX])) +
                            " Reader " + String(Hex2Digit_to_Decimal(&sTagBuf[READERIDX])) + " : Substituting Looked Up Packet " + NewPacket;
                DB->Debug_Print(PrintLine, clRed);
                }
            }
        // Global Counter to keep track of Tags Substituted
        DetectedRFIDTags->TagSubCount->Value++;
        DetectedRFIDTags->TagSubCount->ColorDefault = clRed;
        // process the saved tag packet as if it were just received
        break;
        }
    // Check if its an HELLOPACKET and display a message in the Debug Message Window
    case HELLOPACKET:
        if (DEBUGGING) {
            UP Packet;
            // to display this, we have to convert the packet to binary
            for (int iDex = 0 ; iDex < (48/2); iDex++) {
                Packet.Byte_Array[iDex] = Hex2Digit_to_Decimal(&sTagBuf[iDex * 2]);
                }
			String PrintLine = "Detector " + String(Packet.Detector_Name_Packet.Header.Detector) + " : Hello";
            DB->Debug_Print(PrintLine, clBlue);
            }
        break;

    // Check if it is a PN53xFOUND packet, bump the Number of Readers and display a message in the Debug Message Window
    case PN53xFOUND:
        NumReaders->Value++;
        if (DEBUGGING) {
            UP Packet;
            // to display this, we have to convert the packet to binary
            for (int iDex = 0 ; iDex < (48/2); iDex++) {
                Packet.Byte_Array[iDex] = Hex2Digit_to_Decimal(&sTagBuf[iDex * 2]);
                }
			String PrintLine = "Detector " + String(Packet.Detector_Name_Packet.Header.Detector) + " : PN532 Reader Found";
            DB->Debug_Print(PrintLine, clBlue);
            }
        break;
    // Check if its is a DETECTORNAME packet and display the name in the Debug Message Window
    case DETECTORNAME:
        // Bump the number of Detectors
        NumDetectors->Value++;
        if (DEBUGGING) {
            UP Packet;
            char cmsgBuf[17];
            // to display this, we have to convert the packet to binary
            for (int iDex = 0 ; iDex < (48/2); iDex++) {
                Packet.Byte_Array[iDex] = Hex2Digit_to_Decimal(&sTagBuf[iDex * 2]);
                }
            memcpy(cmsgBuf, &Packet.Detector_Name_Packet.Detector_Name, 16);
            cmsgBuf[16] = '\0';	// Null Terminate
			{	// remove trailing spaces
                int idx = 15;
                while(1) {
                    if (idx == -1) break;	// done
                    if (cmsgBuf[idx] == ' ') {	// is the character a space?
                        cmsgBuf[idx] = '\0';	// yes
                        idx--;
                        continue;
                        }
                    break;	// first non-space and we are done
                    }
			}
			String PrintLine = "Detector " + String(Packet.Detector_Name_Packet.Header.Detector) + " : Name " + String(cmsgBuf);
            DB->Debug_Print(PrintLine, clBlue);
            }
        break;
    // Check if its is a DETECTORIP packet and display the IP Address in the Debug Message Window
    case DETECTORIP:
        if (DEBUGGING) {
            UP Packet;
            char cmsgBuf[17];
            // to display this, we have to convert the packet to binary
            for (int iDex = 0 ; iDex < (48/2); iDex++) {
                Packet.Byte_Array[iDex] = Hex2Digit_to_Decimal(&sTagBuf[iDex * 2]);
                }
            memcpy(cmsgBuf, &Packet.Detector_Name_Packet.Detector_Name, 16);
            cmsgBuf[16] = '\0';	// Null Terminate
			{	// remove trailing spaces
                int idx = 15;
                while(1) {
                    if (idx == -1) break;	// done
                    if (cmsgBuf[idx] == ' ') {	// is the character a space?
                        cmsgBuf[idx] = '\0';	// yes
                        idx--;
                        continue;
                        }
                    break;	// first non-space and we are done
                    }
            }
            String PrintLine = "Detector " + String(Packet.Detector_Name_Packet.Header.Detector) + " : IP " + String(cmsgBuf) +
                " using " + (Packet.Data_Packet.Header.Reader == 0 ? "UDP" : "TCP");
            DB->Debug_Print(PrintLine, clBlue);
            }
        break;
    }
// Count all responses from the Detectors
DetectedRFIDTags->RespCount->Value++;
// add to all of the RFID FIFO's
//
// send the packet to every PC thread that is is active and contains a tag(), nametag() or uidtag() function
// each PC thread has its own FIFO. Individual RFIDReqEvent
// are created by each PC thread. We send that even to each thread. The PC thread
// should ignore any packet that it does care about
//
// Note that each packet has a checksum and an end of packet character (0xF0) appended
//
for (int iPCNo = 1 ; iPCNo <= PCMAX ; iPCNo++) {
    if (ptrProgram_Control[iPCNo - 1] == NULL) continue;							// PC Window must be running, this one isn't, try next one
    if (ptrProgram_Control[iPCNo - 1]->ptrPC_Thread == NULL) continue;				// PC Thread must be running, this one isn't, try next one
    if (ptrProgram_Control[iPCNo - 1]->ptrPC_Thread->RFIDFIFO == NULL) continue;	// RFIDFIFO must be created, this one isn't, try next one
    // make a copy of the tag
    char * sNewBuf = (char *)malloc(EOTIDX + 2 + 1);	// need to allocate at least 57 bytes (56 for the packet and 1 for null);
    memcpy(sNewBuf, sTagBuf, EOTIDX + 2 + 1);

    TList *myList = ptrProgram_Control[iPCNo - 1]->ptrPC_Thread->RFIDFIFO->LockList();	// lock the RFIDFIFO
    assert(myList);
    ptrProgram_Control[iPCNo - 1]->ptrPC_Thread->RFIDFIFO->Add(sNewBuf);
    ptrProgram_Control[iPCNo - 1]->ptrPC_Thread->RFIDFIFO->UnlockList();				// unlock the RFIDFIFO
	// Signal the RFID Request event to all PC threads
	SetEvent(hRFIDReqEvent[iPCNo - 1]);
    }
// each PC thread in RFIDTagHandler() will delete[] the copy of the tag at sNewBuf[] pointed to by the RFIDFIFO.
// We delete the original here.
delete [] sTagBuf;
return;
}
//--------------------------------------------------------------------------
//
// NOT USED - performed in Comm_Thread.cpp
//
void __fastcall TMainForm::CMProcessRadioData(TMessage &Message)
{
// process ASCII Character message received from Radio port - CM_PROCESSRADIODATA
//
assert(false);		// never called
//
/*
char * sResponseBuf = (char *) Message.LParam;	// set address of buffer to process from Long Parameter
assert (sResponseBuf != NULL);
int length = strlen(sResponseBuf);
sResponseBuf[length] = '\0';

if (DEBUGGINGX(6)) {
	String PrintLine;
    PrintLine = "TIU Response WiFi Message " + String((char *)sResponseBuf) + " (" + String(length) + " bytes)";
	//DB->Debug_Print(PrintLine, clBlue);
    ptrTIUCommThread->PrintItem(PrintLine, clBlue);
	}

strncpy((char *)cTIUbuf, sResponseBuf, length + 1);
iTIUbufCount = length;

// We delete the original WiFipacket here.
delete [] sResponseBuf;
return;
*/
}
//--------------------------------------------------------------------------
//
// NOT USED - all sending is done in Comm_Thread which can call RadioSend directly
//
void __fastcall TMainForm::CMSendWiFiPacket(TMessage &Message)
{
// Send a packet to the Radio via WiFi - CM_SENDWIFIPACKET
//
assert(false);		// never called - moved to TCommThread::Execute()
//
char * sSendBuf = (char *) Message.LParam;	// set address of buffer to process from Long Parameter

//RadioSend->Proto 				= "udp";
//RadioSend->Addr  				= Setup->RadioIP->Text;
//RadioSend->Port  				= Setup->RadioPort->Text;
//RadioSend->Connect();
//RadioSend->SendStr(String(sSendBuf));
//RadioSend->Close();

delete sSendBuf;
}
//--------------------------------------------------------------------------

void __fastcall TMainForm::Channel1Click(TObject *Sender)
{
// Activate Channel 1
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// open Var control
StartUpVarControl(1);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Channel2Click(TObject *Sender)
{
// Activate Channel 2
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
// open Var control
StartUpVarControl(2);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineStatus1Click(TObject *Sender)
{
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Engine Status popup menu item
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= ENGINEMAX);	// 0/1-99
//
// Manually create form and show it (menu item is disabled for LashUps and All Engine Operation)
// if the form is not already created, then create it
// ------------ With programmed create using C++ Builder's generated pointer
if (ptrEngineStatus[TIUNo-1][EngineNo] == NULL) {	// Engine 0-99 (0 is factory reset, 1-99)
	ptrEngineStatus[TIUNo-1][EngineNo] = new TEngineStatus(this, Handle, RemoteNo, EngineNo, TIUNo, EngineData->Cells[ENG_NAME][EngineNo]);   // create the Form and show it
								// Pass "this" as the owner
    							// 1-99 Engine Number
	assert(ptrEngineStatus[TIUNo-1][EngineNo] != NULL);	// make sure that construction worked
	} else {    // it's already created
	ptrEngineStatus[TIUNo-1][EngineNo]->Show();
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineTrackSignal1Click(TObject *Sender)
{
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Engine Status popup menu item
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= ENGINEMAX);	// 0/1-99
//
// Manually create form and show it (menu item is disabled for LashUps and All Engine Operation)
// if the form is not already created, then create it
// ------------ With programmed create using C++ Builder's generated pointer
if (ptrTrackSignal[TIUNo-1][EngineNo] == NULL) {	// Engine 0-99 (0 is factory reset, 1-99)
	ptrTrackSignal[TIUNo-1][EngineNo] = new TTrackSignal(this, Handle, RemoteNo, EngineNo, TIUNo, EngineData->Cells[ENG_NAME][EngineNo]);   // create the Form and show it
								// Pass "this" as the owner
    							// 1-99 Engine Number
	assert(ptrTrackSignal[TIUNo-1][EngineNo] != NULL);	// make sure that construction worked
	} else {    // it's already created
	ptrTrackSignal[TIUNo-1][EngineNo]->Show();
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngineInformation1Click(TObject *Sender)
{
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }

// Engine Data popup menu item
assert(EngineNo >= (Setup->AllowZero->Checked? 0 : 1) && EngineNo <= ENGINEMAX);	// 0/1-99
//
// Manually create form and show it (menu item is disabled for LashUps and All Engine Operation)
// if the form is not already created, then create it
// ------------ With programmed create using C++ Builder's generated pointer
if (ptrEngineInfo[TIUNo-1][EngineNo] == NULL) {	// Engine 0-99 (0 is factory reset, 1-99)
	ptrEngineInfo[TIUNo-1][EngineNo] = new TEngineInfo(this, Handle, RemoteNo, EngineNo, TIUNo, EngineData->Cells[ENG_NAME][EngineNo]);   // create the Form and show it
								// Pass "this" as the owner
    							// 1-99 Engine Number
	assert(ptrEngineInfo[TIUNo-1][EngineNo] != NULL);	// make sure that construction worked
	} else {    // it's already created
	ptrEngineInfo[TIUNo-1][EngineNo]->Show();
	}
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::FetchByte(unsigned int addr, byte * value, byte Eng, byte TIU)
{
//
// Fetch one byte of RAM memory
// addr points to the byte address
// byte result in 1 byte
// Never addressed to a Lashup
// returns false on error
//
assert(addr <= 0x03FF);
assert(Eng <= ENGINEMAX);
*value = 0;
byte FBResp[CMDMAX];
char CommandString[15];
snprintf(CommandString, sizeof CommandString, "V1%03X", addr);
bool bretv = ptrTIUCommThread->Send_String_FIFO(Eng, CommandString, TIU, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
if (!bretv) {	// retry once
    if (DEBUGGING) DB->Debug_Print("FetchByte() Retrying " + String(CommandString));
	bretv = ptrTIUCommThread->Send_String_FIFO(Eng, CommandString, TIU, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
	if (!bretv) return false;	// return false with value == 0 if we can't read RAM
    }
*value =  FBResp[5];
if (DEBUGGINGX(1)) DB->Debug_Print("Fetch Byte - address : 0x" + IntToHex((int)addr, 3) + " value : 0x" + IntToHex((int)*value, 2));
return true;
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::FetchWord(unsigned int addr, unsigned short * value, byte Eng, byte TIU)
{
//
// Fetch two bytes of RAM memory
// addr points to the lowest byte address
// unsigned short result in 2 bytes
// Never addressed to a Lashup
//
assert(addr <= 0x03FF);
assert(Eng <= ENGINEMAX);
*value = 0;
byte FBResp[CMDMAX];
char CommandString[15];
snprintf(CommandString, sizeof CommandString, "V2%03X", addr);
bool bretv = ptrTIUCommThread->Send_String_FIFO(Eng, CommandString, TIU, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
if (!bretv) {	// retry once
    if (DEBUGGING) DB->Debug_Print("FetchWord() Retrying " + String(CommandString));
	bretv = ptrTIUCommThread->Send_String_FIFO(Eng, CommandString, TIU, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
	if (!bretv) return false;	// return false with value == 0 if we can't read RAM
    }
*value =  FBResp[5] << 8 | FBResp[6];
if (DEBUGGINGX(1)) DB->Debug_Print("Fetch Word - address : 0x" + IntToHex((int)addr, 3) + " value : 0x" + IntToHex((int)*value, 4));
return true;
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::FetchInt(unsigned int addr, unsigned long * value, byte Eng, byte TIU)
{
//
// Fetch four bytes of RAM memory
// addr points to the lowest byte address
// unsigned long result in 4 bytes
// Never addressed to a Lashup
//
assert(addr <= 0x03FF);
assert(Eng <= ENGINEMAX);
*value = 0;
byte FBResp[CMDMAX];
char CommandString[15];
snprintf(CommandString, sizeof CommandString, "V4%03X", addr);
bool bretv = ptrTIUCommThread->Send_String_FIFO(Eng, CommandString, TIU, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
if (!bretv) {	// retry once
    if (DEBUGGING) DB->Debug_Print("FetchInt() Retrying " + String(CommandString));
	bretv = ptrTIUCommThread->Send_String_FIFO(Eng, CommandString, TIU, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, FBResp);
	if (!bretv) return false;	// return false with value == 0 if we can't read RAM
    }
*value =  FBResp[5] << 24 | FBResp[6] << 16 | FBResp[7] << 8 | FBResp[8];
if (DEBUGGINGX(1)) DB->Debug_Print("Fetch Integer - address : 0x" + IntToHex((int)addr, 3) + " value : 0x" + IntToHex((int)*value, 8));
return true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DumpMemoryButtonClick(TObject *Sender)
{
// Dump Engine memory from 0x00 to 0x6F
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (!DEBUGGING) return;
if (FirstLine->ValueAsInt > LastLine->ValueAsInt) return;	// FirstLine must always be <= LastLine
//
// Dumps from FIRSTLINE to LASTLINE (line = 16 bytes) of either EEPROM or a RAM memory.
// I suspect it is RAM memory as the DCH time is there (at 0x0038) and it is
// incrementing. You really can't do that with a block erase EEPROM.
//
String PLine = "Offset(h)  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F";

DB->Debug_Print(PLine, clGreen);
CRCBuffer->SelAttributes->Color = clGreen;
CRCBuffer->Lines->Append(PLine);
// Method 4 works under CB10.1 but not BCB 4.0
CRCBuffer->HideSelection = false;
CRCBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of CRC Buffer
CRCBuffer->HideSelection = true;

for (int i = FirstLine->ValueAsInt ; i <= LastLine->ValueAsInt ; i++) {
	PLine  = Format((String)"0x0000%s", ARRAYOFCONST((IntToHex((i * 0x10), 4))));
	for (int j = 0x00 ; j < 0x10 ; j = j + 0x04) {
    	unsigned long B;
        FetchInt((i * 0x10) + j, &B, EngineNo, TIUNo);
		PLine = PLine + Format((String)" %s %s %s %s", ARRAYOFCONST((
        	IntToHex((int)(B >> 24) & 0xFF, 2),
            IntToHex((int)(B >> 16) & 0xFF, 2),
            IntToHex((int)(B >>  8) & 0xFF, 2),
            IntToHex((int)(B      ) & 0xFF, 2)
			)));
    	}
    DB->Debug_Print(PLine, clGreen);
    CRCBuffer->SelAttributes->Color = clGreen;
    CRCBuffer->Lines->Append(PLine);
    // Method 4 works under CB10.1 but not BCB 4.0
    CRCBuffer->HideSelection = false;
    CRCBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
    CRCBuffer->HideSelection = true;
    PLine = "";
    }
DB->Debug_Print(PLine, clGreen);
CRCBuffer->SelAttributes->Color = clGreen;
CRCBuffer->Lines->Append("");
//CRCBuffer->Perform(EM_SCROLL,SB_LINEDOWN,0);	// remain at end of CRC Buffer
// Method 4 works under both BCB 4.0 and CB10.1
CRCBuffer->HideSelection = false;
CRCBuffer->Perform(EM_SCROLLCARET,0, 0);    	// remain at end of TIU Buffer
CRCBuffer->HideSelection = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FirstLineValueChanged(TObject *Sender)
{
if (FirstLine->ValueAsInt > LastLine->ValueAsInt) {
	LastLine->ValueAsInt = FirstLine->ValueAsInt;
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::LastLineValueChanged(TObject *Sender)
{
if (LastLine->ValueAsInt < FirstLine->ValueAsInt) {
	FirstLine->ValueAsInt = LastLine->ValueAsInt;
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEnginePS2Click(TObject *Sender)
{
SetEnginePS2->Checked = true;
SetEnginePS3->Checked = false;
MainForm->EngineData->Cells[ENG_PS][EngineNo] = PS2;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SetEnginePS3Click(TObject *Sender)
{
SetEnginePS3->Checked = true;
SetEnginePS2->Checked = false;
MainForm->EngineData->Cells[ENG_PS][EngineNo] = PS3;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::VButtonClick(TObject *Sender)
{
// test the V command
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

byte VResp[CMDMAX];
AnsiString V = Encoded_CommandA->Text;
if (V.Length() != 5) return;
ptrTIUCommThread->Send_String_FIFO(EngineNo, V.c_str(), TIUNo, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, VResp);
unsigned int ival = VResp[5] << 24 | VResp[6] << 16 | VResp[7] << 8 | VResp[8];
char buff[20];
snprintf(buff, sizeof buff, "%08X", ival);
if (DEBUGGING) DB->Debug_Print("V Command : " + String(buff), clGreen);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ComponentButtonClick(TObject *Sender)
{
//
// Can use a scheme like this to check in the Configuration File
// for any component name "Label4", etc and be able to change
// the cmoponent Caption to what is specified in the Configuration File
// http://bcbjournal.org/articles/vol3/9904/Enumerating_components.htm
CRCBuffer->Lines->Clear();
for (int i=0;i<ComponentCount;i++) {
	String S = Components[i]->Name;
	CRCBuffer->Lines->Append(S);
    //TLabel * T;
    //T = dynamic_cast<TLabel *> (Components[i]);
    //if (T) T->Caption = S;	// change all labels to the "Name" of the label
    // http://bcbjournal.org/articles/vol3/9904/Dealing_with_sets.htm?PHPSESSID=2f96cbddf92baa7dc77f7a33591f59b5
    //if (T) T->Font->Style = TFontStyles() << fsBold;	// change all fonts to bold
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EmailAuthorClick(TObject *Sender)
{
ShellExecute(NULL,TEXT("open"),TEXT("mailto:markd@silogic.com"), NULL, NULL, SW_RESTORE);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ShrinkRecordingBufferClick(TObject *Sender)
{
RecordingBuffer->Height = 57;
RecordingBuffer->Top = 406;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ExpandRecordingBufferClick(TObject *Sender)
{
RecordingBuffer->Height = 202;
RecordingBuffer->Top = 260;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ProgramControlClick(TObject *Sender)
{
// Call up Program Control Window based on Tag value
StartProgramControl(Sender, reinterpret_cast<TMenuItem *>(Sender)->Tag);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMPCWindowOpen(TMessage &Message)
{
// Open a window of Program Control
int NewPCNo = Message.WParam;
assert(NewPCNo >= 1 && NewPCNo <= PCMAX);
// Start Program Control and set checkmarks in menu
StartProgramControl(PCPopupMenu->Items->Items[NewPCNo - 1], NewPCNo);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartProgramControl(TObject *Sender, int PCNo)
{
// Show the Detected Tags window if RFID port is connected
// DetectedRFIDTags window is Auto-Create
#if NOTAGCHECKBOX
if (Setup->RFID_Mode->Checked && !DetectedRFIDTags->Visible) {
	DetectedRFIDTags->Show();
    }
#endif
//
// Start the timer thread if it is not already started
	// Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
    // The Execute() routine actually runs before some lower level construction is finished.
    // Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
    // per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0
if (ptrTimer_Thread == NULL) {
	ptrTimer_Thread = new TMyTimerThread(true, NULL, 0);
    assert(ptrTimer_Thread);
	TimerThreadLED->Checked = true;		//turn on Debug LED
	#if __BORLANDC__ >= 0x720
	ptrTimer_Thread->Start();			// start the thread for CB10.1
	#else
	ptrTimer_Thread->Resume();			// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
	#endif
    }
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Program Control Form should be manually created as it is rarely needed
if (!ptrProgram_Control[PCNo - 1]) {
	ptrProgram_Control[PCNo - 1] = new TProgram_Control(this, EngineNo, RemoteNo, TIUNo, PCNo);     // Pass "this" as the owner
	assert(ptrProgram_Control[PCNo - 1] != NULL);			// make sure that construction worked
    ptrProgram_Control[PCNo - 1]->ParentHandle = Handle; 	// for return messages
   	}
ptrProgram_Control[PCNo - 1]->Show();	// since I use a manual pointer, the form must get freed at close
if (Sender) {
	reinterpret_cast<TMenuItem *>(Sender)->Checked = true;	// add check mark to PCPopupMenu item
	}
// show the [Show Tags] button
if ((Setup->RFID_COM_Mode->Checked || Setup->RFID_WiFi_Mode->Checked) ) {
    ptrProgram_Control[PCNo - 1]->ShowDetectedTagsButton->Visible = true;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Prog_Control_ButtonClick(TObject *Sender)
{
// TODO ASSERTING works ok but I need to decide where I want to use it
//ASSERTING(bTIUConnected == true, TEXT("Press the [Connect] button before performing this operation"));
//ASSERTING(bTIUConnected == true, ConnectMsg.c_str());
ProgramControlClick(PCPopupMenu->Items->Items[0]);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormCloseQuery(TObject *Sender, bool &CanClose)
{
// In debug mode, allow disconnect
if (DEBUGGING) {
	CanClose = true;
    return;
    }
// TIU Must be disconnected from the COM ports to allow exit
if (IsRadioCOMConnected() || IsRFIDCOMConnected()){
	DebugMessagePopup("Warning", "[Disconnect] from the TIU before trying to exit", "");
	CanClose = false;
    return;
    }
CanClose = true;	// allow close
}
//---------------------------------------------------------------------------
// An Improved assert() for C++ Builder
// https://edn.embarcadero.com/article/28432
// https://codecentral.embarcadero.com/Item.aspx?id=17563
// By: David Pankhurst
// modified to run under both ANSI (BCB4.0 & BCB6.0) and UNICODE (CB10.x) OS
int HandleAsserting(TCHAR *testStr,TCHAR *msgStr,TCHAR *fileStr,int line,int *callFlag) {
    // assert message and return flag regarding aborting:
    // callFlag set if repeating forbidden
    static TCHAR s_text[199]=TEXT("");	// don't assign dynamically in
                                		// case of 'out of memory' error
	wsprintf(s_text,TEXT("FAILED: %s\r\n")
					TEXT("Error: ( %s )\r\n")
					TEXT("File '%s', Line %d\r\n")
					TEXT("Abort execution, allow assert Retry, or Ignore in future?"),
					msgStr, testStr, fileStr, line);
    switch ( ::MessageBox(NULL, s_text, TEXT("ASSERTION ERROR"),MB_ABORTRETRYIGNORE) ) {
      case IDIGNORE: // prevent calling again - turn off flag
        *callFlag=0; // never call again
        break;
      case IDABORT:  // return flag and break
        return 1;    // abort/break
	  }
return 0;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::BreakButtonClick(TObject *Sender)
{
// Test the DebugBreak using some different TCHAR data types
// This worked but I never used these functions
#ifdef ASSERTINGFUNCTIONS
TCHAR msg[128];
_TCHAR msg1[128];
_sntprintf(msg, sizeof msg, TEXT("Break ?"));
_sntprintf(msg1, sizeof msg1, _T("Break ?"));
// TODO Both of the next two ASSERTING() statements get "W8008 Condition is always true" because they are here for testing only
ASSERTING(0 == 1, msg);					// example of use of ASSERTING with a TCHAR message
ASSERTING(1 == 0, TEXT("2nd Break?"));	// example of use of ASSERTING with a text string

// askDlg() - A convenient message dialog for C++ Builder
// Code near the beginning of Train_Control.cpp

  BOOL isCancel;
  char reportName[25] = {"One"};
  if (askDlg(this, isCancel, "Do you want to print the report %s.", reportName)) {
    //... report printing code
    MyMessageDlg(this, mtInformation, "Report %s printed.", reportName);
  }
  else
    if (isCancel) //You can use this if you want different paths for No and Cancel
      return;     //actions.
  //proceed with further action if any
#endif
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ComboBoxButtonClick(TObject *Sender)
{
// strings for  CarNodelComboBox, EngineModelComboBox and RailroadComboBox
CarModelComboBox->Visible = !CarModelComboBox->Visible;
EngineModelComboBox->Visible = !EngineModelComboBox->Visible;
RailroadNameComboBox->Visible = !RailroadNameComboBox->Visible;
EngineTypeComboBox->Visible = !EngineTypeComboBox->Visible;
TagLocationComboBox->Visible = !TagLocationComboBox->Visible;
CouplerComboBox->Visible = !TagLocationComboBox->Visible;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::NameofEngineChange(TObject *Sender)
{
// Change the backgroup color to red if the string is longer than 16 characters
// or ;ess than or equal to one character
if (NameofEngine->Text.Length() <= 1 || NameofEngine->Text.Length() > 16) {
	// make background red
	NameofEngine->Color = clRed;
	} else {
	// make background normal
	NameofEngine->Color = clWindow;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::NameofEngineExit(TObject *Sender)
{
// when Engine Name edit loses focus, restore its value from the EngineList
NameofEngine->Text = EngineData->Cells[ENG_NAME][EngineNo];
FileNamePrefix->Text = NameofEngine->Text;
NameofEngine->ReadOnly = true;
NameofEngine->Cursor = crNo;
NameofEngine->Hint = "Press [Read] to read names of Engines";
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::NameofEngineKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
// handle Engine Name completion processing
switch (Key) {
	case VK_ESCAPE:
		// on [Esc], cancel edit
    	NameofEngineExit(Sender);
        EngineData->SetFocus();
        return;

	case VK_RETURN:
		// on [Enter], send new engine name to engine
		if (EngineNo > 0) EngineData->Cells[ENG_NAME][EngineNo] = NameofEngine->Text;
		FileNamePrefix->Text = NameofEngine->Text;
        AnsiString NewEngineName = NameofEngine->Text;	// must be an AnsiString because c_str() is taken
		// send Command with exactly 16 characters for new Engine name
		unsigned char Command_String[19];
		// Limit engine name sent to TIU to exactly 16 characters
		snprintf((char *)Command_String, sizeof Command_String, "Ns%-16.16s", NewEngineName.c_str());	// exactly 16 characters
		if (!ptrTIUCommThread->Send_String_FIFO(EngineNo, Command_String, TIUNo, RemoteNo, 11, LONG_WAIT, NO_RECORD)) {
			ShowMessage("Failed to set new Engine name (Engine " + String(EngineNo) + ")");
		    return;
		    }
		SetConsistState(EngineNo, TIUNo, ACTIVE);		// set active
        // update users of the Engine List
		PutEngData();
        //
		NameofEngineExit(Sender);	// complete edit
        EngineData->SetFocus();
    	return;
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ShowDetectedTagsClick(TObject *Sender)
{
// Show the Detected Tags window if RFID port is connected
// DetectedRFIDTags window is Auto-Create
if ((Setup->RFID_COM_Mode->Checked || Setup->RFID_WiFi_Mode->Checked) && !DetectedRFIDTags->Visible) {
	DetectedRFIDTags->Show();
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CMQuick(TMessage &Message)
{
// Open a page of Hot Buttons
int NewTIU = Message.WParam;
assert(NewTIU >= 1 && NewTIU <= TIUMAX);
DoQuickButton(NewTIU);
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::QuickButtonClick(TObject *Sender)
{
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
DoQuickButton(TIUNo);
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::DoQuickButton(int NewTIU)
{
// Create Quick Button page 1 to TIUMAX
//
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Quick Operations Form should be manually created as up to 5 can exist
assert(NewTIU >= 1 && NewTIU <= TIUMAX);

// Create the Lashup window but don't show it.
// This code was copied here so the Quick Controls window has the Lashup name and Lashup Engine List
//
if (ptrLashUp[NewTIU-1] == NULL) { 	// TIU 1-5
	ptrLashUp[NewTIU-1] = new TLashUp(this, RemoteNo, NewTIU);     // Pass "this" as the owner
	assert(ptrLashUp[NewTIU-1] != NULL);		// make sure that construction worked
    ptrLashUp[NewTIU-1]->ParentHandle = Handle; // for return messages
   	//ptrLashUp[NewTIU-1]->Show();	// don't show it here
	}

if (ptrQuick[NewTIU-1] == NULL) { 	// TIU 1-5
	ptrQuick[NewTIU-1] = new TQuickCtrl(this, RemoteNo, NewTIU, Setup->Quick_Ctrl_Scaling->Value);     // Pass "this" as the owner
	assert(ptrQuick[NewTIU-1] != NULL);		// make sure that construction worked
    ptrQuick[NewTIU-1]->ParentHandle = Handle; // for return messages
	} else {   // it's already created
	ptrQuick[NewTIU-1]->Show();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PrintthisWindow1Click(TObject *Sender)
{
MainForm->PrintScale = poProportional;
MainForm->Print();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ShowLayoutWindowClick(TObject *Sender)
{
// Create Layout Window
//
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Quick Operations Form should be manually created
if (LayoutForm == NULL) {
	LayoutForm = new TLayoutForm(this);     // Pass "this" as the owner
	assert(LayoutForm != NULL);		// make sure that construction worked
    LayoutForm->ParentHandle = Handle; // for return messages
	} else {   // it's already created
	LayoutForm->ShowIt();
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::QButtonClick(TObject *Sender)
{
// test the Q "Quality" command
// NOTE: the Q command always fails with a NAK. Either I am not using it correctly
// or it does not work through the radio interface to the TIU
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

byte QResp[CMDMAX];
ptrTIUCommThread->Send_String_FIFO(EngineNo, "Q100", TIUNo, RemoteNo, 21, SHORT_WAIT, NO_RECORD, NULL, NULL, NULL, QResp);
unsigned int ival = QResp[5] << 24 | QResp[6] << 16 | QResp[7] << 8 | QResp[8];
char buff[20];
snprintf(buff, sizeof buff, "%08X", ival);
if (DEBUGGING) DB->Debug_Print("Quality : " + String(buff), clGreen);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngTypeButtonClick(TObject *Sender)
{
// Fetch the Engine Type. This value, 2 bytes starting at address 0x200, is fetched
// by the DCS Loader. I don't know what it means.
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

EngTypeButton->Caption = "Engine Type";

byte QResp[CMDMAX];
unsigned int ival = 0;
bool bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "V2200", TIUNo, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, QResp);
if (!bretv) {	// retry once
    if (DEBUGGING) DB->Debug_Print("EngTypeButtonClick() Retrying V2200");
	bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "V2200", TIUNo, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, QResp);
	if (!bretv) return;	// return with no change if we can't read RAM
    }
if (bretv) ival = QResp[5] << 8 | QResp[6];
char buff[20];
snprintf(buff, sizeof buff, "%04X", ival);
if (DEBUGGING) DB->Debug_Print("Engine Type at 0x200 : " + String(buff), clGreen);
EngTypeButton->Caption = buff;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::EngType2ButtonClick(TObject *Sender)
{
// Fetch the Engine Type. This value, 2 bytes starting at address 0x22, is fetched
// by the start up engine sequence. With it, we can tell if the engine is
// a diesel or steam
//
// I've seen these values for the LO byte:
// 0x00 for steam, 0x05/0x85 for Diesel and Gas/Electric, 0x25 for electric
//
// I don't know what the HO byte means.
//
// if TIU Port is not connected fail
if (!IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (Setup->TXOnlyComm->Checked) {	// cannot do this function without full duplex communications
	MainForm->DebugMessagePopup("Warning", MainForm->TXMsg, "");
    return;
    }

EngType2Button->Caption = "Engine Type";

byte QResp[CMDMAX];
unsigned int ival = 0;
bool bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "V2022", TIUNo, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, QResp);
if (!bretv) {	// retry once
    if (DEBUGGING) DB->Debug_Print("EngType2ButtonClick() Retrying V2022");
	bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "V2022", TIUNo, RemoteNo, 21, QUERY_WAIT, NO_RECORD, NULL, NULL, NULL, QResp);
	if (!bretv) return;	// return with no change if we can't read RAM
    }
if (bretv) ival = QResp[5] << 8 | QResp[6];
char buff[20];
snprintf(buff, sizeof buff, "%04X", ival);
if (DEBUGGING) DB->Debug_Print("Engine Type at 0x22 : " + String(buff), clGreen);
EngType2Button->Caption = buff;
}


void __fastcall TMainForm::MarksTrainLayouts1Click(TObject *Sender)
{
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/layout.html"), NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------



void __fastcall TMainForm::RegButtonClick(TObject *Sender)
{

String Operation = TEXT("open");
String FileName = TEXT("\"c:\\windows\\system32\\reg.exe\"");
String Parameters = TEXT("\"export\" \"HKCU\\Software\\Remote Control\\Serial\" \"RTC.reg\" \"/y\"");
String Directory = cStartDir;
HINSTANCE retv = ShellExecute(NULL, Operation.c_str(), FileName.c_str() , Parameters.c_str(), Directory.c_str(), SW_SHOWNORMAL);
if ((int)retv <= 32) {
	// error
    if (DEBUGGING) DB->Debug_Print("Shell Execute error - return value = " + String((int) retv));
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU11Click(TObject *Sender)
{
AIUPopupMenu->Tag = 1;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU21Click(TObject *Sender)
{
AIUPopupMenu->Tag = 2;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU31Click(TObject *Sender)
{
AIUPopupMenu->Tag = 3;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU41Click(TObject *Sender)
{
AIUPopupMenu->Tag = 4;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU51Click(TObject *Sender)
{
AIUPopupMenu->Tag = 5;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::TIU61Click(TObject *Sender)
{
AIUPopupMenu->Tag = 6;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU71Click(TObject *Sender)
{
AIUPopupMenu->Tag = 7;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIU81Click(TObject *Sender)
{
AIUPopupMenu->Tag = 8;
TPoint p_rel_scr = Mouse->CursorPos;	// get xy Mouse position
AIUPopupMenu->Popup((int) p_rel_scr.x, (int) p_rel_scr.y);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUControl1Click(TObject *Sender)
{
// AIU 1
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 1);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUControl2Click(TObject *Sender)
{
// AIU 2
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 2);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUControl3Click(TObject *Sender)
{
// AIU 3
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 3);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUControl4Click(TObject *Sender)
{
// AIU 4
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 4);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUControl5Click(TObject *Sender)
{
// AIU 5
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 5);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIU6Control1Click(TObject *Sender)
{
// AIU 6
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 6);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIU7Control1Click(TObject *Sender)
{
// AIU 7
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 7);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::AIU8Control1Click(TObject *Sender)
{
// AIU 8
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 8);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIU9Control1Click(TObject *Sender)
{
// AIU 9
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 9);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIU10Control1Click(TObject *Sender)
{
// AIU 10
int MyTIUNo = AIUPopupMenu->Tag;
StartAIUControl(Sender, MyTIUNo, 10);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AIUButtonClick(TObject *Sender)
{
if (AIUNo == 0) {
    //DebugMessage("          No AIU!         ");
    DebugMessagePopup("Warning", "          No AIU!         ");
	return;
    }
// AIUNo from AIU Spinner and TIUNo from TIU Spinner
StartAIUControl(Sender, TIUNo, AIUNo);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::StartAIUControl(TObject *Sender, int iTIUNo, int iAIUNo)
{
if (iAIUNo == 0) {
    //DebugMessage("          No AIU!         ");
    DebugMessagePopup("Warning", "          No AIU!         ");
	return;
    }
// Open accessories / switches form
// manually create form
assert(iTIUNo >= 1 && iTIUNo <= TIUMAX);
assert(iAIUNo >= 1 && iAIUNo <= AIUMAX);
// Many AIU Control Forms
if (ptrAIUControl[iTIUNo-1][iAIUNo-1] == NULL) {
	ptrAIUControl[iTIUNo-1][iAIUNo-1] = new TAIUControl(this, RemoteNo, EngineNo, iTIUNo, iAIUNo);   // create the Form & Pass "this" as the owner
	assert(ptrAIUControl[iTIUNo-1][iAIUNo-1] != NULL);	// make sure that construction worked
    ptrAIUControl[iTIUNo-1][iAIUNo-1]->ParentHandle = Handle; // for return messages
	ptrAIUControl[iTIUNo-1][iAIUNo-1]->Show();
	} else {    // it's already created, just show it
	ptrAIUControl[iTIUNo-1][iAIUNo-1]->Show();
	}
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//
// check that d is a valid Encoded digit, that is, is 0x5,0x6 0x9, or 0xA.
//
inline unsigned char ValidDigit(unsigned char d)
{
if (d == 0x05 || d == 0x06 || d == 0x09 || d == 0x0A) return d;
return 0;
}//---------------------------------------------------------------------------
//
// check that c is a valid encoded byte, that is, is "0x4D" or contains only digits 0x5,0x6 0x9, and 0xA.
//
// W8027 Functions containing some if statements are not expanded inline.
// W8070 Function should return a value
inline unsigned char ValidByte(unsigned char c)
{
if (c == 0x4D) return c;
if (ValidDigit(c & 0x0F))
  if (ValidDigit((c >> 4) & 0x0F))
    return c;
return '\0';
}
//---------------------------------------------------------------------------
/**
 * charToHex
 *
 * 'ch' Character to be converted to hexadecimal
 *
 * Returns:
 *  Hex value, 0x00 on bad hex character
 */
//---------------------------------------------------------------------------
inline byte charToHex(byte ch)
{
byte val;
if (ch >= 'A' && ch <= 'F')
    val = ch - 55;
    else if (ch >= 'a' && ch <= 'f')
      val = ch - 87;
      else if (ch >= '0' && ch <= '9')
         val = ch - 48;
         else
         val = 0x00;
return val;
}
//---------------------------------------------------------------------------
/* Hex2Digit_to_Decimal
 *
 * convert a two digit hex character string into an 8 bit unsigned decimal number
 */
//--------------------------------------------------------------------------
inline unsigned char Hex2Digit_to_Decimal(unsigned char * str) {
// convert a 2 digit char string to decimal
unsigned char val = (charToHex(*str) << 4) & 0xF0;
val |= charToHex(*(str + 1)) & 0x0F;
return val;
}
//---------------------------------------------------------------------------

char * Amp(const char * a)
{
//
// Replace "&" with "&&" in string
//
size_t a_l = strlen(a);
int count = 0, index = 0;
for (size_t i = 0;  i < a_l;  ++i) {
    if (a[i] == '&') {
        count++;
	    }
	}
char *o = (char *) malloc(a_l + 1 * count + 1); // 1 is because you add 2 new symbols, but remove 1 so 2 - 1 = 1
memset((void *)o, 0, sizeof(o));
for (size_t i = 0;  i < a_l;  ++i) {
    if (a[i] != '&')
        o[index++] = a[i];
    else {
        o[index++] = '&';
        o[index++] = '&';
    	}
	}
o[index++] = '\0';
// return pointer value must be free()d by caller once the string is copyed
return o;
}

//---------------------------------------------------------------------------

/* TITLE: A convenient message dialog for C++ Builder

If you often wished for a nice printf like function to display messages 
in C++ Builder, here is some code that I wrote. I suggest that you create 
a separate unit and add this code to it. Copy the function prototypes of 
MyMessageDlg and askDlg to the header file and you are ready to use them. 

Example of usage:
  BOOL isCancel;
  if (askDlg(this, isCancel, "Do you want to print the report %s.", reportName)) {
    //... report printing code
    MyMessageDlg(this, mtInformation, "Report %s printed.", reportName); 
  }
  else
    if (isCancel) //You can use this if you want different paths for No and Cancel
      return;     //actions.
  //proceed with further action if any
  ...

The advantage is that you can use printf like arguments for the message. 
Also, it automatically centers itself in the center of the parent form
passed as the first parameter. Of course, it depends on the values of 
assumedWidthOfMsgDlg and assumedHeightOfMsgDlg. If you make any
improvements to this code, please send them to delphistuff@kanadepro.com.

The second function askDlg can be used to ask a question in the printf 
style arguments. 

Note that this code uses a message buffer of 1024 bytes. If you are 
expecting larger messages, increase the buffer size or change the code
to allocate a larger buffer dynamically.
*/
//---------------------------------------------------------------------------

int msgx, msgy;

//---------------------------------------------------------------------------
BOOL getProperMessageDlgPos(TForm *parent) {
  int assumedWidthOfMsgDlg = 500;
  int assumedHeightOfMsgDlg = 160;

  if (parent==NULL)
    parent = Application->MainForm;
  if (parent==NULL)
    return FALSE;  
  msgx = parent->Left + 100;
  msgy = parent->Top + 100;

  if (msgx + assumedWidthOfMsgDlg > Screen->Width)
    return FALSE;
  if (msgy + assumedHeightOfMsgDlg > Screen->Height)
    return FALSE;

  return TRUE;
}

//---------------------------------------------------------------------------
void MyMessageDlg(TForm *parent, TMsgDlgType DlgType, const char *format,...) {
  char msgBuf[1024];

  va_list arg_ptr;

  va_start(arg_ptr,format);
  vsprintf(msgBuf,format,arg_ptr);
  va_end(arg_ptr);

  if (!getProperMessageDlgPos(parent))
    MessageDlg(msgBuf, DlgType, TMsgDlgButtons() << mbOK, 0);
  else
    MessageDlgPos(msgBuf, DlgType, TMsgDlgButtons() << mbOK, 0, msgx, msgy);
}

//---------------------------------------------------------------------------
BOOL askDlg(TForm *parent, BOOL& isCancel, const char *format,...) {
  char msgBuf[1024];

  va_list arg_ptr;

  va_start(arg_ptr,format);
  vsprintf(msgBuf,format,arg_ptr);
  va_end(arg_ptr);

  isCancel = FALSE;

  int retcode;

  if (!getProperMessageDlgPos(parent))
    retcode = MessageDlg(msgBuf, mtConfirmation,
        TMsgDlgButtons() << mbYes << mbNo << mbCancel, 0);
  else
    retcode = MessageDlgPos(msgBuf, mtConfirmation, 
       TMsgDlgButtons() << mbYes << mbNo << mbCancel, 0,
        msgx, msgy);

  switch (retcode) {
    case mrNo:
      return FALSE;
    case mrCancel:
      isCancel = TRUE;
      return FALSE;
    default:  ;
  }

  return TRUE;
}
//---------------------------------------------------------------------------

/*
DISCLAIMER 

THE INFORMATION AND MATERIAL CONTAINED IN THIS ARTICLE IS PROVIDED "AS 
IS," WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING WITHOUT 
LIMITATION ANY WARRANTY CONCERNING THE ACCURACY, ADEQUACY, OR COMPLETENESS
OF SUCH INFORMATION OR MATERIAL OR THE RESULTS TO BE OBTAINED FROM USING 
SUCH INFORMATION OR MATERIAL. AUTHOR SHALL NOT BE RESPONSIBLE FOR ANY CLAIMS
ATTRIBUTABLE TO ERRORS, OMISSIONS, OR OTHER INACCURACIES IN THE INFORMATION 
OR MATERIAL CONTAINED IN THIS ARTICLE, AND IN NO EVENT SHALL THE AUTHOR BE 
LIABLE FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF SUCH INFORMATION OR MATERIAL.

copyright (C) 2000, Sanjay Kanade
http://kanadepro.com/delphistuff/
*/
//---------------------------------------------------------------------------

void __fastcall TMainForm::TIULineCountClick(TObject *Sender)
{
TIULineCount->Caption = "0";    
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDudpRXChangeState(TObject *Sender,
      TSocketState OldState, TSocketState NewState)
{
// handle change of state
if (DEBUGGINGX(9)) DB->Debug_Print("RFID RX WiFi Change State = " + String(OldState) + " to " + String(NewState));
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDudpRXSessionConnected(TObject *Sender,
      WORD ErrCode)
{
if (ErrCode != 0) {
	DB->Debug_Print("Can't connect to RFID RX WiFi. Error #" + String(ErrCode));
	} else {
	ConnectedRFIDLED->Checked = true;
	ConnectedRFIDLED->Flashing = false;
	bRFIDUDPWiFiConnected = true;
	bRFIDUDPWiFiRXConnected = true;
	if (DEBUGGING) DB->Debug_Print("RFID RX WiFi connected to Port " + RFIDudpRX->Port);
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDudpRXSessionClosed(TObject *Sender,
      WORD ErrCode)
{
ConnectedRFIDLED->Checked = false;
ConnectedRFIDLED->Flashing = false;
bRFIDUDPWiFiConnected = false;
bRFIDUDPWiFiRXConnected = false;
if (DEBUGGING) DB->Debug_Print("RFID RX WiFi disconnected from Port " + RFIDudpRX->Port);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDudpRXDataAvailable(TObject *Sender,
      WORD ErrCode)
{
if (ErrCode != 0) {
	return;
	}

char cBuffer[1024];
int Len;
#if __BORLANDC__ >= 0x720
// Defined for CB10.1 C++ programs that use the VCL.
Overbyteicswinsock::TSockAddrIn Detector;
#else
TSockAddrIn Detector;
#endif
int DetectorLen;

DetectorLen = sizeof(Detector);
Len = RFIDudpRX->ReceiveFrom(cBuffer, sizeof(cBuffer), Detector, DetectorLen);
if (Len <  0) {
    // ignore packets in error
    return;
    }
#ifdef NOTNEEDED
if (DEBUGGINGX(6)) DB->Debug_Print("RFID RX WiFi Dotted Host Address = " + String(WSocket_inet_ntoa(FServerAddr)));
#endif
if (DEBUGGINGX(6)) DB->Debug_Print("RFID RX WiFi Sin Port = " + String(Detector.sin_port));
if (DEBUGGINGX(6)) DB->Debug_Print("RFID RX WiFi Dotted Sin Address = " + String(WSocket_inet_ntoa(Detector.sin_addr)));

if (Len == 0) {
    if (DEBUGGING) DB->Debug_Print("RFID RX WiFi Connection Closed in OnDataAvaiable()");
    return;
    }

if (Len > 0) {
    cBuffer[Len] = '\0';
    //
    if (DEBUGGINGX(8)) DB->Debug_Print(cBuffer);
    // send tag to MainForm
    char * pBuffer = new char[(CMDMAX - 10)];	// NOTE: pbuffer is delete[]d in the CM_PROCESSRFIDDATA message handler CMProcessRFIDData()
    assert(pBuffer != NULL);
    strncpy(pBuffer, cBuffer, (CMDMAX - 10));
    // Send CM_PROCESSRFIDDATA message to MainForm which will handle the incoming data in pBuffer
    PostMessage(MainForm->Handle, CM_PROCESSRFIDDATA, '\n', (long)pBuffer); // WiFi RFID Full Packet
    // pBuffer is deleted in the MainForm Message Handler
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::RFIDtcpRXChangeState(TObject *Sender,
      TSocketState OldState, TSocketState NewState)
{
//OnChangeState 		When the component state changes. 
if (DEBUGGINGX(9)) DB->Debug_Print("RFID TCP RX WiFi Change State = " + String(OldState) + " to " + String(NewState));
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXSessionConnected(TObject *Sender,
      WORD ErrCode)
{
//OnSessionConnected 		Session to host etablished. 
if (ErrCode != 0) {
	if (DEBUGGING) DB->Debug_Print("Can't connect to TCP RFID RX WiFi. Error #" + String(ErrCode));
	} else {
	//ConnectedRFIDLED->Checked = true;
	//ConnectedRFIDLED->Flashing = false;
	//bRFIDTCPWiFiConnected = true;
	//RFIDTCPWiFiRXConnected = true;
	if (DEBUGGING) DB->Debug_Print("RFID TCP RX WiFi connected to Port " + RFIDtcpRX->Port);
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXClientConnect(TObject *Sender,
      TWSocketClient *Client, WORD Error)
{
	//OnClientConnect 		When a new client is connecting.
    // intialize the newly instantiated client component
    if (DEBUGGINGX(3)) DB->Debug_Print("RFID TCP Client connecting: " + ((TTcpSrvClient *)Client)->PeerAddr);
    ((TTcpSrvClient *)Client)->LineMode        = TRUE;
    ((TTcpSrvClient *)Client)->LineEdit        = TRUE;
    ((TTcpSrvClient *)Client)->OnDataAvailable = RFIDtcpRXDataAvailable;
    ((TTcpSrvClient *)Client)->OnBgException   = RFIDtcpRXBgException;
    ((TTcpSrvClient *)Client)->ConnectTime     = Now();
    // NOTE: When using Mircosoft Telnet Client, that client will not show
    // it is connected if the Banner is a blank string. I think that is because
    // it expects a \r\n. That end of line is not sent if Banner is blank.
    // When I press LEDOn or LEDOff, the Microsoft
    // Telnet Client shows the sent string and starts working as expected
    // though, apparently, there is still no end of line.
    // My test ESP8266 script TCPClient.ino works without the eof of line.
   	((TTcpSrvClient *)Client)->Banner     = "";
    if (DEBUGGINGX(8)) DB->Debug_Print("RFID TCP Client #" + String(RFIDtcpRX->ClientCount) + " connect: " + ((TTcpSrvClient *)Client)->PeerAddr + "    " +
        "Duration: " + FormatDateTime("hh:nn:ss",
        Now() - ((TTcpSrvClient *)Client)->ConnectTime));
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXClientDisconnect(TObject *Sender,
      TWSocketClient *Client, WORD Error)
{
//OnClientDisconnect 		When a client disconnects.
    if (DEBUGGINGX(8)) DB->Debug_Print("RFID TCP Client disconnect: " + ((TTcpSrvClient *)Client)->PeerAddr + "    " +
        "Duration: " + FormatDateTime("hh:nn:ss",
		Now() - ((TTcpSrvClient *)Client)->ConnectTime));
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXBgException(TObject *Sender,
      Exception *E, bool &CanClose)
{
//OnBgException 		When a background exception occurs.
    if (DEBUGGING) DB->Debug_Print("RFID TCP Client exception occured: " + E->ClassName() + ": " + E->Message);
    CanClose = TRUE;   // Goodbye client !
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXSessionAvailable(TObject *Sender,
      WORD ErrCode)
{
//OnSessionAvailable 		Client connected to this server.
    if (ErrCode) {
        if (DEBUGGING) DB->Debug_Print("RFID TCP Session available, error #" + IntToStr(ErrCode));
    } else {
    if (DEBUGGINGX(8)) DB->Debug_Print("TCP Session Available");
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXSessionClosed(TObject *Sender,
      WORD ErrCode)
{
//OnSessionClosed 		Socket has closed.
if (ErrCode != 0) {
	if (DEBUGGING) DB->Debug_Print("Can't close RFID TCP RX WiFi Session. Error #" + String(ErrCode));
	} else { 
	//ConnectedRFIDLED->Checked = false;
	//ConnectedRFIDLED->Flashing = false;
	//bRFIDTCPWiFiConnected = false;
	//RFIDTCPWiFiRXConnected = false;
	if (DEBUGGINGX(8)) DB->Debug_Print("RFID TCP RX WiFi disconnected from Port " + RFIDtcpRX->Port);
	}
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::RFIDtcpRXDataAvailable(TObject *Sender,
      WORD ErrCode)
{
//OnDataAvailable 		Data available in internal receive buffer. 
if (ErrCode != 0) {
	return;
	}

TTcpSrvClient *Client;
Client = (TTcpSrvClient *)Sender;						// current client in contact with us
// We use line mode. We will receive complete lines
Client->RcvdLine = Client->ReceiveStr();
// Remove trailing CR/LF
while ((Client->RcvdLine.Length() > 0) &&
        ((Client->RcvdLine[Client->RcvdLine.Length()] == '\r') ||
        (Client->RcvdLine[Client->RcvdLine.Length()] == '\n'))) {
    Client->RcvdLine = Client->RcvdLine.SubString(1, Client->RcvdLine.Length() - 1);
    }

if (Client->RcvdLine.Length() > 0) {
    //
    if (DEBUGGINGX(5)) DB->Debug_Print("RFID TCP Data Available : " + Client->RcvdLine);
    // send tag to MainForm
    char * pBuffer = new char[(1024 - 10)];	// NOTE: pbuffer is delete[]d in the CM_PROCESSRFIDDATA message handler CMProcessRFIDData()
    assert(pBuffer != NULL);
    strncpy(pBuffer, Client->RcvdLine.c_str(), (1024 - 10));
    // Send CM_PROCESSRFIDDATA message to MainForm which will handle the incoming data in pBuffer
    PostMessage(MainForm->Handle, CM_PROCESSRFIDDATA, '\n', (long)pBuffer); // WiFi RFID Full Packet
    // pBuffer is deleted in the MainForm Message Handler
	}
}
//---------------------------------------------------------------------------
// TTcpSrvClient is the class which will be instanciated by server component
// for each new client. N simultaneous clients means N TTcpSrvClient will be
// instanciated. Each being used to handle only a signle client.
// We can add any data that has to be private for each client, such as
// receive buffer or any other data needed for processing.
__fastcall TTcpSrvClient::TTcpSrvClient(TComponent* Owner)
    : TWSocketClient(Owner)
{
}
//---------------------------------------------------------------------------


